if { [info command ::GiD_Process] ne "" } {
    if { ![info exists ::GidPriv(ShowGroupsTab)] } {
        set ::GidPriv(ShowGroupsTab) 0 ;#0 to hide the native groups tab
    }
}

#################################################################################
#    conditions window: open_conditions, open_conditions_draw_post
#################################################################################


# interp alias "" GiD_EnterInBatch "" tk_messageBox -message
# interp alias "" GiD_ModifiedFileFlag "" tk_messageBox -message

# GiD may have loaded the package with an older version
# this should be done at the beginning of compassfem
# package forget snit fulltktree

package require customLib_utils
catch { package require tdom }
package require customLib_utils::c
package require customLib_utils::math

set err [catch { package present Tk }]
if { !$err } {
    set ::has_tk 1
} else {
    set ::has_tk 0
}

if { [catch { package require gid_themes } err] } {
    set exists_gid_themes 0
} else {
    set exists_gid_themes 1
}

if { [info commands ::GidUtils::IsTkDisabled] ne "" && [GidUtils::IsTkDisabled] } {
    set ::has_tk 0
}

if { $::has_tk } {
    package require customLib_utils::img
    package require img::png
    package require img::jpeg
    package require img::gif
    package require fulltktree
    package require tooltip
    package require dialogwin

    if {![package vsatisfies [package provide Tk] 8.5]} {
        if { [info command ttk::style] eq "" } {
            interp alias "" ttk::style "" style
        }
    }
}

package provide customLib 1.22.4

if { [info script] != "" } {
    set dir [file dirname [info script]]
} elseif { [info exists ::GIDDEFAULTTCL] } {
    #to be debugged by ramdebugger inside GiD
    set dir [file join $::GIDDEFAULTTCL gid_groups_conds gid_groups_conds]
}
if { [info exists dir] } {
    msgcat::mcload [file join $dir msgs]
    lappend ::auto_path $dir

    # source is made due to having another version of the package inside GiD
    foreach i [list reports.tcl transform.tcl] {
        catch {source [file join $dir $i]}
    }
}

if { $::has_tk } {
    foreach i [list conditions.tcl groups.tcl function_editor.tcl] {
        catch {source [file join $dir $i]}
    }
    if { $::exists_gid_themes } {
        foreach i [list groups_theme.tcl] {
            if {[file exists [file join $dir $i]]} {
                source [file join $dir $i]
            }
        }
    }
}

namespace eval gid_groups_conds {
    variable groups_file_version 1.0
    variable spd_file
    variable defaults_file

    variable problemtype ""
    variable doc
    variable doc_units
    variable gid_group_var
    variable open_boundary_cond
    variable boundary_cond_handlers ""
    variable open_post
    variable post_handlers ""
    variable last_drawn_id ""
    variable do_change_opengl_rendermode 1
    variable draw_auto_hide_surfaces 1
    variable opengl_rendermode ""
    variable groups_cache
    variable images
    variable images_cache ""
    variable draw_nodes_state ""
    variable program_name
    variable imagesdirList
    variable problemtype_lib_dir
    variable add_report "create_report_xml"
    variable add_data_report ""
    variable is_draw_post 0
    variable openGL_font
    variable origin
    variable dest
    variable id
    variable style "GiD_white_vectorial"
    variable list_w_buttons
    variable side ""
    variable use_pxml 1
    variable status_bar
    variable widgetList ""
    variable gl_colours
    variable delete_gr_from_cond 1
    variable view_entities 0
    variable open_all_var 0
    variable loaded_model 0
    variable close_button_all_gr_lay_window 0
    variable ListboxCont
    variable menu "data_tree"
    array set gl_colours {
        foreground       "0.0 0.0 0.0"
        background       "1.0 1.0 1.0"
        text_foreground  "0.0 0.0 0.0"
        text_background  "1.0 1.0 1.0"
    }
    variable material_import_export_actions_dict

    variable register_popup_help_after
    variable register_popup_help_active
    variable register_popup_help_window

    variable colorList [list magenta green yellow orange pink AntiqueWhite3 aquamarine2 \
        {blue violet} brown4 coral2 grey20 "orange red" "dark olive green" DarkSeaGreen2 \
        salmon RoyalBlue coral DarkKhaki ivory2 gold maroon "sky blue"]

    if { [info script] != "" } {
        set dir [file dirname [info script]]
    } elseif { [info exists ::GIDDEFAULTTCL] } {
        #to be debugged by ramdebugger inside GiD
        set dir [file join $::GIDDEFAULTTCL gid_groups_conds gid_groups_conds]
    }

    if { [info exists dir] } {
        variable topdir $dir
    }

    set register_popup_help_after ""
    set register_popup_help_active 1
    set register_popup_help_window ""

    if { $::tcl_platform(platform) != "windows"} {
        set style "GiD_light_plus"
    }
    if { $::exists_gid_themes } {
        if { ![info exists ::GidPriv(CurrentTheme)] } {
            set theme ""
        } else {
            set theme $::GidPriv(CurrentTheme)
        }
        set style $theme
    }
}


proc gid_groups_conds::parse_XML_data { data } {
    variable use_pxml

    if { $use_pxml } {
        set err [catch {
            package require compass_utils::c
            set doc [p::xml parse $data]
        } ret opts]
        if { $err } {
            set err1 [catch { package require compass_utils::c } ret1]
            if { $err1 } {
                set use_pxml 0
            } elseif { [info commands p::xml] == "" && [info procs p::xml] == "" } {
                set use_pxml 0
            } elseif { $err } {
                return -code $err -errorinfo [dict get $opts -errorinfo] $ret
            }
        }
    }
    if { $use_pxml == 0 } {
        set err [catch { dom parse $data } doc]
        if { $err } {
            error "error parsing data ($data)"
        }
    }
    return $doc
}

proc gid_groups_conds::open_XML_file { file } {
    variable use_pxml

    if { $use_pxml } {
        set err [catch {
            package require compass_utils::c
            set doc [p::xml parse -file $file]
        } ret opts]
        if { $err } {
            set err [catch {
                set fin [open $file r]
                fconfigure $fin -encoding utf-8
                set data [read $fin]
                close $fin
                set doc [gid_groups_conds::parse_XML_data $data]
            } ret opts]
        }
        if { $err } {
            #puts "err=$err opts=$opts"
            return -code $err -errorinfo [dict get $opts -errorinfo] $ret
            set use_pxml 0
        }
    }
    if { $use_pxml == 0 } {
        set err [catch { tDOM::xmlReadFile $file } data]
        if { !$err } {
            set err [catch { dom parse $data } doc]
            if { $err } {
                error "error parsing file ($doc)"
            }
        } else {
            error "error opening file ($data)"
        }
    }
    return $doc
}

proc gid_groups_conds::CurrentTheme {} {
    if { ![info exists ::GidPriv(CurrentTheme)] } {
        set theme ""
    } else {
        set theme $::GidPriv(CurrentTheme)
    }
    return $theme
}

proc gid_groups_conds::GetCurrentTheme {} {
    set theme [gid_groups_conds::CurrentTheme]
    return $theme
}

################################################################################
#    Start & End
################################################################################

proc gid_groups_conds::init_colours { } {
    variable gl_colours

    if { [info command ::GiD_Set] eq "" } {
        return
    }

    # Get Gid graphics background color
    if { [GiD_Set BackColor(Type)] } {
        # i.e. gid is using a 2 color background
        set col1 [gid_groups_conds::_OpenGL_color [GiD_Set BackColor(Top)]]
        set col2 [gid_groups_conds::_OpenGL_color [GiD_Set BackColor(Bottom)]]
        # just do an average
        # set gid_bg [list [expr 0.5 * ( [lindex $col1 0] + [lindex $col2 0])] \
        #                  [expr 0.5 * ( [lindex $col1 1] + [lindex $col2 1])] \
        #                  [expr 0.5 * ( [lindex $col1 2] + [lindex $col2 2])] ]
        # get the darkest one:
        set col1_y [expr 0.2126 * [lindex $col1 0] + 0.7152 * [lindex $col1 1] + 0.0722 * [lindex $col1 2]]
        set col2_y [expr 0.2126 * [lindex $col2 0] + 0.7152 * [lindex $col2 1] + 0.0722 * [lindex $col2 2]]
        if { $col1_y < $col2_y} {
            set gid_bg $col1
        } else {
            set gid_bg $col2
        }
    } else {
        set gid_bg [GiD_Set BackColor(Single)]
        set n_cols [scan $gid_bg %03d#%03d#%03d r g b]
        if { $n_cols == 3} {
            set gid_bg [list [expr $r / 255.0] [expr $g / 255.0] [expr $b / 255.0] ]
        } else {
            set gid_bg [list 1.0 1.0 1.0]
        }
    }

    set gl_colours(background) $gid_bg
    set gl_colours(text_background) $gid_bg
    # calculate the foreground color as in GiD
    lassign $gid_bg r g b
    set lum [expr 0.3086 * $r + 0.6094 * $g + 0.0820 * $b]
    if { $lum < 0.6} {
        set lum 0.0
    } else {
        set lum 1.0
    }
    set fg_comp [expr 1.0 - $lum ]
    set fg [list $fg_comp $fg_comp $fg_comp]
    set gl_colours(foreground) $fg
    set gl_colours(text_foreground) $fg
}

proc gid_groups_conds::init_package {} {
    variable doc
    variable is_draw_post
    variable style
    variable side
    variable use_pxml
    variable doc_units
    variable topdir

    if { [info command ::GiD_Process] eq "" } {
        set is_draw_post 1
    }

    set err [catch { package present Tk }]
    if { !$err } {
        set ::has_tk 1
    } else {
        set ::has_tk 0
    }
    if { ![info exists ::CUSTOM_LIB_USE_NATIVE_GROUPS] } {
        set ::CUSTOM_LIB_USE_NATIVE_GROUPS 0
    }

    if { [file exists [file join $topdir units.xml]] } {
        set doc_units [gid_groups_conds::open_XML_file [file join $topdir units.xml]]
    }


    if { $::has_tk } {
        package require customLib_utils::img
        package require img::png
        package require img::jpeg
        package require img::gif
        package require fulltktree
        package require tooltip
        package require dialogwin

        gid_groups_conds::create_images

        if { [regexp "Tdyn" $style {} curr_theme] } {
            # do nothing
        } else {
            # GiD_white_vectorial, GiD_black, etc
            create_fonts_colors_styles
        }
        init_colours

        if { [info command ::GiD_Process] ne "" || [info command ::draw_full::redraw] ne "" } {
            cu::init_tile_styles
            if { [info exists ::draw_post::data] } {
                draw_post::draw_text init
            } else {
                gid_groups_conds::draw_text init
            }
        }
        foreach i [list Cut Copy Paste] {
            bind TEntry <<$i>> "[bind TEntry <<$i>>];break"
            bind TCombobox <<$i>> "[bind TCombobox <<$i>>];break"
        }
    }
    if { [info command ::ContextualEntity::RegisterUserDataCommand] ne "" } {
        ContextualEntity::RegisterUserDataCommand gid_groups_conds::add_items_to_data_menu
    }
    if { ![info exists doc] } {
        if { $use_pxml == 0 } {
            #set doc [dom parse "<default/>"]
            set doc [dom parse "<default><groups/><groups_types editable='0'/></default>"]
        } else {
            set doc [p::xml parse "<default><groups/><groups_types editable='0'/></default>"]
        }
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
            #invoke it directly now to process the current groups, because before load this file the GiD_AfterCreateGroup event procedure
            #was not defined to syncronize both definitions

            set mff [GiD_ModifiedFileFlag get]
            gid_groups_conds::syncronize_gid_groups_and_doc_groups
            if { !$mff } {
                GiD_ModifiedFileFlag set 0
            }
        }

        if { $::has_tk } {
            if { [regexp "Tdyn" $style {} curr_theme] } {
                create_fonts_colors_styles_compassis
            } else {
                gid_groups_conds::init_style
            }
        }
    }

    # init the material import export window variables
    variable material_import_export_actions_dict
    set material_import_export_actions_dict [dict create \
        global_active [_ "Global database - active"] \
        global_inactive [_ "Global database - inactive"] \
        import_file [_ "Import from file"]... \
        export_file [_ "Export to file"]...]
}

proc gid_groups_conds::end_post {} {
    set theme [gid_groups_conds::GetCurrentTheme]
    if { ![regexp "Tdyn" $theme {} curr_theme] } {
        return
    }
    set w .gid.central.boundaryconds
    if {[winfo exists $w]} {
        gid_groups_conds::open_conditions close
        gid_groups_conds::open_conditions menu
    }
}

proc gid_groups_conds::init_style {} {
    variable style
    variable doc

    if { $::exists_gid_themes } {
        if { ![info exists ::GidPriv(CurrentTheme)] } {
            set theme ""
        } else {
            set theme $::GidPriv(CurrentTheme)
        }
        set style $theme
    }
}

proc gid_groups_conds::draw_text { what args } {
    variable openGL_font

    if { [info command ::ggc::draw::ftgl_text] ne "" } {
        return [gid_groups_conds::draw_text_ftgl $what {*}$args]
    } elseif { [info command ::ggc::draw::glctext] ne "" } {
        return [gid_groups_conds::draw_text_glc $what {*}$args]
    } else {
        return [gid_groups_conds::draw_text_pgf $what {*}$args]
    }
}

proc gid_groups_conds::draw_text_pgf { what args } {
    variable openGL_font
    variable gl_colours

    set default_font_size 13

    switch $what {
        init {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_font_size]
            set err [catch { GiD_OpenGL pgffont dimensions 0 }]
            if { !$err } {
                lassign [GiD_Set PGF(LegendFont)] font_file font_idx old_size fontnameList
                GiD_Set PGF(LegendFont) [list $font_file $font_idx $font_size]
                set openGL_font ""
            } else {
                set openGL_font [gid_groups_conds::font_to_drawopengl_font [list Helvetica $font_size]]
                GiD_OpenGL font push $openGL_font
                GiD_OpenGL font pop
            }
        }
        push {
            if { $openGL_font eq "" } {
                GiD_OpenGL pgffont pushfont legendfont
            } else {
                GiD_OpenGL font push $openGL_font
            }
        }
        pop {
            if { $openGL_font eq "" } {
                GiD_OpenGL pgffont popfont
                # GiD_OpenGL pgffont popfont PSEUDO_WORD_TO_AVOID_BUG
            } else {
                GiD_OpenGL font pop
            }
        }
        set_size {
            lassign $args font_size

            set font_size_old [gid_groups_conds::get_preference draw_post::font_size $default_font_size]
            gid_groups_conds::set_preferenceG draw_post::font_size $font_size

            if { $openGL_font eq "" } {
                set is_disabled [GiD_Project set disable_graphics]
                if { $is_disabled } {
                    enable_graphics
                }
                lassign [GiD_Set PGF(LegendFont)] font_file font_idx old_size fontnameList
                GiD_Set PGF(LegendFont) [list $font_file $font_idx $font_size]
                if { $is_disabled } {
                    enable_graphics
                }
            } else {
                set openGL_font [gid_groups_conds::font_to_drawopengl_font [list Helvetica $font_size]]
            }
            set redo_cmd [list gid_groups_conds::draw_text set_size $font_size]
            set undo_cmd [list gid_groups_conds::draw_text set_size $font_size_old]

            set label [_ "Change font size"]
            start_group_undo_batch $label

            #gid_groups_conds::add_to_undo $label $redo_cmd $undo_cmd ;#do not call this proc because draw_post::add_to_undo_batch second argument will be 1 instead 0, but I don't know its meaning
            variable is_draw_post
            if { !$is_draw_post } {
                GiD_ModifiedFileFlag set 1
                GiD_EnterInBatch "*****MARK START_COMMAND_TCL\n"
                GiD_EnterInBatch "\{-tcl- $redo_cmd\}\n"
                GiD_EnterInBatch "*****MARK END_COMMAND_TCL\n"
            } else {
                draw_post::add_to_undo_batch $label 0 $redo_cmd $undo_cmd
            }
            redraw_all -actualize_information_box
            end_group_undo_batch
        }
        increase {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_font_size]
            gid_groups_conds::draw_text set_size [expr {$font_size+2}]
        }
        decrease {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_font_size]
            gid_groups_conds::draw_text set_size [expr {$font_size-2}]

        }
        reset {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_font_size]
            gid_groups_conds::draw_text set_size $default_font_size
        }
        print {
            if { $openGL_font eq "" } {
                # GiD_OpenGL pgffont foreground 0.0 0.0 0.0 1.0
                GiD_OpenGL pgffont foreground {*}$gl_colours(text_foreground) 1.0
                GiD_OpenGL pgffont print [lindex $args 0]
            } else {
                GiD_OpenGL drawtext [lindex $args 0]
            }
        }
        measure {
            if { $openGL_font eq "" } {
                lassign [GiD_OpenGL pgffont dimensions [lindex $args 0]] width height
                return $width
            } else {
                return [GiD_OpenGL font measure [lindex $args 0]]
            }
        }
        linespace {
            if { $openGL_font eq "" } {
                lassign [GiD_OpenGL pgffont dimensions 0] width height
                return [expr {0.65*$height}]
            } else {
                return [GiD_OpenGL font metrics -linespace]
            }
        }
    }
}

proc gid_groups_conds::draw_text_glc { what args } {
    variable draw_text_stack

    set default_size 10

    switch $what {
        init {
            if { [info command ::ggc::draw::glctext] eq "" } {
                package require gid_groups_conds::c
            }
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_size]
            set face Normal
            foreach family [list FreeSans Verdana Tahoma Arial] {
                set err [catch { ggc::draw::glctext faces $family } facesList]
                if { $err } {
                    continue
                }
                foreach face [list Normal Regular Medium] {
                    if { $face in $facesList } {
                        break
                    }
                }
                break
            }
            ggc::draw::glctext font $family $face $font_size
            set draw_text_stack [list [list $family $face $font_size]]
        }
        push {
            # nothing
        }
        pop {
            # nothing
        }
        set {
            set font_size [lindex $args 0]
            gid_groups_conds::set_preferenceG draw_post::font_size $font_size
            lset draw_text_stack end 2 $font_size
            ggc::draw::glctext font {*}[lindex $draw_text_stack end]
            redraw_all -actualize_information_box
        }
        increase {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_size]
            incr font_size 2
            gid_groups_conds::set_preferenceG draw_post::font_size $font_size
            lset draw_text_stack end 2 $font_size
            ggc::draw::glctext font {*}[lindex $draw_text_stack end]
            redraw_all -actualize_information_box
        }
        decrease {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_size]
            incr font_size -2
            gid_groups_conds::set_preferenceG draw_post::font_size $font_size
            lset draw_text_stack end 2 $font_size
            ggc::draw::glctext font {*}[lindex $draw_text_stack end]
            redraw_all -actualize_information_box
        }
        reset {
            gid_groups_conds::set_preferenceG draw_post::font_size $default_size
            lset draw_text_stack end 2 $default_size
            ggc::draw::glctext font {*}[lindex $draw_text_stack end]
            redraw_all -actualize_information_box
        }
        print {
            ggc::draw::glctext print [lindex $args 0]
        }
        measure {
            return [ggc::draw::glctext measure [lindex $args 0]]
        }
        linespace {
            return [ggc::draw::glctext linespace]
        }
    }
}

proc gid_groups_conds::draw_text_ftgl { what args } {
    variable draw_text_stack

    set default_size 11

    switch $what {
        init {
            if { [info command ::ggc::draw::ftgl_text] eq "" } {
                package require gid_groups_conds::c
            }
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_size]
            if { $::tcl_platform(platform) eq "windows" } {
                set file [cu::give_truetype_font_file "tahoma"]
            } elseif { $::tcl_platform(os) eq "Darwin" } {
                set file [cu::give_truetype_font_file "Tahoma"]
            } else {
                set file [cu::give_truetype_font_file "FreeSans"]
                if { $file eq "" } {
                    set font [font actual TkDefaultFont -family]
                    if { $font eq "" } {
                        error "gid_groups_conds::draw_text_ftgl font does not have any value"
                    }
                    set file [cu::give_truetype_font_file $font]
                }
            }

            if { $file eq "" } {
                error "gid_groups_conds::draw_text_ftgl file does not have any value. font: $font"
            }
            ggc::draw::ftgl_text font $file $font_size
            set draw_text_stack [list [list $file $font_size]]
        }
        push {
            # nothing
        }
        pop {
            # nothing
        }
        set {
            set optional {
                { -redraw boolean 1 }
            }
            set compulsory "font_size"
            parse_args $optional $compulsory $args

            gid_groups_conds::set_preferenceG draw_post::font_size $font_size
            lset draw_text_stack end 1 $font_size
            ggc::draw::ftgl_text font {*}[lindex $draw_text_stack end]
            if { $redraw } {
                redraw_all -actualize_information_box
            }
        }
        increase {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_size]
            incr font_size 2
            gid_groups_conds::set_preferenceG draw_post::font_size $font_size
            lset draw_text_stack end 1 $font_size
            ggc::draw::ftgl_text font {*}[lindex $draw_text_stack end]
            redraw_all -actualize_information_box
        }
        decrease {
            set font_size [gid_groups_conds::get_preference draw_post::font_size $default_size]
            incr font_size -2
            gid_groups_conds::set_preferenceG draw_post::font_size $font_size
            lset draw_text_stack end 1 $font_size
            ggc::draw::ftgl_text font {*}[lindex $draw_text_stack end]
            redraw_all -actualize_information_box
        }
        reset {
            gid_groups_conds::set_preferenceG draw_post::font_size $default_size
            lset draw_text_stack end 1 $default_size
            ggc::draw::ftgl_text font {*}[lindex $draw_text_stack end]
            redraw_all -actualize_information_box
        }
        print {
            ggc::draw::ftgl_text print [lindex $args 0]
        }
        measure {
            set width [ggc::draw::ftgl_text measure [lindex $args 0]]
            return $width
        }
        linespace {
            set linespace [ggc::draw::ftgl_text linespace]
            return [expr {0.8*$linespace}]
        }
    }
}

proc gid_groups_conds::header_title { win txt } {
    variable theme

    foreach j [winfo children $win.f0] {
        grid remove $j
    }
    set fwopen $win.f0

    set st ToolbuttonModBlue
    ttk::label $fwopen.t -text [_ "%s" $txt] -compound left -style $st
    grid $fwopen.t -sticky news

    grid columnconfigure $fwopen 0 -weight 1
    grid $fwopen -sticky w -padx 2 -pady 2

}

proc gid_groups_conds::SetButtonsHeader { fwopen win } {
    variable program_name
    variable list_w_buttons
    variable doc
    variable side
    variable status_bar

    if { ![info exists doc] } { return }

    set programname [gid_groups_conds::GetProgramName]

    set style ToolbuttonModGrey
    if { ![info exists list_w_buttons] } {
        set list_w_buttons ""
    }
    upvar 0 ::ttk::theme::[gid_themes::GetTtkTheme]::colors colors
    frame $fwopen -background $colors(-text_bg)
    $win configure -frame_bg $colors(-text_bg)

    if {$doc != ""} {
        set root [$doc documentElement]
        set styleNode [$root selectNodes style]
        if { $styleNode == ""} {
            set pt [GiD_Info problemtypepath]
            set file_spd [file join $pt compassfem_default.spd]
            if {[file exists $file_spd]} {
                set doc_new [gid_groups_conds::open_XML_file_gzip $file_spd]
                set styleNode [[$doc_new documentElement] selectNodes style]
                if {[$root selectNodes style] == ""} {
                    $root appendChild [$styleNode clone -deep]
                }
                $doc_new delete
            }
        }
        if { $styleNode != ""} {
            set pre_buttonsNode [$styleNode selectNodes pre_buttons]
            if {$pre_buttonsNode != ""} {
                set idx 0
                set col 0
                foreach ivalue [$pre_buttonsNode childNodes] {
                    set style ToolbuttonModGrey
                    set n [$ivalue @n ""]
                    set help [$ivalue @help ""]
                    set help [= [subst -nocommands -novariables $help]]
                    set command [$ivalue @command ""]
                    if {$n eq "programname"} {
                        set v [_ "%s Data" $programname]
                        set style ToolbuttonModBlue
                    } else {
                        set v [_ [$ivalue @pn ""]]
                        set style ToolbuttonModGrey
                    }
                    ttk::button $fwopen.but$idx -text $v -command [list $command $style $fwopen.but$idx $win] \
                        -width 10 -style $style
                    bind $fwopen.but$idx <ButtonPress-1> [list $command $style $fwopen.but$idx $win press]
                    grid $fwopen.but$idx -sticky ewns -row 0 -column [incr col]
                    grid columnconfigure $fwopen $fwopen.but$idx -weight 1
                    cu::tooltip register $fwopen.but$idx "<b>$v</b>\n$help"
                    if {[lsearch $list_w_buttons $fwopen.but$idx] == "-1"} {
                        lappend list_w_buttons $fwopen.but$idx
                    }
                    incr idx
                }
            }

        }
    }

    grid columnconfigure $fwopen 0 -weight 1
    grid $fwopen -sticky w -padx 2 -pady 2

    ####
    # Separator
    frame $win.separator -background $colors(-toolbars_bg) -borderwidth 2 -width 2 -height 1
    grid $win.separator -sticky ew -row 1 -column 0 -padx 5 -pady {5 5}

    gid_groups_conds::create_customized_theme

    set pw [winfo parent $win]
    if { [winfo exists $pw] && [winfo class $pw] eq "Toplevel" } {
        set img "more_options_float"
        if {$side eq "left"} {
            set sd "left"
        } else {
            set sd "right"
        }
        set help [_ "Dock %s panel" $sd]
    } else {
        set img "more_options"
        set help [_ "Float panel"]
    }
    if {$side eq "left"} {
        set sd "right"
    } else {
        set sd "left"
    }
    set data_tree_side [_ "Dock %s panel" $sd]
    set img [gid_themes::GetImage $img.png]
    set cmds [list \
        [list $img $help mbb [list gid_groups_conds::open_conditions toggle -side $side] \
        [list \
            [list $img $help "" [list gid_groups_conds::open_conditions toggle -side $side]] \
            [list $img $data_tree_side "" \
            [list gid_groups_conds::open_conditions_side $sd]] \
        ] \
    ] \
]

set col $idx
foreach i $cmds {
    lassign $i img0 help type cmd cmdList
    set img $img0
    switch $type {
        "" - mb {
            set b [ttk::button $fwopen.but$col -image $img -style ToolbuttonMod]
            if { $type eq "mb" } {
                cu::add_down_arrow_to_image -color black -w $b $img
            }
        }
        mbb {
            set b [cu::menubutton_button $fwopen.but$col -image $img -menu $fwopen.but$col.m \
                -style ToolbuttonMod -direction left]
        }
    }
    set cmd [string map [list %W [list $b]] $cmd]
    switch $type {
        "" {
            $b configure -command $cmd
        }
        mb {
            bind $b <ButtonPress-1> "$cmd ; break"
            bind $b <Return> "$cmd ; break"
            bind $b <space> "$cmd ; break"
        }
        mbb {
            $b configure -command $cmd
            menu $b.m -tearoff 0
            foreach i $cmdList {
                lassign $i img0_i help_i type_i cmd_i
                $b.m add command -label $help_i -command $cmd_i
                if { $img0_i ne "" } {
                    set img_i $img0_i
                    $b.m entryconfigure end -image $img_i -compound left
                }
            }
        }
    }
    grid $b -row 0 -column [incr col] -sticky nsw -padx 5 -pady "0 3"
    set v [_ "Float/dock window"]
    cu::tooltip register $fwopen.but$idx "<b>$v</b>\n$help"
    if { $img0 eq "---" } { grid configure $b -sticky nsw -pady "0 3" }
    if {[lsearch $list_w_buttons $b] == "-1"} {
        lappend list_w_buttons $b
    }
    incr col
}
}

proc gid_groups_conds::status_bar { type } {
    variable status_bar

    if {![info exists status_bar]} {
        set status_bar [gid_groups_conds::get_preference status_bar on]
    }
    set status_bar $type
    gid_groups_conds::set_preference status_bar $status_bar
    GiD_Process 'Redraw
}

proc gid_groups_conds::open_conditions_side { sd } {
    variable side

    set side $sd
    gid_groups_conds::open_conditions close
    gid_groups_conds::open_conditions menu
}

proc gid_groups_conds::_preferences_give_defaults {} {
    return [dict create background_color auto text_color auto]
}

proc gid_groups_conds::setpreferences_do { win w } {
    if { [$w giveaction] < 1 } {
        destroy $w
        return
    }
    if { [$w giveaction] == 3 } {
        set_preferences -to_defaults
        return
    }
    set key $w
    set d [formulae::give_params_dict $key]

    set tree_prefsNode [gid_groups_conds::get_preference_node -local_global global \
        tree_preferences]
    set values [_preferences_give_defaults]
    foreach n [dict keys $values] {
        set v [dict get $d $n]
        $tree_prefsNode setAttribute $n $v
    }
    foreach itree [list $win.ft.t $win.ft.tg $win.ft.tl] {
        gid_groups_conds::_changed_size $itree
    }
    if { [$w giveaction] == 1 } { destroy $w }
}

proc gid_groups_conds::setpreferences { win args } {
    variable doc
    #     set optional {
    #         { -to_defaults "" 0 }
    #     }
    #     set compulsory ""
    #     parse_args $optional $compulsory $args

    set to_defaults 0
    set w $win.tree_prefs
    destroy $w
    dialogwin_snit $w -title [_ "Preferences"] -callback \
        [list gid_groups_conds::setpreferences_do $win] -grab 0 -transient 1 \
        -morebuttons [list [_ "Apply"] [_ "Defaults"]]
    set f [$w giveframe]

    set tree_prefsNode [gid_groups_conds::get_preference_node -local_global global \
        tree_preferences]

    set values [_preferences_give_defaults]
    if { !$to_defaults } {
        dict for "n v" $values {
            if { [$tree_prefsNode hasAttribute $n] } {
                dict set values $n [$tree_prefsNode @$n]
            }
        }
    }
    set xml "<formulae usableviewmode='1' tabstyle='notebook' updatebutton='0'/>"
    set docf [gid_groups_conds::parse_XML_data $xml]
    set root [$docf documentElement]

    # NOTE: this commands should be updated if using p::xml (TODO)

    dom createNodeCmd text t
    foreach node [list title description container param] {
        dom createNodeCmd element $node
    }

    $root appendFromScript {
        title { t [_ "Preferences"] }
        description {}
        container n "graphical" pns [_ "Graphical"] pn [_ "graphical"] {
            param n "background_color" pn [_ "Background color"] \
                field_type color \
                value [dict get $values background_color] \
                default_value white \
                disabled_value [list auto white] \
                help [_ "Background color of the tree. If deactivated, it is chosen automatically"] \
                {}
            param n "text_color" pn [_ "Text color"] \
                field_type color \
                value [dict get $values text_color] \
                default_value darkblue \
                disabled_value [list auto darkblue] \
                help [_ "Color of the text of the tree. If deactivated, it is chosen automatically"] \
                {}
        }
        container n "position" pns [_ "Position"] pn [_ "position"] {
            param n "background_color" pn [_ "Background color"] \
                field_type color \
                value [dict get $values background_color] \
                default_value white
        }
    }
    set f [$w giveframe]
    set key $w
    set f_in [formulae::create_windowD -state disabled \
        $f $key $root]
    grid $f_in -sticky nsew
    grid rowconfigure $f 0 -weight 1
    grid columnconfigure $f 0 -weight 1

    bind $w <Return> [list $w invokeok]
    set action [$w createwindow]
    $docf delete
}

proc gid_groups_conds::ConfigToolbars { {w .gid.toolbars } } {
    global ToolbarsPriv GIDDEFAULT

    InitWindow2 $w -title [_ "Toolbars position"] \
        -geometryvariable PrePostToolbarsPositionWindowGeom \
        -initcommand TOBegin -onlyposition -ontop
    if { ![winfo exists $w] } return ;# windows disabled || UseMoreWindows == 0

    ttk::frame $w.bars

    #pseudottk
    tk::listbox $w.bars.t -relief sunken \
        -yscroll "$w.bars.yscroll set" -width 35 -height 5 -exportselection 0
    $w.bars.t delete 0 end
    ttk::scrollbar $w.bars.yscroll \
        -orient vertical -command "$w.bars.t yview"

    GidHelp "$w.bars.t $w.bars.yscroll" [_ "Select here the toolbar that you want \
        to activate or deactivate"]

    grid $w.bars.t  -row 1 -column 2 -sticky nsew
    grid $w.bars.yscroll  -row 1 -column 3 -sticky ns
    grid rowconf $w.bars 1 -weight 1
    grid columnconf $w.bars 2 -weight 1

    ttk::frame $w.middle
    ttk::frame $w.middle.f1 -style groove.TFrame -borderwidth 2

    ttk::radiobutton $w.middle.f1.b1 -text [_ "Inside"] -variable ToolbarsPriv(inout) -value inside
    ttk::radiobutton $w.middle.f1.b2 -text [_ "Outside"] -variable ToolbarsPriv(inout) -value outside
    ttk::radiobutton $w.middle.f1.b3 -text [_ "None"] -variable ToolbarsPriv(inout) -value none

    GidHelp "$w.middle.f1.b1 $w.middle.f1.b2 $w.middle.f1.b3" \
        [_ "Selects the status of the toolbar. Options are:\n\
        inside: Toolbar will be inside the main program window\n\
        outside: Toolbar will become an independent window\n\
        none: Toolbar will be deactivated"]


    grid $w.middle.f1.b1 -sticky w
    grid $w.middle.f1.b2 -sticky w
    grid $w.middle.f1.b3 -sticky w

    ttk::frame $w.middle.f2 -style groove.TFrame -borderwidth 2
    set im [gid_themes::GetImage point.png medium_icons]
    ttk::radiobutton $w.middle.f2.l -style Toolbutton\
        -image $im -variable ToolbarsPriv(lrtb) -value l
    ttk::radiobutton $w.middle.f2.r -style Toolbutton\
        -image $im -variable ToolbarsPriv(lrtb) -value r
    ttk::radiobutton $w.middle.f2.b -style Toolbutton\
        -image $im -variable ToolbarsPriv(lrtb) -value b
    ttk::radiobutton $w.middle.f2.t -style Toolbutton\
        -image $im -variable ToolbarsPriv(lrtb) -value t

    GidHelp $w.middle.f2 [_ "If 'inside' is selected and the toolbar can adopt\
        different positions inside the main window, these buttons choose\
        if the window is up, left, right or bottom of the main window."]

    grid $w.middle.f2.t -row 1 -column 2
    grid $w.middle.f2.l -row 2 -column 1
    grid $w.middle.f2.r -row 2 -column 3
    grid $w.middle.f2.b -row 3 -column 2

    grid $w.middle.f1 -row 1 -column 1 -sticky wns
    grid $w.middle.f2 -row 1 -column 2 -sticky wns
    grid columnconf $w.middle 1 -weight 2
    grid columnconf $w.middle 2 -weight 1

    #ttk::frame $w.but -style groove.TFrame -borderwidth 2
    ttk::frame $w.but -style BottomFrame.TFrame
    ttk::button $w.but.close -style BottomFrame.TButton -text [_ "Close"] -command [list destroy $w] -underline 0
    grid $w.but.close -padx 5 -pady 5

    grid $w.bars -row 1 -column 1 -sticky nsew
    grid $w.middle -row 2 -column 1 -sticky ew
    grid $w.but -row 3 -column 1 -sticky sew
    grid anchor $w.but center

    grid rowconf $w 1 -weight 1
    grid columnconf $w 1 -weight 1

    TOInitializeToolbars

    foreach i $ToolbarsPriv(toolbars) {
        set name [lindex $i 0]
        if { $name != "MacrosToolbar" && $name != "Right buttons" } { continue }
        if { [info exists ToolbarsPriv(Hide,$name)] && $ToolbarsPriv(Hide,$name) == "1" } {
            continue
        }
        if { [GetCurrentPrePostMode] == "PRE" } {
            #preprocess
            if { ![string match -nocase Pre* [lindex $i 1]] } {
                continue
            }
        } else {
            #postprocess
            if { ![string match -nocase *Post* [string range [lindex $i 1] 0 6]] } {
                continue
            }
        }
        if { [llength $i] < 4 } {
            #old compatibility, previous versions do not exist field 3 title, used field 0 name
            set title [lindex $i 0]
        } else {
            set title [lindex $i 3]
        }
        $w.bars.t insert end $title
    }
    $w.bars.t selection set 0

    TOCheckCurrentBar $w
    bind $w.bars.t <ButtonRelease-1> "TOCheckCurrentBar $w"
    bind $w.bars.t <KeyRelease> "TOCheckCurrentBar $w"
    focus $w.bars.t
}

proc gid_groups_conds::_changed_size { tree } {

    if {![winfo exists $tree]} { return }
    set height [winfo height $tree]
    if { $height < 2 } { return }

    set tree_prefsNode [gid_groups_conds::get_preference_node -local_global global \
        tree_preferences]

    set colors ""
    set theme [gid_groups_conds::GetCurrentTheme]
    if { [regexp "Tdyn" $theme {} curr_theme] } {
        set c "#dbdcde"
        $tree configure -background $c -backgroundimage ""
        lappend colors $c
    } elseif { [$tree_prefsNode @background_color ""] ni [list "" auto] } {
        set c [$tree_prefsNode @background_color]
        $tree configure -background $c -backgroundimage ""
        lappend colors $c
    } elseif { [info command GiD_Set] eq "" } {
        # nothing
    } elseif { [GiD_Set BackColor(Type)] == 0 } {
        $tree configure -background [GiD_Set BackColor(Single)] -backgroundimage ""
        lappend colors [GiD_Set BackColor(Single)]
    } elseif { [GiD_Set BackColor(Sense)] eq "r" } {
        $tree configure -background [GiD_Set BackColor(Top)] -backgroundimage ""
        lappend colors [GiD_Set BackColor(Top)]
    } elseif { [GiD_Set BackColor(Sense)] eq "l" } {
        $tree configure -background [GiD_Set BackColor(Bottom)] -backgroundimage ""
        lappend colors [GiD_Set BackColor(Bottom)]
    } else {
        set img [$tree cget -backgroundimage]
        if { $img eq "" } {
            set img [image create photo]
            gid_groups_conds::add_to_internal_images_cache $img
        }
        lassign [list [GiD_Set BackColor(Top)] [GiD_Set BackColor(Bottom)]] t b
        set m "#"
        set min ""
        foreach i [winfo rgb . $t] j [winfo rgb . $b] {
            append m [format %.4x [expr {round(0.5*($i+$j))}]]
        }
        switch [GiD_Set BackColor(Sense)] {
            u { lassign [list $t $b] b t }
            ul { lassign [list $b $m] t b }
            dl { lassign [list $b $m] b t }
            ur { lassign [list $t $m] b t }
            dr { lassign [list $t $m] t b }
        }
        set h [expr {$height*2}]
        $img configure -width 2 -height $h
        $img put $t -to 0 0
        $img put $t -to 1 0
        $img put $b -to 0 [expr {$h-1}]
        $img put $b -to 1 [expr {$h-1}]
        cu::img::gradient -gradient_type double_vertical $img
        $tree configure -backgroundimage $img
        lappend colors $t $b
    }
    set minmod ""
    foreach c $colors {
        lassign [winfo rgb . $c] rc gc bc
        set mod [expr {sqrt($rc*$rc+$gc*$gc+$bc*$bc)}]
        if { $minmod eq "" || $mod < $minmod } {
            set minmod $mod
        }
    }

    if { [$tree_prefsNode @text_color ""] ni [list "" auto] } {
        set color [$tree_prefsNode @text_color]
        set lines_color $color
    } elseif { $minmod ne "" && $minmod < 20000 } {
        set color white
        set lines_color ""
    } else {
        set color darkblue
        set lines_color ""
    }
    set theme [gid_groups_conds::GetCurrentTheme]
    if { [regexp "Tdyn" $theme {} curr_theme] } {
        set color #000000
    }

    $tree element configure e_text -fill [list $::SystemHighlightText \
        {selected focus} $::SystemHighlightText {selected !focus disabled} \
        gray disabled $color {}]
    if { $lines_color eq "" } {
        $tree configure -linecolor [lindex [$tree configure -linecolor] 3]
    } else {
        $tree configure -linecolor $lines_color
    }
}

proc gid_groups_conds::unselected_buttons { win } {
    variable list_w_buttons
    if { [info exists list_w_buttons] } {
        foreach w $list_w_buttons {
            if { [winfo exists $w] } {
                $w configure -style ToolbuttonModGrey
            }
        }
    }
}

proc gid_groups_conds::gridremove { i } {
    if { [winfo exists $i] } {
        grid remove $i
    }
}

proc gid_groups_conds::update_data_tree { } {
    variable gid_group_var
    variable doc
    variable list_w_buttons

    set win .gid.central.boundaryconds.gg
    set theme [gid_groups_conds::GetCurrentTheme]
    if { ![regexp "Tdyn" $theme {} curr_theme] } {
        return
    }
    if {![winfo exists $win]} {
        return
    }
    set idx 0
    set bw $win.f0.but$idx
    if {![winfo exists $bw]} {
        return
    }
    #     set res [gid_groups_conds::cancel_edit]
    #     if { $res == 0 } { return }

    unselected_buttons $win
    gid_groups_conds::gridremove $win.lt
    gid_groups_conds::gridremove $win.gt
    gid_groups_conds::gridremove $win.grbbox
    gid_groups_conds::gridremove $win.laybbox

    gid_groups_conds::remove_all_search

    $bw configure -style ToolbuttonModBlue
    grid $win.ft -sticky news
    grid $win.ft.t $win.ft.sv -sticky nsew -padx 1 -pady 1
    grid $win.ft.sh -sticky ew -row 3 -column 0
    grid configure $win.ft.t -pady "2 0"
    grid configure $win.ft.sv -row 2 -column 1
}

proc gid_groups_conds::remove_all_search { } {
    set wgid .gid.bitmapsStdBarMod
    set gr_lay "01"
    set lay_search $wgid.search$gr_lay
    set gr_lay "10"
    set gr_search $wgid.search$gr_lay
    set dt_search $wgid.search

    gid_groups_conds::gridremove $lay_search
    gid_groups_conds::gridremove $gr_search
    gid_groups_conds::gridremove $dt_search
}

proc gid_groups_conds::update_layers_tree { bw win } {
    variable gid_group_var
    variable doc
    variable list_w_buttons

    unselected_buttons $win
    gid_groups_conds::gridremove $win.gt
    gid_groups_conds::gridremove $win.ft
    gid_groups_conds::gridremove $win.grbbox

    if { ![winfo exists $win.lt] } {
        set b [gid_groups_theme $win.lt -domdoc $doc -layers 1]
        grid $win.lt -sticky news
    } else {
        grid $win.laybbox -sticky news
        grid $win.lt -sticky news

        set self $win.lt
        $self actualize_layers
    }
    $bw configure -style ToolbuttonModBlue
}

proc gid_groups_conds::open_all_groups_layers { w } {
    variable doc
    variable open_all_var

    set root [$doc documentElement]
    foreach igr [$root selectNodes {groups/group}] {
        set group [$igr @n]
        gid_groups_conds::setAttributes [format_xpath {groups/group[@n=%s]} $group] [list onoff 1]
    }

    GidUtils::DisableGraphics

    array set entities_visible ""
    foreach i [list point line surface volume] {
        set entities_visible($i,$i) ""
    }
    set view_mode [GiD_Info Project ViewMode]
    if { $view_mode eq "MESHUSE" } {
        set geom_mesh mesh
    } else {
        set geom_mesh geometry
    }
    foreach igr [$root selectNodes {groups/group}] {
        set group [$igr @n]
        foreach i [list point line surface volume] {
            if { $view_mode eq "MESHUSE" } {
                switch $i {
                    point { set etype node }
                    default { set etype element }
                }
            } else {
                set etype $i
            }
            if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
                set ent [GiD_EntitiesGroups get $group ${i}s]
                if { $ent ne "" } {
                    lappend entities_visible($i,$etype) $ent
                }
            } else {
                set ent ""
                foreach j [GiD_Info conditions ${i}_groups $geom_mesh] {
                    lassign $j num face - gr
                    if { $num in "E N" } {
                        set num $face
                    } else {
                        set num ""
                    }
                    if { $gr eq $group && $num ne "" } {
                        lappend ent $num
                    }
                }
                if { $ent ne "" } {
                    lappend entities_visible($i,$etype) $ent
                }
            }
        }
    }
    foreach ient [array names entities_visible] {
        if {[regexp {(.*),(.*)} $ient {} i etype] } {
            if { $entities_visible($i,$etype) ne "" } {
                set entities_visible($i,$etype) [join $entities_visible($i,$etype) " "]
                set entities_visible($i,$etype) [lsort -integer -unique $entities_visible($i,$etype)]
                GiD_EntitiesLayers assign_front_layer $geom_mesh ${etype}s $entities_visible($i,$etype)
            }
        }
    }

    GidUtils::DisableWarnLine
    foreach layer [GiD_Info Layers] {
        GiD_Process 'Layers On $layer escape
    }
    GidUtils::EnableWarnLine
    GidUtils::EnableGraphics
    GidUtils::UpdateWindow LAYER
    GidUtils::UpdateWindow GROUPS

    gid_groups_conds::gridremove $w
    set open_all_var 1

    GiD_Redraw
    GiD_Process 'Redraw
}

proc gid_groups_conds::open_groups_layers_window { win w } {
    variable open_all
    variable doc
    variable close_button_all_gr_lay_window

    if {[info exists doc]} {
        set root [$doc documentElement]
        set styleNode [[$doc documentElement] selectNodes style]
        if { $styleNode != ""} {
            set open_closed_layers_groups [get_domnode_attribute \
                $styleNode open_closed_layers_groups 0]
        } else {
            return
        }
    } else {
        return
    }
    if { !$open_closed_layers_groups } { return }
    set is_closed [gid_groups_conds::num_closed_layers_groups]
    if { !$is_closed } { return }
    if { $close_button_all_gr_lay_window } { return }

    upvar 0 ::ttk::theme::[gid_themes::GetTtkTheme]::colors colors
    frame $w -height 150 -width 200 -bd 1 -background $colors(-text_bg)
    set img "close17"
    set img [gid_themes::GetImage $img.png]

    ttk::button $w.close -image $img -command \
        [list gid_groups_conds::gridremove_open_all_window $win 1] -width 20

    frame $w.separator -background $colors(-toolbars_bg) -borderwidth 2 -width 2 -height 1

    ttk::label $w.lab -text \
        [_ "There are many closed layers and groups.\nDo you want to open them?"] \
        -compound left
    ttk::button $w.but -text [_ "Open all groups and layers"] \
        -command [list gid_groups_conds::open_all_groups_layers $w] \
        -width 24

    grid columnconfigure $w 0 -weight 1
    grid $w -sticky nsew -padx 2 -pady 2
    grid $w.separator -sticky ew -row 0 -padx 2 -pady {5 5} -columnspan 2
    grid $w.lab -row 1 -column 0 -sticky w -padx 2 -pady 2
    grid $w.close -row 1 -column 1 -sticky nes -padx 5 -pady "0 5"
    grid $w.but -sticky news -row 2 -column 0 -padx 3 -pady 3

    grid columnconfigure $w $w.lab -weight 1
}

proc gid_groups_conds::gridremove_open_all_window { win {close_but 0} } {
    variable close_button_all_gr_lay_window

    foreach i [list lt.ft.open gt.ft.open ft.open] {
        gid_groups_conds::gridremove $win.$i
    }
    if { $close_but } {
        set close_button_all_gr_lay_window 1
    }
}

proc gid_groups_conds::num_closed_layers_groups { } {
    variable doc

    set root [$doc documentElement]

    set is_closed 0

    set xp {groups/group}
    set num_groups [llength [$root selectNodes $xp]]
    set num_off_groups 0
    foreach igr [$root selectNodes {groups/group}] {
        set onoff [$igr @onoff]
        if { !$onoff } { incr num_off_groups 1 }
    }

    if { $num_off_groups > [expr {abs($num_groups/2)}] } {
        set is_closed 1
    }

    set num_layers [llength [GiD_Info Layers]]

    set num_back_layers 0
    if { [GiD_Info back_layers] ne "" } {
        set num_back_layers [llength [GiD_Info back_layers]]
    }
    if { $num_back_layers > [expr {abs($num_layers/2)}] } {
        set is_closed 1
    }

    set num_off_layers 0
    if { [GiD_Info layers -off] ne "" } {
        set num_off_layers [llength [GiD_Info layers -off]]
    }

    if { $num_off_layers > [expr {abs($num_layers/2)}] } {
        set is_closed 1
    }

    return $is_closed
}

proc gid_groups_conds::data_tree_view { style bw win {press ""}} {
    variable open_all_var
    variable close_button_all_gr_lay_window
    variable menu

    set menu "data_tree"

    unselected_buttons $win
    gid_groups_conds::gridremove $win.lt
    gid_groups_conds::gridremove $win.gt
    gid_groups_conds::gridremove $win.grbbox
    gid_groups_conds::gridremove $win.laybbox

    set wgid .gid.bitmapsStdBarMod
    set gr_lay "01"
    set lay_search $wgid.search$gr_lay
    if {[winfo exists $lay_search] } {
        grid remove $lay_search
    }
    set gr_lay "10"
    set gr_search $wgid.search$gr_lay
    if { [winfo exists $gr_search] } {
        grid remove $gr_search
    }
    set dt_search $wgid.search
    if { [winfo exists $dt_search]} {
        grid remove $dt_search
    }

    grid $win.ft.t -sticky nsew -padx 1 -pady 1

    grid configure $win.ft.t -pady "2 0"

    $bw configure -style ToolbuttonModBlue
    grid $win.ft -sticky news

    if {[winfo exists $win.data] } {
        grid $win.data
    }

    set winbc .gid.central.boundaryconds.gg

    if { ![gid_groups_conds::isdrawpost] } {
        grid $dt_search
    }

    gid_groups_conds::gridremove_open_all_window $win

    if { !$open_all_var } {
        if { ![winfo exists $win.ft.open] } {
            gid_groups_conds::open_groups_layers_window $win $win.ft.open
        } elseif { !$close_button_all_gr_lay_window } {
            grid $win.ft.open -sticky nsew -padx 4 -pady 4
        }
    }
}

proc gid_groups_conds::cancel_edit {} {
    set w .gid.central.boundaryconds.gg
    if { ![winfo exists $w.data] } { return }
    return 1
}

proc gid_groups_conds::update_layers_tree { bw win } {
    variable gid_group_var
    variable doc
    variable list_w_buttons

    #     set res [gid_groups_conds::cancel_edit]
    #     if { $res == 0 } { return }

    unselected_buttons $win
    gid_groups_conds::gridremove $win.gt
    gid_groups_conds::gridremove $win.ft
    gid_groups_conds::gridremove $win.grbbox

    if { ![winfo exists $win.lt] } {
        set b [gid_groups_theme $win.lt -domdoc $doc -layers 1]
        grid $win.lt -sticky news
    } else {
        grid $win.laybbox -sticky news
        grid $win.lt -sticky news
        set self $win.lt
        $self actualize_layers
    }
    $bw configure -style ToolbuttonModBlue

    #     set_pref $win
}

proc gid_groups_conds::update_data_tree_search {} {
    set wgid .gid.bitmapsStdBarMod
    set winbc .gid.central.boundaryconds.gg

    if { [winfo exists $winbc.ft] && ![gid_groups_conds::isdrawpost] } {
        $winbc.ft update_stdbarmod $wgid
    }
}

proc gid_groups_conds::isdrawpost { } {
    set ispost 0
    if { [info command ::draw_post::is_init_postprocess] ne "" } {
        if { [draw_post::is_init_postprocess] } {
            set ispost 1
        }
    }
    return $ispost
}

proc gid_groups_conds::layers_tree_view { style bw win {press ""}} {
    variable gid_group_var
    variable doc
    variable open_all_var
    variable close_button_all_gr_lay_window
    variable menu

    #     set res [gid_groups_conds::cancel_edit]
    #     if { $res == 0 } { return }

    set menu "layers_tree"

    unselected_buttons $win
    gid_groups_conds::gridremove $win.data
    gid_groups_conds::gridremove $win.gt
    gid_groups_conds::gridremove $win.ft
    gid_groups_conds::gridremove $win.grbbox
    set style ToolbuttonModBlue

    if {![winfo exists $win.lt]} {
        $bw configure -style $style
        set b [gid_groups_theme $win.lt -domdoc $doc -layers 1]
        grid $win.lt -sticky news
    } else {
        $bw configure -style $style
        grid $win.laybbox -sticky news
        grid $win.lt -sticky news
        $win.lt actualize_layers

        set wgid .gid.bitmapsStdBarMod
        set gr_lay 01
        set winbc .gid.central.boundaryconds.gg
        if { [winfo exists $winbc.lt] && ![gid_groups_conds::isdrawpost] } {
            $winbc.lt update_stdbarmod_from_gr_or_lay $wgid $gr_lay
        }
    }

    set wgid .gid.bitmapsStdBarMod
    gid_groups_conds::gridremove $wgid.search
    set gr_lay 10
    set gr_search $wgid.search$gr_lay
    gid_groups_conds::gridremove $gr_search

    gid_groups_conds::gridremove_open_all_window $win

    if { !$open_all_var } {
        if { ![winfo exists $win.lt.ft.open] } {
            gid_groups_conds::open_groups_layers_window $win $win.lt.ft.open
        } elseif { !$close_button_all_gr_lay_window } {
            grid $win.lt.ft.open -sticky nsew -padx 4 -pady 4
        }
    }
}

proc gid_groups_conds::set_pref { win } {
    foreach itree [list $win.ft.t $win.ft.tl $win.ft.tg] {
        gid_groups_conds::_changed_size $itree
    }
}

proc gid_groups_conds::groups_tree_view { style bw win {press ""}} {
    variable gid_group_var
    variable doc
    variable open_all_var
    variable close_button_all_gr_lay_window
    variable menu
    #     set res [gid_groups_conds::cancel_edit]
    #     if { $res == 0 } { return }

    set menu "groups_tree"
    unselected_buttons $win
    gid_groups_conds::gridremove $win.data
    gid_groups_conds::gridremove $win.lt
    gid_groups_conds::gridremove $win.ft
    gid_groups_conds::gridremove $win.laybbox
    set style ToolbuttonModBlue
    if {![winfo exists $win.gt] || $open_all_var } {
        $bw configure -style $style
        if { ![winfo exists $win.gt] } {
            set b [gid_groups_theme $win.gt -domdoc $doc -groups 1]
            grid $win.gt -sticky news
        } elseif { $open_all_var } {
            grid $win.gt -sticky news
        }
    } else {
        $bw configure -style $style
        set wgid .gid.bitmapsStdBarMod
        set gr_lay 10
        set winbc .gid.central.boundaryconds.gg
        if { [winfo exists $winbc.gt] && ![gid_groups_conds::isdrawpost] } {
            $winbc.gt update_stdbarmod_from_gr_or_lay $wgid $gr_lay
        }
        set self $win.gt
        grid $win.grbbox -sticky news
        grid $win.gt -sticky news
    }

    set wgid .gid.bitmapsStdBarMod
    gid_groups_conds::gridremove $wgid.search
    set gr_lay 01
    set gr_search $wgid.search$gr_lay
    gid_groups_conds::gridremove $gr_search

    ##
    gid_groups_conds::gridremove_open_all_window $win
    if { !$open_all_var } {
        if { ![winfo exists $win.gt.ft.open] } {
            gid_groups_conds::open_groups_layers_window $win $win.gt.ft.open
        } elseif { !$close_button_all_gr_lay_window } {
            grid $win.gt.ft.open -sticky nsew -padx 4 -pady 4
        }
    }
}

proc gid_groups_conds::begin_problemtype { _spd_file _defaults_file {_spd_module_file ""} {_open_conditions 1} } {
    variable doc
    variable imagesdirList
    variable spd_file
    variable defaults_file
    variable style

    set spd_file $_spd_file
    set defaults_file $_defaults_file
    set spd_module_file $_spd_module_file

    gid_groups_conds::init_package

    set imagesdirList [list [file join [file dirname $spd_file] images]]
    open_spd_file $spd_file LOADING_PROBLEMTYPE
    set basedir [file dirname $spd_file]
    gid_groups_conds::include_from_XML_files $basedir
    merge_spd_file_if_exists $defaults_file
    merge_spd_module_file_if_exists $spd_module_file
    import_export_materials_get_global
    units_get_global
    # the after is necessary for when gid is open with option -p
    if {$_open_conditions} {after 100 [list gid_groups_conds::open_conditions check_default]}
}

proc gid_groups_conds::include_from_XML_files { basedir } {
    variable doc

    if { ![info exists doc] } { return }

    foreach includeNode [$doc selectNodes {.//include}] {
        set active [$includeNode @active 1]
        if { !$active } { continue }
        set parent [$includeNode parent]
        set path [$includeNode @path ""]
        set file [file join $basedir $path]
        set doc_to_include [gid_groups_conds::open_XML_file_gzip $file]
        foreach newnode [$doc_to_include selectNodes {container|blockdata|condition|value|proc}] {
            if { $includeNode eq "" } {
                $parent appendChild $newnode
            } else {
                $parent insertBefore $newnode $includeNode
            }
        }
        $parent removeChild $includeNode
        $doc_to_include delete
    }
    set icount 0
    foreach includeNode [$doc selectNodes {.//include}] {
        set active [$includeNode @active 1]
        if { $active } {
            incr icount
            continue
        }
    }
    if { $icount > 0 } {
        gid_groups_conds::include_from_XML_files $basedir
    }
}

proc gid_groups_conds::ProcessIncludes { dir } {
    set root [$::gid_groups_conds::doc documentElement]
    foreach elem [$root getElementsByTagName "include"] {
        set active 1
        if {[$elem hasAttribute "active"]} {
            set active [$elem getAttribute "active"]
        }
        if {$active} {
            set path [$elem getAttribute "path"]
            set filename [file join $dir $path]
            set processednode [gid_groups_conds::ProcessIncludesRecurse $filename $dir]
            set elemparent [$elem parent]
            $elemparent replaceChild $processednode $elem
            $elem delete
        }
    }
}

proc gid_groups_conds::ProcessIncludesRecurse { filename basedir } {
    set xml [tDOM::xmlReadFile $filename]
    set newnode [dom parse [string trim $xml]]
    set xmlNode [$newnode documentElement]
    foreach elem [$xmlNode getElementsByTagName "include"] {
        set active 1
        if {[$elem hasAttribute "active"]} {
            set active [$elem getAttribute "active"]
        }
        if {$active} {
            set path [$elem getAttribute "path"]
            set f [file join $basedir $path]
            set processednode [gid_groups_conds::ProcessIncludesRecurse $f $basedir]
            set elemparent [$elem parent]
            foreach att [$elem attributes] {
                if {$att ni [list "n" "active" "path"]} {
                    $processednode setAttribute $att [$elem getAttribute $att]
                }
            }
            $elemparent replaceChild $processednode $elem
            $elem delete
        }
    }
    return $xmlNode
}

proc gid_groups_conds::add_images_dir { dir } {
    variable imagesdirList

    lappend imagesdirList $dir
}

proc gid_groups_conds::is_geometry_or_mesh {} {
    variable is_draw_post

    if { $is_draw_post } {
        if { [draw_post::is_geometry] } {
            return geometry
        } else {
            return mesh
        }
    } else {
        switch [GiD_Info Project ViewMode] {
            GEOMETRYUSE { return geometry }
            MESHUSE { return mesh }
        }
    }
}

proc gid_groups_conds::set_modified_flag {} {

    if { [info command ::GiD_ModifiedFileFlag] ne "" } {
        GiD_ModifiedFileFlag set 1
    } else {
        if { [info command ::draw_post::set_modified_flags] ne "" } {
            draw_post::set_modified_flags
        }
    }
}

proc gid_groups_conds::clear_modified_flag {} {

    if { [info command ::GiD_ModifiedFileFlag] ne "" } {
        GiD_ModifiedFileFlag set 0
    } else {
        draw_post::clear_modified_flags
    }
}

proc gid_groups_conds::start_group_undo_batch { label } {
    variable is_draw_post

    if { $is_draw_post } {
        draw_post::start_group_undo_batch $label
    }
}

proc gid_groups_conds::end_group_undo_batch {} {
    variable is_draw_post

    if { $is_draw_post } {
        draw_post::end_group_undo_batch
    }
}

proc gid_groups_conds::add_to_undo { label redo_cmd undo_cmd } {
    variable is_draw_post

    if { !$is_draw_post } {
        GiD_ModifiedFileFlag set 1
        GiD_EnterInBatch "*****MARK START_COMMAND_TCL\n"
        GiD_EnterInBatch "\{-tcl- $redo_cmd\}\n"
        GiD_EnterInBatch "*****MARK END_COMMAND_TCL\n"
    } else {
        draw_post::add_to_undo_batch $label 1 $redo_cmd $undo_cmd
    }
}

proc gid_groups_conds::give_doc {} {
    variable doc
    return [set! doc]
}

proc gid_groups_conds::set_doc { _doc } {
    variable doc

    set doc $_doc
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        #invoke it directly now to process the current groups, because before load this file the GiD_AfterCreateGroup event procedure
        #was not defined to syncronize both definitions
        set mff [GiD_ModifiedFileFlag get]
        gid_groups_conds::syncronize_gid_groups_and_doc_groups
        if { !$mff } {
            GiD_ModifiedFileFlag set 0
        }
    }
}

proc gid_groups_conds::give_root_node {} {
    variable doc

    if { ![info exists doc] } {
        return ""
    }
    return [$doc documentElement]
}

proc gid_groups_conds::give_node_xpath { xpath } {
    variable doc

    return [$doc selectNodes $xpath]
}


proc gid_groups_conds::give_data_version {} {
    variable doc

    return [[$doc documentElement] @version]
}

proc gid_groups_conds::has_conditions_cmd {} {
    variable open_boundary_cond

    if { ![info exists open_boundary_cond] || ![winfo exists $open_boundary_cond] } {
        return 0
    } else {
        return 1
    }
}

proc gid_groups_conds::conditions_cmd { args } {
    variable open_boundary_cond

    if { ![info exists open_boundary_cond] || ![winfo exists $open_boundary_cond] } {
        GID_tk_messageBox -message [_ "It is necessary to open the data window to proceed"] {*}[ GidUtils::GiDDefaultParentIfPresent]
        return
    }
    uplevel #0 [list $open_boundary_cond] $args
}

proc gid_groups_conds::conditions_cmd_if_exists { args } {
    variable open_boundary_cond

    if { ![info exists open_boundary_cond] || ![winfo exists $open_boundary_cond] } {
        return
    }
    uplevel #0 [list $open_boundary_cond] $args
}

proc gid_groups_conds::groups_cmd_if_exists { args } {
    variable gid_group_var

    if { ![info exists gid_group_var] || ![winfo exists $gid_group_var] } {
        return
    }
    uplevel #0 [list $gid_group_var] $args
}

proc gid_groups_conds::has_post_cmd {} {
    variable open_post

    if { ![info exists open_post] || ![winfo exists $open_post] } {
        return 0
    } else {
        return 1
    }
}

proc gid_groups_conds::post_cmd { args } {
    variable open_post

    if { ![info exists open_post] || ![winfo exists $open_post] } {
        GID_tk_messageBox -message [_ "It is necessary to open the post window to proceed"] {*}[ GidUtils::GiDDefaultParentIfPresent]
        return
    }
    uplevel #0 [list $open_post] $args
}

proc gid_groups_conds::post_cmd_if_exists { args } {
    variable open_post

    if { ![info exists open_post] || ![winfo exists $open_post] } {
        return
    }
    uplevel #0 [list $open_post] $args
}

proc gid_groups_conds::give_spd_file {} {
    variable spd_file

    return [set! spd_file]
}

proc gid_groups_conds::give_defaults_file {} {
    variable defaults_file

    return $defaults_file
}

proc gid_groups_conds::internal_data { what args } {
    variable doc

    set optional {
        { -mydoc mydoc "" }
        { -set_modified_flag boolean 1 }
    }
    set compulsory ""
    set args [parse_args -raise_compulsory_error 0 $optional $compulsory $args]

    if { $mydoc eq "" } {
        if { ![info exists doc] } { return "" }
        set mydoc $doc
    }

    switch $what {
        get {
            set name [lindex $args 0]
            set xp [format_xpath {/*/blockdata[@n="Internal data"]/value[@n=%s]/@v} $name]
            return [$mydoc selectNodes string($xp)]
        }
        set {
            lassign [lrange $args 0 1] name value
            set domNodeP [$mydoc selectNodes {/*/blockdata[@n="Internal data"]}]
            if { $domNodeP eq "" } {
                set root [$mydoc selectNodes {/*}]
                set domNodeP [$root appendChildTag blockdata [list attributes() n "Internal data"]]
            }
            set xp [format_xpath {value[@n=%s]} $name]
            set domNode [$domNodeP selectNodes $xp]
            if { $domNode ne "" } {
                $domNode setAttribute v $value
            } else {
                $domNodeP appendChildTag value [list attributes() n $name v $value]
            }
            if { $set_modified_flag } {
                set_modified_flag
            }
        }
        get_list {
            set name [lindex $args 0]
            set domNodeP [$mydoc selectNodes {/*/blockdata[@n="Internal data"]}]
            if { $domNodeP eq "" } {
                set root [$mydoc selectNodes {/*}]
                set domNodeP [$root appendChildTag blockdata [list attributes() n "Internal data"]]
            }
            set xp [format_xpath {container[@n=%s]} $name]
            set domNode [$domNodeP selectNodes $xp]
            if { $domNode eq "" } {
                set domNode [$domNodeP appendChildTag container [list attributes() n $name]]
            }
            return $domNode
        }
        delete {
            set name [lindex $args 0]
            set xp [format_xpath {/*/blockdata[@n="Internal data"]/value[@n=%s]} $name]
            set domNode [$mydoc selectNodes $xp]
            if { $domNode ne "" } {
                if { $set_modified_flag } {
                    set_modified_flag
                }
                $domNode delete
            }
        }
        lognoter_file {
            gid_groups_conds::internal_data get $args
        }
        default {
            error "error in gid_groups_conds::internal_data"
        }
    }
}

proc gid_groups_conds::end_problemtype { filename_user_defaults } {
    variable doc
    variable gid_group_var
    variable open_boundary_cond
    variable boundary_cond_handlers
    variable post_handlers
    variable open_post
    variable images_cache
    variable is_draw_post

    if { !$is_draw_post } {
        gid_groups_conds::end_drawall
        if { [info exists draw_post::data] } {
            draw_post::end_postprocess end
        }
    }
    if { [info exists open_boundary_cond] && [winfo exists $open_boundary_cond] } {
        $open_boundary_cond store_tree_state display_options
        open_conditions end
    }
    if { [info exists open_post] && [winfo exists $open_post] } {
        $open_post store_tree_state display_options
        open_post close
    }
    if { ![info exists ::GID_GROUPS] } {
        if { [info exists gid_group_var] && [winfo exists $gid_group_var] } {
            $gid_group_var store_tree_state display_options
            destroy [winfo toplevel $gid_group_var]
            unset gid_group_var
        }
    }
    close_report_help_window all
    report_clear_default_copy_file

    if { $filename_user_defaults ne "" } {
        if { [GiD_Set SaveGidDefaults] } {
            gid_groups_conds::save_preferences $filename_user_defaults
        }
    }

    if { [info exists doc] } {
        if { [info exists ::GID_GROUPS] } {
            #not delete doc, preserve groups information
            set nodes_to_preserve [list groups display_options]
            set root [$doc documentElement]
            foreach node [$root childNodes] {
                if { [lsearch $nodes_to_preserve [$node nodeName]] == -1 } {
                    $node delete
                }
            }
            if { ![llength [$root childNodes]] } {
                $doc delete
                unset doc
            }

        } else {
            $doc delete
            unset doc
        }
    }

    dict for "n img" $images_cache {
        if { [info exists ::GidPrivImages(tree,$n)] } {
            unset ::GidPrivImages(tree,$n)
        }
        image delete $img
    }
    set images_cache ""
    set boundary_cond_handlers ""
    set post_handlers ""
}

proc gid_groups_conds::close_all_windows {} {
    variable gid_group_var
    variable open_boundary_cond
    variable open_post

    gid_groups_conds::end_drawall

    if { [info exists open_boundary_cond] && [winfo exists $open_boundary_cond] } {
        $open_boundary_cond store_tree_state display_options
        open_conditions close
    }
    if { [info exists open_post] && [winfo exists $open_post] } {
        $open_post store_tree_state display_options
        open_post close
    }
    if { [info exists gid_group_var] && [winfo exists $gid_group_var] } {
        $gid_group_var store_tree_state display_options
        destroy [winfo toplevel $gid_group_var]
    }
}

proc gid_groups_conds::calculate { args } {
    variable calculate_after_id

    set optional {
        { -show_output boolean 0 }
    }
    set compulsory "what"
    parse_args $optional $compulsory $args

    switch -- $what {
        start {
            if { [info exists draw_post::data] } {
                draw_post::end_postprocess end
            }
            set files [PWFIndOutputFilenameCurrent]
            if { [llength $files] } {
                # catch is here because on Windows, if info file is still open it fails
                catch { file delete {*}$files }
            }
            GiD_Process MEscape Utilities Calculate
            if { $show_output } {
                set calculate_after_id [after 500 [list gid_groups_conds::calculate show_output]]
            }
        }
        show_output {
            set exists_file 0
            foreach file [PWFIndOutputFilenameCurrent] {
                if { [file exists $file] } {
                    set exists_file 1
                    break
                }
            }
            if { $exists_file } {
                PWViewOutput
            } else {
                set calculate_after_id [after 500 [list gid_groups_conds::calculate show_output]]
            }
        }
        end {
            if { [info exists calculate_after_id] } {
                after cancel $calculate_after_id
                unset calculate_after_id
            }
        }
    }
}

#################################################################################
#    small utilites
#################################################################################

proc gid_groups_conds::GetCurrentLanguage { args } {
    variable is_draw_post

    if {![info exists is_draw_post]} {
        return [GiD_Set Language]
    } elseif { !$is_draw_post } {
        return [GiD_Set Language]
    } else {
        return [draw_full::get_preference language en]
    }
}

proc gid_groups_conds::give_toplevel {} {
    variable is_draw_post

    if { !$is_draw_post } {
        return ".gid"
    } else {
        return [draw_post::give_toplevel]
    }
}

proc gid_groups_conds::give_model_name {} {
    variable is_draw_post

    if { !$is_draw_post } {
        return [GiD_Info Project ModelName]
    } else {
        return [draw_post::give_model_name]
    }
}

#################################################################################
#    param creator & compassfem & problemtypes
#################################################################################

proc gid_groups_conds::has_default_compassfem {} {
    if { [file exists [file join $draw_full::topdir problemtypes compassfem.gid]] } {
        return 1
    } else {
        return 0
    }
}

proc gid_groups_conds::give_problemtype {} {
    variable problemtype

    return $problemtype
}

proc gid_groups_conds::give_problemtypedir {} {
    variable problemtypedir

    return $problemtypedir
}

proc gid_groups_conds::re_start_problemtype {} {
    variable problemtype0

    if { [set! problemtype0] eq "" } { return }
    start_problemtype $problemtype0
}

proc gid_groups_conds::start_problemtype_check { p } {
    if { $p in "compassfem Code_bright testctb" } {
        return 1
    } else {
        return 0
    }
}

proc gid_groups_conds::start_problemtype { p } {
    variable doc
    variable problemtype0
    variable problemtype
    variable problemtypedir

    if { [set! problemtype0] ne "" } {
        if { [info command ::end_problem_GP] ne "" } {
            ::end_problem_GP
        }
    }

    set problemtypedir ""
    if { [file pathtype $p] ne "relative" } {
        set problemtypedir $p
        set p [file root [file tail $p]]
    }

    set problemtype0 $p

    switch $p {
        compassfem {
            if { $problemtypedir ne "" } {
                set p0 compassfem
                set problemtype compassfem
            } elseif { [info exists draw_full::topdir] && [file exists [file join $draw_full::topdir problemtypes compassfem.gid]] } {
                set problemtypedir [file join $draw_full::topdir problemtypes compassfem.gid]
                set p0 compassfem
                set problemtype compassfem
            } elseif { $::tcl_platform(user) == "danielsa_" } {
                set problemtypedir "C:/Users/danielsa/Documents/Compass_code/compassfem/compassfem.gid"
                set p0 compassfem
                set problemtype compassfem
            } elseif { $::tcl_platform(user) == "peyrau" } {
                set problemtypedir "C:/Users/peyrau/Documents/problemtypes/compassfem/compassfem.gid"
                set p0 compassfem
                set problemtype CompassFEM/compassfem
            } elseif { $::tcl_platform(user) == "Ovidi" } {
                if { $::tcl_platform(platform) == "windows" } {
                    set problemtypedir $::env(COMPASS_SRC)/compassfem/compassfem.gid
                } elseif { $::tcl_platform(platform) == "unix" } {
                    set problemtypedir $::env(COMPASS_SRC)/compassfem/compassfem.gid
                }
                set p0 compassfem
                set problemtype compassfem
            } elseif { $::tcl_platform(user) == "Jaume" } {
                set problemtypedir $::env(COMPASS_SRC)/compassfem/compassfem.gid
                set p0 compassfem
                set problemtype CompassFEM/compassfem
            } elseif { [file isdirectory $::env(HOME)/code/compassfem/compassfem.gid] } {
                #                 global CompassFEMPriv

                #set problemtypedir $CompassFEMPriv(problemtypedir)
                set problemtypedir $::env(HOME)/code/compassfem/compassfem.gid
                set p0 compassfem
                set problemtype compassfem
            } else {
                global CompassFEMPriv
                set problemtypedir $CompassFEMPriv(problemtypedir)
                set p0 compassfem
                set problemtype compassfem
            }
        }
        Code_bright {
            set problemtypedir $::env(HOME)/code/codebright/Code_bright.gid
            set p0 Code_bright
            set problemtype Code_bright
        }
        testctb {
            if { $problemtypedir ne "" } {
                # nothing
            } else {
                set problemtypedir $::env(HOME)/mytcltk/gid_groups_conds_dir/utils/problemtypes/test_CTB.gid
            }
            set p0 test_CTB
            set problemtype test_CTB
        }
        default {
            error [_ "Error: unknown problemtype '%s'" $p]
        }
    }
    uplevel #0 [list source [file join $problemtypedir $p0.tcl]]
    init_problem_GP $problemtypedir

    set def [file join [cu::file::appdatadir $p0] defaults.ini]

    #accept also old name with _default
    set file_spd [file join $problemtypedir ${p0}_default.spd]
    if { ![file exists $file_spd] } {
        set file_spd [file join $problemtypedir $p0.spd]
    }
    gid_groups_conds::begin_problemtype $file_spd $def ""

    set root [give_root_node]
    set domNode [$root selectNodes {container[@n='dir_local_axes']}]
    if { $domNode ne "" } {
        $domNode setAttribute state hidden
    }
    set domNode [$root selectNodes {container[@n='mesh_sizes']}]
    if { $domNode ne "" } {
        $domNode setAttribute state normal
    }
    set domNode [$root selectNodes {condition[@n='local_axes']}]
    if { $domNode ne "" } {
        $domNode setAttribute state normal
    }
}

proc gid_groups_conds::has_default_param_creator {} {

    set dir [file join $draw_full::topdir param_creator]
    if { [cu::islink $dir.lnk] } {
        set dir [cu::readlink $dir.lnk]
    }

    if { [file isdirectory $dir] } {
        return 1
    } else {
        return 0
    }
}

proc gid_groups_conds::set_param_creator_name { name } {
    variable param_creator_name

    set param_creator_name [file normalize $name]
}

proc gid_groups_conds::start_param_creator {} {
    variable param_creator_name

    draw_post::set_geometry_or_mesh -force 1 geometry

    if { [give_problemtype ] eq "" } {
        start_problemtype compassfem
        draw_post::reopen_gui
    }

    set dir_default [file join $draw_full::topdir param_creator]
    if { [cu::islink $dir_default.lnk] } {
        set dir_default [cu::readlink $dir_default.lnk]
    }

    if { [info exists param_creator_name] } {
        set dir $param_creator_name
    } elseif { [file isdirectory $dir_default] } {
        set dir $dir_default
    } elseif { $::tcl_platform(user) eq "danitto" } {
        set dir "C:/Documents and Settings/danitto/Mis documentos/danito/problemtypes/superbam"
    } else {
        set dir $::env(HOME)/code/superbam
    }
    set wp [draw_post::give_toplevel]
    param_creator::create_window $wp $dir
}

proc gid_groups_conds::start_param_creator_frame { wp } {
    variable param_creator_name

    if { [give_problemtype ] eq "" } {
        start_problemtype compassfem
    }
    set dir_default [file join $draw_full::topdir param_creator]
    if { [cu::islink $dir_default.lnk] } {
        set dir_default [cu::readlink $dir_default.lnk]
    }

    if { [info exists param_creator_name] } {
        set dir $param_creator_name
    } elseif { [file isdirectory $dir_default] } {
        set dir $dir_default
    } elseif { $::tcl_platform(user) eq "danitto" } {
        set dir "C:/Documents and Settings/danitto/Mis documentos/danito/problemtypes/superbam"
    } else {
        set dir $::env(HOME)/code/superbam
    }
    set p [param_creator::create_window -frame_toplevel frame $wp $dir]
    return $p
}

#bind all <Control-u><Control-u> "gid_groups_conds::start_param_creator; break"

################################################################################
#    Groups window creation
################################################################################

proc gid_groups_conds::upgrade_problemtype {} {
    variable upgrade_problemtype_automatic
    set project [GiD_Info Project ProblemType]
    set p [split $project \\ ]
    if { [llength $p] == 2 } {
        set project "[lindex $p 0]/[lindex $p 1]"
    }
    if { [info exists upgrade_problemtype_automatic] } {
        if { $upgrade_problemtype_automatic } {
            #after idle [list GiD_Process MEscape Data Defaults TransfProblem $project escape]
            GiD_Project transform_problemtype $project
        }
    } else {
        # Trick to avoid interactive dialogwin when running logTester utilities from GiD
        if { $::has_tk && ![GidUtils::AreWindowsDisabled] } {
            set w [dialogwin_snit .gid._ask -title [_ "Action"] -transient 1 \
                -entrytext [_ "The model needs to be upgraded. Do you want to upgrade to new version?"]]
            set action [$w createwindow]
            destroy $w
            if { $action < 1 } {
                # not upgrade
            } else {
                set mff [GiD_ModifiedFileFlag get]
                if { $mff } {
                    GiD_ModifiedFileFlag set 0
                }
                #after idle [list GiD_Process MEscape Data Defaults TransfProblem $project escape]
                GiD_Project transform_problemtype $project
            }
        }
    }
    return 0
}

proc gid_groups_conds::set_upgrade_problemtype_automatic {} {
    variable upgrade_problemtype_automatic

    set upgrade_problemtype_automatic 1
}

proc gid_groups_conds::unset_upgrade_problemtype_automatic {} {
    variable upgrade_problemtype_automatic

    set upgrade_problemtype_automatic 0
}

################################################################################
#    Groups window creation
################################################################################

proc gid_groups_conds::open_groups { parent what { open_over "" } } {
    variable gid_group_var

    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        if { $::GidPriv(ShowGroupsTab) } {
            if { $what == "window_force" || $what == "window" } {
                Groups::OpenTabIndex 1
                return
            } elseif {$what == "menu"} {
                # with tabs it is the same window for LAYER and GROUPS
                if { [GidUtils::ExistsWindow LAYER] } {
                    GidUtils::CloseWindow LAYER
                } else {
                    Groups::OpenTabIndex 1
                }
                return
            }
        }
    }

    switch $what {
        window_force {
            if { [winfo exists $parent.groups] } {
                raise $parent.groups
                set tree [$gid_group_var givetreectrl]
                focus -force $tree
            } else {
                open_groups $parent window $open_over
            }
        }
        window {
            if { [winfo exists $parent.m] } {
                open_groups_as_menu $parent $open_over
            }
            open_groups_as_window $parent
        }
        menu {
            if { [winfo exists $parent.groups] } {
                # first call closes the window
                open_groups_as_window $parent
            }
            open_groups_as_menu $parent $open_over
        }
        toggle {
            if { [winfo exists $parent.groups] } {
                if { $open_over eq "" } { return }
                # first call closes the window
                open_groups_as_window $parent
                set f [list open_groups_as_menu $parent $open_over]
            } else {
                open_groups_as_menu $parent $open_over
                set f [list open_groups_as_window $parent]
            }
            eval $f
        }
    }
}

proc gid_groups_conds::open_groups_as_window { parent } {
    variable gid_group_var
    variable doc

    set w $parent.groups
    if { [winfo exists $w] } {
        if { [info exists gid_group_var] } {
            $gid_group_var store_tree_state
        }
        destroy $w
        unset -nocomplain gid_group_var
        return
    }
    set cmd [list gid_groups_conds::open_groups_as_window $parent]
    InitWindow2 $w -title [_ "Groups"] \
        -geometryvariable PreGroupsWindowGeom -initcommand $cmd \
        -ontop

    wm protocol $w WM_DELETE_WINDOW [list gid_groups_conds::open_groups_as_window $parent]

    set gid_group_var [gid_groups $w.gg -winstyle toplevel]

    if { [info exists doc] } {
        $gid_group_var configure -domdoc $doc
    }
    set tree [$gid_group_var givetreectrl]

    ttk::frame $w.buts -style BottomFrame.TFrame
    ttk::button $w.buts.close -text [_ "Close"] -width 10 -command \
        [list gid_groups_conds::open_groups_as_window $parent] \
        -style BottomFrame.TButton
    pack $w.buts.close -pady 3

    grid $w.gg - -sticky nsew -pady 2
    grid $w.buts -pady 3 -sticky nsew
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    tk::TabToWindow $w.gg
}

# proc gid_groups_conds::_open_groups_as_window_apply { w } {
#     if { [$w giveaction] < 1 } {
#         open_groups_as_window [winfo parent $w]
#         return
#     }
# }

proc gid_groups_conds::open_groups_as_menu { parent open_over } {
    variable gid_group_var
    variable doc

    set w $parent.m
    if { [winfo exists $w] } {
        if { [info exists gid_group_var] } {
            $gid_group_var store_tree_state
        }
        destroy $w
        return
    }
    set w [toplevel $parent.m -bd 1 -relief raised]
    wm overrideredirect $w 1
    set x [expr {[winfo rootx $open_over]+[winfo width $open_over]}]
    set y [expr {[winfo rooty $open_over]}]
    wm geometry $w +${x}+$y

    set gid_group_var [gid_groups $w.gg -winstyle menu -open_over $open_over]
    if { [info exists doc] } {
        $gid_group_var configure -domdoc $doc
    }

    bind $w <Destroy> [namespace code [list unset -nocomplain gid_group_var]]

    grid $w.gg - -sticky nsew -pady 2
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    tkTabToWindow $w.gg
    grab $w.gg
    bind $w <Escape> [list destroy $w]
    bind $w <1> [list gid_groups_conds::_groups_destroy_if_outside $w %X %Y]
}

proc gid_groups_conds::_groups_destroy_if_outside { w x y } {
    if { ![winfo exists $w] } {
        return
    }
    set wx [winfo rootx $w]
    set wy [winfo rooty $w]
    if { $x < $wx || $x > $wx+[winfo width $w] || $y < $wy || $y > $wy+[winfo height $w] } {
        open_groups_as_menu [winfo parent $w] [winfo parent $w]
    }
}

proc gid_groups_conds::actualize_groups_window {} {
    variable gid_group_var

    set theme [gid_groups_conds::CurrentTheme]
    if { [regexp "Tdyn" $theme {} curr_theme] } {
        set win .gid.central.boundaryconds.gg
        if { [winfo exists $win.gt] } {
            $win.gt actualize_groups
        } elseif { [info command ::draw_post::fill_groups_treeT] ne "" } {
            draw_post::fill_groups_treeT
        }
    } else {
        if { [info exists gid_group_var] && [winfo exists $gid_group_var] } {
            $gid_group_var actualize
        } elseif { [info command ::draw_post::fill_groups_treeT] ne "" } {
            draw_post::fill_groups_treeT
        }
    }
}

################################################################################
#    Generic modification of the DOM tree
################################################################################

proc gid_groups_conds::add { parent_xpath tag_name attList { undo 1 } } {
    return [addF -undo $undo $parent_xpath $tag_name $attList]
}

proc gid_groups_conds::addN { args } {
    variable doc

    set optional {
        { -undo boolean 1 }
        { -resolve_parametric boolean 0 }
        { -ifnotexists "" 0 }
    }
    set compulsory "parentNode tag_name attList"
    parse_args $optional $compulsory $args

    if { [$parentNode ownerDocument] eq $doc } {
        if { !$resolve_parametric } {
            set rp ""
        } else {
            set rp [list -resolve_parametric 1]
        }
        if { !$ifnotexists } {
            set ine ""
        } else {
            set ine -ifnotexists
        }
        return [addF -undo $undo {*}$rp {*}$ine [gid_groups_conds::nice_xpath $parentNode] $tag_name $attList]
    } else {
        if { $ifnotexists && [$parentNode selectNodes $tag_name] ne "" } {
            set newNode [$parentNode selectNodes $tag_name]
        } else {
            set newNode [$doc createElement $tag_name]
            $parentNode appendChild $newNode
        }
        $newNode setAttribute {*}$attList
        return $newNode
    }
}

proc gid_groups_conds::addF { args } {
    variable doc

    set optional {
        { -undo boolean 1 }
        { -resolve_parametric boolean 0 }
        { -ifnotexists "" 0 }
    }
    set compulsory "parent_xpath tag_name attList"
    parse_args $optional $compulsory $args

    set root [$doc documentElement]

    if { $resolve_parametric } {
        if { [info command ::draw_post::resolve_parametric] ne "" } {
            set unit_magnitude [dict_getd $attList unit_magnitude ""]
            set valueN [draw_post::resolve_parametric -ask_user 1 \
                -unit_magnitude $unit_magnitude -units [dict_getd $attList units ""] -- [dict get $attList v]]
        }
    }
    set parentNode [$root selectNodes $parent_xpath]
    if { $parentNode eq "" } {
        error "error in gid_groups_conds::add. parent_xpath=$parent_xpath"
    }
    if { $ifnotexists && [$parentNode selectNodes $tag_name] ne "" } {
        set newNode [$parentNode selectNodes $tag_name]
        set aList ""
        foreach name [$newNode attributes] {
            lappend aList $name [$newNode @$name]
        }
        set undo_cmd [list gid_groups_conds::setAttributes_exclusive [gid_groups_conds::nice_xpath $newNode] $aList]
    } else {
        set newNode [$doc createElement $tag_name]
        $parentNode appendChild $newNode
        set undo_cmd [list gid_groups_conds::delete [gid_groups_conds::nice_xpath $newNode]]
    }
    foreach "name value" $attList {
        if { $name eq "v" && [info exists valueN] } {
            set value $valueN
        }
        $newNode setAttribute $name $value
    }
    if { $undo } {
        if { !$resolve_parametric } {
            set rp ""
        } else {
            set rp [list -resolve_parametric 1]
        }
        if { !$ifnotexists } {
            set ine ""
        } else {
            set ine -ifnotexists
        }
        set redo_cmd [list gid_groups_conds::addF -undo 0 {*}$rp {*}$ine $parent_xpath $tag_name $attList]
        add_to_undo "add" $redo_cmd $undo_cmd
    }
    return $newNode
}

proc gid_groups_conds::addXML { parent_xpath xml } {
    variable doc

    set root [$doc documentElement]
    set parentNode [$root selectNodes $parent_xpath]
    if { $parentNode eq "" } {
        error "error in gid_groups_conds::add. parent_xpath=$parent_xpath"
    }
    $parentNode appendXML $xml
    actualize_conditions_window -domNodeList $parentNode
}

proc gid_groups_conds::deleteN { node } {
    variable doc

    if { [$node ownerDocument] eq $doc } {
        return [delete [gid_groups_conds::nice_xpath $node]]
    } else {
        $node delete
        return 1
    }
}

proc gid_groups_conds::delete { xpath { undo 1 } } {
    variable doc
    set root [$doc documentElement]
    set inum 0
    set undoList ""
    foreach node [$root selectNodes $xpath] {
        lappend undoList [list gid_groups_conds::addXML [gid_groups_conds::nice_xpath [$node parentNode]] [$node asXML]]
        $node delete
        incr inum
    }
    if { $undo != 0 } {
        set undo_cmd [join $undoList ";"]
        set redo_cmd [list gid_groups_conds::delete $xpath 0]
        if { $undo eq "only_undo" } {
            set redo_cmd ""
        }
        add_to_undo "delete" $redo_cmd $undo_cmd
    }
    return $inum
}

#simple proc specialized for nodes 'value' with n==$v set the attribute v to $v
proc gid_groups_conds::setAttributeNodesValue { n v } {
    set xpath "//value\[@n='$n'\]"
    foreach node [gid_groups_conds::GetDomNodes $xpath] {
        $node setAttribute v $v
    }
}

proc gid_groups_conds::setAttributes { xpath attList { undo 1 } } {
    return [gid_groups_conds::setAttributesF -undo $undo $xpath $attList]
}

proc gid_groups_conds::setAttributesN { args } {
    variable doc

    set optional {
        { -undo boolean 1 }
        { -resolve_parametric boolean 0 }
    }
    set compulsory "node attList"
    parse_args $optional $compulsory $args

    if { [$node ownerDocument] eq $doc } {
        return [gid_groups_conds::setAttributesF -undo $undo -resolve_parametric $resolve_parametric \
            [gid_groups_conds::nice_xpath $node] $attList]
    } else {
        $node setAttribute {*}$attList
        return 1
    }
}

proc gid_groups_conds::setAttributesF { args } {
    variable doc

    set optional {
        { -undo boolean 1 }
        { -resolve_parametric boolean 0 }
    }
    set compulsory "xpath attList"
    parse_args $optional $compulsory $args

    set root [$doc documentElement]
    set nodeList [$root selectNodes $xpath]

    if { $resolve_parametric } {
        if { [info command ::draw_post::resolve_parametric] ne "" && [dict get $attList v] ne "" } {
            set unit_magnitude [get_domnode_attribute [lindex $nodeList 0] unit_magnitude]
            set valueN [draw_post::resolve_parametric -ask_user 1 \
                -unit_magnitude $unit_magnitude -units [dict_getd $attList units ""] -- [dict get $attList v]]
        }
    }
    set undoList ""
    set inum 0
    foreach node $nodeList {
        set aList ""
        foreach name [$node attributes] {
            lappend aList $name [$node @$name]
        }
        lappend undoList [list gid_groups_conds::setAttributes_exclusive [gid_groups_conds::nice_xpath $node] $aList]
        foreach "name value" $attList {
            if { $name eq "v" && [info exists valueN] } {
                set value $valueN
            }
            $node setAttribute $name $value
        }
        incr inum
    }
    if { $undo } {
        set undo_cmd [join $undoList ";"]
        if { !$resolve_parametric } {
            set rp ""
        } else {
            set rp [list -resolve_parametric 1]
        }
        set redo_cmd [list gid_groups_conds::setAttributesF -undo 0 {*}$rp $xpath $attList]
        add_to_undo "setattributes" $redo_cmd $undo_cmd
    }
    return $inum
}

proc gid_groups_conds::setAttributes_exclusive { xpath attList } {
    variable doc

    set root [$doc documentElement]
    foreach node [$root selectNodes $xpath] {
        foreach name [$node attributes] {
            $node removeAttribute $name
        }
        foreach "name value" $attList {
            $node setAttribute $name $value
        }
    }
}

proc gid_groups_conds::moveNodeN { node_ref node before_after } {
    variable doc

    if { [$node ownerDocument] eq $doc } {
        return [moveNode [gid_groups_conds::nice_xpath $node_ref] [gid_groups_conds::nice_xpath $node] $before_after]
    } else {
        set parent_ref [$node_ref parentNode]
        if { $before_after eq "after" } {
            set childNodes [$parent_ref childNodes]
            set ipos [lsearch -exact $childNodes $node_ref]
            set node_ref [lindex $childNodes [incr ipos]]
        } elseif { $before_after eq "child" } {
            set parent_ref $node_ref
            set node_ref ""
        }
        [$node parentNode] removeChild $node
        if { $node_ref eq "" } {
            $parent_ref appendChild $node
        } else {
            $parent_ref insertBefore $node $node_ref
        }
    }
}

proc gid_groups_conds::moveNode { xpath_ref xpath before_after { undo 1 } } {
    variable doc

    set root [$doc documentElement]

    set node_ref [$root selectNodes $xpath_ref]
    set parent_ref [$node_ref parentNode]
    set node [$root selectNodes $xpath]

    set node_refU [$node selectNodes {following-sibling::*[1]}]
    if { $node_refU ne "" } {
        set xpath_refU [gid_groups_conds::nice_xpath $node_refU]
        set before_afterU before
    } else {
        set xpath_refU [gid_groups_conds::nice_xpath [$node parentNode]]
        set before_afterU child
    }

    if { $before_after eq "after" } {
        set childNodes [$parent_ref childNodes]
        set ipos [lsearch -exact $childNodes $node_ref]
        set node_ref [lindex $childNodes [incr ipos]]
        if { $node_ref eq $node } {
            set node_ref [lindex $childNodes [incr ipos]]
        }
    } elseif { $before_after eq "child" } {
        set parent_ref $node_ref
        set node_ref ""
    }
    [$node parentNode] removeChild $node
    if { $node_ref eq "" } {
        $parent_ref appendChild $node
    } else {
        $parent_ref insertBefore $node $node_ref
    }
    if { $undo } {
        set undo_cmd [list gid_groups_conds::moveNode $xpath_refU [gid_groups_conds::nice_xpath $node] $before_afterU]
        set redo_cmd [list gid_groups_conds::moveNode $xpath_ref $xpath $before_after 0]
        add_to_undo "movenode" $redo_cmd $undo_cmd
    }
}

proc gid_groups_conds::copyNodeN { node node_to } {
    variable doc

    if { [$node ownerDocument] eq $doc } {
        return [copyNode [gid_groups_conds::nice_xpath $node] [gid_groups_conds::nice_xpath $node_to]]
    } else {
        set newNode [$node_to appendChild [$node cloneNode -deep]]
        return $newNode
    }
}

proc gid_groups_conds::copyNode { xpath xpath_to { undo 1 } } {
    variable doc

    set root [$doc documentElement]
    set node [$root selectNodes $xpath]
    set parent [$root selectNodes $xpath_to]
    set newNode [$parent appendChild [$node cloneNode -deep]]

    if { $undo } {
        set undo_cmd [list gid_groups_conds::delete [gid_groups_conds::nice_xpath $node]]
        set redo_cmd [list gid_groups_conds::copyNode $xpath $xpath_to 0]
        add_to_undo "copynode" $redo_cmd $undo_cmd
    }
    return $newNode
}

proc gid_groups_conds::copyNode_recursive { node node_to { undo 1 } } {

    set attList ""
    foreach i [$node selectNodes @*] {
        lappend attList {*}$i
    }
    set newnode [addN -undo $undo $node_to [$node nodeName] $attList]
    foreach n [$node childNodes] {
        copyNode_recursive $n $newnode $undo
    }
}

################################################################################
#    Group query & modification routines
################################################################################

proc gid_groups_conds::group_exists { group } {
    variable doc
    if { [llength [[$doc documentElement] selectNodes [format_xpath {groups/group[@n=%s]} $group]]] } {
        set group_exists 1
    } else {
        set group_exists 0
    }
    return $group_exists
}

# args: ?-noactualizegroups? ?-exactname? ?-noundo? prefix
proc gid_groups_conds::add_new_group { args } {
    variable doc
    variable colorList
    variable is_draw_post

    set optional {
        { -noactualizegroups "" 0 }
        { -exactname "" 0 }
        { -ifnotexists "" 0 }
        { -noundo "" 0 }
        { -type typeName "" }
        { -permit_extended_name boolean 0 }
        { -color color "" }
        { -copy_group boolean 0 }
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        lappend optional { -noactualize_gid_groups "" 0 }
    }
    set compulsory "prefix"
    parse_args $optional $compulsory $args

    if { $permit_extended_name } {
        set rex0 {^[-.\w\s/:*]+$}
        set rex1 {^[-.\w\s/#]+$}
        set rex2 {^[-.\w\s/]}
    } else {
        set rex0 {^[-.\w\s/]+$}
        set rex1 {^[-.\w\s/#]+$}
        set rex2 {^[-.\w\s/]}
    }
    if { ![regexp $rex0 $prefix] && ![regexp $rex1 $prefix] && ![regexp $rex2 $prefix]} {
        return -code error [_ "error: A group name can only contain letters, digits, space and _-./ (trying to create '%s'" $prefix]
    }

    if { $is_draw_post } {
        if { $exactname || $ifnotexists } {
            set gr [ggc::manage_geometry_groups id $prefix]
            if { $gr != -1 } {
                if { $ifnotexists } { return }
                return -code error [_ "error: Group '%s' already exists" $prefix]
            }
        }
        dict set d Name $prefix
        if { $color ne "" } {
            set colorRGBA [_give_color_RGBA $color]
            dict set d Color [_give_color_geometry_from_RGBA $colorRGBA]
        }
        set gr [draw_post::create_group -make_name_unique 1 $d]
        set name [ggc::manage_geometry_groups name $gr]
        return [list $name $color]
    }

    set actualizegroups [expr {($noactualizegroups)?0:1}]
    set undo [expr {($noundo)?0:1}]

    set root [$doc documentElement]
    if { $color == "" } {
        set found 0
        foreach color $colorList {
            if { [$root selectNodes [format_xpath {groups/group[@color=%s]} $color]] eq "" } {
                set found 1
                break
            }
        }
        if { !$found } {
            set color [lindex $colorList [expr {int(rand()*[llength $colorList])}]]
        }
    }
    if { $type eq "" } {
        set type [$root selectNodes string(groups_types/group_type\[@default='1'\]/@pn)]
        set group_types ""
        foreach node [$root selectNodes groups_types/group_type] {
            lappend group_types [$node @pn]
        }
        if { $type eq "" } { set type [lindex $group_types 0] }
    }
    set idx 0
    set groups ""
    foreach domNode [$root selectNodes groups/group] {
        lappend groups [$domNode @n]
        if { [regexp "^$prefix\(\\d+\)\$" [$domNode @n] {} n] && $n > $idx } {
            set idx $n
        }
    }
    incr idx

    if { $exactname || $ifnotexists } {
        set name $prefix
        if { [gid_groups_conds::group_exists $name] } {
            if { $ifnotexists } { return }
            return -code error [_ "error: Group '%s' already exists" $name]

        }
    } else {
        if { $copy_group } {
            set name "$prefix $idx"
        } else {
            set name $prefix$idx
        }
    }

    GiD_EnterInBatch "*****MARK START_BLOCK CreateCondGroup $name\n"

    if { [$root selectNodes groups] eq "" } {
        gid_groups_conds::add . groups "" $undo
    }

    gid_groups_conds::add groups group [list n $name onoff 1 color $color type $type] $undo

    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        if { ![GiD_Groups exists $name] } {
            GiD_Process 'Groups create $name escape
            #GiD_Groups create $name
            set color_rgb [gid_groups_conds::get_color_rrggbb_from_symbolic_name $color]
            GiD_Process 'Groups Edit Color $name ${color_rgb}ff escape
            #GiD_Groups edit color $name ${color_rgb}ff
            if { !$noactualize_gid_groups } {
                GidUtils::UpdateWindow GROUPS
            }
        }
    }
    if { $actualizegroups } {
        gid_groups_conds::actualize_groups_window
    }

    if { [info commands GiD_RaiseEvent] != "" } {
        GiD_RaiseEvent GiD_Event_AfterCreateCondGroup $name
    }

    GiD_EnterInBatch "*****MARK END_BLOCK CreateCondGroup $name\n"
    return [list $name $color]
}

proc gid_groups_conds::give_groups { {state normal} {condition_disallowed_types ""} {condition_disallowed_element_types ""} } {
    variable doc
    variable is_draw_post

    if { ![info exists doc] } { return "" }

    set groups ""

    if { !$is_draw_post } {
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
            set gid_version [GiD_Info GiDVersion]
            if { [regexp {(.*)d} $gid_version {} version_num] } {
                set version_num $version_num
            } else {
                set version_num [lindex [split $gid_version .] 0]
            }
            if { $version_num >= "14.1.9" } {
                set same_behavior_native_groups_and_cond_groups 1
                if { $same_behavior_native_groups_and_cond_groups } {
                    set groups [gid_groups_conds::ListGroups $state $condition_disallowed_types \
                        $condition_disallowed_element_types]
                } else {
                    set groups [Groups::ListGroups $state $condition_disallowed_types \
                        $condition_disallowed_element_types];
                    #state==normal: hidden groups are not listed
                }
            } else {
                #must store also groups in xml or native names are enougth?
                set groups [Groups::ListGroups normal ""] ;#hidden groups are not listed
            }
        } else {
            set root [$doc documentElement]
            foreach domNode [$root selectNodes groups/group] {
                lappend groups [$domNode @n]
            }
        }
    } else {
        set groups [ggc::manage_geometry_groups list -names]
    }
    return $groups
}

#state: normal, disabled or hidden, use * to list all without take into account its state
#condition_disallowed_types is a list of not allowed types: points lines surfaces volumes elements faces
#condition_disallowed_element_types is a list of not allowed element types: linear triangle ...
#warning: not add nodes to condition_disallowed_types list, because now when meshing the nodes are always
#         added to the group of the source entity!!
proc gid_groups_conds::ListGroups { {state normal} {condition_disallowed_types ""} {condition_disallowed_element_types ""} } {

    set groups ""
    foreach group [GiD_Groups list] {
        if { $state == "*" || [GiD_Groups get state $group] == $state } {
            set over_cond ""
            foreach over [list points lines surfaces volumes nodes elements faces] {
                if { $over in $condition_disallowed_types } { continue }
                lappend over_cond $over
            }
            set over_group ""
            foreach over [list points lines surfaces volumes nodes elements faces] {
                if { [GiD_EntitiesGroups get $group $over -count] } {
                    lappend over_group $over
                }
            }

            set valid 1
            set isfound 0
            foreach over $over_group {
                if { $over in $over_cond } {
                    set isfound 1
                    break
                }
            }
            if { !$isfound } {
                set valid 0
            }

            if { $valid } {
                foreach element_type $condition_disallowed_element_types {
                    if { [GiD_EntitiesGroups get $group elements -count -element_type [list $element_type]] } {
                        set valid 0
                        break
                    }
                }
            }
            if { $valid } {
                lappend groups $group
            }
        }
    }
    return [lsort -unique -dictionary $groups]
}

proc gid_groups_conds::give_groupsP { patternList } {
    variable doc
    variable is_draw_post

    if { ![info exists doc] } { return "" }

    if { !$is_draw_post } {
        set groups ""
        foreach group [gid_groups_conds::give_groups normal "" ""] {
            foreach p $patternList {
                if { [string match $p $group] } {
                    lappend groups $group
                    break
                }
            }
        }
    } else {
        set groups ""
        foreach group [ggc::manage_geometry_groups list -names] {
            foreach p $patternList {
                if { [string match $p $group] } {
                    lappend groups $group
                    break
                }
            }
        }
    }
    return $groups
}


proc gid_groups_conds::give_groups_types {} {
    variable doc

    if { ![info exists doc] } { return "" }

    set group_types ""
    set root [$doc documentElement]
    foreach node [$root selectNodes groups_types/group_type] {
        lappend group_types [$node @pn]
    }
    return $group_types
}

proc gid_groups_conds::apply_entities_to_group { group ov ovName button top scr_ext } {
    variable is_draw_post
    variable theme

    #GiD_EnterInBatch "*****MARK START_BLOCK GroupAssignEntitites $group\n"

    set ov2 $ov
    if { $ov2 eq "surface_as_volume" } { set ov2 volume }

    set scr ""
    set list [list $top]
    for { set i 0 } { $i < [llength $list] } { incr i } {
        set w [lindex $list $i]
        if { $w eq $button } { continue }
        if { $w eq $top } { continue }
        set err [catch { $w instate !disabled } ret]
        if { !$err } {
            if { $ret } {
                $w state disabled
                append scr ";$w state !disabled"
            }
        } else {
            set err [catch { $w cget -state } state]
            if { !$err && $state eq "normal" } {
                $w configure -state disabled
                append scr ";$w configure -state normal"
            }
        }
        eval lappend list [winfo children $w]
    }
    set help [gid_groups_conds::register_popup_help $button]
    append scr ";[list $button configure -image [$button cget -image]];"\
        "[list $button configure -text [$button cget -text]];"\
        "[list $button configure -command [$button cget -command]];"\
        "[list gid_groups_conds::register_popup_help $button $help];"
    append scr $scr_ext

    set textbut [$button cget -text]
    $button configure -image [gid_themes::GetImage cancel small_icons] \
        -text [_ "End"] \
        -command [namespace code [list _end_apply_entities_to_group $scr $button $ov $group]]
    gid_groups_conds::register_popup_help $button \
        [_ "Click on this button or Escape to finish the selection process"]

    if { !$is_draw_post } {
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
            if { $textbut eq [_ "Unselect"] } {
                GiD_Process Mescape Utilities EntitiesGroups Unassign $group ${ovName}s
            } else {
                GiD_Process Mescape Utilities EntitiesGroups Assign $group ${ovName}s
            }
        } else {
            if { $textbut eq [_ "Unselect"] } {
                GidUtils::SetWarnLine [_ "Select %ss to delete from group '%s'" $ovName $group]
                GidUtils::DisableWarnLine
                GiD_Process escape escape escape escape Data Conditions AssignCond ${ov2}_groups UnAssign
                GiD_Process Field name $group
                GidUtils::EnableWarnLine
            } else {
                GidUtils::SetWarnLine [_ "Select %ss to add to group '%s'" $ovName $group]
                GidUtils::DisableWarnLine
                GiD_Process escape escape escape escape Data Conditions AssignCond ${ov2}_groups NoRepeatField name
                GiD_Process Change $group
                GidUtils::EnableWarnLine
            }
        }
        delayedop changefunc [list $button invoke]
    } else {
        set draw_post::selection(elevel) $ov2
        draw_post::begin_or_change_selection -end_cmd [list $button invoke]
    }
}

proc gid_groups_conds::_end_apply_entities_to_group { scr button ov group } {
    variable is_draw_post

    if { !$is_draw_post } {
        set found [delayedop -cancel changefunc [list $button invoke]]
        if { $found } { GiD_Process escape }
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        } else {
            GiD_Process escape
        }
    } else {
        set entitiesDict [dict create $ov [ggc::manage_geometry_entities selection get $ov]]
        draw_post::add_group_entities $group $entitiesDict
        lassign [ggc::manage_geometry_entities list count all $entitiesDict] - - - - num_entities
        draw_post::warnline [_ "%d entities has been applied to group '%s'" $num_entities $group]
        #draw_post::end_draw_graphmode -keep 0
        draw_post::clear_selection
    }

    if { $ov eq "surface_as_volume" } {
        gid_groups_conds::volumes_to_surfaces_in_group $group
    }

    uplevel #0 $scr
    #GiD_EnterInBatch "*****MARK END_BLOCK GroupAssignEntitites $group\n"
}

proc gid_groups_conds::_unfreeze_layers {} {

    set frozen_layers ""
    foreach layer [GiD_Info Layers] {
        set frozen [lindex [GiD_Info Layers $layer] 0 1]
        if { $frozen } {
            GiD_Process Layers UnFreeze $layer escape
            lappend frozen_layers $layer
        }
    }
    return $frozen_layers
}

proc gid_groups_conds::_freeze_layers { frozen_layers } {
    foreach layer $frozen_layers {
        GiD_Process Layers Freeze $layer escape
    }
}

proc gid_groups_conds::volumes_to_surfaces_in_group { group } {
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        set volume_ids [GiD_EntitiesGroups get $group volumes]
        GiD_EntitiesGroups unassing $group volumes
        foreach volume_id $volume_ids {
            set surface_ids [objarray sort -unique [GidUtils::GetVolumeSurfaces $volume_id]]
            GiD_EntitiesGroups assign $group surfaces $surface_ids
        }
    } else {
        GidUtils::DisableWarnLine
        set frozen_layers [_unfreeze_layers]
        set group_slave ""
        foreach j [GiD_Info conditions volume_groups geometry] {
            foreach "- num - groupL" $j break
            if { $groupL ne $group } { continue }
            set surface_ids [GidUtils::GetVolumeSurfaces $num]
            foreach surface_id $surface_ids {
                GiD_Process escape escape escape escape Data Conditions AssignCond surface_groups Change $group $surface_id escape
            }
            GiD_Process escape escape escape escape Data Conditions AssignCond volume_groups UnAssign $num escape
        }
        _freeze_layers $frozen_layers
        GidUtils::EnableWarnLine
    }
}

proc gid_groups_conds::layers_to_groups_do { layer_names } {
    GidUtils::DisableGraphics

    set frozen_layers [gid_groups_conds::_unfreeze_layers]
    set infoProj [GiD_Info Project ViewMode]

    if { $layer_names eq "" } {
        set layer_names [GiD_Layers list]
    }

    foreach layer $layer_names {
        set group_color [string range [GiD_Layers get color $layer] 0 6]
        gid_groups_conds::add_new_group -ifnotexists -noactualizegroups -color $group_color $layer

        foreach geom_mesh [list geometry mesh] {
            if { $infoProj eq "GEOMETRYUSE" && $geom_mesh eq "mesh" } {
                MeshView 0
            } elseif { $infoProj eq "MESHUSE" && $geom_mesh eq "geometry" } {
                GeometryView
            }
            set num 0
            foreach i [list line surface volume] {
                GiD_Process MEscape Data Conditions AssignCond ${i}_groups NoRepeatField name
                GiD_Process Change $layer layer:$layer escape
                incr num [llength [GiD_Info Layers -entities ${i}s $layer]]
            }
            if { !$num } {
                GiD_Process MEscape Data Conditions AssignCond point_groups NoRepeatField name
                GiD_Process Change $layer layer:$layer escape
            }
            if { $infoProj eq "GEOMETRYUSE" && $geom_mesh eq "mesh" } {
                GeometryView
            } elseif { $infoProj eq "MESHUSE" && $geom_mesh eq "geometry" } {
                MeshView 0
            }
        }
    }
    gid_groups_conds::_freeze_layers $frozen_layers
    GidUtils::EnableGraphics
}

proc gid_groups_conds::layers_to_groups { args } {
    set optional {
        { -layerList layerList "" }
    }
    set compulsory ""
    parse_args $optional $compulsory $args

    if { $layerList eq "" } {
        set layerList [GiD_Layers list]
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        Groups::LayersToGroups $layerList
    } else {
        gid_groups_conds::layers_to_groups_do $layerList
    }
}

# args: ?-noactualizegroups? oldname newname
proc gid_groups_conds::rename_group { args } {
    variable is_draw_post
    variable doc

    set optional {
        { -noactualizegroups "" 0 }
        { -noactualizeconds "" 0 }
        { -merge_if_exists "" 0 }
        { -undo boolean 1 }
        { -permit_extended_name boolean 0 }
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        lappend optional { -noactualize_gid_groups "" 0 }
    }
    set compulsory "oldname newname"
    parse_args $optional $compulsory $args

    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        set actualizegroups 0
    } else {
        set actualizegroups [expr {($noactualizegroups)?0:1}]
    }

    if { $oldname eq $newname } { return }

    if { $permit_extended_name } {
        set rex0 {^[-.\w\s/:*]+$}
        set rex1 {^[-.\w\s/#]+$}
        set rex2 {^[-.\w\s/]}
    } else {
        set rex0 {^[-.\w\s/]+$}
        set rex1 {^[-.\w\s/#]+$}
        set rex2 {^[-.\w\s/]}
    }
    if { ![regexp $rex0 $newname] && ![regexp $rex1 $newname] && ![regexp $rex2 $newname]} {
        return -code error [_ "error: A group name can only contain letters, digits, space and _-./ (trying to create '%s'" $newname]
    }

    if { $is_draw_post } {
        set gr_from [ggc::manage_geometry_groups id $oldname]
        set gr_to [ggc::manage_geometry_groups id $newname]
        if { $gr_to != -1 } {
            if { !$merge_if_exists } {
                return -code error [_ "error: group '%s' already exists" $newname]
            } else {
                set entitiesDict [ggc::manage_geometry_groups entities list_group $gr_from all]
                draw_post::add_group_entities $newname $entitiesDict
            }
        } else {
            draw_post::change_group_properties [list $oldname] Name $newname
        }
        return
    }

    if { [gid_groups_conds::group_exists $newname] } {
        if { !$merge_if_exists } {
            return -code error [_ "error: group '%s' already exists" $newname]
        } else {
            set exists_group 1
        }
    } else {
        set exists_group 0
    }

    GiD_EnterInBatch "*****MARK START_BLOCK GroupRename $oldname $newname\n"

    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        set exists_group_native_oldname [GiD_Groups exists $oldname]
        set old_parent [join [lrange [GidUtils::Split $oldname //] 0 end-1] //]
        set new_parent [join [lrange [GidUtils::Split $newname //] 0 end-1] //]
        if { $old_parent == $new_parent } {

            if { !$exists_group } {
                gid_groups_conds::setAttributes [format_xpath {groups/group[@n=%s]} $oldname] [list n $newname]
                gid_groups_conds::rename_group_from_conditions -noactualizeconds=$noactualizeconds -undo $undo $oldname $newname
            } else {
                gid_groups_conds::delete [format_xpath {groups/group[@n=%s]} $oldname]
            }
            #do it in GiD after do it in customlib, to avoid call again this function from GiD_Event
            if { $exists_group_native_oldname } {
                #GiD_Groups edit name $oldname $newname
                GiD_Process 'Groups Edit Name $oldname $newname escape
            }

        } else {
            gid_groups_conds::setAttributes [format_xpath {groups/group[@n=%s]} $oldname] [list n $newname]
            gid_groups_conds::rename_group_from_conditions -noactualizeconds=$noactualizeconds -undo $undo $oldname $newname
            #it is not considered like a rename but a change of parent
            #do it in GiD after do it in customlib, to avoid call again this function from GiD_Event
            if { $exists_group_native_oldname } {
                GiD_Process 'Groups Edit Parent $oldname $new_parent escape
            }
        }
        set oldnameList [cu::splitx $oldname {//}]
        set numoldname [llength $oldnameList]
        foreach i [[$doc documentElement] selectNodes {groups/group}] {
            set n [$i @n]
            set n_splitted [cu::splitx $n {//}]
            set iname [join [lrange $n_splitted 0 [expr {$numoldname-1}]] {//}]
            set child [join [lrange $n_splitted $numoldname end] {//}]
            if { $iname eq $oldname } {
                gid_groups_conds::setAttributes [format_xpath {groups/group[@n=%s]} $n] [list n "$newname//$child"]
                gid_groups_conds::rename_group_from_conditions -noactualizeconds=$noactualizeconds -undo $undo $n "$newname//$child"
            }
        }
        if { !$noactualize_gid_groups } {
            GidUtils::UpdateWindow GROUPS
        }
    } else {
        foreach i [list point line surface volume] {
            GiD_Process MEscape Data Conditions AssignCond ${i}_groups ReplaceField name $oldname $newname escape
        }
        if { !$exists_group } {
            gid_groups_conds::setAttributes [format_xpath {groups/group[@n=%s]} $oldname] [list n $newname]
        } else {
            gid_groups_conds::delete [format_xpath {groups/group[@n=%s]} $oldname]
        }
        gid_groups_conds::rename_group_from_conditions -noactualizeconds=$noactualizeconds -undo $undo $oldname $newname
    }


    if { $actualizegroups } {
        gid_groups_conds::actualize_groups_window
    }

    if { [info commands GiD_RaiseEvent] != "" } {
        GiD_RaiseEvent GiD_Event_AfterRenameCondGroup $oldname $newname
    }
    GiD_EnterInBatch "*****MARK END_BLOCK GroupRename $oldname $newname\n"
}

proc gid_groups_conds::rename_group_from_conditions { args } {
    variable doc

    set optional {
        { -noactualizeconds "" 0 }
        { -undo boolean 1 }
    }
    set compulsory "oldname newname"
    parse_args $optional $compulsory $args

    if { ![info exists doc] } {
        return
    }

    set root [$doc documentElement]
    set actualizeconds [expr {($noactualizeconds)?0:1}]

    set xp [format_xpath {//condition[not(ancestor::blockdata[@n="Internal data"])]/group[@n=%s]} $oldname]
    append xp [format_xpath {|//condition[not(ancestor::blockdata[@n="Internal data"])]/groupList[group/@n=%s]} $oldname]
    if { [llength [$root selectNodes $xp]] } {
        gid_groups_conds::setAttributes $xp [list n $newname] $undo
    }
    if { $actualizeconds } {
        actualize_conditions_window
    }
}

#to get the list of group names of a customlib condition
proc gid_groups_conds::get_groups_from_condition { condition_n } {
    variable doc
    set groups [list]
    set root [$doc documentElement]
    set xp [format_xpath {//condition[@n=%s]//group} $condition_n]
    foreach groupNode [$root selectNodes $xp] {
        lappend groups [$groupNode @n]
    }
    return $groups
}

proc gid_groups_conds::get_group_att { args } {
    variable doc

    set optional {
        { -default value "" }
    }
    set compulsory "name attribute"
    parse_args $optional $compulsory $args

    set root [$doc documentElement]
    set node [$root selectNodes [format_xpath {groups/group[@n=%s]} $name]]
    return [$node getAttribute $attribute $default]
}

# args: ?-noactualizegroups? name attribute newvalue
# attributes can be: onoff, color, type
proc gid_groups_conds::change_group_att { args } {
    variable is_draw_post

    set optional {
        { -noactualizegroups "" 0 }
    }
    set compulsory "name attribute newvalue"
    parse_args $optional $compulsory $args

    GiD_EnterInBatch "*****MARK START_BLOCK GroupChangeAttributes $name\n"
    if { !$is_draw_post } {
        gid_groups_conds::setAttributes [format_xpath {groups/group[@n=%s]} $name] [list $attribute $newvalue]
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
            if { [GiD_Groups exists $name] } {
                if { $attribute == "color" } {
                    set color_rgb [gid_groups_conds::get_color_rrggbb_from_symbolic_name $newvalue]
                    if { [string index $color_rgb 0] == "#" && [string length $color_rgb] == 7 } {
                        if { $color_rgb != [string range [GiD_Groups get color $name] 0 6] } {
                            GiD_Process 'Groups Edit Color $name ${color_rgb}ff escape
                            #GiD_Groups edit color $name ${color_rgb}ff
                            GidUtils::UpdateWindow GROUPS
                        }
                    }
                } elseif { $attribute == "onoff" } {
                    if { $newvalue != [GiD_Groups get visible $name] } {
                        GiD_Process 'Groups Edit Visible $name $newvalue escape
                        #GiD_Groups edit visible $name $newvalue
                        GidUtils::UpdateWindow GROUPS
                    }
                }
            }
        }
    } else {
        draw_post::change_group_properties [list $name] $attribute $newvalue
    }
    if { !$noactualizegroups } {
        gid_groups_conds::actualize_groups_window
    }
    GiD_EnterInBatch "*****MARK END_BLOCK GroupChangeAttributes $name\n"
}

# args: ?-noactualizegroups? sibling name before|after
proc gid_groups_conds::move_group { args } {
    set optional {
        { -noactualizegroups "" 0 }
    }
    set compulsory "sibling name before_after"
    parse_args $optional $compulsory $args
    set actualizegroups [expr {($noactualizegroups)?0:1}]
    gid_groups_conds::moveNode [format_xpath {groups/group[@n=%s]} $sibling] [format_xpath {groups/group[@n=%s]} $name] $before_after
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        GidUtils::UpdateWindow GROUPS
    }
    gid_groups_conds::actualize_groups_window
}

proc gid_groups_conds::delete_special_groups {} {
    variable doc
    if { ![info exists doc] } {
        return ""
    }
    set groups ""
    set root [$doc documentElement]
    foreach domNode [$root selectNodes groups/group] {
        set group [$domNode @n]
        if { [regexp {:\d+$} $group] } {
            delete_group -noactualizegroups -noactualizeconds -check_entities 0 -undo 0 $group
        }
    }
}

proc gid_groups_conds::delete_all_groups { } {
    foreach group [gid_groups_conds::give_groups "*" "" ""] {
        gid_groups_conds::delete_group -noactualizegroups $group
    }
    gid_groups_conds::actualize_groups_window
}

proc gid_groups_conds::delete_group { args } {
    variable is_draw_post
    variable doc

    set optional {
        { -noactualizegroups "" 0 }
        { -noactualizeconds "" 0 }
        { -ifexists "" 0 }
        { -check_entities boolean 1 }
        { -undo boolean 1 }
        { -delete_gr_from_cond boolean 1 }
        { -delete_node boolean 1 }
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        lappend optional { -noactualize_gid_groups "" 0 }
    }
    set compulsory "del_group"
    parse_args $optional $compulsory $args

    if { $is_draw_post } {
        set gr [ggc::manage_geometry_groups id $del_group]
        if { $gr == -1 } {
            if { $ifexists } { return }
            return -code error [_ "error: Group '%s' does not exist" $del_group]
        }
        draw_post::delete_groups [list $del_group]
        return
    }

    if { [info commands GiD_RaiseEvent] != "" } {
        if { [GiD_RaiseEvent GiD_Event_BeforeDeleteCondGroup $del_group] == "-cancel-" } {
            return -cancel-
        }
    }

    set actualizegroups [expr {($noactualizegroups)?0:1}]
    if { ![gid_groups_conds::group_exists $del_group] } {
        if { $ifexists } { return }
        return -code error [_ "error: Group '%s' does not exist" $del_group]
    }

    GiD_EnterInBatch "*****MARK START_BLOCK GroupDelete $del_group\n"

    if { $check_entities } {
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
            if { [GiD_Groups exists $del_group] } {
                GiD_Process 'Groups Delete $del_group escape
                #GiD_Groups delete $del_group
                if { !$noactualize_gid_groups } {
                    GidUtils::UpdateWindow GROUPS
                }
            }
        } else {
            foreach i [list point line surface volume] {
                GiD_Process MEscape Data Conditions AssignCond ${i}_groups UnAssignField name $del_group escape
            }
        }
    }
    if { $delete_node } {
        gid_groups_conds::delete [format_xpath {groups/group[@n=%s]} $del_group] $undo
    }

    if { $delete_gr_from_cond } {
        delete_group_from_conditions -noactualizeconds=$noactualizeconds -undo $undo [list $del_group]
    }

    if { $actualizegroups } {
        gid_groups_conds::actualize_groups_window
    }
    GiD_EnterInBatch "*****MARK END_BLOCK GroupDelete $del_group\n"
}

proc gid_groups_conds::delete_group_from_conditions { args } {
    variable doc
    set optional {
        { -noactualizeconds "" 0 }
        { -undo 0|1|only_undo 1 }
    }
    set compulsory "groupList"
    parse_args $optional $compulsory $args
    if { ![info exists doc] } {
        return
    }
    set root [$doc documentElement]
    set actualizeconds [expr {($noactualizeconds)?0:1}]
    foreach del_group $groupList {
        set xp [format_xpath {//condition[not(ancestor::blockdata[@n="Internal data"])]/group[@n=%s]} $del_group]
        append xp [format_xpath {|//condition[not(ancestor::blockdata[@n="Internal data"])]/groupList[group/@n=%s]} $del_group]
        if { [llength [$root selectNodes $xp]] } {
            gid_groups_conds::delete $xp $undo
        }
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        GidUtils::UpdateWindow GROUPS
    }
    if { $actualizeconds } { actualize_conditions_window }
}

proc gid_groups_conds::give_group_name_auto_from_bc { domNodeCnd } {
    variable doc

    set root [$doc documentElement]
    set group_type [$domNodeCnd @group_type ""]
    if { $group_type ne "" } { return $group_type }
    return [$root selectNodes {string(groups_types/group_type[@auto_from_bc='1']/@pn)}]
}

proc gid_groups_conds::check_groups_auto_from_bc { { groups "" } } {
    variable doc
    if { ![llength $groups] } {
        set groups [gid_groups_conds::give_groups]
    }
    set root [$doc documentElement]
    set xpath {string(groups_types/group_type[@auto_from_bc='1']/@pn)}
    set auto_from_bc_group [$root selectNodes $xpath]
    if { $auto_from_bc_group eq "" } { return }
    set xpath {string(groups_types/group_type[@default='1']/@pn)}
    set default_group [$root selectNodes $xpath]
    if { $default_group eq "" } {
        set default_group [lindex [gid_groups_conds::give_groups_types] 0]
    }
    foreach group $groups {
        set num_defaults 0
        set group_types ""
        set xp [format_xpath {//condition[not(ancestor::blockdata[@n="Internal data"])]/group[@n=%s]} $group]
        append xp [format_xpath {|//condition[not(ancestor::blockdata[@n="Internal data"])]/groupList[group/@n=%s]} $group]
        foreach domNodeCnd [$root selectNodes $xp] {
            if { [$domNodeCnd @group_type ""] ne "" } {
                lappend group_types [$domNodeCnd @group_type]
            } else {
                incr num_defaults
            }
        }
        set group_types [lsort -unique $group_types]
        if { [llength $group_types] } {
            if { [llength $group_types] > 1 || $num_defaults } {
                set newgroup [lindex $group_types 0]...
            } else {
                set newgroup [lindex $group_types 0]
            }
        } elseif { $num_defaults } {
            set newgroup $auto_from_bc_group
        } else {
            set newgroup $default_group
        }
        gid_groups_conds::change_group_att -noactualizegroups $group type $newgroup
    }
    actualize_groups_window
}

proc gid_groups_conds::check_groups_entities { { raise_error "" } } {
    variable doc
    variable is_draw_post

    if { $is_draw_post } { return }

    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        set groups [lsort [GiD_Groups list]]
    } else {
        set groups ""
        foreach i [list point line surface volume] {
            foreach geom_mesh [list geometry mesh] {
                foreach j [GiD_Info conditions ${i}_groups $geom_mesh] {
                    foreach "- num - group" $j break
                    if { [lsearch -exact -sorted $groups $group] == -1 } {
                        lappend groups $group
                        set groups [lsort $groups]
                    }
                }
            }
        }
    }
    set root [$doc documentElement]
    set lostgroups ""
    foreach group $groups {
        if { ![gid_groups_conds::group_exists $group] } {
            lappend lostgroups $group
            gid_groups_conds::add_new_group -exactname -permit_extended_name 1 $group
        }
    }

    if { $lostgroups ne "" } {
        set txt [_ "Lost groups: '%s'. Fixed" $lostgroups]
        if { $raise_error ne "" } {
            error $txt
        } else {
            WarnWin $txt
        }
    }
}

proc gid_groups_conds::add_entities_to_groups { entities gList ovtype } {
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        foreach i $gList {
            lassign $i - group
            GiD_Process MEscape Utilities EntitiesGroups Assign $group ${ovtype}s {*}$entities escape
        }
    } else {
        GidUtils::DisableGraphics
        GiD_Process MEscape Data Conditions AssignCond ${ovtype}_groups NoRepeatField name
        foreach i $gList {
            lassign $i - group
            GiD_Process Change $group {*}$entities
        }
        GiD_Process escape
        GidUtils::EnableGraphics
    }
    return
}

proc gid_groups_conds::delete_entities_from_groups { entities gList ovtype } {
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        foreach i $gList {
            lassign $i - group
            GiD_Process MEscape Utilities EntitiesGroups Unassign $group ${ovtype}s {*}$entities escape
        }
    } else {
        GidUtils::DisableGraphics
        GiD_Process MEscape Data Conditions AssignCond ${ovtype}_groups UnAssign
        foreach i $gList {
            lassign $i - group
            GiD_Process Field name $group {*}$entities
        }
        GiD_Process escape
        GidUtils::EnableGraphics
    }
}


proc gid_groups_conds::delete_entities_from_layers { entities gList ovtype } {

    #     GidUtils::DisableGraphics
    #     GiD_Process MEscape Data Conditions AssignCond ${ovtype}_groups UnAssign
    #     foreach i $gList {
    #         lassign $i - group
    #         GiD_Process Field name $group {*}$entities
    #     }
    #     GiD_Process escape
    #     GidUtils::EnableGraphics
}

################################################################################
#    List entities
################################################################################

proc gid_groups_conds::_simple_contextualmenu { w tree items x y } {

    catch { destroy $tree._cmenu }
    set menu [menu $tree._cmenu -tearoff 0]

    $menu add command -label [_ "Select all"] \
        -command [list gid_groups_conds::_list_entities_contextualmenu_cmd \
        $tree selectall]
    $menu add separator
    $menu add command -label [_ "Copy"] \
        -command [list gid_groups_conds::_list_entities_contextualmenu_cmd \
        $tree copy]
    tk_popup $menu $x $y
}

proc gid_groups_conds::_list_entities_contextualmenu { w tree items x y } {

    catch { destroy $tree._cmenu }
    set menu [menu $tree._cmenu -tearoff 0]

    $menu add command -label [_ "Select all"] \
        -command [list gid_groups_conds::_list_entities_contextualmenu_cmd \
        $tree selectall]
    $menu add separator
    $menu add command -label [_ "Copy"] \
        -command [list gid_groups_conds::_list_entities_contextualmenu_cmd \
        $tree copy]

    $menu add separator
    if { [gid_groups_conds::report_can_add] } {
        $menu add command -label [_ "Add to report"] -command \
            [list gid_groups_conds::_list_entities_cnd_output $w report]
    }
    $menu add command -label [_ "Export to file"] -command \
        [list gid_groups_conds::_list_entities_cnd_output $w file]

    tk_popup $menu $x $y
}

proc gid_groups_conds::_list_entities_no_group_contextualmenu { w tree items x y } {

    catch { destroy $tree._cmenu }
    set menu [menu $tree._cmenu -tearoff 0]

    $menu add command -label [_ "Select all"] \
        -command [list gid_groups_conds::_list_entities_contextualmenu_cmd \
        $tree selectall]
    $menu add separator
    $menu add command -label [_ "Copy"] \
        -command [list gid_groups_conds::_list_entities_contextualmenu_cmd \
        $tree copy]

    $menu add separator
    if { [gid_groups_conds::report_can_add] } {
        $menu add command -label [_ "Add to report"] -command \
            [list gid_groups_conds::_list_entities_no_group_output $w report]
    }
    $menu add command -label [_ "Export to file"] -command \
        [list gid_groups_conds::_list_entities_no_group_output $w file]

    tk_popup $menu $x $y
}

proc gid_groups_conds::_list_entities_contextualmenu_cmd { tree what } {

    switch $what {
        selectall {
            $tree selection add all
        }
        copy {
            set data ""
            foreach i [$tree selection get] {
                lappend data [join [$tree item text $i] \t]
            }
            clipboard clear
            clipboard append [join $data \n]
        }
    }
}

proc gid_groups_conds::_list_entities_editbegin { columnList col0 tree item col } {
    variable doc

    if { [lsearch -exact [$tree item state forcolumn $item $col] disabled] != -1 } { return 0 }

    if { $col == 3 } {
        set values ""
        foreach node [$doc selectNodes /*/groups/group] {
            lappend values [$node @n]
        }
        #return [list combo 1 $values]
        set values_tree ""
        foreach i $values {
            lappend values_tree [list 0 $i $i "" 1]
        }
        return [list combotree 1 $values_tree]
    }
    set valueNode [lindex $columnList [expr {$col-$col0}] 3]
    set values [split [get_domnode_attribute $valueNode values] ,]
    set values_tree [split [get_domnode_attribute $valueNode values_tree] ,]
    set editable [$valueNode @editable 0]

    set dict ""
    foreach "ni vi" [split [get_domnode_attribute $valueNode dict] ,] {
        dict set dict $ni [_ $vi]
    }

    if { [llength $values] } {
        #return [list combo $editable $values $dict]
        set values_tree ""
        foreach i $values {
            lappend values_tree [list 0 $i $i "" 1]
        }
        return [list combotree $editable $values_tree]
    } elseif { [llength $values_tree] } {
        set idx 0
        set imgs ""
        foreach i $values_tree {
            if { [dict exists $imgs [lindex $i 3]] } {
                set img [dict get $imgs [lindex $i 3]]
            } else {
                set img [giveimageL [lindex $i 3]]
                dict set imgs [lindex $i 3] $img
            }
            lset values_tree $idx 3 $img
            incr idx
        }
        return [list combotree $editable $values_tree]
    } else {
        return entry
    }
}

proc gid_groups_conds::_list_entities_editaccept { columnList col0 tree item col value } {
    variable list_entities_cnd_data

    set item [$tree item id $item]

    if { ![info exists list_entities_cnd_data($item)] } {
        set list_entities_cnd_data($item) [$tree item text $item]
    }
    $tree item text $item $col $value

    if { $list_entities_cnd_data($item) eq [$tree item text $item] } {
        unset list_entities_cnd_data($item)
    }

    if { $col<$col0 } { return }
    set valueNode [lindex $columnList [expr {$col-$col0}] 3]

    set valueNodes ""
    foreach i $columnList { lappend valueNodes [lindex $i 3] }

    set defaults_d ""
    dict set defaults_d state normal

    foreach depNode [give_dependencies_nodes -value $value $valueNode] {
        set tomodNodes [$valueNode selectNodes [$depNode @node]]
        if { [llength $tomodNodes] == 0 } {
            if { [$depNode @check 1] == 0 } { continue }
            error "error in _list_entities_editaccept. xpath='[$depNode @node]'"
        }
        $depNode setAttribute v $value
        foreach tomodNode $tomodNodes {
            if { [$tomodNode nodeName] eq "condition" } { continue }
            set ipos [lsearch -exact $valueNodes $tomodNode]
            if { $ipos == -1 } { continue }
            set icol [expr {$col0+$ipos}]
            set changed_state 0

            if {[$depNode hasAttribute att]} {
                set iiList [list 1]
            } else {
                set iiList [list 1 2]
            }

            foreach i $iiList {
                if { [$depNode hasAttribute att$i] } {
                    set v [get_domnode_attribute $depNode v$i]
                    set depatt [$depNode @att$i]
                } elseif { [$depNode hasAttribute att] } {
                    set v [get_domnode_attribute $depNode v]
                    set depatt [$depNode @att]
                } else {
                    set v [get_domnode_attribute $depNode v]
                    set depatt "v"
                }
                switch $depatt {
                    state {
                        switch $v {
                            normal { set state !disabled }
                            default { set state disabled }
                        }
                        $tree item state forcolumn $item $icol $state
                        set changed_state 1
                    }
                    v {
                        $tree item text $item $icol $v
                    }
                    units {
                        set units0 [lindex $columnList $ipos 2]
                        set units $v
                        set v [$tree item text $item $icol]
                        set unit_magnitude [get_domnode_attribute $tomodNode unit_magnitude]
                        if { [string is double -strict $v] } {
                            set v [gid_groups_conds::convert_unit_value $unit_magnitude \
                                $v $units $units0]
                        }
                        $tree item text $item $icol $v
                    }
                }
            }
            if { !$changed_state } {
                switch [get_domnode_attribute $tomodNode state] {
                    normal - "" { set state !disabled }
                    default { set state disabled }
                }
                $tree item state forcolumn $item $icol $state
            }
        }
        $depNode removeAttribute v
    }
}

proc gid_groups_conds::_list_entities_no_group_editbegin { columnList tree item col } {
    variable doc

    if { [lsearch -exact [$tree item state forcolumn $item $col] disabled] != -1 } { return 0 }

    #     if { $col == 3 } {
    #         set values ""
    #         foreach node [$doc selectNodes /*/groups/group] {
    #             lappend values [$node @n]
    #         }
    #         #return [list combo 1 $values]
    #         set values_tree ""
    #         foreach i $values {
    #             lappend values_tree [list 0 $i $i "" 1]
    #         }
    #         return [list combotree 1 $values_tree]
    #     }
    set valueNode [lindex $columnList $col 3]
    set values [split [get_domnode_attribute $valueNode values] ,]
    set values_tree [split [get_domnode_attribute $valueNode values_tree] ,]
    set editable [$valueNode @editable 0]

    set dict ""
    foreach "ni vi" [split [get_domnode_attribute $valueNode dict] ,] {
        dict set dict $ni [_ $vi]
    }

    if { [llength $values] } {
        #return [list combo $editable $values $dict]
        set values_tree ""
        foreach i $values {
            set err [catch {set i [dict get $dict $i]} errstring]
            lappend values_tree [list 0 $i $i "" 1]
        }
        return [list combotree $editable $values_tree]
    } elseif { [llength $values_tree] } {
        set idx 0
        set imgs ""
        foreach i $values_tree {
            if { [dict exists $imgs [lindex $i 3]] } {
                set img [dict get $imgs [lindex $i 3]]
            } else {
                set img [giveimageL [lindex $i 3]]
                dict set imgs [lindex $i 3] $img
            }
            lset values_tree $idx 3 $img
            incr idx
        }
        return [list combotree $editable $values_tree]
    } else {
        return entry
    }
}

proc gid_groups_conds::_list_entities_no_group_editaccept { columnList tree item col value } {
    variable list_entities_no_group_data

    set item [$tree item id $item]

    if { ![info exists list_entities_no_group_data($item)] } {
        set list_entities_no_group_data($item) [$tree item text $item]
    }
    $tree item text $item $col $value

    if { $list_entities_no_group_data($item) eq [$tree item text $item] } {
        unset list_entities_no_group_data($item)
    }


    set valueNode [lindex $columnList $col 3]

    set valueNodes ""
    foreach i $columnList { lappend valueNodes [lindex $i 3] }

    set defaults_d ""
    dict set defaults_d state normal

    set col0 0

    foreach depNode [give_dependencies_nodes -value $value $valueNode] {
        set tomodNodes [$valueNode selectNodes [$depNode @node]]
        if { [llength $tomodNodes] == 0 } {
            if { [$depNode @check 1] == 0 } { continue }
            error "error in _list_entities_editaccept. xpath='[$depNode @node]'"
        }
        $depNode setAttribute v $value
        foreach tomodNode $tomodNodes {
            #             if { [$tomodNode nodeName] eq "condition" } { continue }
            set ipos [lsearch -exact $valueNodes $tomodNode]
            if { $ipos == -1 } { continue }
            set icol [expr {$col0+$ipos}]
            set changed_state 0

            if {[$depNode hasAttribute att]} {
                set iiList [list 1]
            } else {
                set iiList [list 1 2]
            }

            foreach i $iiList {
                if { [$depNode hasAttribute att$i] } {
                    set v [get_domnode_attribute $depNode v$i]
                    set depatt [$depNode @att$i]
                } elseif { [$depNode hasAttribute att] } {
                    set v [get_domnode_attribute $depNode v]
                    set depatt [$depNode @att]
                } else {
                    set v [get_domnode_attribute $depNode v]
                    set depatt "v"
                }
                switch $depatt {
                    state {
                        switch $v {
                            normal { set state !disabled }
                            default { set state disabled }
                        }
                        $tree item state forcolumn $item $icol $state
                        set changed_state 1
                    }
                    v {
                        $tree item text $item $icol $v
                    }
                    units {
                        set units0 [lindex $columnList $ipos 2]
                        set units $v
                        set v [$tree item text $item $icol]
                        set unit_magnitude [get_domnode_attribute $tomodNode unit_magnitude]
                        if { [string is double -strict $v] } {
                            set v [gid_groups_conds::convert_unit_value $unit_magnitude \
                                $v $units $units0]
                        }
                        $tree item text $item $icol $v
                    }
                }
            }
            if { !$changed_state } {
                switch [get_domnode_attribute $tomodNode state] {
                    normal - "" { set state !disabled }
                    default { set state disabled }
                }
                $tree item state forcolumn $item $icol $state
            }
        }
        $depNode removeAttribute v
    }
}

proc gid_groups_conds::list_entities_cnd_switch_view { w tree_dict cndList what } {
    variable list_entities_cnd_data

    set tree [$w give_uservar_value tree]

    switch $what {
        tree {
            $tree configure -treecolumn 1 -showlines 1 -showbuttons 1
            $tree column configure 0 -visible 0
            $tree column configure 1 -width 200

            foreach item [$tree item children root] {
                if { [$tree item state get $item condition] } {
                    $tree item configure $item -visible 1
                    foreach item_in [$tree item children $item] {
                        $tree item configure $item_in -visible 1
                    }
                } elseif { [$tree item state get $item group] } {
                    $tree item configure $item -visible 1
                } else {
                    if { [info exists list_entities_cnd_data($item)] } {
                        foreach "cndName - - group" $list_entities_cnd_data($item) break
                    } else {
                        foreach "cndName - - group" [$tree item text $item] break
                    }
                    if { [string trim $cndName] eq "" } {
                        set idp 0
                    } else {
                        set idp [dict get $tree_dict $cndName]
                    }
                    set idp1 [dict get $tree_dict $idp $group]
                    $tree item lastchild $idp1 $item
                }
            }
        }
        list {
            if { [llength $cndList] > 1 } {
                $tree column configure 0 -visible 1
            } else {
                $tree column configure 0 -visible 0
            }
            $tree column configure 1 -width 40
            $tree configure -showlines 0 -showbuttons 0

            foreach item [$tree item children root] {
                $tree item configure $item -visible 0
                set prev $item
                if { [$tree item state get $item condition] } {
                    foreach item_in [$tree item children $item] {
                        $tree item configure $item_in -visible 0
                        foreach item_in_in [$tree item children $item_in] {
                            $tree item nextsibling $prev $item_in_in
                            set prev $item_in_in
                        }
                    }
                } else {
                    foreach item_in [$tree item children $item] {
                        $tree item nextsibling $prev $item_in
                        set prev $item_in
                    }
                }
            }
        }
    }
}

proc gid_groups_conds::list_entities_no_group_switch_view { w tree_dict cndList what } {
    variable list_entities_cnd_data

    set tree [$w give_uservar_value tree]

    return

    switch $what {
        tree {
            $tree configure -treecolumn 1 -showlines 1 -showbuttons 1
            $tree column configure 0 -visible 0
            $tree column configure 1 -width 200

            foreach item [$tree item children root] {
                if { [$tree item state get $item container] } {
                    $tree item configure $item -visible 1
                    foreach item_in [$tree item children $item] {
                        $tree item configure $item_in -visible 1
                    }
                } elseif { [$tree item state get $item group] } {
                    $tree item configure $item -visible 1
                } else {
                    if { [info exists list_entities_cnd_data($item)] } {
                        foreach "cndName - - group" $list_entities_cnd_data($item) break
                    } else {
                        foreach "cndName - - group" [$tree item text $item] break
                    }
                    if { [string trim $cndName] eq "" } {
                        set idp 0
                    } else {
                        set idp [dict get $tree_dict $cndName]
                    }
                    set idp1 [dict get $tree_dict $idp $group]
                    $tree item lastchild $idp1 $item
                }
            }
        }
        list {
            if { [llength $cndList] > 1 } {
                $tree column configure 0 -visible 1
            } else {
                $tree column configure 0 -visible 0
            }
            $tree column configure 1 -width 40
            $tree configure -showlines 0 -showbuttons 0

            foreach item [$tree item children root] {
                $tree item configure $item -visible 0
                set prev $item
                if { [$tree item state get $item condition] } {
                    foreach item_in [$tree item children $item] {
                        $tree item configure $item_in -visible 0
                        foreach item_in_in [$tree item children $item_in] {
                            $tree item nextsibling $prev $item_in_in
                            set prev $item_in_in
                        }
                    }
                } else {
                    foreach item_in [$tree item children $item] {
                        $tree item nextsibling $prev $item_in
                        set prev $item_in
                    }
                }
            }
        }
    }
}

proc gid_groups_conds::list_entities_give_data_data { domNodes geom_mesh columnList groupFullList } {
    variable is_draw_post
    set data ""
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        if { !$is_draw_post } {
            if { $geom_mesh == "geometry" } {
                foreach ov [list point line surface volume] {
                    foreach group [GiD_Groups list] {
                        set gidx [lsearch -sorted -index 0 $groupFullList $group]
                        if { $gidx == -1 } { continue }
                        set cndName [lindex $groupFullList $gidx 1]
                        set d [lindex $groupFullList $gidx 2]
                        foreach num [GiD_EntitiesGroups get $group ${ov}s] {
                            set l [list $cndName $num $ov $group]
                            foreach col $columnList {
                                set n [lindex $col 0]
                                if { [dict exists $d $n] } {
                                    lappend l [lindex [dict get $d $n] 0]
                                } else {
                                    lappend l ""
                                }
                            }
                            lappend data $l
                        }
                    }
                }
            } elseif { $geom_mesh == "mesh" } {
                foreach ov [list node element] {
                    foreach group [GiD_Groups list] {
                        set gidx [lsearch -sorted -index 0 $groupFullList $group]
                        if { $gidx == -1 } { continue }
                        set cndName [lindex $groupFullList $gidx 1]
                        set d [lindex $groupFullList $gidx 2]
                        foreach num [GiD_EntitiesGroups get $group ${ov}s] {
                            set l [list $cndName $num $ov $group]
                            foreach col $columnList {
                                set n [lindex $col 0]
                                if { [dict exists $d $n] } {
                                    lappend l [lindex [dict get $d $n] 0]
                                } else {
                                    lappend l ""
                                }
                            }
                            lappend data $l
                        }
                    }
                }
                set ov face
                foreach group [GiD_Groups list] {
                    set gidx [lsearch -sorted -index 0 $groupFullList $group]
                    if { $gidx == -1 } { continue }
                    set cndName [lindex $groupFullList $gidx 1]
                    set d [lindex $groupFullList $gidx 2]
                    set entity_ids [GiD_EntitiesGroups get $group ${ov}s]
                    foreach element_id [lindex $entity_ids 0] face_id [lindex $entity_ids 1] {
                        set num "$element_id - $face_id"
                        set l [list $cndName $num $ov $group]
                        foreach col $columnList {
                            set n [lindex $col 0]
                            if { [dict exists $d $n] } {
                                lappend l [lindex [dict get $d $n] 0]
                            } else {
                                lappend l ""
                            }
                        }
                        lappend data $l
                    }
                }
            } else {
                W "gid_groups_conds::list_entities_give_data_data. unexpected geom_mesh=$geom_mesh"
            }
        }
    } else {
        foreach ovG [list point line surface volume] {
            set entsDict ""
            if { !$is_draw_post } {
                # WARNING: what happens with element's faces?
                set err [catch {set info_cond [GiD_Info conditions ${ovG}_groups $geom_mesh]}]
                if { $err } {
                    W "The model needs to be upgraded."
                    break
                }
                foreach j [GiD_Info conditions ${ovG}_groups $geom_mesh] {
                    lassign $j num face - group
                    if { $num in "E N" } {
                        set num $face
                    } else {
                        set num "$num - $face"
                    }
                    if { [lsearch -sorted -index 0 $groupFullList $group] == -1 } { continue }
                    dict lappend entsDict $group $num
                }
            } else {
                foreach grFull $groupFullList {
                    lassign $grFull group
                    set gr [ggc::manage_geometry_groups id $group]
                    dict set entsDict $group [ggc::manage_geometry_groups entities list_group $gr $ovG]
                }
            }
            switch $geom_mesh {
                geometry { set type $ovG }
                mesh {
                    switch $ovG {
                        point { set type node }
                        default { set type element }
                    }
                }
            }
            dict for "group numList" $entsDict {
                foreach num $numList {
                    foreach gidx [lsearch -sorted -all -index 0 $groupFullList $group] {
                        set cndName [lindex $groupFullList $gidx 1]
                        set l [list $cndName $num $type $group]
                        set d [lindex $groupFullList $gidx 2]
                        foreach col $columnList {
                            set n [lindex $col 0]
                            if { [dict exists $d $n] } {
                                lappend l [lindex [dict get $d $n] 0]
                            } else {
                                lappend l ""
                            }
                        }
                        lappend data $l
                    }
                }
            }
        }
    }
    return $data
}

proc gid_groups_conds::list_entities_give_data_container { domNodes geom_mesh columnList groupFullList dataList } {

    set data ""
    set valList ""
    foreach idata $dataList {
        lassign $idata n v state
        if {$state eq "hidden"} { continue }
        foreach icol $columnList {
            lassign $icol n_col pn_col - node_col
            if {$n_col ne $n} { continue }
            set dict ""
            foreach "ni vi" [split [get_domnode_attribute $node_col dict] ,] {
                dict set dict $ni [= $vi]
            }
            set err [catch {set val [dict get $dict $v]} errstring]
            if {!$err} {
                lappend valList $val
            } else {
                lappend valList $v
            }
        }
    }

    lappend data $valList

    return $data
}

proc gid_groups_conds::list_entities_give_data_header { domNodes geom_mesh } {
    variable is_draw_post

    ################################################################################
    #    intermediate
    ################################################################################

    set groupNodes ""
    foreach domNode $domNodes {
        switch [$domNode nodeName] {
            blockdata {
                set xp {.//condition/group|.//condition/groupList/group}
                eval lappend groupNodes [$domNode selectNodes $xp]
            }
            condition {
                set xp {group|groupList/group}
                eval lappend groupNodes [$domNode selectNodes $xp]
            }
            container {
                set xp {.//condition/group|.//condition/groupList/group}
                set ret [$domNode selectNodes $xp]
                if { ![llength $ret] } {
                    set xp {ancestor::condition/group|ancestor::condition/groupList/group}
                    set ret [$domNode selectNodes $xp]
                }
                eval lappend groupNodes $ret
            }
            group {
                lappend groupNodes $domNode
            }
            value - groupList {
                set xp {ancestor::condition/group|ancestor::condition/groupList/group}
                eval lappend groupNodes [$domNode selectNodes $xp]
            }
        }
    }
    foreach "columnList cndList groupFullList" [list "" "" ""] break
    foreach groupNode $groupNodes {
        set cndNode [$groupNode selectNodes {ancestor::condition}]
        if { $cndNode ne "" } {
            set cndName [get_domnode_attribute $cndNode pn]
            set bdNode [$cndNode selectNodes {ancestor::blockdata[@sequence='1']}]
            if { $bdNode ne "" } {
                set bdName [get_domnode_attribute $bdNode name]
                append cndName " - $bdName"
            }
            if { [lsearch -exact $cndList $cndName] == -1 } {
                lappend cndList $cndName
            }
        } else {
            set cndName ""
        }
        set valuesDict ""
        foreach valueNode [$groupNode selectNodes .//value] {
            foreach i "n pn v units unit_magnitude state" {
                set $i [get_domnode_attribute $valueNode $i]
            }
            set ipos [lsearch -exact -index 0 $columnList $n]
            if { $ipos == -1 } {
                if { $units ne "" } {
                    set unitN [gid_groups_conds::units_to_nice_units $units]
                    set pn "$pn ($unitN)"
                }
                lappend columnList [list $n $pn $units $valueNode]
                set ipos [expr {[llength $columnList]-1}]
            }
            set unitsG [lindex $columnList $ipos 2]

            if { $units ne $unitsG } {
                set v [gid_groups_conds::convert_unit_value $unit_magnitude \
                    $v $units $unitsG]
            }
            dict set valuesDict $n [list $v $state]
        }
        lappend groupFullList [list [$groupNode @n] $cndName $valuesDict]
    }

    set groupFullList [lsort -index 0 $groupFullList]
    return [list $columnList $cndList $groupFullList]
}

proc gid_groups_conds::list_entities_container_give_data_header { domNodes geom_mesh } {
    variable is_draw_post

    foreach "columnList cndList groupFullList dataList" [list "" "" ""] break
    foreach domNode $domNodes {
        set cndName ""
        set valuesDict ""
        foreach valueNode [$domNode selectNodes .//value] {
            foreach i "n pn v units unit_magnitude state" {
                set $i [get_domnode_attribute $valueNode $i]
            }
            if {$state eq "hidden"} { continue }
            set ipos [lsearch -exact -index 0 $columnList $n]
            if { $ipos == -1 } {
                if { $units ne "" } {
                    set unitN [gid_groups_conds::units_to_nice_units $units]
                    set pn "$pn ($unitN)"
                }
                lappend columnList [list $n $pn $units $valueNode]
                set ipos [expr {[llength $columnList]-1}]
            }
            set unitsG [lindex $columnList $ipos 2]

            if { $units ne $unitsG } {
                set v [gid_groups_conds::convert_unit_value $unit_magnitude \
                    $v $units $unitsG]
            }
            dict set valuesDict $n [list $v $state]
            lappend dataList "$n [list $v] $state"
        }
        lappend groupFullList [list [$domNode @n] $cndName $valuesDict]
    }

    set groupFullList [lsort -index 0 $groupFullList]
    return [list $columnList $cndList $groupFullList $dataList]
}

proc gid_groups_conds::list_entities_give_data_full { domNodes geom_mesh } {
    lassign [gid_groups_conds::list_entities_give_data_header $domNodes $geom_mesh] \
        columnList cndList groupFullList
    set data [gid_groups_conds::list_entities_give_data_data $domNodes $geom_mesh $columnList $groupFullList]
    set  headers [list [_ "Condition"] [_ "Entity"] [_ "Type"] [_ "Group"]]
    foreach j $columnList {
        lappend headers [lindex $j 1]
    }
    return [list $headers $cndList $data]
}

proc gid_groups_conds::list_entities_give_data_no_group_full { domNodes geom_mesh } {
    lassign [gid_groups_conds::list_entities_container_give_data_header $domNodes $geom_mesh] \
        columnList cndList groupFullList dataList
    set data [gid_groups_conds::list_entities_give_data_container $domNodes $geom_mesh $columnList $groupFullList $dataList]
    set  headers ""
    foreach j $columnList {
        lappend headers [lindex $j 1]
    }
    return [list $headers $cndList $data]
}

proc gid_groups_conds::list_entities_no_group { args } {
    variable list_entities_no_group_data
    variable is_draw_post
    variable name_page

    set optional {
        { -geom_mesh geom|mesh|auto auto }
    }
    set compulsory "wp domNodes"
    parse_args $optional $compulsory $args

    if { $geom_mesh eq "auto" } {
        set geom_mesh [is_geometry_or_mesh]
    }
    lassign [gid_groups_conds::list_entities_container_give_data_header \
        $domNodes $geom_mesh] columnList cndList groupFullList dataList

    if { !$is_draw_post } {
        set okname [_ "Ok"]
    } else {
        set okname -
    }
    set nName [$domNodes @n ""]
    set pnName [$domNodes @pn ""]
    set nameName [$domNodes @name ""]
    if {$pnName eq ""} {
        set name_page $nameName
    } else {
        set name_page $pnName
    }

    destroy $wp._ask
    set w [dialogwin_snit $wp._ask -title [_ "List entities"] \
        -callback [list gid_groups_conds::list_entities_no_group_do $domNodes $columnList] \
        -okname $okname -cancelname [_ "Cancel"] -transient 1 -grab 1]
    set f [$w giveframe]

    #     foreach "ni vi" [split [get_domnode_attribute $valueNode dict] ,] {
    #         dict set dict $ni [= $vi]
    #     }
    #     set values [split [get_domnode_attribute $valueNode values] ,]
    #     set value [get_domnode_attribute $valueNode v]

    set columns ""
    foreach i $columnList {
        lappend columns [list 18 [lindex $i 1] left text 1]
    }

    package require fulltktree
    fulltktree $f.tree -width 600 -height 60 \
        -columns $columns -expand 0 \
        -selectmode extended -showlines 0 -indent 0 \
        -sensitive_cols all \
        -bd 1 -relief solid \
        -have_search_button automatic \
        -contextualhandler [list gid_groups_conds::_list_entities_no_group_contextualmenu $w] \
        -editbeginhandler [list gid_groups_conds::_list_entities_no_group_editbegin $columnList] \
        -editaccepthandler [list gid_groups_conds::_list_entities_no_group_editaccept $columnList]

    $w set_uservar_value tree $f.tree
    $f.tree state define container

    set tree_dict ""
    set data [gid_groups_conds::list_entities_give_data_container \
        $domNodes $geom_mesh $columnList $groupFullList $dataList]

    foreach l $data {
        set ncolend [llength $l]
        if { $nName eq "" } {
            set idp 0
        } elseif { ![dict exists $tree_dict $nName] } {
            set idp [$f.tree insert end $l]
            #             $f.tree item configure $idp -button 1
            $f.tree item state set $idp container
            #             $f.tree item state forcolumn $idp 1 emphasis
            dict set tree_dict $nName $idp
        } else {
            set idp [dict get $tree_dict $nName]
        }
        set idx 0
        foreach id $dataList {
            lassign $id n v state
            if { $state eq "disabled" } {
                $f.tree item state forcolumn $idp $idx disabled
            }
            incr idx
        }

        #         set gidx [lsearch -sorted -index 0 $groupFullList $group]
        #         set d [lindex $groupFullList $gidx 2]
        #
        #         if { $cndName eq "" } {
        #             set idp 0
        #         } elseif { ![dict exists $tree_dict $cndName] } {
        #             set l0 [list "" $cndName]
        #             eval lappend l0 [lrepeat $ncolend ""]
        #             set idp [$f.tree insert end $l0]
        #             $f.tree item configure $idp -button 1
        #             $f.tree item state set $idp condition
        #             $f.tree item state forcolumn $idp 1 emphasis
        #             foreach idx [range 2 $ncolend] {
        #                 $f.tree item state forcolumn $idp $idx disabled
        #             }
        #             dict set tree_dict $cndName $idp
        #         } else {
        #             set idp [dict get $tree_dict $cndName]
        #         }
        #
        #         if { ![dict exists $tree_dict $idp $group] } {
        #             set l0 [list "" $group]
        #             eval lappend l0 [lrepeat $ncolend ""]
        #             set idp1 [$f.tree insert end $l0 $idp]
        #             $f.tree item configure $idp1 -button 1
        #             $f.tree item state set $idp1 group
        #             $f.tree item state forcolumn $idp1 1 emphasis
        #             foreach idx [range 2 $ncolend] {
        #                 $f.tree item state forcolumn $idp1 $idx disabled
        #             }
        #             dict set tree_dict $idp $group $idp1
        #         } else {
        #             set idp1 [dict get $tree_dict $idp $group]
        #         }
        #
        #         $f.tree insert end $l $idp1
        #
        #         set idx $col0
        #         foreach col $columnList {
        #             set n [lindex $col 0]
        #             if { ![dict exists $d $n] } {
        #                 $f.tree item style set end $idx full_disabled
        #             } elseif { [lindex [dict get $d $n] 1] eq "disabled" } {
        #                 $f.tree item state forcolumn end $idx disabled
        #             }
        #             incr idx
        #         }
    }
    #     set fvs [ttk::labelframe $f.fvs -text [_ "View style:"]]
    #     ttk::radiobutton $fvs.r1 -text [_ "Tree"] -variable [$w give_uservar view_style] \
    #         -value tree
    #     register_popup_help $fvs.r1 [_ "View data as tree"]
    #     ttk::radiobutton $fvs.r2 -text [_ "Table"] -variable [$w give_uservar view_style] \
    #         -value list
    #     register_popup_help $fvs.r2 [_ "View data as table"]

    $w set_uservar_value domNodes $domNodes
    $w set_uservar_value geom_mesh $geom_mesh

    #     $w add_traceN_to_uservar view_style [list gid_groups_conds::list_entities_no_group_switch_view \
    #             $w $tree_dict $cndList]
    #     $w set_uservar_value view_style list

    set fu [ttk::labelframe $f.fu -text [_ "Utilities:"]]

    if { [gid_groups_conds::report_can_add] } {
        cu::menubutton_button $fu.mb -text [_ "Add to report"] -width 12 -menu $fu.mb.mr \
            -command  [list gid_groups_conds::_list_entities_no_group_output $w report]
        menu $fu.mb.mr -tearoff 0
        $fu.mb.mr add command -label [_ "Add to report"] -command \
            [list gid_groups_conds::_list_entities_no_group_output $w report]
        $fu.mb.mr add command -label [_ "Export to file"] -command \
            [list gid_groups_conds::_list_entities_no_group_output $w file]
        set help [_ "Put data from the table into a report"]
        register_popup_help $fu.mb $help
    } else {
        ttk::button $fu.mb -text  [_ "Export to file"] -command \
            [list gid_groups_conds::_list_entities_no_group_output $w file]
    }
    grid $f.tree - - - -sticky nsew -padx 2 -pady 2
    grid $fu -sticky nsew -padx 2 -pady 2
    #     grid $fvs.r1 $fvs.r2 -sticky w -padx 2 -pady 2
    grid $fu.mb -sticky w -padx 2 -pady 2
    grid configure $fu.mb -padx 2 -pady 2
    grid columnconfigure $f 0 -weight 1
    grid rowconfigure $f 0 -weight 1

    tk::TabToWindow $f.tree
    bind $w <Return> [list $w invokeok]
    $w createwindow
}

proc gid_groups_conds::list_entities_cnd { args } {
    variable list_entities_cnd_data
    variable is_draw_post

    set optional {
        { -geom_mesh geom|mesh|auto auto }
    }
    set compulsory "wp domNodes"
    parse_args $optional $compulsory $args

    if { $geom_mesh eq "auto" } {
        set geom_mesh [is_geometry_or_mesh]
    }
    lassign [gid_groups_conds::list_entities_give_data_header $domNodes $geom_mesh] columnList cndList groupFullList

    if { !$is_draw_post } {
        set okname [_ "Ok"]
    } else {
        set okname -
    }
    destroy $wp._ask
    set w [dialogwin_snit $wp._ask -title [_ "List groups"] \
        -callback [list gid_groups_conds::list_entities_cnd_do $columnList] \
        -okname $okname -cancelname [_ "Cancel"] -transient 1 -grab 1]
    set f [$w giveframe]

    set columns [list \
        [list 15 [_ "Condition"] left text 0] \
        [list 5 [_ "Entity"] left text 0] \
        [list 6 [_ "Type"] left text 0] \
        [list 15 [_ "Group"] left text 1] \
    ]
set col0 4
foreach i $columnList {
    lappend columns [list 12 [lindex $i 1] left text 1]
}
package require fulltktree
# GiD change: added option -showlines 0
fulltktree $f.tree -width 700 -height 500 \
    -columns $columns -expand 0 \
    -selectmode extended -showlines 0 -indent 0 \
    -sensitive_cols all \
    -bd 1 -relief solid \
    -have_search_button automatic \
    -contextualhandler [list gid_groups_conds::_list_entities_contextualmenu $w] \
    -editbeginhandler [list gid_groups_conds::_list_entities_editbegin $columnList $col0] \
    -editaccepthandler [list gid_groups_conds::_list_entities_editaccept $columnList $col0]

$w set_uservar_value tree $f.tree
$f.tree state define condition
$f.tree state define group

set tree_dict ""
set data [gid_groups_conds::list_entities_give_data_data $domNodes $geom_mesh $columnList $groupFullList]
foreach l $data {
    lassign $l cndName num type group
    set gidx [lsearch -sorted -index 0 $groupFullList $group]
    set d [lindex $groupFullList $gidx 2]
    set ncolend [expr {[llength $l]-2}]
    if { $cndName eq "" } {
        set idp 0
    } elseif { ![dict exists $tree_dict $cndName] } {
        set l0 [list "" $cndName]
        eval lappend l0 [lrepeat $ncolend ""]
        set idp [$f.tree insert end $l0]
        $f.tree item configure $idp -button 1
        $f.tree item state set $idp condition
        $f.tree item state forcolumn $idp 1 emphasis
        foreach idx [range 2 $ncolend] {
            $f.tree item state forcolumn $idp $idx disabled
        }
        dict set tree_dict $cndName $idp
    } else {
        set idp [dict get $tree_dict $cndName]
    }

    if { ![dict exists $tree_dict $idp $group] } {
        set l0 [list "" $group]
        eval lappend l0 [lrepeat $ncolend ""]
        set idp1 [$f.tree insert end $l0 $idp]
        $f.tree item configure $idp1 -button 1
        $f.tree item state set $idp1 group
        $f.tree item state forcolumn $idp1 1 emphasis
        foreach idx [range 2 $ncolend] {
            $f.tree item state forcolumn $idp1 $idx disabled
        }
        dict set tree_dict $idp $group $idp1
    } else {
        set idp1 [dict get $tree_dict $idp $group]
    }

    $f.tree insert end $l $idp1

    set idx $col0
    foreach col $columnList {
        set n [lindex $col 0]
        if { ![dict exists $d $n] } {
            $f.tree item style set end $idx full_disabled
        } elseif { [lindex [dict get $d $n] 1] eq "disabled" } {
            $f.tree item state forcolumn end $idx disabled
        }
        incr idx
    }
}
set fvs [ttk::labelframe $f.fvs -text [_ "View style:"]]
ttk::radiobutton $fvs.r1 -text [_ "Tree"] -variable [$w give_uservar view_style] \
    -value tree
register_popup_help $fvs.r1 [_ "View data as tree"]
ttk::radiobutton $fvs.r2 -text [_ "Table"] -variable [$w give_uservar view_style] \
    -value list
register_popup_help $fvs.r2 [_ "View data as table"]

#     set f1 [ttk::frame $f.f1]

$w set_uservar_value domNodes $domNodes
$w set_uservar_value geom_mesh $geom_mesh

$w add_traceN_to_uservar view_style [list gid_groups_conds::list_entities_cnd_switch_view \
    $w $tree_dict $cndList]
$w set_uservar_value view_style list

set fu [ttk::labelframe $f.fu -text [_ "Utilities:"]]

cu::menubutton_button $fu.mb1 -text [_ "Signal entity"] -width 12 -menu $fu.mb1.m \
    -command  [list gid_groups_conds::_list_entities_cnd_signal $w signal]
menu $fu.mb1.m -tearoff 0
$fu.mb1.m add command -label [_ "Signal entity"] -command \
    [list gid_groups_conds::_list_entities_cnd_signal $w signal]
$fu.mb1.m add command -label [_ "Label on"] -command \
    [list gid_groups_conds::_list_entities_cnd_signal $w labelon]
$fu.mb1.m add command -label [_ "Label off"] -command \
    [list gid_groups_conds::_list_entities_cnd_signal $w labeloff]
set help [_ "Signaling entities"]
register_popup_help $fu.mb1 $help

if { [gid_groups_conds::report_can_add] } {
    cu::menubutton_button $fu.mb -text [_ "Add to report"] -width 12 -menu $fu.mb.mr \
        -command  [list gid_groups_conds::_list_entities_cnd_output $w report]
    menu $fu.mb.mr -tearoff 0
    $fu.mb.mr add command -label [_ "Add to report"] -command \
        [list gid_groups_conds::_list_entities_cnd_output $w report]
    $fu.mb.mr add command -label [_ "Export to file"] -command \
        [list gid_groups_conds::_list_entities_cnd_output $w file]
    set help [_ "Put data from the table into a report"]
    register_popup_help $fu.mb $help
} else {
    ttk::button $fu.mb -text  [_ "Export to file"] -command \
        [list gid_groups_conds::_list_entities_cnd_output $w file]
}

grid $f.tree - - - -sticky nsew -padx 2 -pady 2
grid $fvs $fu -sticky nsew -padx 2 -pady 2
grid $fvs.r1 $fvs.r2 -sticky w -padx 2 -pady 2
grid $fu.mb1 $fu.mb -sticky w -padx 2 -pady 2
grid configure $fu.mb1 -padx 2 -pady 2
grid columnconfigure $f 0 -weight 1
grid rowconfigure $f 0 -weight 1

tk::TabToWindow $f.tree
bind $w <Return> [list $w invokeok]
$w createwindow
}

proc gid_groups_conds::list_entities_no_group_do { domNodes columnList wp } {
    variable doc
    variable list_entities_no_group_data
    variable name_page

    if { [$wp giveaction] < 1 } {
        unset -nocomplain list_entities_no_group_data
        destroy $wp
        return
    }
    set haschanges 0

    set default_change_values ask
    set do_not_ask 0
    if {[llength $domNodes] == "1"} {
        set do_not_ask 1
    }

    set tree [$wp give_uservar_value tree]
    foreach i [lsort -integer [array names list_entities_no_group_data]] {
        set l_old $list_entities_no_group_data($i)
        set values_old $l_old
        set l [$tree item text $i]
        set values $l

        set question_type ""
        unset -nocomplain opt
        if { $values ne $values_old } {
            set txt [_ "There are changes in values for item '%s'" $name_page]
            set opt(1) [list [_ "Modify item '%s' with new values for all its parameters" \
                $name_page] modify]
            set opt(2) [list [_ "Do not do any action for this entity"] nothing]

            set question_type new_values
            if { !$do_not_ask } {
                set answer $default_change_values
            } else {
                set answer "modify"
            }
        }


        if { $question_type ne "" && $answer eq "ask" } {
            destroy $wp._ask
            set w [dialogwin_snit $wp._ask -title [_ "Action"]]
            set f [$w giveframe]

            ttk::label $f.cl0 -text $txt -wraplength 280 -justify left
            grid $f.cl0 -sticky nw -padx "10 2" -pady "2 2"

            set columns ""
            lappend columns [list 18 "" left text 1]
            foreach j $columnList {
                lappend columns [list 18 [lindex $j 1] left text 1]
            }
            fulltktree $f.tree -width 500 -height 60 \
                -columns $columns -expand 0 \
                -sensitive_cols all \
                -selectmode extended -showlines 0 -indent 0 \
                -contextualhandler [list gid_groups_conds::_list_entities_no_group_contextualmenuu $w]
            grid $f.tree -sticky nsew -padx "10 10" -pady 5

            $f.tree insert end [concat [list [_ "Old"]] $l_old]
            $f.tree item state forcolumn end 0 emphasis
            $f.tree insert end [concat [list [_ "New"]] $l]
            $f.tree item state forcolumn end 0 emphasis

            ttk::label $f.cl1 -text [_ "Please select an option:"]
            grid $f.cl1 -sticky w -padx "10 5" -pady "5 2"

            $w set_uservar_value options ""
            for { set j 1 } { $j <= 10 } { incr j } {
                if { ![info exists opt($j)] } { break }
                foreach "q v" $opt($j) break
                ttk::radiobutton $f.cb$v -text $q -variable [$w give_uservar options] -value $v
                grid $f.cb$v -sticky nw -padx "12 10" -pady "4 0"

                if { [$w give_uservar_value options] eq "" } {
                    $w set_uservar_value options $v
                }
            }
            ttk::checkbutton $f.cba -text [_ "Answer the same for similar changes"] \
                -variable [$w give_uservar answer_the_same 0]
            grid $f.cba -sticky w -padx "10 2" -pady "20 2"

            grid columnconfigure $f 0 -weight 1
            grid rowconfigure $f 1 -weight 1

            bind [winfo toplevel $f] <Return> [list $w invokeok]
            set action [$w createwindow]
            set answer [$w give_uservar_value options]
            set answer_the_same [$w give_uservar_value answer_the_same]
            destroy $w
            if { $action < 1 } {
                destroy $w
                return
            }
        } else {
            set answer_the_same ""
        }
        if { $question_type ne "" } {
            set ret [list_entities_no_group_action $tree $i $columnList $answer $domNodes]

            if { $answer_the_same == 1 } {
                switch $question_type {
                    new_values { set default_change_values $answer }
                }
            }
            incr haschanges
        }
    }
    if { $haschanges } {
        actualize_conditions_window
    }
    unset -nocomplain list_entities_no_group_data
    destroy $wp
}

proc gid_groups_conds::list_entities_cnd_do { columnList wp } {
    variable doc
    variable list_entities_cnd_data


    if { [$wp giveaction] < 1 } {
        unset -nocomplain list_entities_cnd_data
        destroy $wp
        return
    }
    set haschanges 0

    set groups ""
    foreach node [$doc selectNodes /*/groups/group] {
        lappend groups [$node @n]
    }
    set groups [lsort $groups]
    set new_groups ""

    set default_change_values ask
    set default_change_group ask

    set tree [$wp give_uservar_value tree]
    foreach i [lsort -integer [array names list_entities_cnd_data]] {
        set l_old $list_entities_cnd_data($i)
        foreach "cndName_old num_old type_old group_old" [lrange $l_old 0 3] break
        set values_old [lrange $l_old 4 end]
        set l [$tree item text $i]
        foreach "cndName num type group" [lrange $l 0 3] break
        set group [string trim $group]
        set values [lrange $l 4 end]

        set question_type ""
        unset -nocomplain opt
        if { $group eq $group_old && $values ne $values_old } {
            set txt [_ "There are changes in values for entity %s\nCondition: '%s'\nGroup: '%s'" \
                $num $cndName $group_old]

            if { [regexp {^(.*)\s+(\d+)$} $group {} group inum] } {
                incr inum
            } else {
                set inum 2
            }
            while 1 {
                set new_group "$group $inum"
                if { [lsearch -sorted $groups $new_group] == -1 } { break }
                incr inum
            }
            set opt(1) [list [_ "Create a new group '%s' and move entity %s into it with the new values" \
                $new_group $num] new_group]
            set opt(2) [list [_ "Modify group '%s' with new values for all its entities" \
                $group] modify_group]
            set opt(3) [list [_ "Do not do any action for this entity"] nothing_group]

            set question_type new_values
            set answer $default_change_values
        } elseif { $group ne $group_old } {
            set txt [_ "There are changes in group name for entity %s condition '%s', old group: '%s' new group: '%s'" \
                $num $cndName $group_old $group]

            if { [lsearch -exact $new_groups $group] != -1 } {
                set group_exists new_created
            } elseif { [lsearch -exact $groups $group] == -1 } {
                set group_exists no
            } else {
                set group_exists yes
            }
            set new_group $group

            set idx 1
            if { $group_exists eq "no" } {
                set opt($idx) [list [_ "Create new group '%s' and move entity %s into it with new values" \
                    $new_group $num] new_group]
                incr idx
            } elseif { $group_exists eq "new_created" && $values eq $values_old } {
                set opt($idx) [list [_ "Create new group '%s' and move entity %s into it" \
                    $new_group $num] move_to_group]
                incr idx
            } elseif { $group_exists eq "yes" && $values eq $values_old } {
                set opt($idx) [list [_ "Move entity %s into existing group '%s'" \
                    $num $new_group] move_to_group]
                incr idx
            }
            set opt($idx) [list [_ "Do nothing for this entity"] nothing_group]
            incr idx
            set question_type new_group
            set answer $default_change_group
        }

        if { $question_type ne "" && $answer eq "ask" } {
            destroy $wp._ask
            set w [dialogwin_snit $wp._ask -title [_ "Action"]]
            set f [$w giveframe]

            ttk::label $f.cl0 -text $txt -wraplength 280 -justify left
            grid $f.cl0 -sticky nw -padx "10 2" -pady "2 2"

            set columns [list \
                [list 9 "" left text 0] \
                [list 15 [_ "Condition"] left text 0] \
                [list 5 [_ "Entity"] left text 0] \
                [list 6 [_ "Type"] left text 0] \
                [list 15 [_ "Group"] left text 1] \
            ]
        foreach j $columnList {
            lappend columns [list 12 [lindex $j 1] left text 1]
        }
        fulltktree $f.tree -width 350 -height 60 \
            -columns $columns -expand 0 \
            -sensitive_cols all \
            -selectmode extended -showlines 0 -indent 0 \
            -contextualhandler [list gid_groups_conds::_list_entities_contextualmenu $w]
        grid $f.tree -sticky nsew -padx "10 10" -pady 5

        $f.tree insert end [concat [list [_ "Old"]] $l_old]
        $f.tree item state forcolumn end 0 emphasis
        $f.tree insert end [concat [list [_ "New"]] $l]
        $f.tree item state forcolumn end 0 emphasis

        ttk::label $f.cl1 -text [_ "Please select an option:"]
        grid $f.cl1 -sticky w -padx "10 5" -pady "5 0"

        $w set_uservar_value options ""
        for { set j 1 } { $j <= 10 } { incr j } {
            if { ![info exists opt($j)] } { break }
            foreach "q v" $opt($j) break
            ttk::radiobutton $f.cb$v -text $q -variable [$w give_uservar options] -value $v
            grid $f.cb$v -sticky nw -padx "12 10" -pady "4 0"

            if { [$w give_uservar_value options] eq "" } {
                $w set_uservar_value options $v
            }
        }
        ttk::checkbutton $f.cba -text [_ "Answer the same for similar changes"] \
            -variable [$w give_uservar answer_the_same 0]
        grid $f.cba -sticky w -padx "10 2" -pady "20 2"

        grid columnconfigure $f 0 -weight 1
        grid rowconfigure $f 1 -weight 1

        bind [winfo toplevel $f] <Return> [list $w invokeok]
        set action [$w createwindow]
        set answer [$w give_uservar_value options]
        set answer_the_same [$w give_uservar_value answer_the_same]
        destroy $w
        if { $action < 1 } {
            destroy $w
            return
        }
    } else {
        set answer_the_same ""
    }
    if { $question_type ne "" } {
        set ret [list_entities_cnd_action $tree $i $new_group $columnList $answer]
        eval lappend groups $ret
        eval lappend new_groups $ret

        if { $answer_the_same == 1 } {
            switch $question_type {
                new_values { set default_change_values $answer }
                new_group { set default_change_group $answer }
            }
        }
        incr haschanges
    }
}
if { $haschanges } {
    actualize_groups_window
    actualize_conditions_window
    check_draw_nodes_state
}
unset -nocomplain list_entities_cnd_data
destroy $wp
}

proc gid_groups_conds::_list_entities_cnd_signal { w what } {
    variable is_draw_post

    set tree [$w give_uservar_value tree]

    set idx 0
    foreach item [$tree selection get] {
        lassign [lrange [$tree item text $item] 1 2] num type
        if { [llength $num] == 3 } {
            #face elements num instead a single number is "num_elem - num_face"
            #store now only the element id to point and discard num_face...
            set num [lindex $num 0]
        }
        dict lappend entities $type $num
        incr idx
    }
    if { !$idx && $what ne "labeloff" } {

        return
    }
    switch $what {
        signal {
            if { !$is_draw_post } {
                if { $idx > 1 } {
                    snit_messageBox -message [_ "It is necessary to select only one entity"] -parent $w
                    return
                }
                set txt [_ "Signaling %s %s" $type $num]
                set cmd [list GiD_Process MEscape Utilities SignalEntities ${type}s $num]
                select_apply_entities $w $cmd $txt
            } else {
                draw_post::draw_signal_do -draw_lines 1 $type [dict get $entities $type]
            }
        }
        labelon {
            if { !$is_draw_post } {
                dict for "n v" $entities {
                    lassign [list GiD_Process Label Select $n] $v escape
                }
                select_apply_entities $w "" ""
            } else {
                draw_post::draw_signal_do -draw_lines 0 $type [dict get $entities $type]
            }
        }
        labeloff {
            if { !$is_draw_post } {
                GiD_Process Label Off
            } else {
                draw_post::refresh
            }
        }
    }

}

proc gid_groups_conds::_list_entities_from_data_to_xml { domNodes geom_mesh domNode } {

    $domNode setAttribute geom_mesh $geom_mesh

    set nodesNode [$domNode appendChildTag nodes]
    foreach node $domNodes {
        set xpath [gid_groups_conds::nice_xpath $node]
        $nodesNode appendChildTag node [list attributes() xpath $xpath]
    }
}

proc gid_groups_conds::_list_entities_from_xml_to_data { domNode } {
    variable doc

    set geom_mesh [$domNode getAttribute geom_mesh]

    set domNodes ""
    foreach nodeNode [$domNode selectNodes nodes/node] {
        set node [$doc selectNodes /*/[$nodeNode @xpath]]
        lappend domNodes $node
    }
    return [list $domNodes $geom_mesh]
}


proc gid_groups_conds::_list_entities_cnd_output { w output } {
    variable list_entities_cnd_data

    if { [array size list_entities_cnd_data] } {
        snit_messageBox -parent $w -message [_ "You can only export if there are no changes. Save and open again to export."]
        return
    }

    set domNodes [$w give_uservar_value domNodes]
    set geom_mesh [$w give_uservar_value geom_mesh]

    switch $output {
        report {
            set data [gid_groups_conds::create_report_part_xml lognoter list_entities \
                [_ "List groups"] $domNodes $geom_mesh]

            set cmd [list createpage -ask_user -goto -accept_existing_pages \
                [_ "Report"]>[_ "List groups"] - $data ""]

            set err [catch { view_report -raise_error -execute_commands $cmd} errstring]
            if { $err } {
                snit_messageBox -message $errstring -parent $w
                return
            }
        }
        file {
            _list_entities_cnd_export_file $w $domNodes $geom_mesh
        }
    }
}

proc gid_groups_conds::_list_entities_no_group_output { w output } {
    variable list_entities_no_group_data
    variable name_page

    if { [array size list_entities_no_group_data] } {
        snit_messageBox -parent $w -message [_ "You can only export if there are no changes. Save and open again to export."]
        return
    }

    set domNodes [$w give_uservar_value domNodes]
    set geom_mesh [$w give_uservar_value geom_mesh]

    switch $output {
        report {
            set data [gid_groups_conds::create_report_part_xml lognoter list_entities_no_group \
                [_ "%s" $name_page] $domNodes $geom_mesh]

            set cmd [list createpage -ask_user -goto -accept_existing_pages \
                [_ "Report"]>[_ "%s" $name_page] - $data ""]
            set err [catch { view_report -raise_error -execute_commands $cmd -deiconify_win 0 } errstring]
            if { $err } {
                snit_messageBox -message $errstring -parent $w
                return
            }
            GidUtils::SetWarnLine [_ "Added data to report OK."]
        }
        file {
            _list_entities_no_group_export_file $w $domNodes $geom_mesh
        }
    }
}

proc gid_groups_conds::_list_entities_no_group_export_file { wp domNodes geom_mesh } {
    variable is_draw_post
    variable name_page

    set reports_post_export_dir [get_preference reports_post_export_dir ""]
    if { ![file isdirectory $reports_post_export_dir] } {
        set reports_post_export_dir [pwd]
    }
    set default_ext [get_preference reports_post_def_extension .xml]
    switch $default_ext {
        ".xml" {
            set types [list \
                [list [_ "Excel XML files "] ".xml"] \
                [list [_ "CSV files"] ".csv"] \
                [list [_ "All files"] "*"] \
            ]
    }
    default {
        set types [list \
            [list [_ "CSV files"] ".csv"] \
            [list [_ "Excel XML files "] ".xml"] \
            [list [_ "All files"] "*"] \
        ]
}
}
if { !$is_draw_post } {
    set model_name [GiD_Info Project ModelName]
} else {
    set model_name [draw_post::give_model_name]
}
set initial_file "[file tail $model_name]_list_entitities$default_ext"
if { $initial_file eq "UNNAMED_list_entitities$default_ext" } {
    set initial_file "list_entitities$default_ext"
}
set file [tk_getSaveFile -filetypes $types -parent $wp -title \
    [_ "Export CSV file"] -initialdir $reports_post_export_dir \
    -initialfile $initial_file -defaultextension $default_ext]
if { $file == "" } { return }
set_preference reports_post_export_dir [file dirname $file]
if { [string equal -nocase [file extension $file] ".xml"] } {
    set_preference reports_post_def_extension ".xml"
} else {
    set_preference reports_post_def_extension ".csv"
}
lassign [gid_groups_conds::list_entities_give_data_no_group_full $domNodes $geom_mesh] columnList cndList rows

set fout [open $file w]
if { [string equal -nocase [file extension $file] ".xml"] } {
    fconfigure $fout -encoding utf-8
    puts $fout {<?xml version="1.0"?>
        <?mso-application progid="Excel.Sheet"?>
        <Workbook xmlns="urn:schemas-microsoft-com:office:spreadsheet"
        xmlns:o="urn:schemas-microsoft-com:office:office"
        xmlns:x="urn:schemas-microsoft-com:office:excel"
        xmlns:ss="urn:schemas-microsoft-com:office:spreadsheet"
        xmlns:html="http://www.w3.org/TR/REC-html40">
        <Worksheet ss:Name="post report">
        <Table >}
        puts $fout "<Row>"
        puts $fout "<Cell><Data ss:Type='String'>[xml_map $name_page]</Data></Cell>"
        puts $fout "</Row>"
        puts $fout "<Row>"
        foreach h $columnList {
        puts $fout "<Cell><Data ss:Type='String'>[xml_map $h]</Data></Cell>"
    }
    puts $fout "</Row>"
    foreach row $rows {
        puts $fout "<Row>"
        foreach c $row {
            if { [string is double -strict $c] } {
                puts $fout "<Cell><Data ss:Type='Number'>[xml_map $c]</Data></Cell>"
            } else {
                puts $fout "<Cell><Data ss:Type='String'>[xml_map $c]</Data></Cell>"
            }
        }
        puts $fout "</Row>"
    }
    puts $fout "</Table></Worksheet></Workbook>"
} else {
    package require csv
    set sep ";"
    puts $fout [csv::join $columnList $sep]
    foreach row $rows {
        if { $first_col_out } { set row [lrange $row 1 end] }
        puts $fout [csv::join $row $sep]
    }
}
close $fout
}


proc gid_groups_conds::_list_entities_cnd_export_file { wp domNodes geom_mesh } {
    variable is_draw_post

    set reports_post_export_dir [get_preference reports_post_export_dir ""]
    if { ![file isdirectory $reports_post_export_dir] } {
        set reports_post_export_dir [pwd]
    }
    set default_ext [get_preference reports_post_def_extension .xml]
    switch $default_ext {
        ".xml" {
            set types [list \
                [list [_ "Excel XML files "] ".xml"] \
                [list [_ "CSV files"] ".csv"] \
                [list [_ "All files"] "*"] \
            ]
    }
    default {
        set types [list \
            [list [_ "CSV files"] ".csv"] \
            [list [_ "Excel XML files "] ".xml"] \
            [list [_ "All files"] "*"] \
        ]
}
}
if { !$is_draw_post } {
    set model_name [GiD_Info Project ModelName]
} else {
    set model_name [draw_post::give_model_name]
}
set initial_file "[file tail $model_name]_list_entitities$default_ext"
if { $initial_file eq "UNNAMED_list_entitities$default_ext" } {
    set initial_file "list_entitities$default_ext"
}
set file [tk_getSaveFile -filetypes $types -parent $wp -title \
    [_ "Export CSV file"] -initialdir $reports_post_export_dir \
    -initialfile $initial_file -defaultextension $default_ext]
if { $file == "" } { return }
set_preference reports_post_export_dir [file dirname $file]
if { [string equal -nocase [file extension $file] ".xml"] } {
    set_preference reports_post_def_extension ".xml"
} else {
    set_preference reports_post_def_extension ".csv"
}
lassign [gid_groups_conds::list_entities_give_data_full $domNodes $geom_mesh] columnList cndList rows

if { [llength $cndList] < 2 } {
    set columnList [string range $columnList 1 end]
    set first_col_out 1
} else {
    set first_col_out 0
}

set fout [open $file w]
if { [string equal -nocase [file extension $file] ".xml"] } {
    fconfigure $fout -encoding utf-8
    puts $fout {<?xml version="1.0"?>
        <?mso-application progid="Excel.Sheet"?>
        <Workbook xmlns="urn:schemas-microsoft-com:office:spreadsheet"
        xmlns:o="urn:schemas-microsoft-com:office:office"
        xmlns:x="urn:schemas-microsoft-com:office:excel"
        xmlns:ss="urn:schemas-microsoft-com:office:spreadsheet"
        xmlns:html="http://www.w3.org/TR/REC-html40">
        <Worksheet ss:Name="post report">
        <Table >}
        puts $fout "<Row>"
        foreach h $columnList {
        puts $fout "<Cell><Data ss:Type='String'>[xml_map $h]</Data></Cell>"
    }
    puts $fout "</Row>"
    foreach row $rows {
        if { $first_col_out } { set row [lrange $row 1 end] }
        puts $fout "<Row>"
        foreach c $row {
            if { [string is double -strict $c] } {
                puts $fout "<Cell><Data ss:Type='Number'>[xml_map $c]</Data></Cell>"
            } else {
                puts $fout "<Cell><Data ss:Type='String'>[xml_map $c]</Data></Cell>"
            }
        }
        puts $fout "</Row>"
    }
    puts $fout "</Table></Worksheet></Workbook>"
} else {
    package require csv
    set sep ";"
    puts $fout [csv::join $columnList $sep]
    foreach row $rows {
        if { $first_col_out } { set row [lrange $row 1 end] }
        puts $fout [csv::join $row $sep]
    }
}
close $fout
}

proc gid_groups_conds::list_entities_no_group_action { tree item columnList what domNode } {
    variable doc
    variable list_entities_no_group_data

    set new_item ""
    switch $what {
        modify {
            set values [$tree item text $item]
            set n [$domNode @n]
            switch [$domNode nodeName] {
                blockdata {
                    set node $domNode
                    set name [$node @name ""]
                    if {$name eq ""} {
                        set name [$node @n ""]
                    }
                }
                container {
                    set node [$domNode selectNodes {ancestor::blockdata[@sequence='1']}]
                    if {$node eq ""} {
                        set xp [format_xpath {.//container[@n=%s]} $n]
                        set node [$doc selectNodes $xp]
                        if {[llength $node] > 1} {
                            set node $domNode
                        }
                        set name [$node @n]
                    } else {
                        set name [$node @name ""]
                        if {$name eq ""} {
                            set name [$node @n ""]
                        }
                    }
                }
            }
            set values_dict ""
            foreach valueNode [$node selectNodes {.//value}] {
                set n [get_domnode_attribute $valueNode n]
                dict set values_dict $n v [get_domnode_attribute $valueNode v]
                if { [get_domnode_attribute $valueNode unit_magnitude] ne "" } {
                    dict set values_dict $n units [get_domnode_attribute $valueNode units]
                }
                set fNode [$valueNode selectNodes function]
                if { [llength $fNode] == 1 } {
                    dict set values_dict $n function [$fNode asXML]
                }
            }
            set idx 0
            foreach col $columnList {
                lassign $col n - units valueNode
                set v [lindex $values $idx]
                dict set values_dict $n v $v
                if { [get_domnode_attribute $valueNode unit_magnitude] ne "" } {
                    dict set values_dict $n units $units
                }
                incr idx
            }
            gid_groups_conds::list_entities_no_group_ok [gid_groups_conds::nice_xpath $node] $name $values_dict
        }
    }
    return $new_item
}

proc gid_groups_conds::list_entities_no_group_ok { args } {
    variable doc

    set optional ""
    set compulsory "xpathCnd name values_dict"
    parse_args $optional $compulsory $args

    set actualize_nodesList ""
    set att [list n $name]

    set root [$doc documentElement]
    set domNodeCnd [$root selectNodes $xpathCnd]

    set domNodeBD [$domNodeCnd selectNodes {ancestor::blockdata[1]}]

    if { $domNodeBD ne "" && [$domNodeBD @state normal] eq "disabled" && [$domNodeBD @sequence_type any] eq "non_void_disabled" } {
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNodeBD] [list state normal]
        lappend actualize_nodesList $domNodeBD
    } elseif { $domNodeBD ne "" && [$domNodeBD @active 1] == 0 && [$domNodeBD @sequence_type any] eq "non_void_deactivated" } {
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNodeBD] [list active 1]
        lappend actualize_nodesList $domNodeBD
    }
    set ret1 [_apply_values $domNodeCnd $domNodeBD "" $values_dict normal]
    lappend actualize_nodesList {*}$ret1
    return [lsort -unique $actualize_nodesList]
}


#by default assumed attribute named ov, but some special condition handle two groups
#and then index must be 1 or 2
#return sublist of: points lines surfaces volumes
proc gid_groups_conds::get_condition_disallowed_types_geometry { domNodeCnd index } {
    set condition_disallowed_types [list]
    set attribute ov$index
    set attribute_value [get_domnode_attribute $domNodeCnd $attribute ""]
    set condition_allowed_types [split $attribute_value ,]
    if { [llength $condition_allowed_types] == 0 } {
        #unexpected empty, assumed all types allowed
        W "Wrong condition [$domNodeCnd @n]. Expected not empty $attribute attribute. Assuming all types allowed"
    } else {
        foreach type {points lines surfaces volumes} type_to_find {point line surface volume} {
            if { [lsearch $condition_allowed_types $type_to_find] == -1 } {
                lappend condition_disallowed_types $type
            }
        }
    }
    return $condition_disallowed_types
}

proc gid_groups_conds::get_condition_allowed_types_geometry { domNodeCnd index } {
    set condition_disallowed_types_geometry [gid_groups_conds::get_condition_disallowed_types_geometry $domNodeCnd $index]
    set condition_allowed_types_geometry [list]
    foreach type {points lines surfaces volumes} {
        if { [lsearch $condition_disallowed_types_geometry $type] == -1 } {
            lappend condition_allowed_types_geometry $type
        }
    }
    return $condition_allowed_types_geometry
}

# by default assumed attribute named ov, but some special condition handle two groups
# and then index must be 1 or 2
# return sublist of: nodes elements faces
proc gid_groups_conds::get_condition_disallowed_types_mesh { domNodeCnd {index ""} } {
    set condition_disallowed_types [list]
    set attribute ovm$index
    set attribute_value [get_domnode_attribute $domNodeCnd $attribute ""]
    set condition_allowed_types [split $attribute_value ,]
    if { [llength $condition_allowed_types] == 0 } {
        #empty, assumed all types allowed, don not show error message in this case
    } else {
        foreach type {nodes elements faces} type_to_find {node element face_element} {
            if { [lsearch $condition_allowed_types $type_to_find] == -1 } {
                lappend condition_disallowed_types $type
            }
        }
    }
    return $condition_disallowed_types
}

proc gid_groups_conds::get_condition_allowed_types_mesh { domNodeCnd index } {
    set condition_disallowed_types_mesh [gid_groups_conds::get_condition_disallowed_types_mesh $domNodeCnd $index]
    set condition_allowed_types_mesh [list]
    foreach type {nodes elements faces} {
        if { [lsearch $condition_disallowed_types_mesh $type] == -1 } {
            lappend condition_allowed_types_mesh $type
        }
    }
    return $condition_allowed_types_mesh
}

#by default assumed attribute named ov, but some special condition handle two groups
#and then index must be 1 or 2
#return sublist of: points lines surfaces volumes nodes elements faces
proc gid_groups_conds::get_condition_disallowed_types { domNodeCnd index } {
    set condition_disallowed_types_geometry [gid_groups_conds::get_condition_disallowed_types_geometry $domNodeCnd $index]
    set condition_disallowed_types_mesh [gid_groups_conds::get_condition_disallowed_types_mesh $domNodeCnd $index]
    return [concat $condition_disallowed_types_geometry $condition_disallowed_types_mesh]
}

proc gid_groups_conds::get_condition_allowed_types { domNodeCnd index } {
    set condition_allowed_types_geometry [gid_groups_conds::get_condition_allowed_types_geometry $domNodeCnd $index]
    set condition_allowed_types_mesh [gid_groups_conds::get_condition_allowed_types_mesh $domNodeCnd $index]
    return [concat $condition_allowed_types_geometry $condition_allowed_types_mesh]
}

#return sublist of: linear triangle quadrilateral ...
proc gid_groups_conds::get_condition_disallowed_element_types { domNodeCnd index } {
    set condition_disallowed_element_types [list]
    set attribute ov_element_types$index
    set attribute_value [get_domnode_attribute $domNodeCnd $attribute ""]
    set condition_allowed_element_types [split $attribute_value ,]
    if { [llength $condition_allowed_element_types] == 0 } {
        #ov_element_types empty, assumed all element_types allowed
    } else {
        foreach element_type [GidUtils::GetElementTypesLowercase all] {
            if { [lsearch $condition_allowed_element_types $element_type] == -1 } {
                lappend condition_disallowed_element_types $element_type
            }
        }
    }
    return $condition_disallowed_element_types
}

proc gid_groups_conds::get_condition_allowed_element_types { domNodeCnd index } {
    set condition_disallowed_element_types [gid_groups_conds::get_condition_disallowed_element_types $domNodeCnd $index]
    set condition_allowed_element_types [list]
    foreach element_type [GidUtils::GetElementTypesLowercase all] {
        if { [lsearch $condition_disallowed_element_types $element_type] == -1 } {
            lappend condition_allowed_element_types $element_type
        }
    }
    return $condition_allowed_element_types
}

proc gid_groups_conds::_apply_values { domNode_from domNode_to ov values_dict \
    parent_state { level 0 } } {

    set actualize_nodesList ""
    set valueNodes [$domNode_from selectNodes value|container]
    if { ![llength $valueNodes] && [$domNode_from @isvalue 0] } {
        lappend valueNodes $domNode_from
    }
    foreach valueNode $valueNodes {
        set state [get_domnode_attribute $valueNode state normal]
        if { $parent_state eq "hidden" } { set state hidden }
        switch [$valueNode nodeName] {
            value - condition {
                set n [$valueNode @n]
                set v [dict get $values_dict $n v]
                set value $v
                if {[get_domnode_attribute $valueNode dict] != ""} {
                    set dict ""
                    foreach "ni vi" [split [get_domnode_attribute $valueNode dict] ,] {
                        dict set dict $ni [= $vi]
                        if {[= $vi] eq [= $v]} {
                            set value $ni
                            break
                        }
                    }
                }
                $valueNode setAttribute v $value
            }
            container {
                #                 set attDict ""
                #                 foreach i [$valueNode selectNodes @*] { lappend attDict {*}$i }
                #                 set c [gid_groups_conds::add [gid_groups_conds::nice_xpath $domNode_to] container \
                #                 $attDict]
                #
                _apply_values $valueNode $domNode_to "" $values_dict $state \
                    [expr {$level+1}]
            }
        }
    }
    if { $level == 0 && $domNode_to ne ""} {
        foreach valueNode [$domNode_to selectNodes {.//value}] {
            set hc [gid_groups_conds::check_node_dependencies $valueNode]
            lappend actualize_nodesList {*}$hc
        }
    }
    return [lsort -unique $actualize_nodesList]
}

proc gid_groups_conds::list_entities_cnd_action { tree item new_group columnList what } {
    variable doc
    variable list_entities_cnd_data
    set new_groups ""
    switch $what {
        new_group {
            gid_groups_conds::add_new_group -ifnotexists -noactualizegroups -exactname $new_group
            lappend new_groups $new_group
            set l_old $list_entities_cnd_data($item)
            lassign [lrange $l_old 0 3] cndName num type group
            set l [$tree item text $item]
            set values [lrange $l 4 end]
            gid_groups_conds::delete_entities_from_groups [list $num] [list [list "" $group]] $type
            gid_groups_conds::add_entities_to_groups [list $num] [list [list "" $new_group]] $type
            set xp [format_xpath {//condition[not(ancestor::blockdata[@n="Internal data"])]/group[@n=%s]} $group]
            append xp [format_xpath {|//condition[not(ancestor::blockdata[@n="Internal data"])]/groupList[group/@n=%s]} $group]
            foreach groupNode [$doc selectNodes $xp] {
                set cndNode [$groupNode selectNodes ancestor::condition]
                set values_dict ""
                foreach valueNode [$groupNode selectNodes .//value] {
                    set n [get_domnode_attribute $valueNode n]
                    dict set values_dict $n v [get_domnode_attribute $valueNode v]
                    if { [get_domnode_attribute $valueNode unit_magnitude] ne "" } {
                        dict set values_dict $n units [get_domnode_attribute $valueNode units]
                    }
                    set fNode [$valueNode selectNodes function]
                    if { [llength $fNode] == 1 } {
                        dict set values_dict $n function [$fNode asXML]
                    }
                }
                if { $cndName eq [get_domnode_attribute $cndNode pn] } {
                    set idx 0
                    foreach col $columnList {
                        lassign $col n - units valueNode
                        set v [lindex $values $idx]
                        dict set values_dict $n v $v
                        if { [get_domnode_attribute $valueNode unit_magnitude] ne "" } {
                            dict set values_dict $n units $units
                        }
                        incr idx
                    }
                }
                apply_condition -ov $type [gid_groups_conds::nice_xpath $cndNode] $new_group $values_dict
            }
        }
        move_to_group {
            set l_old $list_entities_cnd_data($item)
            lassign [lrange $l_old 0 3] cndName num type group_old
            delete_entities_from_groups [list $num] [list [list "" $group_old]] $type
            add_entities_to_groups [list $num] [list [list "" $new_group]] $type
        }
        modify_group {
            set l [$tree item text $item]
            lassign [lrange $l 0 3] cndName num type group
            set values [lrange $l 4 end]
            set xp [format_xpath {//condition[not(ancestor::blockdata[@n="Internal data"])]/group[@n=%s]} $group]
            append xp [format_xpath {|//condition[not(ancestor::blockdata[@n="Internal data"])]/groupList[group/@n=%s]} $group]
            foreach groupNode [$doc selectNodes $xp] {
                set cndNode [$groupNode selectNodes ancestor::condition]
                if { $group ne [get_domnode_attribute $groupNode n] } { continue }
                break
            }
            set values_dict ""
            foreach valueNode [$groupNode selectNodes {.//value}] {
                set n [get_domnode_attribute $valueNode n]
                dict set values_dict $n v [get_domnode_attribute $valueNode v]
                if { [get_domnode_attribute $valueNode unit_magnitude] ne "" } {
                    dict set values_dict $n units [get_domnode_attribute $valueNode units]
                }
                set fNode [$valueNode selectNodes function]
                if { [llength $fNode] == 1 } {
                    dict set values_dict $n function [$fNode asXML]
                }
            }
            if { $cndName eq [get_domnode_attribute $cndNode pn] } {
                set idx 0
                foreach col $columnList {
                    lassign $col n - units valueNode
                    set v [lindex $values $idx]
                    dict set values_dict $n v $v
                    if { [get_domnode_attribute $valueNode unit_magnitude] ne "" } {
                        dict set values_dict $n units $units
                    }
                    incr idx
                }
            }
            gid_groups_conds::clear_condition_group [gid_groups_conds::nice_xpath $cndNode] $group
            gid_groups_conds::apply_condition -ov $type [gid_groups_conds::nice_xpath $cndNode] $group $values_dict
        }
    }
    return $new_groups
}

################################################################################
#    Draw groups & conditions routines
################################################################################

proc gid_groups_conds::_OpenGL_color { tcl_color } {
    set ocolor ""
    foreach i [winfo rgb . $tcl_color] {
        lappend ocolor [expr {$i/65535.0}]
    }
    return $ocolor
}

proc gid_groups_conds::_OpenGL_color4 { tcl_color } {
    set ocolor ""
    foreach i [winfo rgb . $tcl_color] {
        lappend ocolor [expr {$i/65535.0}]
    }
    lappend ocolor 1.0
    return $ocolor
}

proc gid_groups_conds::_tcl_color_from_OpenGL { r g b } {
    set color "#"
    foreach i [list $r $g $b] {
        append color [format "%02x" [expr {int(255*$i)}]]
    }
    return $color
}

proc gid_groups_conds::draw { condition geometry_mesh num group } {
    variable last_drawn_id
    variable groups_cache
    if { $last_drawn_id eq "" || $last_drawn_id > $num } {
        set groups_cache [lsort [give_groups]]
    }
    set last_drawn_id $num
    set ovwhat surface
    set color [list .8 .8 .8]
    if { [lsearch -sorted -exact $groups_cache $group] != -1 } {
        GiD_OpenGL draw -color $color
        GiD_OpenGL drawentity -mode filled surface $num
    }
    return 1
}

proc gid_groups_conds::_valuesList { groupNode } {
    set list ""
    foreach valueNode [$groupNode selectNodes .//value|ancestor::groupList//value] {
        if { [get_domnode_attribute $valueNode state normal] eq "hidden" } { continue }
        set ishidden 0
        foreach ancgroupNode [$valueNode selectNodes ancestor-or-self::value] {
            if {[$ancgroupNode @state normal] eq "hidden"} { set ishidden 1; break }
        }
        if { !$ishidden }  {
            lappend list [list [$valueNode @n] [$valueNode @v] [$valueNode @units ""]]
        }
    }
    if { ![llength $list] } {
        set cndNode [$groupNode selectNodes ancestor::condition]
        lappend list [list [$cndNode @n] [get_domnode_attribute $cndNode @pn] ""]
    }
    return $list
}

proc gid_groups_conds::draw_nodes_xp { whatG xpath } {
    variable doc
    variable open_boundary_cond
    if { [info exists open_boundary_cond] && [winfo exists $open_boundary_cond] } {
        return [$open_boundary_cond draw_groups_do -domNodes [$doc selectNodes $xpath] $whatG]
    } else {
        return [gid_groups_conds::draw_nodes -auto_finish -domNodes [$doc selectNodes $xpath] $whatG]
    }
}

proc gid_groups_conds::give_draw_nodes_state {} {
    variable draw_nodes_state
    return $draw_nodes_state
}

proc gid_groups_conds::set_draw_nodes_state { state } {
    variable draw_nodes_state
    if { $state ne $draw_nodes_state } {
        set draw_nodes_state $state
        gid_groups_conds::check_draw_nodes_state -force
    }
}

proc gid_groups_conds::check_draw_nodes_state { args } {
    variable doc
    variable draw_nodes_state

    set optional {
        { -force "" 0 }
    }
    set compulsory ""
    parse_args $optional $compulsory $args

    if { $draw_nodes_state eq "" && !$force } { return }

    set root [$doc documentElement]
    set domNodes ""
    foreach xp [lindex $draw_nodes_state 1] {
        lappend domNodes [$root selectNodes $xp]
    }
    gid_groups_conds::draw_nodes -domNodes $domNodes [lindex $draw_nodes_state 0]
}

# what: groupnames, values, symbols, symbols_all
proc gid_groups_conds::draw_nodes { args } {
    variable doc
    variable draw_nodes_state
    variable is_draw_post

    set optional {
        { -domNodes domNodes "" }
        { -auto_finish "" 0 }
    }
    set compulsory "whatG"
    parse_args $optional $compulsory $args

    set ns ""
    foreach i $domNodes {
        lappend ns [gid_groups_conds::nice_xpath $i]
    }
    set draw_nodes_state [list $whatG $ns]

    if { $whatG eq "symbols_all" } {
        set domNodes [$doc selectNodes /*/*]
        set what symbols
    } else {
        set what $whatG
    }
    if { ![info exists ::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE] } {
        set ::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE 0
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE && $what != "symbols" } {
        #draw color
    } else {
        gid_groups_conds::end_drawall
    }
    set groupList ""
    foreach domNode $domNodes {
        switch [$domNode nodeName] {
            blockdata {
                set xp {.//condition/group|.//condition/groupList/group}
                lappend groupList {*}[$domNode selectNodes $xp]
            }
            condition {
                set xp {group|groupList/group}
                lappend groupList {*}[$domNode selectNodes $xp]
            }
            container {
                set xp {.//condition/group|.//condition/groupList/group}
                set ret [$domNode selectNodes $xp]
                if { ![llength $ret] } {
                    set xp {ancestor::condition/group|ancestor::condition/groupList/group}
                    set ret [$domNode selectNodes $xp]
                }
                lappend groupList {*}$ret
            }
            group {
                lappend groupList $domNode
            }
            groupList {
                set xp {group}
                lappend groupList {*}[$domNode selectNodes $xp]
            }
            value {
                set xp {parent::group|parent::groupList/group}
                lappend groupList {*}[$domNode selectNodes $xp]
            }
        }
    }
    set groups ""
    foreach gr $groupList {
        set cndNode [$gr selectNodes ancestor::condition]
        set xp {ancestor::groupList/group}
        set grs [$gr selectNodes $xp]
        if { [llength $grs] } {
            set ipos [lsearch -exact $grs $gr]
        } else {
            set ipos ""
        }
        set ov [$gr @ov ""]
        if { $ov eq "" } {
            set ov [get_domnode_attribute $cndNode ov$ipos]
        }
        if { $what eq "groupnames" } {
            set elm [list [$gr @n] $ov ""]
        } else {
            set elm [list [$gr @n] $ov [_valuesList $gr]]
        }
        if { $ipos eq "" } {
            set symbolNode ""
            foreach symbolNode [$cndNode selectNodes symbol] {
                set o [$symbolNode @orientation]
                if { $o in "section shell_thickness loads" && [info command ::draw_model_conditions] eq "" } { continue }
                break
            }
        } else {
            set symbolNode [$cndNode selectNodes [format_xpath {symbol[@pos=%s]} \
                [expr {$ipos+1}]]]
            if { $symbolNode eq "" } {
                set symbolNode [$cndNode selectNodes symbol]
            }
        }
        if { $symbolNode ne "" } {
            lappend elm [list [$symbolNode @proc] [$symbolNode @orientation]]
        }
        lappend groups $elm
    }
    if { ![llength $groups] } { return }

    if { $what eq "symbols" } {
        set drawtype symbols
    } else {
        set drawtype colors
    }

    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE && $what != "symbols" } {
        #draw color native
        #GiD_Groups draw $fullnames
        set fullnames [list]
        foreach item $groups {
            set group [lindex $item 0]
            set ov [lindex $item 1]
            if { [GiD_Groups get state $group] == "hidden" && [GiD_Groups get parent $group] != "" && ![GiD_EntitiesGroups get $group ${ov}s -count -visible] } {
                #dark kratos trick using a parent group with entities and empty
                set group [GiD_Groups get parent $group]
            }
            lappend fullnames $group
        }
        GiD_Process 'Groups Draw {*}$fullnames escape
    } else {
        gid_groups_conds::drawall $drawtype $groups
    }

    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE && $what != "symbols" } {
        #if { $auto_finish } {
        #    delayedop changefunc [list GiD_Groups end_draw]
        #}
    } else {
        if { !$is_draw_post } {
            GiD_Process 'Redraw
            if { 0 } {
                # dirty trick to make delayedop work correctly
                GidUtils::DisableWarnLine
                GiD_Process Mescape 'SelectEntities
                GidUtils::EnableWarnLine
            }
            if { $auto_finish } {
                delayedop changefunc gid_groups_conds::end_drawall
            }
        }
    }
    return $groups
}

proc gid_groups_conds::register_drawopenGLvector { } {
    variable id

    set id [GiD_OpenGL register gid_groups_conds::drawopenGLvector]
}

proc gid_groups_conds::drawopenGLvector { } {
    variable origin
    variable dest

    # Drawing Points
    GiD_OpenGL draw -color "0 0.6 0.8" -pointsize 5
    GiD_OpenGL draw -begin points
    GiD_OpenGL draw -vertex $origin
    GiD_OpenGL draw -vertex $dest
    GiD_OpenGL draw -end

    # Drawing vector
    set p1 $origin
    set p2 $dest

    set vn [m::norm [m::sub $p2 $p1]]
    set vv [m::unitLengthVector [m::sub $p2 $p1]]

    GiD_OpenGL draw -color "0 1 0"
    ggc::draw::draw_vector -draw_style lines $vn $p1 $vv
}

proc gid_groups_conds::end_register_drawopenGLvector { } {
    variable id

    GiD_Process 'Redraw
    GiD_OpenGL unregister $id
}

# drawtype can be: colors or symbols
proc gid_groups_conds::drawall { drawtype groups } {
    variable opengl_id
    variable opengl_rendermode
    variable do_change_opengl_rendermode
    variable draw_auto_hide_surfaces
    variable show_figures_surfaces
    variable is_draw_post

    if { [info exists opengl_id] } {
        gid_groups_conds::end_drawall
    }
    switch [is_geometry_or_mesh] {
        "geometry" { set ents Surfaces }
        "mesh" { set ents Elements }
    }
    if { !$is_draw_post } {
        set show_figures_surfaces [GiD_Set ShowFigures($ents)]
        if { $drawtype eq "colors" || $drawtype eq "symbols" } {
            if { $drawtype eq "colors" || $draw_auto_hide_surfaces } {
                GiD_Set ShowFigures($ents) 0
            }
            if { [GiD_Info Project RenderMode] eq "normal" && $do_change_opengl_rendermode } {
                set opengl_rendermode normal
                #                 GiD_Process 'Render Flat
            }
        }
        set opengl_id [GiD_OpenGL register [list gid_groups_conds::drawall_do $drawtype $groups]]
    } else {
        if { $drawtype eq "colors" || $drawtype eq "symbols" } {
            if { $drawtype eq "colors" || $draw_auto_hide_surfaces } {
                set style boundaries
            }
            if { [draw_post::give_current_view_style] ne $style } {
                set opengl_rendermode [draw_post::give_current_view_style]
                draw_post::change_view_style $style
            }
        }
        set opengl_id [list gid_groups_conds::drawall_do $drawtype $groups]
        draw_post::draw_point_line_indicator add $opengl_id
    }

}

proc gid_groups_conds::drawall_do { drawtype groups } {
    variable opengl_object
    variable show_figures_surfaces

    #     if { ![info exists show_figures_surfaces] } {
    #         set show_figures_surfaces [GiD_Set ShowFigures(Surfaces)]
    #     }
    if { ![info exists opengl_object] } {
        switch $drawtype {
            colors {
                gid_groups_conds::drawall_create_object $groups
                #GiD_Set ShowFigures(Surfaces) 0
            }
            symbols {
                gid_groups_conds::draw_entities $groups
            }
        }
    }
    GiD_OpenGL draw -call $opengl_object
}

proc gid_groups_conds::end_drawall_maintain {} {
    gid_groups_conds::end_drawall_do
}

proc gid_groups_conds::end_drawall_do {} {
    variable show_figures_surfaces
    variable opengl_rendermode
    variable is_draw_post

    if { [info exists show_figures_surfaces] } {
        if { !$is_draw_post } {
            switch [GiD_Info Project ViewMode] {
                "GEOMETRYUSE" { set ents Surfaces }
                "MESHUSE" { set ents Elements }
            }
            GiD_Set ShowFigures($ents) $show_figures_surfaces
        }
        unset show_figures_surfaces
    }
    if { $opengl_rendermode ne "" } {
        if { !$is_draw_post } {
            if { [GiD_Info Project RenderMode] eq "polygons" } {
                # Added to avoid that the redraw calls again gid_groups_conds::drawall_create_object,
                # that creates the OpenGL drawing lists.
                after idle {GiD_Process 'Render Normal}
            }
        } else {
            draw_post::change_view_style $opengl_rendermode
        }
        set opengl_rendermode ""
    }
    if { !$is_draw_post } {
        set found [delayedop -cancel changefunc gid_groups_conds::end_drawall]
        if { $found } { GiD_Process escape escape }
        if { ![info exists ::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE] } {
            set ::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE 0
        }
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS_DRAW_NATIVE } {
            GiD_Process Mescape
        }
    }
}

proc gid_groups_conds::end_drawall {} {
    variable opengl_object
    variable opengl_id
    variable show_figures_surfaces
    variable opengl_groupList
    variable opengl_rendermode
    variable draw_nodes_state
    variable is_draw_post

    if { [info exists opengl_groupList] } {
        foreach "n v" [array get opengl_groupList] {
            foreach d $v {
                if { [dict exists $d obj] } {
                    GiD_OpenGL draw -deletelists [dict get $d obj] 1
                }
            }
        }
        unset opengl_groupList
    }
    if { [info exists opengl_object] } {
        GiD_OpenGL draw -deletelists $opengl_object 1
        unset opengl_object
    }

    # GID change: moved here so that the 'redraw renders geometry correctly
    # and does not call an empty & deleted object
    gid_groups_conds::end_drawall_do

    if { [info exists opengl_id] } {
        if { !$is_draw_post } {
            GiD_OpenGL unregister $opengl_id
        } else {
            draw_post::draw_point_line_indicator unset $opengl_id
        }
        unset opengl_id
        if { !$is_draw_post } {
            GiD_Process 'Redraw
        } else {
            draw_post::redraw_all
        }
    }
    set draw_nodes_state ""
}

proc gid_groups_conds::drawall_create_object { groups } {
    variable doc
    variable opengl_object
    variable is_draw_post

    set opengl_object [GiD_OpenGL draw -genlists 1]
    GiD_OpenGL draw -newlist $opengl_object compile

    set root [$doc documentElement]
    if { [lindex $groups 0] eq "ALL" } {
        set ov [lindex $groups 1]
        set groups ""
        if { !$is_draw_post } {
            foreach domNode [$root selectNodes groups/group] {
                if { [$domNode @onoff 1] } {
                    lappend groups [list [$domNode @n] $ov]
                }
            }
        } else {
            foreach "gr d" [ggc::manage_geometry_groups list] {
                if { [dict get $d IsOn] } {
                    lappend groups [list [dict get $d Name] $ov]
                }
            }
        }
    }
    set groups [lsort -dictionary -index 0 $groups]
    set groups_names ""
    foreach i $groups {
        lappend groups_names [lindex $i 0]
    }
    if { !$is_draw_post } {
        foreach domNode [$root selectNodes groups/group] {
            set ipos [lsearch -exact -sorted -dictionary $groups_names [$domNode @n]]
            if { $ipos == -1 } { continue }
            if { [$domNode @onoff 1] } {
                set colorgroup([$domNode @n]) [$domNode @color blue]
                set ovs([$domNode @n]) [lindex $groups $ipos 1]
                set numents([$domNode @n]) 0
            } else {
                set colorgroup([$domNode @n]) [$domNode @color blue]
                set ovs([$domNode @n]) [lindex $groups $ipos 1]
                set numents([$domNode @n]) 0
            }
        }
    } else {
        foreach "gr d" [ggc::manage_geometry_groups list] {
            set ipos [lsearch -exact -sorted -dictionary $groups_names [dict get $d Name]]
            if { $ipos == -1 } { continue }
            if { [dict get $d IsOn] } {
                set colorgroup([dict get $d Name]) [draw_post::_give_color_tk_from_geometry [dict get $d Color]]
                set ovs([dict get $d Name]) [lindex $groups $ipos 1]
                set numents([dict get $d Name]) 0
            }
        }
    }


    switch [is_geometry_or_mesh] {
        geometry {
            if { !$is_draw_post } {
                if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
                    foreach ov {point line surface volume} {
                        foreach group $groups_names {
                            if { [info exists colorgroup($group)] && [lsearch -exact $ovs($group) $ov] != -1 } {
                                set err [catch { set group_entities [GiD_EntitiesGroups get $group ${ov}s -visible]}]
                                if { $err } {
                                    set group_entities [GiD_EntitiesGroups get $group ${ov}s]
                                }
                                if { ![llength $group_entities] && [GiD_Groups get state $group] == "hidden" && [GiD_Groups get parent $group] != "" } {
                                    #dark kratos trick using a parent group with entities and empty
                                    # If the group to draw has state equal to hidden (only this state can be set from code, and the group
                                    # is not shown in the group window), and it also has an empty parent group,
                                    # then the entities of the parent are drawn.
                                    set group_parent [GiD_Groups get parent $group]
                                    set group_entities [GiD_EntitiesGroups get $group_parent ${ov}s -visible]
                                }
                                foreach entity_id $group_entities {
                                    GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $colorgroup($group)]
                                    GiD_OpenGL drawentity -mode filled $ov $entity_id
                                    incr numents($group)
                                }
                            }
                        }
                    }
                } else {
                    foreach ov [list point line surface volume] {
                        foreach j [GiD_Info conditions ${ov}_groups geometry] {
                            foreach "- num - group" $j break
                            if { [info exists colorgroup($group)] && [lsearch -exact $ovs($group) $ov] != -1 } {
                                GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $colorgroup($group)]
                                GiD_OpenGL drawentity -mode filled $ov $num
                                incr numents($group)
                            }
                        }
                    }
                }
            } else {
                foreach "group color" [array get colorgroup] {
                    GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $colorgroup($group)]
                    foreach ov $ovs($group) {
                        ggc::draw::geometry_group $group $ov
                        set gr [ggc::manage_geometry_groups id $group]
                        lassign [ggc::manage_geometry_groups entities count $gr] nP nL nS nV nDim nAll
                        switch $ov {
                            point { incr numents($group) $nP }
                            line { incr numents($group) $nL }
                            surface { incr numents($group) $nS }
                            volume { incr numents($group) $nV }
                        }
                    }
                }
            }
        }
        mesh {
            if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
                foreach ov {node element} {
                    foreach group $groups_names {
                        if { [info exists colorgroup($group)] } {
                            if { ($ovs($group) != "point" && $ov != "node") || ($ovs($group) == "point" && $ov == "node") } {
                                #try to draw if ov==node and the condition is ov="point"
                                #            or ov==element and  the condition is ov="line" or "surface" or "volume"
                                #must check for each element if it is of level type point, linear, surfacic or volumic...
                                set err [catch { set group_entities [GiD_EntitiesGroups get $group ${ov}s -visible]}]
                                if { $err } {
                                    set group_entities [GiD_EntitiesGroups get $group ${ov}s]
                                }
                                if { ![llength $group_entities] && [GiD_Groups get state $group] == "hidden" && [GiD_Groups get parent $group] != "" } {
                                    #dark kratos trick using a parent group with entities and empty
                                    set group_parent [GiD_Groups get parent $group]
                                    set err [catch { set group_entities [GiD_EntitiesGroups get $group_parent ${ov}s -visible]}]
                                    if { $err } {
                                        set group_entities [GiD_EntitiesGroups get $group_parent ${ov}s]
                                    }
                                }
                                foreach entity_id $group_entities {
                                    GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $colorgroup($group)]
                                    GiD_OpenGL drawentity -mode filled $ov $entity_id
                                    incr numents($group)
                                }
                            }
                        }
                    }
                }
                set ov face
                foreach group $groups_names {
                    if { [info exists colorgroup($group)] } {
                        if { $ovs($group) != "point" } {
                            #try tp draw face if the condition is ov="line" or "surface" or "volume"
                            #must check for each element if it is of level type point->linear, line->surfacic, surface->volumic
                            set err [catch { set elements_and_faces [GiD_EntitiesGroups get $group ${ov}s -visible]}]
                            if { $err } {
                                set elements_and_faces [GiD_EntitiesGroups get $group ${ov}s]
                            }
                            foreach element_id [lindex $elements_and_faces 0] face_id [lindex $elements_and_faces 1] {
                                set entity_id [list [list $element_id $face_id]]
                                GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $colorgroup($group)]
                                GiD_OpenGL drawentity -mode filled element $entity_id
                                incr numents($group)
                            }
                        }
                    }
                }
            } else {
                foreach ov [list point line surface volume] {
                    switch $ov {
                        point { set ov_p node }
                        line - surface - volume { set ov_p element }
                    }
                    foreach j [GiD_Info conditions ${ov}_groups mesh] {
                        foreach "num face - group" $j break
                        if { $num eq "E" } {
                            set num $face
                            set face null
                            set ov_pL $ov_p
                        } elseif { $num eq "N" } {
                            set num $face
                            set face null
                            set ov_pL node
                        } else {
                            set num [list [list $num $face]]
                            set ov_pL $ov_p
                        }
                        if { [info exists colorgroup($group)] && [lsearch -exact $ovs($group) $ov] != -1 } {
                            GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $colorgroup($group)]
                            GiD_OpenGL drawentity -mode filled $ov_pL $num
                            incr numents($group)
                        }
                    }
                }
            }
        }
    }

    draw_box_for_bc $groups colorgroup numents

    GiD_OpenGL draw -endlist
}

################################################################################
#
#  To create a symbol for the problemtype, follow these steps:
#
#  1- Create a model in GiD centered in point 0,0,0 and contained in a box
#     of size 1 for every of its dimensions (it can be smaller or bigger to
#     reflect a different size of the symbol
#
#  2- Generate a mesh. It can contain line elements, triangles and quadrilaterals
#     do use as few elements as possible
#
#  3- Export a GiD mesh (Files->Export->GiD mesh) and give it a name of:
#     "symbol.msh".
#
#  4- call proc: gid_groups_conds::import_gid_mesh_as_openGL with arguments:
#     file color_lines color_surfaces
#
################################################################################

proc gid_groups_conds::import_gid_mesh_as_openGL { file color_lines color_surfaces } {
    set fin [open $file r]
    set elemtype ""
    set needs_pop_att 0
    while { [gets $fin line] != -1 } {
        if { [string match -nocase "MESH*" $line] } {
            if { $elemtype ne "" } { GiD_OpenGL draw "-end" }
            if { $needs_pop_att } {
                GiD_OpenGL draw -popattrib
                set needs_pop_att 0
            }
            scan $line "%s %s %d %s %s %s %d" - - - - elemtype - nnode
            switch $elemtype {
                Linear {
                    if { $color_lines ne "" } {
                        GiD_OpenGL draw -pushattrib currentbit \
                            -color [gid_groups_conds::_OpenGL_color $color_lines]
                        set needs_pop_att 1
                    }
                    GiD_OpenGL draw -begin lines
                }
                Triangle {
                    if { $color_surfaces ne "" } {
                        GiD_OpenGL draw -pushattrib currentbit \
                            -color [gid_groups_conds::_OpenGL_color $color_surfaces]
                        set needs_pop_att 1
                    }
                    GiD_OpenGL draw -begin triangles
                }
                Quadrilateral {
                    if { $color_surfaces ne "" } {
                        GiD_OpenGL draw -pushattrib currentbit \
                            -color [gid_groups_conds::_OpenGL_color $color_surfaces]
                        set needs_pop_att 1
                    }
                    GiD_OpenGL draw -begin quads
                }
                default {
                    error "error in gid_groups_conds::import_gid_mesh_as_openGL. Elemtype '$elemtype' not accepted"
                }
            }
        } elseif { [string equal -nocase $line "Coordinates"] } {
            while { [gets $fin line] != -1 } {
                if { [string match -nocase "end*" $line] } { break }
                scan $line "%d %f %f %f" num x y z
                set point($num) [format "%.3g %.3g %.3g" $x $y $z]
            }
        } elseif { [string equal -nocase $line "Elements"] } {
            switch $elemtype {
                Linear { set format "%d %d %d" }
                Triangle { set format "%d %d %d %d" }
                Quadrilateral { set format "%d %d %d %d %d" }
            }
            while { [gets $fin line] != -1 } {
                set ret [scan $line $format]
                if { [lindex $ret 0] eq "" } { break }
                if { [llength $ret] > 3 } {
                    foreach "p1 p2 p3" [lrange $ret 1 3] break
                    set n [m::vectorprod3 [m::sub $point($p2) $point($p1)] \
                        [m::sub $point($p3) $point($p1)]]
                    GiD_OpenGL draw -normal [m::unitLengthVector $n]
                }
                foreach i [lrange $ret 1 end] {
                    GiD_OpenGL draw -vertex $point($i)
                }
            }
        }
    }
    if { $elemtype ne "" } { GiD_OpenGL draw "-end" }
    if { $needs_pop_att } {
        GiD_OpenGL draw -popattrib
        set needs_pop_att 0
    }
    close $fin
}

proc gid_groups_conds::draw_symbol_image { image valuesList } {
    variable draw_symbol_image_list

    if { ![info exists draw_symbol_image_list($image)] } {
        set img [giveimageL $image]
        set dest_img [image create photo]
        cu::img::swapy $dest_img $img
        set draw_symbol_image_list($image) $dest_img
    }
    set l [GiD_OpenGL draw -genlists 1]
    GiD_OpenGL draw -newlist $l compile
    GiD_OpenGL draw -rasterpos "0 0 0"
    GiD_OpenGL draw -drawpixels $draw_symbol_image_list($image)
    GiD_OpenGL draw -endlist
    return $l
}

proc gid_groups_conds::draw_symbol_text { txt valuesList } {

    set l [GiD_OpenGL draw -genlists 1]
    GiD_OpenGL draw -newlist $l compile
    GiD_OpenGL draw -rasterpos "0 0 0"
    GiD_OpenGL drawtext "$txt"
    GiD_OpenGL draw -endlist
    return $l
}

proc gid_groups_conds::_opengl_rotation_mat { LAmat } {

    if { $LAmat eq "" } { return "" }

    foreach "xvec yvec zvec" [list "" "" ""] break
    foreach "c1 c2 c3" $LAmat {
        lappend xvec $c1
        lappend yvec $c2
        lappend zvec $c3
    }
    # mat is 4x4 = 16 in column order (OpenGL matrix)
    set mat [concat $xvec 0.0]
    eval lappend mat $yvec 0.0
    eval lappend mat $zvec 0.0
    lappend mat 0.0 0.0 0.0 1.0
    return $mat
}

proc gid_groups_conds::_line_opengl_rotation_mat { p1 p2 } {

    set iprima [m::unitLengthVector [m::sub $p2 $p1]]
    if { [lindex $iprima 0] != 0 } {
        set jprima [list [expr {-1*[lindex $iprima 1]/double([lindex $iprima 0])}] 1.0 0.0]
        set jprima [m::unitLengthVector $jprima]
    } else {
        set jprima [list 1.0 0.0 0.0]
    }
    set kprima [m::vectorprod3 $iprima $jprima]

    # mat is 4x4 = 16 in column order (OpenGL matrix)
    set mat [concat $iprima 0.0]
    eval lappend mat $jprima 0.0
    eval lappend mat $kprima 0.0
    lappend mat 0.0 0.0 0.0 1.0
    return $mat
}

proc gid_groups_conds::_surface_opengl_rotation_mat { normal } {

    set kprima [m::unitLengthVector $normal]
    if { [lindex $kprima 0] != 0 || [lindex $kprima 2] != 0 } {
        set iprima [m::unitLengthVector [m::vectorprod3 [list 0.0 1.0 0.0] $kprima]]
    } else {
        set iprima [m::unitLengthVector [m::vectorprod3 [list 0.0 0.0 1.0] $kprima]]
    }
    set jprima [m::vectorprod3 $kprima $iprima]

    # mat is 4x4 = 16 in column order
    set mat [concat $iprima 0.0]
    eval lappend mat $jprima 0.0
    eval lappend mat $kprima 0.0
    lappend mat 0.0 0.0 0.0 1.0
    return $mat
}

proc gid_groups_conds::give_draw_boundary_for_group { args } {
    variable doc

    set optional {
        { -give_section_shell boolean 1 }
        { -give_valuesList boolean 1 }
        { -scale double "" }
    }
    set compulsory "groupList"
    set args [parse_args -raise_compulsory_error 0 $optional $compulsory $args]

    set grNodes ""
    foreach group $groupList {
        set xp [format_xpath {//group[@n=%s and ancestor::condition]} $group]
        lappend grNodes {*}[$doc selectNodes $xp]
    }

    set dList ""
    foreach grNode $grNodes {
        set cndNode [$grNode selectNodes ancestor::condition]
        set xp {ancestor::groupList/group}
        set grs [$grNode selectNodes $xp]
        if { [llength $grs] } {
            set ipos [lsearch -exact $grs $grNode]
        } else {
            set ipos ""
        }
        set ov [$grNode @ov ""]
        if { $ov eq "" } {
            set ov [get_domnode_attribute $cndNode ov$ipos]
        }
        if { $ipos eq "" } {
            set symbolNode [lindex [$cndNode selectNodes symbol] 0]
        } else {
            set symbolNode [$cndNode selectNodes [format_xpath {symbol[@pos=%s]} \
                [expr {$ipos+1}]]]
            if { $symbolNode eq "" } {
                set symbolNode [$cndNode selectNodes symbol]
            }
        }
        if { $symbolNode eq "" } { continue }
        if { [$symbolNode @orientation] eq "free" } { continue }
        if { !$give_section_shell && [$symbolNode @orientation] in "section shell_thickness" } { continue }

        set valuesList [_valuesList $grNode]
        set orientation [$symbolNode @orientation]

        if { [string match "localLA*" $orientation] } {
            lassign $orientation - nn v
            set ipos [lsearch -index 0 $valuesList $nn]
            if { [lindex $valuesList $ipos 1] != $v } {
                set orientation "global"
            } else {
                set orientation "localLA"
            }
        }
        set d [dict create ov $ov orientation $orientation]
        if { $give_valuesList } {
            dict set d valuesList $valuesList
        }
        if { $scale ne "" } {
            dict set d scale $scale
        }
        set ret [uplevel #0 [$symbolNode @proc] [list $valuesList]]
        if { [string is integer -strict $ret] } {
            dict set d obj $ret
        } else {
            dict for "n v" $ret {
                dict set d $n $v
            }
        }
        lappend dList $d
    }
    return $dList
}

proc gid_groups_conds::draw_entities { groups } {
    variable doc
    variable opengl_object
    variable opengl_groupList
    variable is_draw_post

    set root [$doc documentElement]
    foreach i $groups {
        lassign $i group ov valuesList symbol
        if { !$is_draw_post } {
            set groupNode [$root selectNodes [format_xpath {groups/group[@n=%s]} $group]]
            if { !$gid_groups_conds::view_entities } {
                if { [$groupNode @onoff 1] == 0 } { continue }
            }
            set color [$groupNode @color blue]
        } else {
            set gr [ggc::manage_geometry_groups id $group]
            lassign [ggc::manage_geometry_groups list $gr] - d_gr
            if { [dict get $d_gr IsOn] == 0 } { continue }
            set color [draw_post::_give_color_tk_from_geometry [dict get $d_gr Color]]
        }
        if { ![info exists opengl_groupList($group)] } {
            set opengl_groupList($group) ""
        }
        lassign $symbol proc orientation
        if { $proc eq "" } { continue }

        set d [dict create ov $ov color $color orientation $orientation valuesList $valuesList]
        if { $orientation ne "free" } {
            set ret [uplevel #0 $proc [list $valuesList]]
            if { [string is integer -strict $ret] } {
                dict set d obj $ret
            } else {
                dict for "n v" $ret {
                    dict set d $n $v
                }
            }
        } else {
            dict set d proc [list {*}$proc $valuesList]
        }
        lappend opengl_groupList($group) $d
    }
    set opengl_object [GiD_OpenGL draw -genlists 1]
    GiD_OpenGL draw -newlist $opengl_object compile

    set scale [projection_scale 15]

    switch [is_geometry_or_mesh] {
        geometry {
            set err [catch { gid_groups_conds::draw_entities_geometry $scale } errstring]
        }
        mesh {
            set err [catch { gid_groups_conds::draw_entities_mesh $scale } errstring]
        }
    }
    GiD_OpenGL draw -endlist
    if { $err } {
        GID_tk_messageBox -message $::errorInfo {*}[ GidUtils::GiDDefaultParentIfPresent]
    }
}

proc gid_groups_conds::projection_scale { num_pixels } {

    set P1 [GiD_OpenGL project "0 0 0"]
    set P2 [GiD_OpenGL project "1 0 0"]
    set P3 [GiD_OpenGL project "0 1 0"]
    set P4 [GiD_OpenGL project "0 0 1"]

    set norm [m::norm [m::sub $P2 $P1]]
    set normT [m::norm [m::sub $P3 $P1]]
    if { $normT > $norm } { set norm $normT }

    set normT [m::norm [m::sub $P4 $P1]]
    if { $normT > $norm } { set norm $normT }

    return [expr {$num_pixels/$norm}]
}

proc gid_groups_conds::draw_entities_geometry { scale } {
    if { 1 } {
        #it seems that this procedure draw using a GiD C++ command "draw_model_conditions"
        #and then variable opengl_groupList become empty and the rest of the code to does nothing !!
        #move this rest of code to gid_groups_conds::draw_entities_old_style_geometry  !!
        gid_groups_conds::draw_entities_new_style geometry $scale
    } else {
        gid_groups_conds::draw_entities_old_style_geometry $scale
    }
    return
}

proc gid_groups_conds::draw_entities_old_style_geometry { scale } {
    variable opengl_object
    variable opengl_groupList
    variable is_draw_post
    foreach ovG [list point line surface volume] {
        set entsDict ""
        if { !$is_draw_post } {
            if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
                foreach group [GiD_Groups list] {
                    if { ![info exists opengl_groupList($group)] } { continue }
                    foreach num [GiD_EntitiesGroups get $group ${ovG}s -visible] {
                        dict lappend entsDict $group $num
                    }
                }
            }
        } else {
            foreach group [array names opengl_groupList] {
                set gr [ggc::manage_geometry_groups id $group]
                dict set entsDict $group [ggc::manage_geometry_groups entities list_group $gr $ovG]
            }
        }
        dict for "group numList" $entsDict {
            foreach num $numList {
                foreach d $opengl_groupList($group) {
                    dict for "n v" $d { set $n $v }
                    if { [lsearch $ov $ovG] == -1 } { continue }
                    GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $color]
                    lassign "" entities pointsList ent_type mat
                    switch $ovG {
                        point {
                            if { !$is_draw_post } {
                                set P [lrange [GiD_Geometry get point $num] 1 3]
                            } else {
                                lassign [ggc::manage_geometry_entities list details point $num] - d_geom
                                set P [dict get $d_geom PointCoords]
                            }
                            if { [string match "localLA*" $orientation] } {
                                lassign $orientation - n v
                                set ipos [lsearch -index 0 $valuesList $n]
                                if { [lindex $valuesList $ipos 1] == $v } {
                                    set mat [_opengl_rotation_mat [local_axes_matrix point geometry $num]]
                                }
                            }
                        }
                        line {
                            if { !$is_draw_post } {
                                set list_ent [GiD_Info list_entities Lines $num]
                                regexp {Points:\s+([0-9]+)\s+([0-9]+)} $list_ent {} P1 P2
                                foreach j "P1 P2" {
                                    set list_pnt [GiD_Info list_entities Points [set $j]]
                                    regexp {Coord:\s+([^\n]+)} $list_pnt {} coords
                                    set points([set $j]) $coords
                                }
                                regexp {^\S+} $list_ent ent_type
                                set P [GiD_Info parametric line $num coord 0.5]
                            } else {
                                lassign [ggc::manage_geometry_entities list geometry line $num] - d_geom
                                lassign [dict get $d_geom ContourEntities] P1 P2
                                if { [dict get $d_geom Degree] ==1 && [dict get $d_geom NumCtrlPoints] == 2 } {
                                    set ent_type STLINE
                                } else {
                                    set ent_type NURBLINE
                                }
                                set P [dict get $d_geom Center]
                                foreach j "P1 P2" {
                                    lassign [ggc::manage_geometry_entities list details point [set $j]] - d_geomP
                                    set points([set $j]) [dict get $d_geomP PointCoords]
                                }
                            }
                            if { $orientation eq "local" } {
                                set mat [_opengl_rotation_mat [local_axes_matrix line geometry $num]]
                                if { $mat eq "" } {
                                    set mat [_line_opengl_rotation_mat $points($P1) \
                                        $points($P2)]
                                }
                            } elseif { [string match "localLA*" $orientation] } {
                                lassign $orientation - n v
                                set ipos [lsearch -index 0 $valuesList $n]
                                if { [lindex $valuesList $ipos 1] == $v } {
                                    set mat [_opengl_rotation_mat [local_axes_matrix line geometry $num]]
                                }
                            }
                            set entities [list $P1 $P2]
                            set pointsList [list $points($P1) $points($P2)]
                        }
                        surface {
                            lassign "" entities pointsList
                            if { !$is_draw_post } {
                                set list_ent [GiD_Info list_entities Surfaces $num]
                                regexp {Center:\s+(\S+)\s+(\S+)\s+(\S+)} $list_ent {} x y z
                                set P [list $x $y $z]
                                regexp {Normal:\s+(\S+)\s+(\S+)\s+(\S+)} $list_ent {} x y z
                                set normal [list $x $y $z]
                                foreach "- l or" [regexp -all -inline -line \
                                    {^Line:\s+(\d+)\s+Orientation:\s+(\S+)} $list_ent] {
                                        lappend entities [list $l $or]
                                    set list_ent [GiD_Info list_entities Lines $l]
                                    regexp {Points:\s+([0-9]+)\s+([0-9]+)} $list_ent {} P1 P2
                                    switch $or {
                                        SAME1ST { set p $P1 }
                                        DIFF1ST { set p $P2 }
                                    }
                                    set list_pnt [GiD_Info list_entities Points $p]
                                    regexp {Coord:\s+([^\n]+)} $list_pnt {} coords
                                    lappend pointsList $coords
                                }
                            } else {
                                lassign [ggc::manage_geometry_entities list geometry surface $num] - d_geom
                                set P [dict get $d_geom Center]
                                set normal [dict get $d_geom Normal]
                                foreach l [dict get $d_geom ContourEntities] or [dict get $d_geom ContourEntitiesOrient] {
                                    switch $or 0 { set orG SAME1ST } 1 { set orG DIFF1ST }
                                    lappend entities [list $l $orG]
                                    lassign [ggc::manage_geometry_entities list geometry line $l] - d_geomL
                                    lassign [dict get $d_geomL ContourEntities] P1 P2
                                    switch $or 0 { set p $P1 } 1 { set p $P2 }
                                    lassign [ggc::manage_geometry_entities list details point $p] - d_geomP
                                    lappend pointsList [dict get $d_geomP PointCoords]
                                }
                            }
                            if { $orientation eq "local" } {
                                set mat [_opengl_rotation_mat [local_axes_matrix surface geometry $num]]
                                if { $mat eq "" } {
                                    set mat [_surface_opengl_rotation_mat $normal]
                                }
                            } elseif { [string match "localLA*" $orientation] } {
                                lassign $orientation - n v
                                set ipos [lsearch -index 0 $valuesList $n]
                                if { [lindex $valuesList $ipos 1] == $v } {
                                    set mat [_opengl_rotation_mat [local_axes_matrix surface geometry $num]]
                                }
                            }
                        }
                        volume {
                            if { !$is_draw_post } {
                                set list_ent [GiD_Info list_entities Volumes $num]
                                regexp {Center:\s+(\S+)\s+(\S+)\s+(\S+)} $list_ent {} x y z
                                set P [list $x $y $z]
                            } else {
                                lassign [ggc::manage_geometry_entities list geometry volume $num] - d_geom
                                set P [dict get $d_geom Center]
                            }
                            if { [string match "localLA*" $orientation] } {
                                lassign $orientation - n v
                                set ipos [lsearch -index 0 $valuesList $n]
                                if { [lindex $valuesList $ipos 1] == $v } {
                                    set mat [_opengl_rotation_mat [local_axes_matrix volume geometry $num]]
                                }
                            }
                        }
                    }
                    if { $orientation ne "free" } {
                        GiD_OpenGL draw -pushmatrix -translate $P
                        if { $mat ne "" } {
                            GiD_OpenGL draw -multmatrix $mat
                        }
                        GiD_OpenGL draw -scale [list $scale $scale $scale]
                        GiD_OpenGL draw -call $obj
                        GiD_OpenGL draw -popmatrix
                    } else {
                        uplevel #0 $proc [list GEOMETRYUSE $ovG $num \
                            $entities $pointsList $ent_type $P $scale]
                    }
                }
            }
        }
    }
    return
}

set FaceDescriptorTetra {
    { 1 2 3 5 6 7 }
    { 2 4 3 9 10 6 }
    { 3 4 1 10 8 7 }
    {4 2 1 9 5 8 }
}

set FaceDescriptorHexa {
    { 1 2 3 4 9 10 11 12 21 }
    { 1 4 8 5 12 16 20 13 25 }
    { 1 5 6 2 13 17 14 9 22 }
    { 2 6 7 3 14 18 15 10 23 }
    { 3 7 8 4 15 19 16 11 24 }
    { 5 8 7 6 20 19 18 17 26 }
}

proc gid_groups_conds::give_face_nodes { nodes quadratic elemtype face } {

    incr face -1
    set ret ""
    switch $elemtype {
        Triangle - Quadrilateral {
            switch $elemtype Triangle { set nnodec 3 } Quadrilateral { set nnodec 4 }
            set ret [lrange $nodes $face [expr {$face+1}]]
            eval lappend ret [lrange $nodes [expr {$face-$nnodec}] \
                [expr {$face+1-$nnodec}]]
            if { $quadratic > 0 } {
                lappend ret [lindex $nodes [expr {$face+$nnodec}]]
            }
        }
        Tetrahedra {
            switch $quadratic 0 { set n 2 } default { set n 5 }
            foreach i [lrange [lindex $::FaceDescriptorTetra $face] 0 $n] {
                lappend ret [lindex $nodes [expr {$i-1}]]
            }
        }
        Hexahedra {
            switch $quadratic 0 { set n 3 } 1 { set n 7 } 2 { set n 8 }
            foreach i [lrange [lindex $::FaceDescriptorHexa $face] 0 $n] {
                lappend ret [lindex $nodes [expr {$i-1}]]
            }
        }
    }
    return $ret
}

proc gid_groups_conds::draw_entities_new_style { geometry_mesh scale } {
    variable opengl_groupList

    if { [info command ::draw_model_conditions] ne "" } {
        set groupsDict ""
        foreach group [array names opengl_groupList] {
            lassign "" new_list old_list
            foreach d $opengl_groupList($group) {
                dict for "n v" $d { set $n $v }
                if { [string match "localLA*" $orientation] } {
                    lassign $orientation - nn v
                    set ipos [lsearch -index 0 $valuesList $nn]
                    if { [lindex $valuesList $ipos 1] != $v } {
                        set orientation "global"
                    } else {
                        set orientation "localLA"
                    }
                }
                if { $orientation ni "global local localLA section shell_thickness loads" } {
                    lappend old_list $d
                    continue
                }
                set dL [dict create ov $ov color [_OpenGL_color4 $color] orientation $orientation]
                if { [dict exists $d obj] } { dict set dL object $obj }

                if { $orientation eq "shell_thickness" } {
                    dict set dL thickness $thickness
                    if { [dict exists $d cdg_pos] } {
                        dict set dL cdg_pos [dict get $d cdg_pos]
                    }
                }
                if { $orientation eq "loads" } {
                    dict set dL load_type $load_type
                    dict set dL load_vector $load_vector
                }
                lappend new_list $dL
            }
            if { $new_list ne "" } {
                dict set groupsDict $group $new_list
            }
            set opengl_groupList($group) $old_list
        }
        draw_model_conditions $geometry_mesh $scale $groupsDict
    }
}

proc gid_groups_conds::draw_entities_mesh { scale } {
    if { 1 } {
        #it seems that this procedure draw using a  GiD C++ command "draw_model_conditions"
        #and then variable opengl_groupList become empty and the rest of the code to does nothing
        #move this rest of code to gid_groups_conds::draw_entities_old_style_mesh  !!
        gid_groups_conds::draw_entities_new_style mesh $scale
    } else {
        gid_groups_conds::draw_entities_old_style_mesh $scale
    }
}

proc gid_groups_conds::draw_entities_old_style_mesh { scale } {
    variable opengl_object
    variable opengl_groupList


    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        set element_level_type(linear) line
        foreach element_type [GidUtils::GetElementTypesLowercase surface] {
            set element_level_type($element_type) surface
        }
        foreach element_type [GidUtils::GetElementTypesLowercase volume] {
            set element_level_type($element_type) volume
        }

        set ov node
        set ovG point ;#??
        foreach group [GiD_Groups list] {
            if { ![info exists opengl_groupList($group)] } { continue }
            foreach node_id [GiD_EntitiesGroups get $group ${ov}s -visible] {
                set nodes $node_id
                gid_groups_conds::draw_entities_mesh_part $scale $ovG $nodes $group
            }
        }
        set ov element
        foreach group [GiD_Groups list] {
            if { ![info exists opengl_groupList($group)] } { continue }
            foreach element_id [GiD_EntitiesGroups get $group ${ov}s -visible] {
                set element_data [GiD_Mesh get element $element_id]
                set element_type [string tolower [lindex $element_data 1]]
                set nodes [lrange $element_data 3 end]
                if { ![info exists element_level_type($element_type)] } {
                    W "gid_groups_conds::draw_entities_mesh. unexpected element_type $element_type"
                } else {
                    set ovG $element_level_type($element_type)
                    gid_groups_conds::draw_entities_mesh_part $scale $ovG $nodes $group
                }
            }
        }
        set ov face
        foreach group [GiD_Groups list] {
            if { ![info exists opengl_groupList($group)] } { continue }
            set elements_and_faces [GiD_EntitiesGroups get $group ${ov}s -visible]
            foreach element_id [lindex $elements_and_faces 0] face_id [lindex $elements_and_faces 1] {
                set element_data [GiD_Mesh get element $element_id]
                set element_type [string tolower [lindex $element_data 1]]
                set ns [lrange $element_data 3 end]
                set nodes [give_face_nodes $ns 0 $element_type $face_id]
                if { ![info exists element_level_type($element_type)] } {
                    W "gid_groups_conds::draw_entities_mesh. unexpected element_type $element_type"
                } else {
                    set ovG $element_level_type($element_type)
                    gid_groups_conds::draw_entities_mesh_part $scale $ovG $nodes $group
                }
                gid_groups_conds::draw_entities_mesh_part $scale $ovG $nodes $group
            }
        }
    } else {
        foreach ovG [list point line surface volume] {
            foreach j [GiD_Info conditions ${ovG}_groups mesh] {
                lassign $j num face - group
                if { $num eq "E" || $num eq "N" } {
                    set num $face
                    set face null
                }
                if { ![info exists opengl_groupList($group)] } { continue }
                if { $ovG eq "point" } {
                    set nodes $num
                } elseif { $face eq "null" } {
                    set nodes [lrange [GiD_Info Mesh Elements Any $num] 1 end-1]
                } else {
                    set ns [lrange [GiD_Info Mesh Elements Any $num] 1 end-1]
                    switch $ovG {
                        line {
                            switch [llength $ns] {
                                3 - 6 { set elemtype Triangle }
                                4 - 8 - 9 { set elemtype Quadrilateral }
                            }
                        }
                        surface {
                            switch [llength $ns] {
                                4 - 10 { set elemtype Tetrahedra }
                                8 - 20 - 27 { set elemtype Hexahedra }
                            }
                        }
                    }
                    set nodes [give_face_nodes $ns 0 $elemtype $face]
                }
                gid_groups_conds::draw_entities_mesh_part $scale $ovG $nodes $group
            }
        }
    }
}

#auxiliary funcion
proc gid_groups_conds::draw_entities_mesh_part { scale ovG nodes group } {
    variable opengl_object
    variable opengl_groupList

    foreach d $opengl_groupList($group) {
        dict for "nd vd" $d { set $nd $vd }
        if { [lsearch $ov $ovG] == -1 } { continue }
        GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $color]
        set mid [list 0 0 0]
        set idx 1
        set pointsList ""
        foreach node $nodes {
            set n($idx) [lrange [GiD_Info Mesh Nodes $node] 1 end]
            set mid [m::add $mid $n($idx)]
            lappend pointsList $n($idx)
            incr idx
        }
        if { [llength $nodes] <= 1 } {
            set P $mid
        } elseif { [llength $nodes] == 2 } {
            set P [m::scale [expr {1.0/[llength $nodes]}] $mid]
        } else {
            set P [m::scale [expr {1.0/[llength $nodes]}] $mid]
        }
        if { $orientation ne "free" } {
            set mat ""
            if { $orientation eq "local" } {
                set ret [GiD_Info conditions -localaxesmat ${ovG}_Local_axes mesh $num]

                set mat [_opengl_rotation_mat [lindex $ret 0 3]]
                if { $mat eq "" } {
                    if { [llength $nodes] == 2 } {
                        set mat [_line_opengl_rotation_mat $n(1) $n(2)]
                    } else {
                        set normal [m::vectorprod3 [m::sub $n(2) $n(1)] [m::sub $n(3) $n(1)]]

                        set mat [_surface_opengl_rotation_mat $normal]
                    }
                }
            } elseif { [string match "localLA*" $orientation] } {
                foreach "- nn v" $orientation break
                set ipos [lsearch -index 0 $valuesList $nn]
                if { [lindex $valuesList $ipos 1] == $v } {
                    set ret [GiD_Info conditions -localaxesmat ${ovG}_Local_axes mesh $num]

                    set mat [_opengl_rotation_mat [lindex $ret 0 3]]
                }
            }
            GiD_OpenGL draw -pushmatrix -translate $P
            if { $mat ne "" } {
                GiD_OpenGL draw -multmatrix $mat
            }
            GiD_OpenGL draw -scale [list $scale $scale $scale]
            GiD_OpenGL draw -call $obj
            GiD_OpenGL draw -popmatrix
        } else {
            uplevel #0 $proc [list MESHUSE $ovG $num $nodes $pointsList ELEMENT $P $scale]
        }
    }
}


proc gid_groups_conds::_values_to_string { valuesList } {
    set len 0
    set vstring ""
    foreach i $valuesList {
        set val [lindex $i 1]
        incr len [string length $val]
        if { $len > 100 } {
            break
        }
        if { $vstring ne "" } { append vstring " " }
        append vstring $val
    }
    return $vstring
}

proc gid_groups_conds::give_graph_window {} {
    if { [winfo exists .gid] } {
        return .gid.central.s
    } else {
        return [draw_full::give_graph_window]
    }
}

proc gid_groups_conds::draw_box_for_bc { groups colorgroup_name numents_name } {
    variable gl_colours

    upvar 1 $colorgroup_name colorgroup
    upvar 1 $numents_name numents

    set wg [gid_groups_conds::give_graph_window]

    set sizew [winfo width $wg]
    set sizeh [winfo height $wg]

    GiD_OpenGL draw -pushattrib "viewportbit enablebit" -disable lighting \
        -disable depthtest

    gid_groups_conds::draw_text push

    set maxwidth 255
    set linespace [gid_groups_conds::draw_text linespace]
    set boxsep [expr {$linespace/2}]

    set boxwidth 0
    set boxheight 3
    foreach i $groups {
        lassign $i group - valuesList
        set txt [_values_to_string $valuesList]
        if { $txt eq "" } { set txt $group }
        set txt "$txt ($numents($group))"
        set w [expr {[gid_groups_conds::draw_text measure "$txt"]+3.0*$boxsep}]
        if { $w > $boxwidth } { set boxwidth $w }
        set boxheight [expr {$boxheight+$linespace}]
        if { $w > $maxwidth } {
            set boxwidth $maxwidth
            set boxheight [expr {$boxheight+$linespace}]
        }
    }
    if { $boxwidth < 70 } { set boxwidth 70 }
    if { $boxwidth > $maxwidth } { set boxwidth $maxwidth }

    GiD_OpenGL draw -viewport [list 0 0 $sizew $sizeh] \
        -scissor [list 0 0 $sizew $sizeh] -matrixmode projection \
        -pushmatrix -loadidentity -ortho [list 0 $sizew 0 $sizeh -1 1] \
        -matrixmode modelview -pushmatrix -loadidentity

    set status_bar_height 26
    set box {
        $sizew-$boxwidth-$boxsep $boxsep+$status_bar_height
        $sizew-$boxwidth-$boxsep $boxsep+$boxheight+$status_bar_height
        $sizew-$boxsep $boxsep+$boxheight+$status_bar_height
        $sizew-$boxsep $boxsep+$status_bar_height
    }

    set isbox 0
    for { set i 0 } { $i < [llength $groups] } { incr i } {
        lassign [lindex $groups $i] group - valuesList
        if { ![info exists colorgroup($group)] } { continue }
        set isbox 1
        break
    }

    if { $isbox } {
        GiD_OpenGL draw -color $gl_colours(foreground) -begin lineloop
        foreach "x y" $box { GiD_OpenGL draw -vertex [list [expr $x] [expr $y]] }
        GiD_OpenGL draw -end
    }

    set box {
        $sizew-$boxwidth-.5*$boxsep $boxsep+$boxheight+$status_bar_height-$idx*$linespace-.5*$boxsep
        $sizew-$boxwidth-.5*$boxsep $boxsep+$boxheight+$status_bar_height-$idx*$linespace-1.5*$boxsep
        $sizew-$boxwidth+.5*$boxsep $boxsep+$boxheight+$status_bar_height-$idx*$linespace-1.5*$boxsep
        $sizew-$boxwidth+.5*$boxsep $boxsep+$boxheight+$status_bar_height-$idx*$linespace-.5*$boxsep
    }
    set idx 0
    for { set i 0 } { $i < [llength $groups] } { incr i } {
        lassign [lindex $groups $i] group - valuesList
        set txt [_values_to_string $valuesList]
        if { $txt eq "" } { set txt $group }

        if { ![info exists colorgroup($group)] } { continue }
        GiD_OpenGL draw -color $gl_colours(foreground) -begin lineloop
        foreach "x y" $box { GiD_OpenGL draw -vertex [list [expr $x] [expr $y]] }
        GiD_OpenGL draw -end
        GiD_OpenGL draw -color [gid_groups_conds::_OpenGL_color $colorgroup($group)] -begin quads
        foreach "x y" $box { GiD_OpenGL draw -vertex [list [expr $x] [expr $y]] }
        GiD_OpenGL draw -end
        GiD_OpenGL draw -color $gl_colours(foreground) -rasterpos [list [expr {$sizew-$boxwidth+$boxsep}] \
            [expr {$boxsep+$boxheight+$status_bar_height-$idx*$linespace-1*$boxsep-6}]]

        set txt "$txt ($numents($group))"

        if { $::tcl_platform(osVersion) >= 6.0 } {
            # WARNING: emergency solution. This is not printing well
            #             set txtnew ""
            #             foreach c [split $txt ""] {
            #                 append txtnew [format %c [expr {[scan $c %c]+1}]]
            #             }
            #             set txt $txtnew
        }
        set w [expr {[gid_groups_conds::draw_text measure "$txt    "]+3.0*$boxsep}]

        if { $w <= $maxwidth } {
            gid_groups_conds::draw_text print "$txt"
        } else {
            lassign "" txt1 txt2
            foreach sp [split $txt] {
                if {$txt1 eq "" } {
                    append txt1 $sp
                } elseif { $txt2 eq "" } {
                    set w [expr {[gid_groups_conds::draw_text measure "$txt1 $sp"]+3.0*$boxsep}]
                    if { $w <= $maxwidth } {
                        append txt1 " $sp"
                    } else {
                        append txt2 " $sp"
                    }
                } else {
                    set w [expr {[gid_groups_conds::draw_text measure "$txt2 $sp"]+3.0*$boxsep}]
                    if { $w <= $maxwidth } {
                        append txt2 " $sp"
                    } else {
                        set w [expr {[gid_groups_conds::draw_text measure "$txt2..."]+3.0*$boxsep}]
                        if { $w > $maxwidth } {
                            set txt2 "[join [lrange [split $txt2] 0 end-1] { }]..."
                        } else {
                            set txt2 "$txt2..."
                        }
                        break
                    }
                }
            }
            gid_groups_conds::draw_text print "$txt1"
            if { $txt2 ne "" } {
                incr idx
                GiD_OpenGL draw -rasterpos [list [expr {$sizew-$boxwidth+$boxsep}] \
                    [expr {$boxsep+$boxheight+$status_bar_height-$idx*$linespace-1*$boxsep-6}]]
                gid_groups_conds::draw_text print "$txt2"
            }
        }
        incr idx
    }
    gid_groups_conds::draw_text pop
    GiD_OpenGL draw -popattrib -matrixmode projection -popmatrix \
        -matrixmode modelview -popmatrix
}

################################################################################
#    Conditions windows creation & opening
################################################################################

# proc gid_groups_conds::IsCompassLIB_academic {} {
#     variable CompassLIBid
#     variable program_name
#
#     if { ![info exists CompassLIBid] || ![info exists program_name] } {
#         return 1
#     }
#     set id [cu::sha1 compass99-$program_name-compass99]
#     if { $id eq $CompassLIBid } {
#         return 0
#     } else {
#         return 1
#     }
# }


proc gid_groups_conds::setCustomLib_window { wp } {

    set w $wp.scl
    set title [_ "CustomLib"]

    set win [dialogwin_snit $w -title $title -okname - -cancelname "Close"]

    set f1 [$win giveframe]

    text $f1.t -bg #dbdcde -width 45 -height 10 \
        -padx 4 -pady 4 -wrap word

    if {[winfo depth $f1] > 1} {
        set bold "-background #60a8ff -relief raised -borderwidth 1"
        set normal "-background {} -relief flat"
    } else {
        set bold "-foreground white -background black"
        set normal "-foreground {} -background {}"
    }

    set d [list \
        "[_ "Cimne's copyrighted CustomLib is powered by Compass Ingenieria y Sistemas."]" "" \
        "[_ " CustomLib is the library included in GiD for creating advanced customization problemtypes."]" "" \
        "[_ " For more information about CustomLib, please visit our website at: "]" "" \
    ]

eval [list $f1.t insert end] $d

$f1.t insert end \
    {www.compassis.com} d1

$f1.t insert end \n\n

set company "Compass Ingeniería y Sistemas"

$f1.t insert end \
    $company d2
$f1.t insert end \n

$f1.t insert end \
    {http://www.compassis.com} d3

$f1.t insert end \n

$f1.t insert end \
    {info@compassis.com} d4


set ContactUs "http://www.compassis.com/compass/en/Soporte/Ayuda"
set cmd_contact "VisitWeb $ContactUs"

set CompassISWeb "http://www.compassis.com"
set cmd "VisitWeb $CompassISWeb"

foreach tag { d1 d2 d3  d4 } {
    $f1.t tag bind $tag <Any-Enter> "$f1.t tag configure $tag $bold"
    $f1.t tag bind $tag <Any-Leave> "$f1.t tag configure $tag $normal"
}

$f1.t tag bind d1 <1> $cmd
$f1.t tag bind d2 <1> $cmd
$f1.t tag bind d3 <1> $cmd
$f1.t tag bind d4 <1> $cmd_contact

set font [font actual [$f1.t cget -font]]
set bfont [concat $font -weight bold]
set ufont [concat $font -underline 1]

ttk::label $f1.img -image "CompassIS_logo_img"

$f1.t tag bind link <Enter> "
[list $f1.t tag configure link -font $ufont]
[list $f1.t configure -cursor hand2]
[list tooltip::show $f1.t]
"
$f1.t tag bind link <Leave> "
[list $f1.t tag configure link -font $font]
[list $f1.t configure -cursor ""]
[list tooltip::hide]
"

grid $f1.img -row 0 -column 0 -padx 20 -pady 20
grid $f1.t -row 0 -column 1

$f1.t configure -state disabled
bind $f1.t <1> [list focus $f1.t]

grid columnconf $f1 1 -weight 1
grid rowconf $f1 1 -weight 1

bind [winfo toplevel $f1] <Return> [list $w invokeok]
set action [$w createwindow]
destroy $w
}

proc gid_groups_conds::SetProgramName { name } {
    variable program_name

    set program_name $name
}

proc gid_groups_conds::SetCompassLIBid { id } {
    variable CompassLIBid

    set CompassLIBid $id
}

proc gid_groups_conds::BackgroundColor {} {
    variable HeaderBackground

    upvar 0 ::ttk::theme::[gid_themes::GetTtkTheme]::colors colors
    set HeaderBackground $colors(-text_bg)
    set ::SystemBackgroundHi $HeaderBackground
}

proc gid_groups_conds::HeaderBackgroundColor { } {
    variable HeaderBackground

    if { [info command ::GiD_Set] eq "" || [info command ::GiD_Info] eq ""} {
        # nothing
    } else {
        if { [GiD_Set BackColor(Type)] == 0 } {
            set bg "BackColor(Single)"
            set HeaderBackground [GiD_Set $bg]
        } elseif { [GiD_Set BackColor(Sense)] eq "r" } {
            set bg "BackColor(Top)"
            set HeaderBackground [GiD_Set $bg]
        } elseif { [GiD_Set BackColor(Sense)] eq "l" } {
            set bg "BackColor(Bottom)"
            set HeaderBackground [GiD_Set $bg]
        } else {
            set bg "BackColor(Bottom)"
            set HeaderBackground [GiD_Set $bg]
        }
    }
}

proc gid_groups_conds::setHeaderBackground { } {
    variable HeaderBackground

    if { ![info exists HeaderBackground]} { return }
    if { $HeaderBackground ne ""} {
        set ::SystemBackgroundHi $HeaderBackground
    }
}

proc gid_groups_conds::SetDescription { id } {
    variable Description

    set Description $id
}

proc gid_groups_conds::SetHeaderName { w } {
    variable HeaderImage

    set txt "[gid_groups_conds::GetProgramName]"

    set fg "white"
    if {$::SystemBackgroundHi eq "#ffffff"} {
        set fg "black"
    }

    if {![info exists HeaderImage]} { set HeaderImage "" }

    label $w -text $txt -bg $::SystemBackgroundHi -image $HeaderImage \
        -cursor hand2 -fg $fg -compound left -pady 2
}

proc gid_groups_conds::setHeaderCompassIS { w } {
    if { [info command CompassIS_logo_img16] eq "" } { return }

    set height [image height CompassIS_logo_img16]
    label $w -image CompassIS_logo_img16 -bg $::SystemBackgroundHi \
        -cursor hand2 -fg white -compound left -height [expr {$height + 1}]

    register_popup_help $w [_ "CustomLIB powered by CompassIS. Press here for details"]
    bind $w <1> [list gid_groups_conds::setCustomLib_window $w]
}

proc gid_groups_conds::SetHeaderDescription { id } {
    variable HeaderDescription

    set HeaderDescription $id
}

proc gid_groups_conds::GetProgramName {} {
    variable program_name

    if { ![info exists program_name] } { return "" }
    return $program_name
}

proc gid_groups_conds::SetLibDir { dir } {
    variable problemtype_lib_dir

    set problemtype_lib_dir $dir
}

proc gid_groups_conds::SetHeaderImage { id } {
    variable HeaderImage

    set HeaderImage $id
}

proc gid_groups_conds::setHeaderImage { } {
    variable HeaderImage

    if {![info exists HeaderImage]} { set HeaderImage "" }

    if { $HeaderImage ne "" } {
        if {[file exists $HeaderImage]} {
            set HeaderImage [image create photo -file $HeaderImage]
        }
    }
}

proc gid_groups_conds::add_open_conditions_handler { args } {
    variable boundary_cond_handlers

    set optional {
        { -if_not_exists "" 0 }
    }
    set compulsory "handler"
    parse_args $optional $compulsory $args

    if { $if_not_exists } {
        set ipos [lsearch -exact $boundary_cond_handlers $handler]
        if { $ipos != -1 } { return }
    }
    lappend boundary_cond_handlers $handler
}

proc gid_groups_conds::remove_open_conditions_handler { handler } {
    variable boundary_cond_handlers

    set ipos [lsearch -exact $boundary_cond_handlers $handler]
    if { $ipos != -1 } {
        set boundary_cond_handlers [lreplace $boundary_cond_handlers $ipos $ipos]
    }
}

proc gid_groups_conds::open_conditions { what args } {
    variable open_boundary_cond
    variable doc
    variable is_draw_post
    variable side
    variable style
    variable open_all_var
    variable menu

    if { ( [info commands ::GidUtils::IsTkDisabled] ne "") && [GidUtils::IsTkDisabled] } {
        #e.g. batch mode without windows
        return
    }
    if { [info command ::GidUtils::AreWindowsDisabled] ne "" && [::GidUtils::AreWindowsDisabled] } {
        return
    }

    if { $is_draw_post } { return }
    set w .gid.central.boundaryconds

    if { [info exists doc] && $side eq "" } {
        set xp {string(/*/display_options/@frame_side)}
        set side [$doc selectNodes $xp]
        if { $side eq "" } {
            if { [regexp "Tdyn" $style {} curr_theme] } {
                set side "right"
            } else {
                set side "left"
            }
        }
    }

    switch $what {
        window_type {
            if { ![winfo exists $w] } {
                return none
            } elseif { [winfo class $w] eq "Toplevel" } {
                return window
            } else {
                return menu
            }
        }
        window {
            set optional {
                { -select_xpath xpath_expr "" }
            }
            set compulsory ""
            parse_args $optional $compulsory $args

            if { [winfo exists $w] && [winfo class $w] ne "Toplevel" } {
                open_conditions close
            }
            if { $select_xpath ne "" } {
                set wbc [open_conditions_window -force_open $w]
                $wbc select_domNode_xpath $select_xpath
            } else {
                open_conditions_window $w
            }
        }
        menu {
            set optional {
                { -select_xpath xpath_expr "" }
                { -force_open boolean 0 }
                { -select_tree select_tree data_tree }
            }
            set compulsory ""
            parse_args $optional $compulsory $args
            set isclose 1
            if { [winfo exists $w] && [winfo class $w] eq "Toplevel" } {
                if { $force_open } {
                    return
                }
                open_conditions close
            }
            set menu "data_tree"
            if { $select_xpath ne "" } {
                set wbc [open_conditions_inside -force_open -select_tree=$select_tree $w]
                $wbc select_domNode_xpath $select_xpath
            } else {
                open_conditions_inside -force_open=$force_open -select_tree=$select_tree $w
            }
            gid_groups_conds::gridremove $w.gg.ft.open
            if { !$open_all_var } {
                if { ![winfo exists $w.gg.ft.open] && [winfo exists $w] } {
                    gid_groups_conds::open_groups_layers_window $w.gg $w.gg.ft.open
                } elseif { [winfo exists $w] } {
                    grid $w.gg.ft.open -sticky nsew -padx 4 -pady 4
                }
            }
        }
        menu_or_any {
            gid_groups_conds::update_data_tree
            if { [winfo exists $w] && [winfo class $w] eq "Toplevel" } {
                return [eval open_conditions window $args]
            }
            return [eval open_conditions menu $args]
        }
        toggle {
            if { ![winfo exists $w] } {
                set f open_conditions_window
                $f $w
            } elseif { [winfo class $w] eq "Toplevel" } {
                # first call closes the window
                open_conditions_window $w
                open_conditions_inside $w
            } else {
                open_conditions_inside $w
                set f open_conditions_window
                $f $w
            }
        }
        check_default {
            if { ![info exists doc] } { return }
            set xp {string(/*/display_options/@is_frame_open)}
            set is_frame_open [$doc selectNodes $xp]
            if { $is_frame_open == 1 } {
                open_conditions menu -force_open 1 {*}$args
            }
        }
        close {
            if { ![winfo exists $w] } {
                return
            } elseif { [winfo class $w] eq "Toplevel" } {
                # first call closes the window
                open_conditions_window $w
            } else {
                open_conditions_inside $w
            }
        }
        end {
            if { ![winfo exists $w] } {
                return
            } elseif { [winfo class $w] eq "Toplevel" } {
                # first call closes the window
                open_conditions_window -update_preferences 0 $w
            } else {
                open_conditions_inside -update_preferences 0 $w
            }
        }
        info_frame_width {
            if { ![winfo exists $w] || [winfo class $w] eq "Toplevel" } {
                return ""
            }
            return [$open_boundary_cond info_frame_width]
        }
        info_frame_side {
            return $side
        }
        give_paned {
            return $w
        }
        show_window {
            set optional {
                { -select_xpath xpath_expr "" }
                { -del_cancel_button boolean 0 }
            }
            set compulsory ""
            parse_args $optional $compulsory $args
            gid_groups_conds::update_data_tree
            if { $select_xpath ne "" } {
                set wbc [open_conditions_window -force_open $w]
                set id [$wbc select_domNode_xpath $select_xpath]
                set tree $wbc.ft.t
                if { [$tree item state get $id condition] } {
                    set domNodeCnd [$tree item element cget $id 0 e_text -data]
                    set a0 [dict create add_group 1 edit_group 1 delete_group 1]
                    set actions [dict merge $a0 [split [$domNodeCnd @actions ""] ,]]
                    if { [dict get $actions add_group] } {
                        $wbc apply_cond -what apply
                    }
                } elseif { [$tree item state get $id blockdata] } {
                    $wbc apply_blockdata -del_cancel_button $del_cancel_button
                }
            } else {
                open_conditions_window $w
            }
        }
    }

    if { [winfo exists $w] } {
        $w configure -cursor arrow
    }
}

proc gid_groups_conds::update_toolbar_more_options_button { w cmds lay gr } {

    set width [winfo width $w]

    if { $lay eq "1" } {
        set ww $w.laybbox
    } else {
        set ww $w.grbbox
    }

    set px_tot [winfo reqwidth $ww]

    set wwList [winfo children $ww]
    set wwlength [llength $wwList]

    set isadded 0
    set isdeleted 0
    set deleted_list ""
    set added_list ""

    for { set i [expr {$wwlength-1}] } { $i > 0 } {incr i -1} {
        set iw [lindex $wwList $i]
        set name [winfo name $iw]
        if { [winfo class $iw] == "TButton" ||  [winfo class $iw] == "TMenubutton" } {
            set width_iw [expr {[winfo width $iw]+5*2}]
        } elseif { [winfo class $iw] eq "TSeparator" } {
            set width_iw [expr {[winfo width $iw]+5*2}]
            set img_i "---"
        }
        if { $px_tot > $width && [winfo exists $iw] && !$isadded } {
            if {[winfo class $iw] eq "TSeparator"} {
                set img "---"
            } else {
                set img [$iw cget -image]
            }
            lappend deleted_list $img
            grid remove $iw
            set isdeleted 1
            incr px_tot [expr {-$width_iw}]
        } elseif { [winfo exists $iw] == "0" && !$isdeleted } {
            grid $iw
            set isadded 1
        } elseif { $isdeleted } {
            if {[winfo class $iw] == "TButton"} {
                set val [regexp {f(.*)} $iw {} col]
                set cmd [$iw cget -command]
                set img "dots-18"
                set image [gid_themes::GetImage $img.png]
                grid remove $iw
                set b [cu::menubutton_button $ww.menu$col -image $image \
                    -menu $ww.menu$col.m -style ToolbuttonMod -direction left]
                set help [_ "More options"]
                cu::tooltip register $b "$help"
                menu $b.m -tearoff 0
                set deleted_list [lreverse $deleted_list]
                foreach iimg $deleted_list {
                    foreach j $cmds {
                        lassign $j img0 help_j type cmd_j cmdList
                        if { $img0 eq $iimg } {
                            break
                        }
                    }
                    if {$iimg eq "---"} {
                        $b.m add separator
                    } else {
                        $b.m add command -label "" -command $cmd_j
                        if { $img0 ne "" } {
                            # GiD change: only next line
                            set img_i [groups_theme::GetThemedImage $img0]
                            $b.m entryconfigure end -image $img_i -compound left
                        }
                    }
                }
                grid $b -row 0 -column $col -sticky nsw -padx 5 -pady "0 3"
                set isdeleted 0
            } elseif {[winfo class $iw] eq "TSeparator"} {
                lappend deleted_list "---"
                grid remove $iw
            }
        } elseif { [winfo class $iw] == "TMenubutton" && [regexp {menu(.*)} $iw {} original_col] && $px_tot < $width } {
            if { [expr {$px_tot + [winfo width $iw]+5*2}] < $width } {
                set jcol 0
                foreach j $cmds {
                    lassign $j img0 help_j type cmd_j cmdList
                    if {$jcol > [expr {$original_col-1}]} {
                        grid $ww.f$jcol
                        #                         if {[winfo class $ww.f$jcol] eq "TSeparator"} {
                        #                             set img "---"
                        #                         } else {
                        #                             set img [$iw cget -image]
                        #                         }
                        #                         lappend added_list [$ww.f$jcol cget -image]
                        set isadded 1
                        grid configure $iw -row 0 -column [expr {$jcol+1}]
                    }
                    incr jcol
                }
            }
        }
    }
}

proc gid_groups_conds::open_conditions_window { args } {
    variable doc
    variable open_boundary_cond
    variable boundary_cond_handlers
    variable open_all_var
    variable close_button_all_gr_lay_window

    set optional {
        { -force_open "" 0 }
        { -update_preferences boolean 1 }
    }
    set compulsory "w"
    parse_args $optional $compulsory $args

    if { [winfo exists $w] } {
        if { $force_open } { return $w.gg }
        if { $update_preferences && [info exists open_boundary_cond] && [winfo exists $open_boundary_cond] } {
            $open_boundary_cond store_tree_state
        }
        destroy $w
        if { [info exists open_boundary_cond] } {
            unset open_boundary_cond
        }
        return
    }
    if { ![info exists doc] } { return }
    if { [GiD_Info Project ViewMode] ni "GEOMETRYUSE MESHUSE" } {
        return
    }
    set cmd [list gid_groups_conds::open_conditions_window $w]
    InitWindow2 $w -title [_ "Boundary conditions"] \
        -geometryvariable PreBoundaryConditionsWindowGeom -initcommand $cmd \
        -ontop

    wm protocol $w WM_DELETE_WINDOW [list gid_groups_conds::open_conditions_window $w]

    boundary_conds $w.gg

    set open_boundary_cond $w.gg

    if { [info exists doc] } {
        $open_boundary_cond configure -domdoc $doc
    }

    bind $w.gg <Destroy> [namespace code [list gid_groups_conds::destroy_conditions]]

    ttk::frame $w.buts -style BottomFrame.TFrame
    #     ttk::button $w.buts.close -text [_ "Close"] -command \
    #         [list gid_groups_conds::open_conditions_window $w] \
    #         -style BottomFrame.TButton
    #     pack $w.buts.close -pady 3

    grid $w.gg -sticky nsew -pady 0
    grid $w.buts -pady 0 -sticky nsew
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    foreach i $boundary_cond_handlers { eval $i }

    tk::TabToWindow $w.gg


    gid_groups_conds::gridremove_open_all_window $w.gg

    if { !$open_all_var } {
        if { ![winfo exists $w.gg.ft.open] } {
            gid_groups_conds::open_groups_layers_window $w $w.gg.ft.open
        } elseif { !$close_button_all_gr_lay_window } {
            grid $w.gg.ft.open -sticky nsew -padx 4 -pady 4
        }
    }

    return $w.gg
}

proc gid_groups_conds::apply_frame_width_from_doc { } {
    variable doc
    variable open_boundary_cond
    set width ""
    if { [info exists doc] && [info exists open_boundary_cond] && [winfo exists $open_boundary_cond] } {
        $open_boundary_cond configure -domdoc $doc
        set root [$doc documentElement]
        set width [$root selectNodes string(display_options/@frame_width)]
        if { $width ne "" && $width > 1 } {
            $open_boundary_cond configure -width $width
        }
    }
    return $width
}

########################
#   Procedure to open window inside, include creation of panedwindow to manage several windows
########################

proc gid_groups_conds::open_conditions_inside { args } {
    set result_boundary_conds_widged ""

    variable doc
    variable open_boundary_cond
    variable boundary_cond_handlers
    global GidPriv
    variable style
    variable side

    set optional {
        { -force_open "" 0 }
        { -update_preferences boolean 1 }
        { -select_tree select_tree data_tree }
    }
    set compulsory "p"
    parse_args $optional $compulsory $args

    #######
    # pwCentral is a panedwindow to manage $grWindow with other windows in GiD
    if {![info exists GidPriv(pwCentral)]} {
        set GidPriv(pwCentral) .gid.central.pwCentral
    }
    #######

    if  { [info procs GidUtils::GetMainDrawAreaWidget] != "" } {
        set grWindow [GidUtils::GetMainDrawAreaWidget]
    } else {
        #to be deleted when compass use the current dev_kit.tcl scripts of developer version
        if { [winfo exists .gid.central.wins] } {
            set grWindow .gid.central.wins
        } else {
            set grWindow .gid.central.s
        }
    }
    if { [winfo exists $p] } {
        #tree pane already exits, delete it (or simply return if force_open)
        if { $force_open } {
            set result_boundary_conds_widged $p.gg
            return $result_boundary_conds_widged
        }
        if { $update_preferences } {
            $open_boundary_cond store_tree_state
            $open_boundary_cond store_tree_state_close_frame
        }
        if { [winfo exists $GidPriv(pwCentral)] } {
            if { [winfo class $GidPriv(pwCentral)] in "Panedwindow TPanedwindow" } {
                set panes [$GidPriv(pwCentral) panes]
                set maxpane [expr [llength $panes]-1]
                set _pos [lsearch $panes $p]
                if {$_pos >= 0} {
                    if { [llength $panes] == 2 } {
                        set oldwin [lreplace $panes $_pos $_pos]
                        set mg [winfo manager $GidPriv(pwCentral)]
                        set mginfo [$mg info $GidPriv(pwCentral)]
                        $GidPriv(pwCentral) forget $p
                        $mg $oldwin {*}$mginfo
                        update idletask
                        destroy $GidPriv(pwCentral)
                        unset GidPriv(pwCentral)
                    } else {
                        $GidPriv(pwCentral) forget $p
                    }
                }
            }
        } else {
            unset GidPriv(pwCentral)
        }

        update idletasks
        destroy $p
        if { [regexp "Tdyn" $style {} curr_theme] } {
            gid_groups_conds::remove_all_search
        }
        if { [info exists open_boundary_cond] } {
            unset open_boundary_cond
        }
    } else {
        #tree pane doesn't exits, create it

        if { ![info exists doc] } {
            return ""
        }
        if { [GiD_Info Project ViewMode] ni "GEOMETRYUSE MESHUSE" } {
            return ""
        }

        set focus [focus]

        panedwindow $p -orient vertical -opaqueresize 0

        set open_boundary_cond [boundary_conds $p.gg -select_tree $select_tree ]
        $p add $p.gg -sticky nsew -stretch always -minsize 60

        set width [gid_groups_conds::apply_frame_width_from_doc]

        if {![winfo exists $GidPriv(pwCentral)]} {
            set mg [winfo manager $grWindow]
            set mginfo [$mg info $grWindow]
            $mg forget $grWindow

            panedwindow $GidPriv(pwCentral) -bd 0 -showhandle 0 -sashpad 1 -sashrelief sunken -opaqueresize 0

            if { [regexp "Tdyn" $style {} curr_theme] } {
                upvar 0 ::ttk::theme::[gid_themes::GetTtkTheme]::colors colors
                set color $colors(-text_bg)
                $GidPriv(pwCentral) configure -background $color
            }

            $mg $GidPriv(pwCentral) {*}$mginfo
            $GidPriv(pwCentral) add $grWindow -stretch always -minsize 60
            raise $grWindow
            if {$grWindow == ".gid.central.wins"} {
                raise .gid.central.s
            }
        }

        set maxpane [expr [llength [$GidPriv(pwCentral) panes]]-1]
        if {$maxpane < 0} {
            set maxpane 0
        }

        update idletasks
        if {$side eq "left"} {
            if {$width != ""} {
                $GidPriv(pwCentral) add $p -before $grWindow -width $width -stretch never -minsize 60
            } else {
                $GidPriv(pwCentral) add $p -before $grWindow -stretch never -minsize 60
            }
        } else {
            if {$width != ""} {
                set sepw 2
                set w1 [winfo width $GidPriv(pwCentral)]
                set w [expr {$w1-$width}]
                if { $w <= 0 || $w1 <= 0 } {
                    set w 360
                }
                set gr [lindex [$GidPriv(pwCentral) panes] 0]
                $GidPriv(pwCentral) forget $gr
                $GidPriv(pwCentral) add $gr -minsize 60 -width [expr {$w-2}]
                $GidPriv(pwCentral) add $p -minsize 60 -width [expr {$width+2}]
            } else {
                $GidPriv(pwCentral) add $p -after $grWindow -stretch never -minsize 60
            }
        }

        raise $p

        bind $p <Destroy> [namespace code [list gid_groups_conds::destroy_conditions]]

        if {$side eq "right"} {
            $GidPriv(pwCentral) paneconfigure $grWindow -stretch always -minsize 60
        }

        if { $focus ne "" } {
            update
            if { [catch { focus -force $focus } msg] } {
                set to_stop 1
            }
        }

        foreach i $boundary_cond_handlers {
            eval $i
        }
        set result_boundary_conds_widged $p.gg
    }
    return $result_boundary_conds_widged
}

proc gid_groups_conds::open_conditions_draw_post { args } {
    variable doc
    variable open_boundary_cond
    variable boundary_cond_handlers

    set focus [focus]

    set optional {
        { -what auto|toggle|open|close auto }
        { -check_paned "" 0 }
    }
    set compulsory "frame paned"
    parse_args $optional $compulsory $args

    if { [winfo exists $frame ] } {
        destroy {*}[winfo children $frame]
    } else {
        ttk::frame $frame
    }

    set pref [gid_groups_conds::get_preference draw_post::conditions_state tree]
    if { $what eq "auto" } {
        set what $pref
    } elseif { $what eq "toggle" } {
        set what [expr {($pref ne "tree")?"tree":"close"}]
    }
    if { $what eq "close" } {
        #         set fr [cu::handle $frame.fr -orient vertical -length 40]
        #         draw_post::set_handle_events init $fr $frame $paned \
        #             [list gid_groups_conds::open_conditions_draw_post -what tree -check_paned \
        #                 $frame $paned] [_ "Open conditions & data selector"] close "tree close"
        #         grid $fr -sticky nsw -padx "2 1"
        set pref close
        set stretch never
        destroy $frame
    } else {
        set fr [cu::handle $frame.fr]
        draw_post::set_handle_events init $fr $frame $paned \
            [list gid_groups_conds::open_conditions_draw_post -what close \
            -check_paned $frame $paned] [_ "Minimize conditions & data selector"] \
            tree "tree close"

        set open_boundary_cond [boundary_conds $frame.gg -auto_take_focus 0 -width 120]

        set width [gid_groups_conds::apply_frame_width_from_doc]

        bind $frame.gg <Destroy> [namespace code [list gid_groups_conds::destroy_conditions]]

        if { $focus ne "" } {
            update
            catch { focus -force $focus }
        }

        foreach i $boundary_cond_handlers { eval $i }

        grid $fr $frame.gg -sticky nsew -padx 1 -pady 1
        grid configure $fr -sticky nsw -padx "2 1"
        grid columnconfigure $frame 1 -weight 1
        grid rowconfigure $frame 0 -weight 1
        set pref tree
        set stretch always
    }
    gid_groups_conds::set_preferenceG draw_post::conditions_state $pref

    dict set draw_post::data conditions_frame $frame

    if { [winfo exists $frame] } {
        dict set draw_post::data conditions_tree_widget $frame.gg
        dict set draw_post::data stretchs $frame $stretch
        $paned paneconfigure $frame -stretch $stretch -minsize 60
    }
    if { $check_paned } {
        draw_post::correct_panes_heights -added_this $frame
    }
    return [list gid_groups_conds::open_conditions_draw_post $frame $paned conditions_state "tree close"]
}

proc gid_groups_conds::destroy_conditions {} {
    variable open_boundary_cond

    if { [info exists open_boundary_cond] } {
        unset open_boundary_cond
    }
}

proc gid_groups_conds::actualize_conditions_window { args } {
    variable open_boundary_cond

    set optional {
        { -do_focus boolean 1 }
        { -domNodeList nodes "" }
    }
    set compulsory ""
    parse_args $optional $compulsory $args

    if { [info exists open_boundary_cond] && [winfo exists $open_boundary_cond] } {
        if { $domNodeList ne "" } {
            $open_boundary_cond actualize_domNode $domNodeList
        } else {
            $open_boundary_cond actualize_do -do_focus $do_focus 0
        }
    }
}

################################################################################
#    Post windows creation & opening
################################################################################

proc gid_groups_conds::add_open_post_handler { args } {
    variable post_handlers

    set optional {
        { -if_not_exists "" 0 }
    }
    set compulsory "handler"
    parse_args $optional $compulsory $args

    if { $if_not_exists } {
        set ipos [lsearch -exact $post_handlers $handler]
        if { $ipos != -1 } { return }
    }
    lappend post_handlers $handler
}

proc gid_groups_conds::remove_open_post_handler { handler } {
    variable post_handlers

    set ipos [lsearch -exact $post_handlers $handler]
    if { $ipos != -1 } {
        set post_handlers [lreplace $post_handlers $ipos $ipos]
    }
}

proc gid_groups_conds::open_post { what args } {
    variable open_post
    variable doc

    set w .gid.central.post
    switch $what {
        window_type {
            if { ![winfo exists $w] } {
                return none
            } elseif { [winfo class $w] eq "Toplevel" } {
                return window
            } else {
                return menu
            }
        }
        window {
            set optional {
                { -select_xpath xpath_expr "" }
            }
            set compulsory ""
            parse_args $optional $compulsory $args

            if { [winfo exists $w] && [winfo class $w] ne "Toplevel" } {
                open_post close
            }
            if { $select_xpath ne "" } {
                set wbc [open_post_window -force_open $w]
                $wbc select_domNode_xpath $select_xpath
            } else {
                open_post_window $w
            }
        }
        menu {
            set optional {
                { -select_xpath xpath_expr "" }
                { -force_open boolean 0 }
            }
            set compulsory ""
            parse_args $optional $compulsory $args

            set isclose 1
            if { [winfo exists $w] && [winfo class $w] eq "Toplevel" } {
                if { $force_open } {
                    return
                }
                open_post close
            }
            if { $select_xpath ne "" } {
                set wbc [open_post_inside -force_open $w]
                $wbc select_domNode_xpath $select_xpath
            } else {
                open_post_inside -force_open=$force_open $w
            }
        }
        menu_or_any {
            if { [winfo exists $w] && [winfo class $w] eq "Toplevel" } {
                return [eval open_post window $args]
            }
            return [eval open_post menu $args]
        }
        toggle {
            if { ![winfo exists $w] } {
                set f open_post_window
            } elseif { [winfo class $w] eq "Toplevel" } {
                # first call closes the window
                open_post_window $w
                set f open_post_inside
            } else {
                open_post_inside $w
                set f open_post_window
            }
            $f $w
        }
        check_default {
            if { ![info exists doc] } { return }
            set xp {string(/*/display_options/@is_post_frame_open)}
            set is_post_frame_open [$doc selectNodes $xp]
            if { $is_post_frame_open == 1 } {
                open_post menu -force_open 1 {*}$args
            }
        }
        close {
            if { ![winfo exists $w] } {
                return
            } elseif { [winfo class $w] eq "Toplevel" } {
                # first call closes the window
                open_post_window $w
            } else {
                open_post_inside $w
            }
        }
        info_frame_width {
            if { ![winfo exists $w] || [winfo class $w] eq "Toplevel" } {
                return ""
            }
            return [$open_post info_frame_width]
        }
    }
}

proc gid_groups_conds::open_post_window { args } {
    variable doc
    variable open_post
    variable post_handlers

    set optional {
        { -force_open "" 0 }
    }
    set compulsory "w"
    parse_args $optional $compulsory $args

    if { [winfo exists $w] } {
        if { $force_open } { return $w.gg }
        $open_post store_tree_state
        destroy $w
        if { [info exists open_post] } {
            unset open_post
        }
        return
    }
    if { ![info exists doc] } { return }
    if { [GiD_Info Project ViewMode] in "GEOMETRYUSE MESHUSE" } {
        return
    }
    set cmd [list gid_groups_conds::open_post_window $w]
    InitWindow2 $w -title [_ "Results"] \
        -geometryvariable PreBoundaryConditionsWindowGeom -initcommand $cmd \
        -ontop

    wm protocol $w WM_DELETE_WINDOW [list gid_groups_conds::open_post_window $w]

    post_tree $w.gg

    set open_post $w.gg

    if { [info exists doc] } {
        $open_post configure -domdoc $doc
    }

    bind $w.gg <Destroy> [namespace code [list gid_groups_conds::destroy_post]]

    ttk::frame $w.buts -style BottomFrame.TFrame
    ttk::button $w.buts.close -text [_ "Close"] -command \
        [list gid_groups_conds::open_post_window $w] \
        -style BottomFrame.TButton
    pack $w.buts.close -pady 3

    grid $w.gg -sticky nsew -pady 0
    grid $w.buts -pady 0 -sticky nsew
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    foreach i $post_handlers { eval $i }

    tk::TabToWindow $w.gg
    return $w.gg
}

########################
#   Procedure to open window inside [post], include creation of panedwindow to manage several windows
########################
proc gid_groups_conds::open_post_inside { args } {
    variable doc
    variable open_post
    variable post_handlers
    global GidPriv

    set optional {
        { -force_open "" 0 }
    }
    set compulsory "p"
    parse_args $optional $compulsory $args

    # pwCentral is a panedwindow to manage $grWindow with other windows in GiD
    if {![info exists GidPriv(pwCentral)]} {
        set GidPriv(pwCentral) .gid.central.pwCentral
    }

    if  { [info procs GidUtils::GetMainDrawAreaWidget] != "" } {
        set grWindow [GidUtils::GetMainDrawAreaWidget]
    } else {
        #to be deleted when compass use the current dev_kit.tcl scripts of developer version
        if { [winfo exists .gid.central.wins] } {
            set grWindow .gid.central.wins
        } else {
            set grWindow .gid.central.s
        }
    }
    if { [winfo exists $p] } {
        if { $force_open } { return $p.gg }
        $open_post store_tree_state
        $open_post store_tree_state_close_frame

        if { [winfo exists $GidPriv(pwCentral)] } {
            if { [winfo class $GidPriv(pwCentral)] in "Panedwindow TPanedwindow" } {
                set panes [$GidPriv(pwCentral) panes]
                set maxpane [expr [llength $panes]-1]
                set _pos [lsearch $panes $p]
                if {$_pos >= 0} {
                    if { [llength $panes] == 2 } {
                        set oldwin [lreplace $panes $_pos $_pos]
                        set mg [winfo manager $GidPriv(pwCentral)]
                        set mginfo [$mg info $GidPriv(pwCentral)]
                        $GidPriv(pwCentral) forget $p
                        $mg $oldwin {*}$mginfo
                        update idletask
                        destroy $GidPriv(pwCentral)
                        unset GidPriv(pwCentral)
                    } else {
                        $GidPriv(pwCentral) forget $p
                    }
                }
            }
        } else {
            unset GidPriv(pwCentral)
        }

        update idletasks
        destroy $p
        if { [info exists open_post] } {
            unset open_post
        }
        return ""
    }
    if { ![info exists doc] } { return }
    if { [GiD_Info Project ViewMode] in "GEOMETRYUSE MESHUSE" } {
        return
    }

    set focus [focus]

    ttk::frame $p

    set open_post [post_tree $p.gg -width 120]
    pack $p.gg -expand 1 -fill both

    set width ""
    if { [info exists doc] } {
        $open_post configure -domdoc $doc
        set root [$doc documentElement]
        set width [$root selectNodes string(display_options/@post_frame_width)]
        if { $width ne "" && $width > 1 } {
            $open_post configure -width $width
        }
    }

    if {![winfo exists $GidPriv(pwCentral)]} {
        set mg [winfo manager $grWindow]
        set mginfo [$mg info $grWindow]
        $mg forget $grWindow
        panedwindow $GidPriv(pwCentral) -bd 0 -showhandle 0 -sashpad 1 -sashrelief sunken -opaqueresize 0;#-background #880000
        #ttk::panedwindow $GidPriv(pwCentral) -orient horizontal

        $mg $GidPriv(pwCentral) {*}$mginfo
        $GidPriv(pwCentral) add $grWindow -minsize 60 -stretch always
        raise $grWindow
        if {$grWindow == ".gid.central.wins"} {
            raise .gid.central.s
        }
    }

    set maxpane [expr [llength [$GidPriv(pwCentral) panes]]-1]
    if { $width eq "" } {
        $GidPriv(pwCentral) add $p -before $grWindow -stretch never -minsize 60
    } else {
        $GidPriv(pwCentral) add $p -before $grWindow -width $width -stretch never -minsize 60
    }

    raise $p

    bind $p <Destroy> [namespace code [list gid_groups_conds::destroy_post]]
    if { $focus ne "" } {
        update
        catch { focus -force $focus }
    }

    foreach i $post_handlers { eval $i }

    return $p.gg
}

proc gid_groups_conds::destroy_post {} {
    variable open_post

    if { [info exists open_post] } {
        unset open_post
    }
}

proc gid_groups_conds::actualize_post_window { args } {
    variable open_post

    set optional {
        { -do_focus boolean 1 }
    }
    set compulsory ""
    parse_args $optional $compulsory $args

    if { [info exists open_post] && [winfo exists $open_post] } {
        $open_post actualize_do -do_focus $do_focus 0
    }
}

################################################################################
#    Conditions dependencies
################################################################################

proc gid_groups_conds::give_dependencies_nodes { args } {

    set optional {
        { -value node_value }
    }
    set compulsory "domNode"
    parse_args $optional $compulsory $args
    if { [info command $domNode] == "" } {
        return ""
    }
    if { [get_domnode_attribute $domNode state ""] eq "hidden" } {
        return ""
    }

    set domNodeS $domNode
    if { ![info exists value] } {
        set value [get_domnode_attribute $domNode v ""]
    }
    if { [$domNodeS selectNodes "dependencies"] eq "" } {
        set groupNode [$domNode selectNodes ancestor::group]
        if { $groupNode == "" } {
            set groupNode [$domNode selectNodes ancestor::groupList]
        }
        if { $groupNode ne "" } {
            set xp [format_xpath {../value[@n=%s]} [$domNode @n]]
            append xp [format_xpath {|../container//value[@n=%s]} [$domNode @n]]
            set domNodeS [$groupNode selectNodes $xp]
        }
        if { $domNodeS eq "" } {
            return
        }
    }
    set depNodes ""
    foreach depNode [$domNodeS selectNodes "dependencies"] {
        if { [$depNode @node ""] eq "" } {
            continue
        }
        set cnd [$depNode @condition ""]
        if { [$depNode hasAttribute value] && [get_domnode_attribute $depNode value] eq $value } {
            lappend depNodes $depNode
        } elseif { [$depNode hasAttribute not_value] && [get_domnode_attribute $depNode not_value] ne $value } {
            lappend depNodes $depNode
        } elseif { $cnd ne "" && [$domNode selectNodes $cnd] } {
            lappend depNodes $depNode
        }
    }
    if { ![llength $depNodes] } {
        foreach depNode [$domNodeS selectNodes "dependencies\[@default='1'\]"] {
            if { [$depNode @node ""] eq "" } {
                continue
            }
            lappend depNodes $depNode
        }
    }
    foreach depNode [$domNodeS selectNodes {dependencies[not(@value) and not(@not_value) and not(@default) and not(@condition)]}] {
        if { [$depNode @node ""] eq "" } {
            continue
        }
        lappend depNodes $depNode
    }
    return $depNodes
}

proc gid_groups_conds::check_node_dependencies { valueNode { num 0 } } {

    # this is to stop circular references
    if { $num > 500 } { return 0 }

    set defaults_d ""
    dict set defaults_d state normal

    set retnodesList ""
    foreach depNode [give_dependencies_nodes $valueNode] {
        set tomodNodes [$valueNode selectNodes [$depNode @node]]
        if { [llength $tomodNodes] == 0 } {
            if { [$depNode @check 1] == 0 } { continue }
            error "error in check_node_dependencies. xpath='[$depNode @node]'"
        }
        $depNode setAttribute v [get_domnode_attribute $valueNode v ""]
        if { [$valueNode hasAttribute units] } {
            $depNode setAttribute units [get_domnode_attribute $valueNode units ""]
        }
        foreach tomodNode $tomodNodes {

            if { [$depNode @actualize 0] } {
                lappend retnodesList $tomodNode
            }

            # if { [$tomodNode nodeName] eq "condition" } { continue }
            set haschangesL 0
            if { [$depNode hasAttribute att1] } {
                set v1 [get_domnode_attribute $depNode v1]
                set att1 [$depNode @att1]

                if { $att1 eq "change_units_system" } {
                    set units_system [get_domnode_attribute $depNode v1]
                } elseif { ![$tomodNode hasAttribute $att1] } {
                    if { [dict exists $defaults_d $att1] } {
                        set v1_dep [dict get $defaults_d $att1]
                    } else {
                        unset -nocomplain v1_dep
                    }
                } else {
                    set v1_dep [get_domnode_attribute $tomodNode $att1]
                }
                if { $att1 ne "change_units_system" && (![info exists v1_dep] || $v1_dep ne $v1) } {
                    $tomodNode setAttribute [$depNode @att1] $v1
                    #setAttributes [gid_groups_conds::nice_xpath $tomodNode] [list [$depNode @att1] $v1] 0
                    lappend retnodesList $tomodNode
                    set haschangesL 1
                }
            }
            if { [$depNode hasAttribute att2] } {
                set v2 [get_domnode_attribute $depNode v2]
                set att2 [$depNode @att2]
                if { $att2 eq "change_units_system" } {
                    set units_system [get_domnode_attribute $depNode v1]
                } elseif { ![$tomodNode hasAttribute $att2] } {
                    if { [dict exists $defaults_d $att2] } {
                        set v2_dep [dict get $defaults_d $att2]
                    } else {
                        unset -nocomplain v2_dep
                    }
                } else {
                    set v2_dep [get_domnode_attribute $tomodNode $att2]
                }
                if { $att2 ne "change_units_system" && (![info exists v2_dep] || $v2_dep ne $v2) } {
                    $tomodNode setAttribute [$depNode @att2] $v2
                    #setAttributes [gid_groups_conds::nice_xpath $tomodNode] [list [$depNode @att2] $v2] 0
                    lappend retnodesList $tomodNode
                    set haschangesL 1
                }
            }
            if { [get_domnode_attribute $tomodNode values] ne "" && [$tomodNode @editable 0] == 0} {
                set value [get_domnode_attribute $tomodNode v]
                set values [split [get_domnode_attribute $tomodNode values] ,]
                if { [lsearch -exact $values $value] == -1 } {
                    set value [lindex $values 0]
                    setAttributes [gid_groups_conds::nice_xpath $tomodNode] \
                        [list v $value] 0
                    lappend retnodesList $tomodNode
                    set haschangesL 1
                }
            }
            set xp {self::*[@unit_definition or @unit_mesh_definition='1']}
            if { [$tomodNode selectNodes $xp] ne "" } {
                lassign [gid_groups_conds::give_units_actualize_value \
                    -perform_the_change 1 $tomodNode] newunit haschanges

                if { $haschanges } {
                    lappend retnodesList $tomodNode
                    set haschangesL 1
                }
            }
            if { $haschangesL && [$tomodNode nodeName] eq "value" && [$tomodNode selectNodes dependencies] ne "" } {
                set hc [check_node_dependencies $tomodNode [incr num]]
                lappend retnodesList {*}$hc
            }
            #             if { $haschangesL } {
            #                 set xp {(ancestor::group|ancestor::groupList|ancestor::blockdata)[1]}
            #                 if { $haschanges < 2 && $haschangesL < 2 && [$valueNode selectNodes $xp] eq [$tomodNode selectNodes $xp] } {
            #                     set haschanges 1
            #                 } else {
            #                     set haschanges 2
            #                 }
            #             }
        }
        if { [$depNode hasAttribute v] } {
            $depNode removeAttribute v
        }
        if { [$depNode hasAttribute units] } {
            $depNode removeAttribute units
        }
    }
    return [lsort -unique $retnodesList]
}

# in this moment it is not used
proc gid_groups_conds::check_dependencies {} {
    variable doc

    set root [$doc documentElement]
    foreach valueNode [$root selectNodes {//value[dependencies]}] {
        check_node_dependencies $valueNode
    }
}

proc gid_groups_conds::_check_condition_dependencies1_do { domNodeCnd tomodNodeCnd subtreeNode } {

    set actualizeNodes ""
    if { $subtreeNode ne "" } {
        set ancestors [$tomodNodeCnd selectNodes ancestor::*]
        if { [lsearch -exact $ancestors $subtreeNode] == -1 } {
            return ""
        }
    }
    # WARNING: it is necessary to deal also with "grouplist"
    foreach groupNodeFrom [$domNodeCnd selectNodes group] {
        set n [$groupNodeFrom @n]
        if { [$tomodNodeCnd selectNodes [format_xpath {group[@n=%s]} $n]] ne "" } {
            continue
        }
        set ov [$groupNodeFrom @ov ""]
        set values_dict ""
        foreach valueNode [$groupNodeFrom selectNodes .//value] {
            set n_in [get_domnode_attribute $valueNode n]
            dict set values_dict $n_in v [get_domnode_attribute $valueNode v]
            if { [get_domnode_attribute $valueNode unit_magnitude] ne "" } {
                dict set values_dict $n_in units [get_domnode_attribute $valueNode units]
            }
            set fNode [$valueNode selectNodes function]
            if { [llength $fNode] == 1 } {
                dict set values_dict $n_in function [$fNode asXML]
            }
        }
        set ret [gid_groups_conds::apply_condition -ov $ov [gid_groups_conds::nice_xpath $tomodNodeCnd] $n $values_dict]
        lappend actualizeNodes {*}$ret $tomodNodeCnd
    }
    # WARNING: it is necessary to deal also with "grouplist"
    foreach groupNodeTo [$tomodNodeCnd selectNodes group] {
        set n [$groupNodeTo @n]
        if { [$domNodeCnd selectNodes [format_xpath {group[@n=%s]} $n]] ne "" } {
            continue
        }
        gid_groups_conds::delete [gid_groups_conds::nice_xpath $groupNodeTo]
        lappend actualizeNodes $tomodNodeCnd
    }
    return $actualizeNodes
}

proc gid_groups_conds::check_condition_dependencies1 { domNodeCnd { subtreeNode "" } } {
    variable doc

    if { $domNodeCnd ne "" } {
        set domNodeCndList [list $domNodeCnd]
    } else {
        set xp {//condition[not(ancestor::blockdata[@n="Internal data"]) and dependencies[@update_condition_groups]]}
        set domNodeCndList [$doc selectNodes $xp]
    }
    set foreach_nodeDict ""
    foreach domNodeCnd $domNodeCndList {
        foreach depNode [$domNodeCnd selectNodes {dependencies[@update_condition_groups]}] {
            set foreach_node [$depNode @foreach /*]
            set l [list $domNodeCnd [$depNode @update_condition_groups]]
            dict lappend foreach_nodeDict $foreach_node $l
        }
    }
    set actualizeNodes ""
    dict for "n v" $foreach_nodeDict {
        foreach faNode [$doc selectNodes $n] {
            set unc [uncompress_subtree $faNode]
            foreach l $v {
                lassign $l domNodeCnd update_condition_groups
                set tomodNodesCnd [$faNode selectNodes $update_condition_groups]
                foreach tomodNodeCnd $tomodNodesCnd {
                    set ret [_check_condition_dependencies1_do $domNodeCnd $tomodNodeCnd \
                        $subtreeNode]
                    lappend actualizeNodes {*}$ret
                }
            }
            set found 0
            foreach i $actualizeNodes {
                if { $faNode in [$i selectNodes ancestor::*] } {
                    set found 1
                    break
                }
            }
            if { $unc && !$found } { compress_subtree $faNode }
        }
    }
    return [lsort -unique $actualizeNodes]
}

################################################################################
#    Utility popup help routines
################################################################################

proc gid_groups_conds::_popup_help_activate_motion { args } {

    set optional {
        { -strict boolean 1 }
    }
    set compulsory "l X0 Y0"
    parse_args $optional $compulsory $args

    if { ![winfo exists $l] } { return }
    set w [winfo toplevel $l]
    bind $w <Motion> [list gid_groups_conds::_popup_help_check_destroy -strict $strict \
        $w %x %y %X %Y $X0 $Y0]
}

proc gid_groups_conds::_popup_help_check_destroy { args } {

    set optional {
        { -strict boolean 1 }
    }
    set compulsory "w x y X Y X0 Y0"
    parse_args $optional $compulsory $args

    set top [winfo toplevel $w]
    set focus [focus]
    if { $focus ne "" && [winfo toplevel $focus] ne [winfo toplevel [winfo parent $top]] } {
        destroy $w
        return
    }
    if { $strict } {
        set radius 10
    } else {
        set radius 100
    }
    if { $strict && abs($X-$X0)>$radius || abs($Y-$Y0)>$radius } {
        destroy $w
        return
    }
    if { $x < -$radius || $x > [winfo width $w]+$radius || $y < -$radius || $y > $radius+[winfo height $w] } {
        destroy $w
    }
}

proc gid_groups_conds::_popup_help_button_event { w event X Y } {
    destroy $w
    if { $event ne "" } {
        set win [winfo containing $X $Y]
        if { $win eq "" } { return }
        set x [expr {$X-[winfo rootx $win]}]
        set y [expr {$Y-[winfo rooty $win]}]
        event generate $win $event -x $x -y $y -rootx $X -rooty $Y
    }
}

proc gid_groups_conds::is_popup_help_active { wp } {
    return [winfo exists $wp.tootip]
}

proc gid_groups_conds::text_html_init { text } {

    $text configure -tabs "10mm right"

    $text tag configure bold -font SystemBoldFont
    $text tag configure italic -font SystemItalicsFont
    $text tag configure sub -font SystemTinyFont -offset [expr {-0.5*[font metrics SystemTinyFont -linespace]}]
    $text tag configure sup -font SystemTinyFont -offset [expr {0.5*[font metrics SystemTinyFont -linespace]}]
    $text tag configure small -font SystemTinyFont
    $text tag configure small_italic -font SystemItalicsTinyFont
    $text tag configure blockquote -lmargin1 6 -lmargin2 6 -rmargin 6 -spacing1 6
    $text tag configure item -lmargin1 15 -lmargin2 24
}

proc gid_groups_conds::text_html_add_tag_process_cmd { cmd } {
    popup_help_cancel
    uplevel #0 $cmd
}

# move to end proc gid_groups_conds::text_html_add_tag_do

proc gid_groups_conds::text_html_add_tag { args } {
    variable register_popup_help_handler

    set optional {
        { -cmd_handler cmd "" }
    }
    set compulsory "text idx1 idx_end"
    parse_args $optional $compulsory $args

    if { $cmd_handler eq "" } {
        set cmd_handler [set! register_popup_help_handler]
    }

    while {  $idx1 ne "" } {
        set tagsList [lassign [gid_groups_conds::text_html_add_tag_do $cmd_handler $text $idx1 $idx_end] idx1]
        set idx 0
        foreach i $tagsList {
            lassign $i i1 i2 t
            if { $t in "small italic" } {
                switch $t small { set t_alt italic } italic { set t_alt small }
                foreach ipos [lsearch -index 2 -exact -all $tagsList $t_alt] {
                    lassign [lindex $tagsList $ipos] i1_alt i2_alt
                    if { $i1 eq $i1_alt && $i2 eq $i2_alt } {
                        lset tagsList $idx 2 small_italic
                        lset tagsList $ipos 2 small_italic
                    }
                }
            }
            incr idx
        }
        foreach i $tagsList {
            lassign $i i1 i2 t
            $text tag add $t $i1 $i2
        }
    }

    #small_italic

}

proc gid_groups_conds::popup_help { args } {
    variable register_popup_help_window
    variable register_popup_help_handler

    set optional {
        { -minimum_time miliseconds 0 }
        { -width auto|free|size auto }
        { -buttons buttonsList "" }
        { -image_tag image_tag "" }
    }
    set compulsory "wp text x y"
    parse_args $optional $compulsory $args

    set focus [focus]
    if { $focus ne ""  && [winfo toplevel $focus] ne [winfo toplevel $wp] } {
        return
    }
    destroy $wp.tootip

    set focus [focus]
    set grab [grab current $wp]

    if { $grab ne "" } {
        if { ![string equal -length [string length $grab] $grab $wp] } {
            return
        }
    }
    set w [cu::create_tooltip_toplevel -withdraw $wp.tootip]
    set register_popup_help_window $w

    if { [llength $buttons] } {
        ttk::frame $w.buttons
        set cmd0 [list destroy $w]
        set inum 0
        foreach b $buttons {
            dict set b -command "$cmd0; [dict_getd $b -command ""]"
            ttk::button $w.buttons.b$inum {*}$b
            foreach i [list <ButtonPress-1> <ButtonRelease-1>] {
                bind $w.buttons.b$inum $i "[bind TButton $i]; break"
            }
            grid $w.buttons.b$inum -sticky w -padx 2 -pady 2 -row 0 -column $inum
            incr inum
        }
    }
    if 0 {
        label $w.l -text $text -highlightthickness 1 -background white -foreground black \
            -justify left -wraplength 300
    }

    text $w.l -highlightthickness 0 -background white -foreground black \
        -wrap none -width 40 -height 4 -bd 1 -tabs "3mm" -relief solid \
        -font SystemFont

    $w.l insert end $text

    if {$image_tag != ""} {
        $w.l image create end -image $image_tag -align center -padx 5
        set im_width [image width $image_tag]
        set im_height [image height $image_tag]
    }

    if { [regexp {<a\s+href} $text] || [llength $buttons] } {
        set strict 0
    } else {
        set strict 1
    }
    text_html_init $w.l
    text_html_add_tag $w.l 1.0 end
    $w.l configure -state disabled

    set irow 0
    grid $w.l -padx 1 -pady 1 -sticky nsew
    if { [winfo exists $w.buttons] } {
        grid $w.buttons -sticky e
        incr irow
    }
    grid columnconfigure $w 0 -weight 1
    grid rowconfigure $w 0 -weight 1

    update idletasks
    set width_line 0
    set nlines [expr {[llength [split $text \n]]+1}]
    for { set line 1 } { $line <=  $nlines } { incr line } {
        set xi [$w.l count -update -xpixels "$line.0" "$line.0 lineend"]
        if { $xi > $width_line } { set width_line $xi }
    }
    set width_line [expr {$width_line+10}]

    if { $width eq "auto" } {
        if { $width_line > 220 } { set width_line 220 }
    } elseif { [string is integer -strict $width] } {
        set width_line $width
    }

    $w.l configure -wrap word
    wm geometry $w ${width_line}x200+${x}+$y

    update
    if { ![winfo exists $w.l] } { return }

    if { $::tcl_platform(platform) eq "windows" } {
        set nlines [scan [$w.l index end-1c] "%d"]
        set n 0
        for { set i 1 } { $i <= $nlines } { incr i } {
            set t [$w.l get "$i.0" "$i.0 lineend"]
            incr n [expr {int([font measure [$w.l cget -font] $t]/$width_line)+1}]
        }
        set height [expr {$n*[font metrics [$w.l cget -font] -linespace]}]
        set height [expr {$height+$n*4}]
    } else {
        set height [$w.l count -update -ypixels 1.0 end]
        set height [expr {$height+5}]
    }
    if { [winfo exists $w.buttons] } {
        if { $width_line < [winfo reqwidth $w.buttons] } {
            set width_line [winfo reqwidth $w.buttons]
        }
        incr height [winfo reqheight $w.buttons]
    }
    #   incr y 24
    if {$image_tag != ""} {
        set height [expr {$height + [image height $image_tag]}]
        set width_line [image width $image_tag]
        if { $width eq "auto" && $width_line > 220 } {
            set width_line 220
        }
    }
    incr y [expr {-24-$height}]

    # [winfo reqheight $w.l]
    if { $y+$height > [winfo screenheight $wp]-40 } {
        set y [expr {[winfo screenheight $wp]-$height-40}]
    }
    if { $y < 0 } { set y 0 }

    wm overrideredirect $w 1
    wm geometry $w ${width_line}x$height+${x}+$y
    wm deiconify $w

    bind $w <ButtonPress-1> "[list gid_groups_conds::_popup_help_button_event \
        $w <ButtonPress-1> %X %Y];break"
    bind $w <ButtonPress-3> "[list gid_groups_conds::_popup_help_button_event \
        $w <ButtonPress-3> %X %Y];break"
    bind $w <ButtonPress-4> "[list gid_groups_conds::_popup_help_button_event \
        $w <ButtonPress-4> %X %Y];break"
    bind $w <ButtonPress-5> "[list gid_groups_conds::_popup_help_button_event \
        $w <ButtonPress-5> %X %Y];break"
    bind $w <MouseWheel> "[list gid_groups_conds::_popup_help_button_event \
        $w <MouseWheel> %X %Y];break"

    after $minimum_time [namespace code [list _popup_help_activate_motion -strict $strict $w.l $x $y]]
    bind $w.l <KeyPress> [list gid_groups_conds::_popup_help_button_event \
        $w "" "" ""]
    bind $w <Destroy> [list gid_groups_conds::popup_help_destroy $focus $grab]

    set err [catch { grab $w }]
    if { $err } { destroy $w }
    return $w
}

proc gid_groups_conds::popup_help_destroy { focus grab } {
    catch { focus $focus }
    catch { grab $grab }
}


proc gid_groups_conds::popup_help_cancel {} {
    variable register_popup_help_after
    variable register_popup_help_window

    if { $register_popup_help_after ne "" } {
        after cancel $register_popup_help_after
        set register_popup_help_after ""
    }
    if { [winfo exists $register_popup_help_window] } {
        destroy $register_popup_help_window
    }
}

proc gid_groups_conds::_register_popup_help_do { args } {
    variable register_popup_help_after
    variable register_popup_help_active

    set optional {
        { -before_pop_cmd cmd_prefix "" }
        { -text_tag tagname "" }
        { -image_tag image_tag "" }
        { -width auto|free|size auto }
    }
    set compulsory "w text x_old y_old"
    parse_args $optional $compulsory $args

    if { !$register_popup_help_active } { return }

    if { $register_popup_help_after ne "" } {
        after cancel $register_popup_help_after
        set register_popup_help_after ""
    }
    if { ![winfo exists $w] } { return }
    set x [winfo pointerx $w]
    set y [winfo pointery $w]
    if { $x != $x_old || $y != $y_old } { return }
    if { $text_tag ne "" } {

    } else {
        if { [winfo containing $x $y] ne $w } { return }
    }
    if { $before_pop_cmd ne "" } {
        set text [uplevel #0 $before_pop_cmd [list $w $text]]
    }

    popup_help -width $width -image_tag $image_tag $w $text $x $y
}

proc gid_groups_conds::_register_popup_help_start { args } {
    variable register_popup_help_after
    variable register_popup_help_active

    set optional {
        { -before_pop_cmd cmd_prefix "" }
        { -text_tag tagname "" }
        { -image_tag image_tag "" }
        { -width auto|free|size auto }
    }
    set compulsory "w text x y"
    parse_args $optional $compulsory $args

    if { !$register_popup_help_active } { return }

    if { $register_popup_help_after ne "" } {
        after cancel $register_popup_help_after
        set register_popup_help_after ""
    }
    set register_popup_help_after [after 600 \
        [list gid_groups_conds::_register_popup_help_do \
        -before_pop_cmd $before_pop_cmd -text_tag $text_tag \
        -image_tag $image_tag -width $width \
        $w $text $x $y]]
}

proc gid_groups_conds::_register_popup_help_add_contents { w text } {
    if { [winfo exists $w] } {
        set xe [expr {[lindex [$w bbox end] 0]+[lindex [$w bbox end] 2]}]
        set wd [winfo width $w]
        if { [winfo class $w] eq "TCombobox" } {
            incr wd -16
        }
        set contents [$w get]
        if { [lindex [$w bbox 0] 0] >= 0 && $xe <= $wd } {
            set contents ""
        }
        if { $contents ne "" } {
            if { [string length $contents] > 400 } {
                set contents [string range $contents 0 396]...
            }
            return "$contents\n\n$text"
        }
    }
    return $text
}

proc gid_groups_conds::register_popup_help { args } {

    set optional {
        { -before_pop_cmd cmd_prefix "" }
        { -check_bbox "" 0 }
        { -text_tag tagname "" }
        { -image_tag image_tag "" }
        { -list "" 0 }
        { -width auto|free|size auto }
    }
    set compulsory "w"
    set args [parse_args -raise_compulsory_error 0 $optional $compulsory $args]

    if { $check_bbox } {
        set before_pop_cmd [list gid_groups_conds::_register_popup_help_add_contents]
    }

    if { $list } {
        set wList $w
    } else {
        set wList [list $w]
    }
    foreach w $wList {
        switch [llength $args] {
            0 {
                return [lindex [bind $w <Motion>] 4]
            }
            1 {
                set text [string map [list % %%] [lindex $args 0]]
                if { $text_tag ne "" } {
                    $w tag bind $text_tag <Motion> [list gid_groups_conds::_register_popup_help_start \
                        -before_pop_cmd $before_pop_cmd -text_tag $text_tag -image_tag $image_tag \
                        -width $width $w $text %X %Y]
                } else {
                    bind $w <Motion> [list gid_groups_conds::_register_popup_help_start \
                        -before_pop_cmd $before_pop_cmd -width $width -image_tag $image_tag $w \
                        $text %X %Y]
                }
            }
            default {
                error "error in gid_groups_conds::register_popup_help"
            }
        }
    }
}

proc gid_groups_conds::register_popup_help_link_handler { handler } {
    variable register_popup_help_handler
    set register_popup_help_handler $handler
}

proc gid_groups_conds::register_popup_help_deactivate {} {
    variable register_popup_help_active

    set register_popup_help_active 0
}

proc gid_groups_conds::register_popup_help_reactivate {} {
    variable register_popup_help_active

    set register_popup_help_active 1
}

################################################################################
#    Open & save spd file
################################################################################

proc gid_groups_conds::open_XML_file_gzip { file } {

    set err [catch { cu::unzipfile -encoding utf-8 -data $file } ret]
    if { !$err && [llength $ret] == 1 } {
        set data [lindex $ret 0 2]
        set doc [gid_groups_conds::parse_XML_data $data]
    } else {
        set doc [gid_groups_conds::open_XML_file $file]
    }
    return $doc
}

# used in makefile.tcl when generating distribution
proc gid_groups_conds::convert_XML_file_to_zip { file } {

    set doc [gid_groups_conds::open_XML_file_gzip $file]
    cu::zipfile -encoding utf-8 -data $file [list [file tail $file] \
        [$doc asXML -indent none]]
    $doc delete
    unset doc
}

#action could be READING_MODEL or LOADING_PROBLEMTYPE to provide more information
proc gid_groups_conds::open_spd_file { file {action READING_MODEL} } {
    return [gid_groups_conds::open_spd_file_do -action $action $file]
}

proc gid_groups_conds::open_spd_file_do { args } {
    variable doc
    variable open_boundary_cond
    variable gid_group_var

    set optional {
        { -action READING_MODEL|LOADING_PROBLEMTYPE READING_MODEL }
        { -return_on_version_change boolean 0 }
    }
    set compulsory "file"
    parse_args $optional $compulsory $args

    if { [info exists doc] } {
        set doc_old $doc
        set old_version [$doc selectNodes {string(/*/@version)}]
    } else {
        set old_version ""
    }
    set doc_new [gid_groups_conds::open_XML_file_gzip $file]
    set root [$doc_new documentElement]

    if { $return_on_version_change && $old_version ne "" && [$root @version] ne $old_version } {
        return -1
    }

    set doc $doc_new
    if { [info exists doc_old] } {
        set nodes_to_preserve [list display_options global_preferences]
        if { [info exists ::GID_GROUPS] && $action == "LOADING_PROBLEMTYPE" } {
            #when READING_MODEL want the read ones (doc)
            #when LOADING_PROBLEMTYPE want to maintain the old ones (doc_old)
            lappend nodes_to_preserve groups
        }
        foreach i $nodes_to_preserve {
            set node [$doc_old selectNodes /*/$i]
            if { $node ne "" } {
                set nodeold [$root selectNodes [$node nodeName]]
                if { $nodeold ne "" } {
                    $nodeold delete
                }
                $root appendChild [$node clone -deep]
            }
        }
        $doc_old delete
    }

    #check_dependencies

    if { [info exists open_boundary_cond] && [winfo exists $open_boundary_cond] } {
        set width [gid_groups_conds::apply_frame_width_from_doc]
    }
    if { [info exists gid_group_var]  && [winfo exists $gid_group_var] } {
        $gid_group_var configure -domdoc $doc
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        #invoke it directly now to process the current groups, because before load this file the GiD_AfterCreateGroup event procedure
        #was not defined to syncronize both definitions
        set mff [GiD_ModifiedFileFlag get]
        gid_groups_conds::syncronize_gid_groups_and_doc_groups
        if { !$mff } {
            GiD_ModifiedFileFlag set 0
        }
    }
    return 0
}

# file will be typically the defaults file and will have higher priority
# as it only contains default values
proc gid_groups_conds::merge_spd_file_if_exists { file } {
    variable doc
    if { ![info exists doc] } {
        return
    }
    if { [catch { set tdoc [gid_groups_conds::open_XML_file_gzip $file] } err_msg] } {
        return
    }
    set root [$doc documentElement]
    set troot [$tdoc documentElement]
    foreach node [$troot selectNodes *] {
        set nodeold [$root selectNodes [$node nodeName]]
        if { $nodeold ne "" } {
            $nodeold delete
        }
        $root appendChild [$node clone -deep]
    }
    $tdoc delete
    #check_dependencies
}

proc gid_groups_conds::merge_spd_module_file_if_exists { file } {
    variable doc
    if { ![info exists doc] } {
        return
    }
    if { $file eq "" } {
        return
    }
    if { [catch { set tdoc [gid_groups_conds::open_XML_file_gzip $file] } error_msg] } {
        return
    }
    set root [$doc documentElement]
    set troot [$tdoc documentElement]
    foreach node [$troot selectNodes *] {
        set nodeName [$node nodeName]
        if { [$node hasAttribute n] } {
            set name [$node @n]
            set nodeold [$root selectNodes "$nodeName\[@n='$name'\]"]
        } else {
            set nodeold [$root selectNodes [$node nodeName]]
        }
        if { $nodeold ne "" } {
            set root [gid_groups_conds::mergeNodes-r $node $nodeold $root]
        } else {
            $root appendChild [$node clone -deep]
        }
    }
    $tdoc delete
    #check_dependencies
}

#Recursive mergeNodes
proc gid_groups_conds::mergeNodes-r { node nodeold root } {

    foreach attribute [$nodeold attributes] {
        if { [$node hasAttribute $attribute] } {
            set v [$node @$attribute]
            $nodeold setAttribute $attribute $v
        }
    }
    foreach nnode [$node selectNodes *] {
        set nnodeName [$nnode nodeName]
        if { [$nnode hasAttribute n] } {
            set name [$nnode @n]
            if { $name eq "material" } {
                set matname [$nnode @name]
                set nnodeold [$nodeold selectNodes "$nnodeName\[@n='$name' and @name='$matname'\]"]
            } elseif { $name eq "materials_types" } {
                set type [$nnode @pn]
                set nnodeold [$nodeold selectNodes "$nnodeName\[@n='$name' and @pn='$type'\]"]
            } else {
                set nnodeold [$nodeold selectNodes "$nnodeName\[@n='$name'\]"]
            }
        } else {
            set nnodeold [$nodeold selectNodes [$node nodeName]]
        }
        if { $nnodeold ne "" } {
            foreach attribute [$nnodeold attributes] {
                if { [$nnode hasAttribute $attribute] } {
                    set v [$nnode @$attribute]
                    $nnodeold setAttribute $attribute $v
                }
            }
            if { ![$nnodeold hasChildNodes] } {
                $nnodeold delete
                $nodeold appendChild [$nnode clone -deep]
            } else {
                gid_groups_conds::mergeNodes-r $nnode $nnodeold $root
            }
        } else {
            $nodeold appendChild [$nnode clone -deep]
        }
    }
    return $root
}

proc gid_groups_conds::activated_cond_groups { filename } {

    set GidData(CND,nameslist) ""

    set f [open $filename r]
    while { ![eof $f] } {
        gets $f aa
        if { $aa == "" || [string index $aa 0] == "#" } { continue }
        if { ![regexp {NUMBER:[ ]*[0-9]*[ ]*CONDITION:?[ ]*([^ ]*)} \
            $aa {} cndname] } { break }
        lappend GidData(CND,nameslist) $cndname
    }
    set activated_cond_groups 0
    foreach i [list point_groups line_groups surface_groups volume_groups] {
        if { $i in $GidData(CND,nameslist) } {
            set activated_cond_groups 1
            break
        }
    }
    return $activated_cond_groups
}

proc gid_groups_conds::save_spd_file { args } {
    variable doc
    variable open_boundary_cond

    set optional {
        { -compress boolean 1 }
        { -save_post boolean 1 }
    }
    #if { [info exists ::debug] && $::debug == 1 } { lset optional 0 2 0 }

    set compulsory "file"
    parse_args $optional $compulsory $args

    if { ![info exists doc] } { return [dict create result fail] }

    set n_file [file normalize $file]
    set s_file [internal_data get spd_file]
    if { $s_file ne "" && $n_file ne $s_file } {
        set is_save_as 1
    } else {
        set is_save_as 0
    }
    internal_data set spd_file $n_file

    if { [info exists open_boundary_cond] } {
        $open_boundary_cond store_tree_state
    }

    set xp {/*/container/blockdata[@di_ic="0"]}
    foreach domNode [$doc selectNodes $xp] {
        if { [info exists open_boundary_cond] } {
            set is_closed [$open_boundary_cond check_if_closed_and_clear $domNode]
        } else {
            set is_closed 1
        }
        if { $is_closed } {
            compress_subtree $domNode
        }
    }

    if { [info exists gid_group_var] && [winfo exists $gid_group_var] } {
        $gid_group_var store_tree_state
    }

    if { !$compress } {
        set fout [open $file w]
        fconfigure $fout -encoding utf-8
        puts -nonewline $fout {<?xml version="1.0" encoding="utf-8"?>}
        puts $fout [$doc asXML -indent 2]
        close $fout
    } else {
        cu::zipfile -encoding utf-8 -data $file [list [file tail $file] \
            [$doc asXML -indent none]]
    }

    set old_model_name [file root [file dirname $s_file]]
    set new_model_name [file root [file dirname $file]]

    if { !$save_post } {
        return [dict create result ok is_save_as $is_save_as old_model_name $old_model_name \
            new_model_name $new_model_name]
    }

    set directory [file dirname $file]
    set old_directory [file dirname $s_file]
    if { [file tail $directory] eq "backup-[pid].gid" || [file tail $old_directory] eq "backup-[pid].gid"} {
        # In this case it is not necessary to execute SaveAs
        set is_save_as 0
        return [dict create result ok is_save_as $is_save_as old_model_name $old_model_name \
            new_model_name $new_model_name]
    }

    if { ![info exists draw_post::data] } {
        return [dict create result fail is_save_as $is_save_as old_model_name $old_model_name \
            new_model_name $new_model_name]
    }
    if { [draw_post::is_readonly] && !$is_save_as } {
        set ret "ok"
    } else {
        set ret [draw_post::save -directory $directory -save_as $is_save_as -save_also_geometry 0]
    }

    if { $ret eq "cancel" && $is_save_as } {
        set err [catch { draw_post::select_default_postprocess_filename $old_model_name 0 } \
            old_pfile opts]
        if { $err } {
            return -options $opts $old_pfile
        }
        if { $old_pfile eq "" } {
            return [dict create result ok is_save_as $is_save_as old_model_name $old_model_name \
                new_model_name $new_model_name]
        }

        switch -- [file extension $old_pfile] {
            .res { set ext ".flavia.res" }
            default { set ext [file extension $old_pfile] }
        }
        set new_pfile [file join [file dirname $file] [file root [file tail $file]]$ext]
        file copy -force $old_pfile $new_pfile
    }
    return [dict create result ok is_save_as $is_save_as old_model_name $old_model_name \
        new_model_name $new_model_name]
}

#this is similar to gid_groups_conds::save_spd_file, gid_groups_conds::open_spd_file
#but now if groups information is managed by GiD ( if { [info exists ::GID_GROUPS] }
#then GiD read/write the groups in the .grp file calling gid_groups_conds::open_grp_file/gid_groups_conds::save_grp_file
#and the problemtype read/write in the .spd the rest of its data calling gid_groups_conds::open_spd_file/gid_groups_conds::save_spd_file
#in fact the spd is saving also a copy of the groups information and then and old GiD is able to read new models
proc gid_groups_conds::save_grp_file { args } {
    variable groups_file_version
    variable doc
    variable gid_group_var

    set optional {
        { -compress boolean 1 }
    }
    #if { [info exists ::debug] && $::debug == 1 } { lset optional 0 2 0 }

    set compulsory "file"
    parse_args $optional $compulsory $args

    if { ![info exists doc] } { return }

    if { [info exists gid_group_var] && [winfo exists $gid_group_var] } {
        $gid_group_var store_tree_state
    }

    set root [$doc documentElement]
    set newdoc [gid_groups_conds::parse_XML_data "<GID_GROUPS_FILE/>"]
    #dom createDocument GID_GROUPS_FILE newdoc
    set newroot [$newdoc documentElement]
    $newroot setAttribute version $groups_file_version
    set node_names {groups groups_types}
    set preserve_something 1;#is it necessary to preserve this when reading model groups?
    if { $preserve_something } {
        #lappend node_names display_options global_preferences
        lappend node_names display_options
    }
    foreach node_name $node_names {
        set dNode [$root selectNodes $node_name]
        if { $dNode ne "" } {
            $newroot appendChild [$dNode cloneNode -deep]
        }
    }

    if { !$compress } {
        set fout [open $file w]
        fconfigure $fout -encoding utf-8
        puts -nonewline $fout {<?xml version="1.0" encoding="utf-8"?>}
        puts $fout [$newdoc asXML -indent 2]
        close $fout
    } else {
        cu::zipfile -encoding utf-8 -data $file [list [file tail $file] \
            [$newdoc asXML -indent none]]
    }

    $newdoc delete
}

proc gid_groups_conds::open_grp_file { file } {
    variable doc
    variable gid_group_var

    if { ![file exists $file] } {
        WarnWinText [_ "File %s not found" $file]
        return
    }

    if { [info exists doc] } {
        set doc_old $doc
    }
    set doc [gid_groups_conds::open_XML_file_gzip $file]
    set root [$doc documentElement]
    if { [$root nodeName] != "GID_GROUPS_FILE" } {
        $doc delete
        if { [info exists doc_old] } {
            set doc $doc_old
        } else {
            unset doc
        }
        WarnWinText [_ "File %s is not a valid groups file" $file]
        return
    }

    if { [info exists doc_old] } {
        set preserve_something 1;#is it necessary to preserve this when reading model groups?
        if { $preserve_something } {
            #set nodes_to_preserve [list display_options global_preferences]
            set nodes_to_preserve [list display_options]
            foreach i $nodes_to_preserve {
                set node [$doc_old selectNodes /*/$i]
                if { $node ne "" } {
                    set nodeold [$root selectNodes [$node nodeName]]
                    if { $nodeold ne "" } {
                        $nodeold delete
                    }
                    $root appendChild [$node clone -deep]
                }
            }
        }
        $doc_old delete
    }

    #check_dependencies
    if { [info exists gid_group_var]  && [winfo exists $gid_group_var] } {
        $gid_group_var configure -domdoc $doc
    }
    if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
        #invoke it directly now to process the current groups, because before load this file the GiD_AfterCreateGroup event procedure
        #was not defined to syncronize both definitions
        set mff [GiD_ModifiedFileFlag get]
        gid_groups_conds::syncronize_gid_groups_and_doc_groups
        if { !$mff } {
            GiD_ModifiedFileFlag set 0
        }
    }
}



proc gid_groups_conds::save_preferences { file } {
    variable doc

    if { ![info exists doc] } { return }
    set root [$doc documentElement]
    if { [$root nodeName] eq "" } {
        error "error in gid_groups_conds::save_preferences. root not ok"
    }

    set newdoc [gid_groups_conds::parse_XML_data "<[$root nodeName]/>"]
    set newroot [$newdoc documentElement]
    foreach i [$root selectNodes @*] {
        eval [list $newroot setAttribute] $i
    }
    set dNode [$root selectNodes display_options]
    if { $dNode ne "" } {
        $newroot appendChild [$dNode cloneNode -deep]
    }
    set pNode [$root selectNodes global_preferences]
    if { $pNode ne "" } {
        $newroot appendChild [$pNode cloneNode -deep]
    }
    set fout [open $file w]
    fconfigure $fout -encoding utf-8
    puts $fout {<?xml version="1.0" encoding="utf-8"?><!-- -*- coding: utf-8;-*- -->}
    puts $fout [$newdoc asXML -indent none]
    close $fout
    $newdoc delete
}

proc gid_groups_conds::set_preferenceG { args } {
    return [set_preference -local_global global {*}$args]
}

proc gid_groups_conds::set_preference { args } {
    variable doc
    if { ![info exists doc] } {
        return
    }
    set optional {
        { -local_global local|global|all all }
    }
    set compulsory "name value"
    parse_args $optional $compulsory $args
    set root [$doc documentElement]
    foreach lg [list local_preferences global_preferences] {
        set pNode [$root selectNodes $lg]
        if { $pNode eq "" } {
            set pNode [$root appendChildTag $lg]
        }
        set p1Node [$pNode selectNodes [format_xpath {preference[@n=%s]} $name]]
        if { $p1Node ne "" } {
            $p1Node delete
        }
        if { $lg eq "local_preferences" && $local_global eq "global" } {
            continue
        }
        if { $lg eq "global_preferences" && $local_global eq "local" } {
            continue
        }
        $pNode appendChildTag preference [list attributes() n $name v $value]
    }
}

proc gid_groups_conds::get_preference { args } {
    variable doc

    set optional {
        { -check cmd_prefix "" }
    }
    set compulsory "name default"
    parse_args $optional $compulsory $args

    if { ![info exists doc] } { return $default }

    set root [$doc documentElement]

    foreach lg [list local_preferences global_preferences] {
        set pNode [$root selectNodes $lg]
        if { $pNode eq "" } { continue }
        set p1Node [$pNode selectNodes [format_xpath {preference[@n=%s]} $name]]
        if { $p1Node eq "" } { continue }
        set v [$p1Node @v]
        if { $check ne "" } {
            set cmd [list {*}$check $v]
            set err [catch $cmd ret]
            if { $err || !$ret } { return $default }
        }
        return $v
    }
    return $default
}

proc gid_groups_conds::get_preference_node { args } {
    variable doc

    set optional {
        { -clear "" 0 }
        { -local_global local|global local }
    }
    set compulsory "name"
    parse_args $optional $compulsory $args

    set root [$doc documentElement]

    switch $local_global {
        local { set lp local_preferences }
        global { set lp global_preferences }
    }

    set pNode [$root selectNodes $lp]
    if { $pNode eq "" } {
        set pNode [$root appendChildTag $lp]
    }
    set p1Node [$pNode selectNodes $name]
    if { !$clear && $p1Node eq "" && $local_global eq "local" } {
        set p2Node [$root selectNodes global_preferences/$name]
        if { $p2Node ne "" } {
            $pNode appendChild [$p2Node cloneNode -deep]
            set p1Node [$pNode selectNodes $name]
        }
    }
    if { $clear && $p1Node ne "" } {
        $p1Node delete
        set p1Node ""
    }
    if { $p1Node eq "" } {
        return [$pNode appendChildTag $name]
    } else {
        return $p1Node
    }
}

proc gid_groups_conds::update_global_preference_node { name } {
    variable doc

    set root [$doc documentElement]

    set pNode [$root selectNodes local_preferences]
    if { $pNode eq "" } {
        set pNode [$root appendChildTag local_preferences]
    }
    set p1Node [$pNode selectNodes $name]
    if { $p1Node eq "" } {
        $pNode appendChildTag $name
        set p1Node [$pNode selectNodes $name]
    }
    set pNodeG [$root selectNodes global_preferences]
    if { $pNodeG eq "" } {
        set pNodeG [$root appendChildTag global_preferences]
    }
    set p2Node [$pNodeG selectNodes $name]
    if { $p2Node ne "" } {
        $p2Node delete
    }
    $pNodeG appendChild [$p1Node cloneNode -deep]
}

################################################################################
#    images creation
################################################################################

proc gid_groups_conds::create_fonts_colors_styles {} {

    #set SystemButtonFace [$w cget -highlightbackground]
    #set SystemHighlight [$w cget -selectbackground]
    #set SystemHighlightText [$w cget -selectforeground]

    set inside_gid 0
    if { [info exists ::GIDDEFAULT] } {
        set inside_gid 1
    }

    set w [listbox .listbox]
    if { $inside_gid && [info command GiD_Info] ne "" } {
        set currentGiDVer [lindex [split [GiD_Info GiDVersion] .] 0]
    } else {
        set currentGiDVer ""
    }
    if { $::exists_gid_themes } {
        upvar 0 ::ttk::theme::[gid_themes::GetTtkTheme]::colors colors
        set ::SystemBackgroundSearchOpener #9114cd
        set ::SystemBackgroundFactorOpener blue
        set ::SystemBackgroundHi $colors(-highlight)
        set ::SystemBackground $colors(-frame_bg)
        set ::SystemHighlightNoFocus gray
        set ::SystemHighlight $colors(-highlight)
        set ::SystemHighlightText white
    } elseif { !$inside_gid || $currentGiDVer >= "12"} {
        set ::SystemBackgroundSearchOpener #9114cd
        set ::SystemBackgroundFactorOpener blue
        set ::SystemBackgroundHi #880000
        set ::SystemBackground #e9dfcd
        set ::SystemHighlightNoFocus gray
        set ::SystemHighlight #880000
        set ::SystemHighlightText white
    }

    if { [lsearch -exact [font names] SystemFont] == -1 } {
        font create SystemFont {*}[font actual [$w cget -font]]
        font create SystemBoldFont {*}[concat [font actual SystemFont] -weight bold]
        font create SystemItalicsFont {*}[concat [font actual SystemFont] -slant italic]
        set size [expr {round(0.85*[font actual SystemFont -size])}]
        font create SystemTinyFont -family Arial -size $size
        font create SystemItalicsTinyFont -family Arial -size $size -slant italic
    }
    destroy $w

    #style layout TLabelframeG [style layout TLabelframe]
    # following line is useful for winnative theme
    #style configure Toolbutton -padding 1
    ttk::style layout ToolbuttonT [ttk::style layout Toolbutton]
    eval ttk::style map ToolbuttonT [ttk::style map Toolbutton]
    eval ttk::style configure ToolbuttonT [ttk::style configure Toolbutton]
    ttk::style configure ToolbuttonT -font SystemTinyFont
}

proc gid_groups_conds::create_fonts_colors_styles_compassis {} {

    #set SystemButtonFace [$w cget -highlightbackground]
    #set SystemHighlight [$w cget -selectbackground]
    #set SystemHighlightText [$w cget -selectforeground]

    set inside_gid 0
    if { [info exists ::GIDDEFAULT] } {
        set inside_gid 1
    }

    set w [listbox .listbox]
    if { $inside_gid } {
        set currentGiDVer [lindex [split [GiD_Info GiDVersion] .] 0]
    } else {
        set currentGiDVer ""
    }
    if { !$inside_gid || $currentGiDVer == "12"} {
        set ::SystemBackgroundSearchOpener #9114cd
        set ::SystemBackgroundFactorOpener blue
        set ::SystemBackgroundHi #880000
        set ::SystemBackground #e9dfcd
        set ::SystemHighlightNoFocus gray
        set ::SystemHighlight #880000
        set ::SystemHighlightText white
    } else {
        upvar 0 ::ttk::theme::[gid_themes::GetTtkTheme]::colors colors
        set ::SystemBackgroundSearchOpener #9114cd
        set ::SystemBackgroundFactorOpener blue
        set ::SystemBackgroundHi $colors(-highlight)
        set ::SystemBackground $colors(-frame_bg)
        set ::SystemHighlightNoFocus #b4b5b2
        set ::SystemHighlight $colors(-highlight)
        set ::SystemHighlightText white
    }
    if { [lsearch -exact [font names] SystemFont] == -1 } {
        font create SystemFont {*}[font actual [$w cget -font]]
        font create SystemBoldFont {*}[concat [font actual SystemFont] -weight bold]
        font create SystemItalicsFont {*}[concat [font actual SystemFont] -slant italic]
        set size [expr {round(0.85*[font actual SystemFont -size])}]
        font create SystemTinyFont -family Arial -size $size
        font create SystemItalicsTinyFont -family Arial -size $size -slant italic
    }
    if { $::tcl_platform(platform) eq "unix" } {
        if { [font actual SystemFont -size] > 10 } {
            font configure SystemFont -size 10
            font configure SystemBoldFont -size 10
        }
    }
    destroy $w

    #style layout TLabelframeG [style layout TLabelframe]
    # following line is useful for winnative theme
    #style configure Toolbutton -padding 1
    ttk::style layout ToolbuttonT [ttk::style layout Toolbutton]
    eval ttk::style map ToolbuttonT [ttk::style map Toolbutton]
    eval ttk::style configure ToolbuttonT [ttk::style configure Toolbutton]
    ttk::style configure ToolbuttonT -font SystemTinyFont
}

proc gid_groups_conds::giveimage { name } {
    if { [info exists ::GidPrivImages(tree,$name)] } {
        return $::GidPrivImages(tree,$name)
    }
    set ret ""
    set err [catch { icon_chooser::giveimage $name } ret]
    if { $err } {
        set err [catch { cu::get_image $name } ret]
        if { $err } {
            set ret [giveimageL $name]
        }
    }
    set ::GidPrivImages(tree,$name) $ret
    return $ret
}

proc gid_groups_conds::create_images {} {
    variable images
    variable topdir
    variable style

    if { [info exists images] } { return }

    set dir_images [file join $topdir images]

    cu::get_images_add_dir $dir_images

    foreach i [list results_contourfill results_linediagram results_vectors results_mainstresses results_deformed results_localaxes] {
        if { ![info exists ::GidPriv(image_${i})] } {
            set ::GidPriv(image_${i}) [image create photo -file $dir_images/${i}.gif]
        }
    }

    set imagesList [list gear-22 edit-22 forward-16 up-16 down-16 \
        finish-16 top-16 bottom-16 back-16 filenew-16 filesave-16 \
        mail_get-16 removecomment-16 colorize-16 locationbar_erase-22 \
        bookmark_folder-16 queue-16 filefind-16 editcut-16 editcopy-16 \
        editpaste-16 attach-16 view_text-16 view_detailed-16 \
        forward-22 back-22]

    if { [info command ::icon_chooser::giveimage] eq "" } {
        # it is necessary when executing gid with option -n2
        catch { auto_load ::icon_chooser::giveimage }
    }
    if { [info command ::icon_chooser::giveimage] ne "" } {
        foreach i $imagesList {
            set images($i) [icon_chooser::giveimage $i]
        }
    } else {
        # catch is here for cases where we are using the package without gid and without full
        catch {
            foreach i $imagesList {
                set err [catch {
                    set images($i) [cu::get_image $i]
                } ret]
            }
        }
    }


    set inside_gid 0
    if { [info exists ::GIDDEFAULT] } {
        set inside_gid 1
    }
    if { $inside_gid && [regexp "Tdyn" $style {} curr_theme] } {
        create_customized_theme
    } else {
        foreach name {mac-collapse mac-collapse2 mac-expand mac-expand2} {
            image create photo $name -file [file join $dir_images $name.png]
        }
    }
    foreach name { appbook16 folder-open folder-closed folder-magenta-open folder-magenta-closed \
        boundary_conds-16 boundary_conds_yellow-16 local_axes-16 blank-16 \
            CompassIS_logo_img16 CompassIS_logo_img CompassIS_logo_img128 } {
            image create photo $name -file [file join $dir_images $name.png]
    }
}

proc gid_groups_conds::create_customized_theme { } {
    global GidPriv

    set topdir $GidPriv(CurrentTheme,path,0,generic)

    foreach i [list mac-collapse mac-collapse2 mac-expand mac-expand2 mac-up arrow_left arrow_right] {
        catch { image delete $i }
        set file [file join $topdir $i.png]
        image create photo $i -file $file
    }
}

proc gid_groups_conds::giveimageL { name } {
    # GiD change
    set img [gid_groups_conds::giveimageL-original $name]
    if { $::exists_gid_themes} {
        set img [gid_themes::HighResolutionImage $img]
    }
    return $img
}

proc gid_groups_conds::giveimageL-original { name } {
    # end GiD change
    variable imagesdirList
    variable images_cache

    if { $name eq "" } { return "" }
    if { [info command $name] ne "" } { return $name }
    #if { [lsearch -exact [image names] $name] != -1 } { return $name }
    if { [dict exists $images_cache $name] } {
        return [dict get $images_cache $name]
    }
    set err 1
    if { ![info exists imagesdirList] } {
        set imagesdirList ""
    }
    foreach dir $imagesdirList {
        set files [glob -nocomplain -directory $dir $name]
        lappend files {*}[glob -nocomplain -directory $dir $name.*]
        set file [lindex $files 0]
        set err [catch { image create photo -file $file } img]
        if { !$err } { break }
    }
    if { $err && [info exists icons_lib_db] } {
        set err [catch { cu::get_image_or_icon $name } img]
    }
    if { $err } {
        error [_ "error: image '%s' not found or not correct in directories '%s'" $name $imagesdirList]
    }
    dict set images_cache $name $img
    return $img
}

proc gid_groups_conds::create_give_image { name data } {
    variable images_cache

    if { $data eq "" } {
        return ""
    }
    if { [dict exists $images_cache $name] } {
        return [dict get $images_cache $name]
    }

    if { [regexp {^\s*data:([^;]+);base64,(.*)} $data {} mtype data] } {
        unset -nocomplain type
        switch -- $mtype {
            image/png { set type png }
            image/gif { set type gif }
            image/jpeg { set type jpeg }
        }
        if { [info exists type] } {
            set img [image create photo -format $type -data $data]
            dict set images_cache $name $img
            return $img
        }
    }
    return [giveimageL $data]
}

proc gid_groups_conds::add_to_internal_images_cache { img } {
    dict set images_cache internal_image::$img $img
}

################################################################################
#    Generic utilities
################################################################################

if { [info command lrepeat] eq "" } {
    proc lrepeat {count value args} {
        set values [linsert $args 0 $value]
        set result {}
        for {set i 0} {$i < $count} {incr i} {
            eval [list lappend result] $values
        }
        return $result
    }
}
namespace eval ::dom::domNode {}

# proc ::dom::domNode::appendChildTag { node tagName } {
#     set newnode [[$node ownerDocument] createElement $tagName]
#     $node appendChild $newnode
#     return $newnode
# }

proc ::dom::domNode::appendText { node text } {
    foreach childNode [$node SelectNodes text()] {
        $childNode delete
    }
    set newnode [[$node ownerDocument] createTextNode $text]
    $node appendChild $newnode
    return $newnode
}

proc ::dom::domNode::getAttributeP { node args } {

    set optional {
        { -asList "" 0 }
        { -defaultValue value "" }
    }
    set compulsory "xpath attributeName"
    parse_args $optional $compulsory $args

    set domNode [$node selectNodes $xpath]
    if { !$asList } {
        if { [llength $domNode] != 1 } {
            error "error in dom::domNode::getAttributeP. The xpath expression returns [llength $domNode] nodes. It should return one and only one"
        }
        return [get_domnode_attribute $domNode $attributeName $defaultValue]
    } else {
        set ret ""
        foreach domNode [$node selectNodes $xpath] {
            lappend ret [get_domnode_attribute $domNode $attributeName $defaultValue]
        }
        return $ret
    }
}

# arguments: domNode xpath attributeName { defaultValue "" }
proc get_domnode_attribute_path { args } {

    set defaultValue ""
    lassign $args domNode xpath attributeName defaultValue
    set domNode [$domNode selectNodes $xpath]
    return [get_domnode_attribute $domNode $attributeName $defaultValue]
}

# arguments: domNode xpath attributeName { defaultValue "" }
proc get_domnode_attribute_path_list { domNode xpath attributeName { defaultValue "" } } {
    set ret ""
    foreach domNode [$domNode selectNodes $xpath] {
        lappend ret [get_domnode_attribute $domNode $attributeName $defaultValue]
    }
    return $ret
}

# arguments: domNode attributeName { defaultValue "" }
proc get_domnode_attribute { domNode attributeName { defaultValue "" } } {
    set res [string trim [$domNode getAttribute $attributeName $defaultValue]]
    set res [_get_attribute_value $domNode $res]
    if { $attributeName eq "pn" } {
        set res [subst -nocommands -novariables [= $res]]
    }
    return $res
}

proc _get_attribute_value { domNode res } {
    if { [gid_groups_conds::string_is_proc $res] } {
        set proc_name [string range $res 1 end-1]
        return [gid_groups_conds::eval_proc $proc_name $domNode]
    } elseif { [string index $res 0] eq "\{" && [string index $res end] eq "\}" } {
        set list [split $res ","]
        if { [llength $list] == 1 } {
            set xpath [string range $res 1 end-1]
            return [_get_attribute_xpath $domNode $xpath]
        } else {
            for { set i 0 } { $i < [llength $list]-1 } { incr i 2 } {
                set xpath [string range [lindex $list $i] 1 end-1]
                set ret [_get_attribute_xpath $domNode $xpath]
                if { $ret } {
                    return [lindex $list $i+1]
                }
            }
            if { $i == [llength $list]-1 } {
                return [lindex $list end]
            }
        }
    }
    return $res
}

proc _get_attribute_xpath { domNode xpath } {
    set ret [$domNode selectNodes $xpath typevar]
    switch $typevar {
        empty - bool - number - string {
            return $ret
        }
        nodes {
            set resList ""
            foreach node $ret {
                lappend resList [$node @n]
            }
            return [join $resList ,]
        }
        attrnodes {
            set resList ""
            foreach val $ret {
                set v [lindex $val 1]
                lappend resList [_get_attribute_value $domNode $v]
            }
            return [join $resList ,]
        }
        mixed {
            error "error with expression '$xpath'"
        }
    }
    return ""
}

#to not get attribute 'v' empty because the GUI tree node was never opened and then filled
#this proc evaluat attribute 'values' if exists to get the first items as default value
# e.g.
# a domNode value defined by this .spd:
#
# <container n="units" pn="Units" icon="darkorange-length-18">
#   <value n="test" pn="My test" v='' values='[join {one two three} ,]'/>
# </container>
#
#if the container tree is not opened in the GUI the attribute v remain ''
# set domNode [[$::gid_groups_conds::doc documentElement] selectNodes {container/value[@n='test']}]
# W [get_domnode_attribute $domNode v]
# --> ""
#
#and unfolding the container the value is filled to 'uno'
# W [get_domnode_attribute $domNode v]
# --> "one"
#
# this behavior become a problem to write the calculation file without have visited as user the tree!!
# the new proc gid_groups_conds::get_domnode_attribute_2 in both scenaries do what we want
#
# W [gid_groups_conds::get_domnode_attribute_2 $domNode v]
# --> "one"
#
proc gid_groups_conds::get_domnode_attribute_2 { domNode attributeName { defaultValue "" } } {
    set v [get_domnode_attribute $domNode $attributeName $defaultValue]
    if { $attributeName == "v" } {
        if { $v == "" } {
            # attribute v could depend of other attributes with Tcl code like values
            # Maybe others special attribuges (dict value state ...) must be taken into account!!
            if { [$domNode hasAttribute values] } {
                #this eval the att code and can modiy v
                set values [get_domnode_attribute $domNode values]
                set v [lindex [split $values ,] 0]
                #$domNode setAttribute v $v
                #do not change the attribute to this default value, to left get_XX a proc const (read only)
            }
        }
    }
    return $v
}

proc gid_groups_conds::nice_xpath { domNode } {

    set doc [$domNode ownerDocument]
    set root [$doc documentElement]

    set nodeList ""
    while { $domNode ne $root } {
        set parentNode [$domNode parentNode]
        if { $parentNode eq "" } {
            break
        }
        if { [$domNode @n ""] ne "" } {
            set xp [$domNode nodeName]
            append xp [format_xpath {[@n=%s]} [$domNode @n]]
        } else {
            set xp [$domNode nodeName]
        }
        if { $xp eq "" } {
            set NL ""
        } else {
            set NL [$parentNode selectNodes $xp]
        }
        if { [llength $NL] > 1 || [$domNode @sequence 0] == 1 } {
            set ipos [lsearch -exact $NL $domNode]
            append xp \[[expr {$ipos+1}]\]
        }
        set nodeList [linsert $nodeList 0 $xp]
        set domNode $parentNode
    }
    return [join $nodeList /]
}

#to create the list of maps only once, not for each item of the tree each search
proc gid_groups_conds::get_fcompare_str_maps { } {
    variable _fcompare_str_maps
    if { ![info exists _fcompare_str_maps] } {
        set groups [list a\ue1\ue2\ue3\ue4\ue5 e\ue8\ue9\uea\ueb \
            i\uec\ued\uee\uef o\uf2\uf3\uf4\uf5\uf6 u\ufa\ufb\ufc]

        set _fcompare_str_maps [list]
        foreach group $groups {
            set rest [lassign [split $group ""] base]
            foreach i $rest {
                lappend _fcompare_str_maps $i $base
            }
        }
    }
    return $_fcompare_str_maps
}

proc gid_groups_conds::fcompare_str { str } {
    set maps [gid_groups_conds::get_fcompare_str_maps]
    return [string map $maps [string tolower $str]]
}

proc gid_groups_conds::_eval_proc_xpath { args } {
    set optional {
        { -string "" 0 }
    }
    set compulsory "domNode xpath"
    parse_args $optional $compulsory $args
    if { $string } {
        set xpath "string($xpath)"
    }
    return [$domNode selectNodes $xpath]
}

proc gid_groups_conds::_eval_value_cond { args } {
    set optional {
    }
    set compulsory "domNode xpath groupName valueName"
    parse_args $optional $compulsory $args
    append xpath [format_xpath {/group[@n=%s]/value[@n=%s]/@v} $groupName $valueName]
    return [$domNode selectNodes "string($xpath)"]
}

proc gid_groups_conds::eval_proc { args } {
    set optional {
        { -tree tree "" }
        { -boundary_conds boundary_conds "" }
        { -item item "" }
        { -dict dict "" }
        { -dict_units dict_units "" }
    }
    set compulsory "procScript domNode"
    parse_args $optional $compulsory $args

    if { $domNode eq "" } {
        variable doc
        set domNode [$doc documentElement]
    }

    uncompress_subtree $domNode

    set err [catch { lindex $procScript 0 } proc0]
    if { !$err } {
        set xp [format_xpath {/*/procs/proc[@n=%s]} $proc0]
        set procNode [$domNode selectNodes $xp]

        if { $procNode eq "" } {
            variable doc
            set procNode [$doc selectNodes $xp]
            if { $procNode ne "" } {
                set domNode $procNode
            }
        }
    } else {
        set procNode ""
    }
    if { $procNode eq "" } {
        interp alias "" xpath "" gid_groups_conds::_eval_proc_xpath $domNode
        interp alias "" xpath_s "" gid_groups_conds::_eval_proc_xpath -string $domNode
        interp alias "" value_cond "" gid_groups_conds::_eval_value_cond $domNode

        set procScript [string map [list %W [list $domNode] %TREE [list $tree] %BC [list $boundary_conds] %ITEM $item %DICT [list $dict] %DICT_UNITS [list $dict_units]] $procScript]
        set err [catch {{*}$procScript} ret opts]
        if { $err == 0 } {
            return $ret
        } elseif { $err == 2 } {
            #kike: why this err==2 is considered like no error??
            return $ret
        } else {
            return -options $opts $ret
        }
    }
    #     if { $procNode eq "" } {
    #         error "error in getAttribute : proc '$proc0' does not exist"
    #     }
    set args [$procNode @args ""]
    set pargs [lrange $procScript 1 end]
    if { [lindex $args end] eq "args" } {
        if { [llength $pargs] < [llength $args]-1 } {
            error "error in eval_proc. procScript=$procScript procNode=[$procNode asXML]"
        }
        foreach i [lrange $args 0 end-1] j $pargs {
            if { $i eq "" } { break }
            set $i $j
        }
        set args [lrange $pargs [expr {[llength $args]-1}] end]
    } else {
        if { [llength $pargs] != [llength $args] } {
            error "error in eval_proc. procScript=$procScript procNode=[$procNode asXML]"
        }
        if { [llength $args] } {
            foreach $args $pargs break
        }
    }
    return [eval [$procNode text]]
}

proc gid_groups_conds::eval_transform { args } {

    set optional {
    }
    set compulsory "procNode"
    parse_args $optional $compulsory $args

    return [eval [$procNode text]]
}

################################################################################
#    nice print things
################################################################################

proc gid_groups_conds::nice_number { args } {
    set optional  {
        { -precision integer 6 }
    }

    set compulsory "number"
    parse_args $optional $compulsory $args

    set format "%."
    append format $precision
    append format g
    set number [format $format $number]
    regsub {([eE][+-])0(\d{2})} $number {\1\2} number
    #if { $number eq "0" } { set number 0.0 }
    return $number
}

proc gid_groups_conds::nice_number5 { number } {
    set number [format %.5g $number]
    regsub {([eE][+-])0(\d{2})} $number {\1\2} number
    return $number
}

proc gid_groups_conds::nice_number_print { number } {
    set number [format %.4g $number]
    regsub {([eE][+-])0(\d{2})} $number {\1\2} number
    if { abs($number) < 1e-6 } { set number 0.0 }
    return $number
}

proc gid_groups_conds::nice_number_printList { numberList } {
    set ret ""
    foreach n $numberList {
        lappend ret [nice_number_print $n]
    }
    return [join $ret " "]
}

proc gid_groups_conds::give_printable_value { domNode } {

    if { [get_domnode_attribute $domNode units_system_definition] == 1 } {
        lassign [give_active_units_system] - value
    } elseif { [get_domnode_attribute $domNode unit_definition] ne "" } {
        set magnitude [get_domnode_attribute $domNode unit_definition]
        set value [give_active_unit $magnitude]
    } elseif { [get_domnode_attribute $domNode unit_mesh_definition] == 1 } {
        set value [give_mesh_unit]
    } else {
        set value [get_domnode_attribute $domNode v]
    }

    if { [get_domnode_attribute $domNode values] ne "" && [$domNode @editable 0] == 0 } {
        set values [split [get_domnode_attribute $domNode values] ,]
        if { [$domNode @check_listbox 0] == 0 } {
            if { [lsearch -exact $values $value] == -1 } {
                set value [lindex $values 0]
                gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNode] [list v $value]
            }
        }
    }
    set unit_magnitude [get_domnode_attribute $domNode unit_magnitude]
    if { $unit_magnitude ne "" && $value ne "" } {
        set unit [$domNode @units ""]
        if { ![is_unit_correct $unit_magnitude $unit] } {
            set unit [give_active_unit $unit_magnitude]
            #             set old_unit $unit
            #             set unit [lindex $units 0]
            #             set value [gid_groups_conds::convert_unit_value $unit_magnitude $value $old_unit $unit]
            gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNode] [list v $value units $unit]
        }
    }
    array set dict [split [get_domnode_attribute $domNode dict] ,]
    if { [info exists dict($value)] } {
        set value [= $dict($value)]
    }
    if { $unit_magnitude ne "" && $value ne "" } {
        set unitN [gid_groups_conds::units_to_nice_units $unit]
        return "$value $unitN"
    } else {
        return $value
    }
}

################################################################################
#    modify_value_node (can be used in the problemtype TCL files)
################################################################################

proc gid_groups_conds::modify_value_nodeXP { args } {
    variable doc

    set optional {
        { -use_dict boolean 1 }
        { -units_type units|nice_units nice_units }
        { -noundo "" 0 }
    }
    set compulsory "xpathNode newvalue"
    parse_args $optional $compulsory $args

    set root [$doc documentElement]
    set domNode [$root selectNodes $xpathNode]

    return [modify_value_node -use_dict $use_dict -units_type $units_type \
        -noundo=$noundo $domNode $newvalue]
}

proc gid_groups_conds::modify_value_node { args } {

    set optional {
        { -use_dict boolean 1 }
        { -units_type units|nice_units nice_units }
        { -noundo "" 0 }
    }
    set compulsory "domNode newvalue"
    parse_args $optional $compulsory $args

    set unit_magnitude [get_domnode_attribute $domNode unit_magnitude]
    set unit_definition [get_domnode_attribute $domNode unit_definition]
    set units_system_definition [get_domnode_attribute $domNode units_system_definition]
    set unit_mesh_definition [get_domnode_attribute $domNode unit_mesh_definition]
    set function [get_domnode_attribute $domNode function]

    if { $unit_definition ne "" || $unit_mesh_definition == 1 } {
        if { $units_type eq "nice_units" } {
            set unit [gid_groups_conds::nice_units_to_units $newvalue]
        } else {
            set unit $newvalue
        }
    } elseif { $unit_magnitude ne "" && $function ne ""} {
        lassign $newvalue newvalue unit func
        if { $units_type eq "nice_units" } {
            set unit [nice_units_to_units $unit]
        }
    } elseif { $unit_magnitude ne "" } {
        lassign $newvalue newvalue unit
        if { $units_type eq "nice_units" } {
            set unit [nice_units_to_units $unit]
        }
    }
    if { $use_dict } {
        array unset dict_inv
        foreach "n v" [split [get_domnode_attribute $domNode dict] ,] {
            set dict_inv([= $v]) $n
        }
        if { $units_system_definition == 1 } {
            # the value for units_system_definition has already been converted to original
            # by caller
        }
        if { [info exists dict_inv($newvalue)] } {
            set newvalue $dict_inv($newvalue)
        }
    }
    if { $unit_definition ne "" } {
        if { $noundo } { error "error in gid_groups_conds::modify_value_node 1" }
        gid_groups_conds::set_active_unit $unit_definition $unit
    } elseif { $unit_mesh_definition ne "" } {
        if { $noundo } { error "error in gid_groups_conds::modify_value_node 2" }
        gid_groups_conds::set_mesh_unit $unit
    } elseif { $units_system_definition == 1 } {
        if { $noundo } { error "error in gid_groups_conds::modify_value_node 3" }
        gid_groups_conds::set_active_units_system $newvalue
    } else {
        if { $noundo == 0 } {
            setAttributes [gid_groups_conds::nice_xpath $domNode] [list v $newvalue]
        } else {
            $domNode setAttribute v $newvalue
        }
        if { $unit_magnitude ne "" } {
            if { $noundo == 0 }  {
                setAttributes [gid_groups_conds::nice_xpath $domNode] [list units $unit]
            } else {
                $domNode setAttribute units $unit
            }
        }
        if { $function ne "" } {
            if {[$domNode hasChildNodes]} {
                set f [$domNode selectNodes function]
                if {[$f nodeName] == "function"} {
                    $f delete
                }
            }
            $domNode appendChildTag function
            set f [$domNode selectNodes function]
            set fvNode [$func selectNodes .//functionVariable]
            set n_fvNode [$fvNode @n]
            set pn_fvNode [$fvNode @pn]
            set var_fvNode [$fvNode @variable]
            $f appendChildTag functionVariable [list attributes() \
                n $n_fvNode pn $pn_fvNode variable $var_fvNode]
            set fv [$f selectNodes functionVariable]
            foreach v_point [$func selectNodes .//value] {
                set pnt [$v_point @v]
                $fv appendChildTag value [list attributes() \
                    n point pn [_ "Point"] v $pnt]
            }
        }
    }
    set actualize_nodesList ""
    if { [$domNode @actualize_tree 0] } {
        lappend actualize_nodesList 0
    }
    set hc [check_node_dependencies $domNode]
    lappend actualize_nodesList {*}$hc

    set bdNode [lindex [$domNode selectNodes ancestor::blockdata] 0]
    if { $bdNode ne "" && [$bdNode @state normal] eq "disabled_edit" } {
        if { $noundo == 0 }  {
            setAttributes [gid_groups_conds::nice_xpath $bdNode] [list state normal]
        } else {
            $bdNode setAttribute state normal
        }
        lappend actualize_nodesList $bdNode
    }
    if { [$domNode @state normal] eq "disabled_on_void" && [$domNode @v ""] eq "" } {
        if { $noundo == 0 }  {
            setAttributes [gid_groups_conds::nice_xpath $domNode] [list state disabled]
        } else {
            $domNode setAttribute state disabled
        }
        lappend actualize_nodesList $domNode
    }
    return [lsort -unique $actualize_nodesList]
}

################################################################################
#    Edit tree parts window
################################################################################

proc gid_groups_conds::edit_tree_parts_window { args } {
    variable doc

    set optional {
        { -accepted_n name "" }
        { -select_only_one boolean 0 }
        { -apply_cond_group groupName "" }
    }
    set compulsory "wp title xpath xpath_sel"
    parse_args $optional $compulsory $args

    set w $wp.edit_tree_parts_window
    destroy $w
    if { $apply_cond_group eq "" } {
        set w [dialogwin_snit $w -title $title -okname [_ "Select"] -cancelname \
            [_ "Close"] -morebuttons [list [_ "Create new"] [_ "Edit"]] -transient 1]
    } else {
        set w [dialogwin_snit $w -title $title -transient 1]
    }
    set f [$w giveframe]
    set domNodes [$doc selectNodes $xpath]
    boundary_conds $f.bc -domnodes $domNodes -edit_window_position up -height 300 -title $title -title_active 0
    if { $xpath_sel ne "" } {
        set dom_nodes [$doc selectNodes $xpath_sel]
        foreach dom_node $dom_nodes {
            set id [$f.bc select_domNode $dom_node]
            $f.bc item collapse $id
        }
    }
    grid $f.bc -sticky nsew
    grid columnconfigure $f 0 -weight 1
    grid rowconfigure $f 0 -weight 1

    tk::TabToWindow $f.bc

    if { $apply_cond_group ne "" } {
        $f.bc configure -edit_window_position full
        set what apply
        set domNode [lindex $domNodes 0]
        $f.bc notify bind DontDelete <Selection> ""
        set cndid [$f.bc select_domNode $domNode]
        set groupNode [$domNode selectNodes [format_xpath {group[@n=%s]} $apply_cond_group]]
        if { $groupNode ne "" } {
            set what edit
            set domNode $groupNode
        }
        update
        # very ugly, but if updates are not here, this does not work. To review in the future
        update
        set id [$f.bc select_domNode $domNode]
        $f.bc apply_cond -what $what -show_buttons 0
        $f.bc configure -height [winfo reqheight $f.bc.data]
        if { $groupNode eq "" } {
            $f.bc _apply_cond_set_group "" $apply_cond_group
        }
    }

    bind $w <Return> [list $w invokeok]
    set action [$w createwindow]

    while 1 {
        set domNodes ""
        if { $action == 1 } {
            if { $apply_cond_group ne "" } {
                set existing_ov [list ""]
                set err [catch { $f.bc apply_cond_ok_do -actualize_tree 0 $existing_ov $cndid $id $what } errstring]
                if { $err } {
                    if { $errstring ne "" } {
                        GID_tk_messageBox -message $errstring -parent $w
                    }
                } else {
                    set domNode [$f.bc item element cget $id 0 e_text -data]
                    if { $domNode ne "" } {
                        lappend domNodes $domNode
                    }
                    #actualize_conditions_window -do_focus 0
                    break
                }
            } else {
                foreach id [$f.bc selection get] {
                    set domNode [$f.bc item element cget $id 0 e_text -data]
                    if { $accepted_n ne "" && [$domNode @n] ne $accepted_n } {
                        set xp [format_xpath {ancestor::*[@n=%s]} $accepted_n]
                        set domNode [$domNode selectNodes $xp]
                    }
                    if { $domNode ne "" } {
                        lappend domNodes $domNode
                    }
                }
                if { [llength $domNodes] == 0 } {
                    snit_messageBox -message [_ "It is necessary to select one item"] -parent $w
                } elseif { $select_only_one && [llength $domNodes] > 1 } {
                    snit_messageBox -message [_ "It is necessary to select only one item"] -parent $w
                } else {
                    #actualize_conditions_window -do_focus 0
                    break
                }
            }
        } elseif { $action == 2 } {
            $f.bc copy_block_data
            set idx 0
            while { [winfo exists $wp.f$idx] } {
                set id [$f.bc selection get]
                set selNode [$f.bc item element cget $id 0 e_text -data]
                set icon [$selNode @icon ""]
                set err [catch {gid_groups_conds::giveimageL $icon]
                } image]
                if { $err || $image == "" } {
                set image [gid_groups_conds::giveimage contents2-16]
            }
            set name [$selNode @name ""]
            if {![winfo exists $wp.data.f0.e$idx]} { incr idx; continue }
            if {[winfo class $wp.data.f0.e$idx] eq "TCombobox"} {
                $wp.data.f0.e$idx tree_insert -image $image \
                    -active 1 end $name $name root
            }
            incr idx
        }
    } elseif { $action == 3} {
        foreach id [$f.bc selection get] {
            set domNode [$f.bc item element cget $id 0 e_text -data]
            if { $accepted_n ne "" && [$domNode @n] ne $accepted_n } {
                set xp [format_xpath {ancestor::*[@n=%s]} $accepted_n]
                set domNode [$domNode selectNodes $xp]
            }
            if { $domNode ne "" } {
                lappend domNodes $domNode
            }
        }
        if { [llength $domNodes] == 0 } {
            snit_messageBox -message [_ "It is necessary to select one item"] -parent $w
        } elseif { $select_only_one && [llength $domNodes] > 1 } {
            snit_messageBox -message [_ "It is necessary to select only one item"] -parent $w
        } elseif { [llength $domNodes] == 1 } {
            if { [$domNodes nodeName] eq "container"} {
                snit_messageBox -message \
                    [_ "It is necessary to select an item that can be edited instead of an item of type container"] -parent $w
            } elseif { [$domNodes nodeName] eq "blockdata"} {
                set exp {edit_command="@edit_type='exclusive' or @edit_type='exclusive_blockdata'"}
                if { [$domNodes selectNodes $exp] ne ""} {
                    $w.f.bc _edit_command_eval -item $id $domNodes
                } else {
                    $w.f.bc apply_blockdata
                }
            } else {
                $w.f.bc apply_blockdata
            }
        } else {
            #actualize_conditions_window -do_focus 0
            $w.f.bc apply_blockdata
        }
    } elseif { $apply_cond_group ne "" } {
        $f.bc apply_cond_cancel
        break
    } else {
        break
    }

    set action [$w waitforwindow]
}
destroy $w
return $domNodes
}

proc gid_groups_conds::edit_tree_parts_frame { args } {
    variable doc

    set optional {
        { -accepted_n name "" }
        { -select_only_one boolean 0 }
        { -apply_cond_group groupName "" }
        { -donotshow 0 }
    }
    set compulsory "wp frame title xpath xpath_sel"
    parse_args $optional $compulsory $args

    #     set f [ttk::frame $frame.f]
    set domNodes [$doc selectNodes $xpath]

    boundary_conds $frame.bc -domnodes $domNodes \
        -edit_window_position up -height 300 -title $title \
        -title_active 0

    if { $xpath_sel ne "" } {
        set id [$frame.bc select_domNode [$doc selectNodes $xpath_sel]]
        $frame.bc item collapse $id
    }
    if {!$donotshow} {
        grid $frame.bc -sticky nsew -padx 2 -pady 2 -row 0 -column 0 -columnspan 2
        grid columnconfigure $frame.bc 1 -weight 1
        grid rowconfigure $frame.bc 1 -weight 1
    }
    tk::TabToWindow $frame.bc

    if { $apply_cond_group ne "" } {
        $frame.bc configure -edit_window_position full
        set what apply
        set domNode [lindex $domNodes 0]
        $frame.bc notify bind DontDelete <Selection> ""
        set cndid [$frame.bc select_domNode $domNode]
        set groupNode [$domNode selectNodes [format_xpath {group[@n=%s]} $apply_cond_group]]
        if { $groupNode ne "" } {
            set what edit
            set domNode $groupNode
        }
        update
        # very ugly, but if updates are not here, this does not work. To review in the future
        update
        set id [$frame.bc select_domNode $domNode]
        $frame.bc apply_cond -what $what -show_buttons 0
        $frame.bc configure -height [winfo reqheight $frame.bc.data]
        if { $groupNode eq "" } {
            $frame.bc _apply_cond_set_group "" $apply_cond_group
        }
    }
    while 1 {
        set domNodes ""

        if { $apply_cond_group ne "" } {
            set existing_ov [list ""]
            set err [catch { $frame.bc apply_cond_ok_do -actualize_tree 0 $existing_ov $cndid $id $what } errstring]
            if { $err } {
                if { $errstring ne "" } {
                    GID_tk_messageBox -message $errstring -parent $wp
                }
            } else {
                set domNode [$frame.bc item element cget $id 0 e_text -data]
                if { $domNode ne "" } {
                    lappend domNodes $domNode
                }
                break
            }
        } else {
            foreach id [$frame.bc selection get] {
                set domNode [$frame.bc item element cget $id 0 e_text -data]
                if { $accepted_n ne "" && [$domNode @n] ne $accepted_n } {
                    set xp [format_xpath {ancestor::*[@n=%s]} $accepted_n]
                    set domNode [$domNode selectNodes $xp]
                }
                if { $domNode ne "" } {
                    lappend domNodes $domNode
                }
            }
            if { [llength $domNodes] == 0 } {
                snit_messageBox -message [_ "It is necessary to select one item"] -parent $wp
            } elseif { $select_only_one && [llength $domNodes] > 1 } {
                snit_messageBox -message [_ "It is necessary to select only one item"] -parent $wp
            } else {
                break
            }
        }
    }
    if {!$donotshow} {
        return $frame
    } else {
        return $domNodes
    }
}

################################################################################
#    Apply condition
################################################################################

proc gid_groups_conds::clear_condition { xpathCnd } {
    variable doc

    set root [$doc documentElement]
    set domNodeCnd [$root selectNodes $xpathCnd]
    foreach g [$domNodeCnd selectNodes group|groupList] {
        gid_groups_conds::delete [gid_groups_conds::nice_xpath $g]
    }
}

proc gid_groups_conds::clear_condition_group { xpathCnd group } {
    variable doc
    set root [$doc documentElement]
    set domNodeCnd [$root selectNodes $xpathCnd]
    set xp [format_xpath {group[@n=%s]} $group]
    set gNode [$domNodeCnd selectNodes $xp]
    gid_groups_conds::delete [gid_groups_conds::nice_xpath $gNode]
}

proc gid_groups_conds::has_condition_group { xpathCnd group } {
    variable doc
    set root [$doc documentElement]
    set domNodeCnd [$root selectNodes $xpathCnd]
    set xp [format_xpath {group[@n=%s]} $group]
    set gNode [$domNodeCnd selectNodes $xp]
    if { $gNode ne "" } {
        return 1
    } else {
        return 0
    }
}

proc gid_groups_conds::apply_condition { args } {
    variable doc

    set optional {
        { -ov point|line|surface|volume "" }
        { -noactualizegroups "" 0 }
    }
    set compulsory "xpathCnd group_name values_dict"
    parse_args $optional $compulsory $args

    set actualize_nodesList ""
    if { $ov ne "" } {
        set att [list n $group_name ov $ov]
    } else {
        set att [list n $group_name]
    }
    set gn [gid_groups_conds::add $xpathCnd group $att]

    set root [$doc documentElement]
    set domNodeCnd [$root selectNodes $xpathCnd]
    set newtype [gid_groups_conds::give_group_name_auto_from_bc $domNodeCnd]
    if { $newtype ne "" } {
        gid_groups_conds::change_group_att -noactualizegroups=$noactualizegroups $group_name type $newtype
    }
    set domNodeBD [$domNodeCnd selectNodes {ancestor::blockdata[1]}]

    if { $domNodeBD ne "" && [$domNodeBD @state normal] eq "disabled" && [$domNodeBD @sequence_type any] eq "non_void_disabled" } {
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNodeBD] [list state normal]
        lappend actualize_nodesList $domNodeBD
    } elseif { $domNodeBD ne "" && [$domNodeBD @active 1] == 0 && [$domNodeBD @sequence_type any] eq "non_void_deactivated" } {
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNodeBD] [list active 1]
        lappend actualize_nodesList $domNodeBD
    }
    set ret1 [_copy_cond_values $domNodeCnd $gn $ov $values_dict normal]
    set ret2 [check_condition_dependencies1 $domNodeCnd]
    lappend actualize_nodesList {*}$ret1 {*}$ret2

    return [lsort -unique $actualize_nodesList]
}

proc gid_groups_conds::_copy_cond_values { domNode_from domNode_to ov values_dict parent_state { level 0 } } {
    set actualize_nodesList ""
    set valueNodes [$domNode_from selectNodes value|container]
    if { ![llength $valueNodes] && [$domNode_from @isvalue 0] } {
        lappend valueNodes $domNode_from
    }
    foreach valueNode $valueNodes {
        set state [get_domnode_attribute $valueNode state normal]
        if { $parent_state eq "hidden" } { set state hidden }
        switch [$valueNode nodeName] {
            value - condition {
                set n [$valueNode @n]
                set unit_magnitude [get_domnode_attribute $valueNode unit_magnitude]

                set functionNode ""
                set attDict ""
                foreach i [$valueNode selectNodes @*] {
                    lappend attDict {*}$i
                }
                if { [dict exists $values_dict $n] } {
                    dict set attDict v [dict get $values_dict $n v]
                    if { $unit_magnitude ne "" } {
                        dict set attDict units [dict get $values_dict $n units]
                    }
                    if { [dict exists $values_dict $n function] } {
                        set docFunc [gid_groups_conds::parse_XML_data [dict get $values_dict $n function]]
                        $docFunc documentElement functionNode
                    }
                } else {
                    if { $unit_magnitude ne "" && ![dict exists $attDict units] } {
                        lassign [gid_groups_conds::give_value_in_active_unit \
                            -ov $ov $valueNode] value unit
                        dict set attDict v $value
                        dict set attDict units $unit
                    }
                    set functionNode [$valueNode selectNodes function]
                }
                if { [llength [split $unit_magnitude ,]] > 1 } {
                    dict set attDict unit_magnitude [dict get [split $unit_magnitude ,] $ov]
                }
                foreach i [list ov isvalue] {
                    dict unset attDict $i
                }
                set v [gid_groups_conds::add [gid_groups_conds::nice_xpath $domNode_to] \
                    value $attDict]

                if { $functionNode ne "" } {
                    gid_groups_conds::copyNode_recursive $functionNode $v
                }
            }
            container {
                set attDict ""
                foreach i [$valueNode selectNodes @*] { lappend attDict {*}$i }
                set c [gid_groups_conds::add [gid_groups_conds::nice_xpath $domNode_to] container \
                    $attDict]

                _copy_cond_values $valueNode $c $ov $values_dict $state \
                    [expr {$level+1}]
            }
        }
    }
    if { $level == 0 } {
        foreach valueNode [$domNode_to selectNodes ".//value"] {
            set hc [gid_groups_conds::check_node_dependencies $valueNode]
            lappend actualize_nodesList {*}$hc
        }
    }
    return [lsort -unique $actualize_nodesList]
}

proc gid_groups_conds::copy_block_data { args } {

    set optional {
        { -copy_cond_groups "" 0 }
        { -newname name "" }
    }
    set compulsory "domNode"
    parse_args $optional $compulsory $args

    if { ![$domNode @sequence 0] } {
        error [_ "Blockdata item cannot be copied"]
    }
    if { [$domNode @state normal] eq "disabled" && [$domNode @sequence_type any] eq "non_void_disabled" } {
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNode] [list state normal]
    } elseif { [$domNode @active 1] == 0 && [$domNode @sequence_type any] eq "non_void_deactivated" } {
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $domNode] [list active 1]
    }
    set newNode [gid_groups_conds::copyNode [gid_groups_conds::nice_xpath $domNode] [gid_groups_conds::nice_xpath [$domNode parentNode]]]
    if { !$copy_cond_groups } {
        set xp {condition/group|condition/groupList|}
        append xp {container/condition/group|} \
            {container/condition/groupList}
        foreach d [$newNode selectNodes $xp] {
            gid_groups_conds::delete [gid_groups_conds::nice_xpath $d]
        }
    }

    gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $newNode] [list state normal tree_state close]

    if { $newname ne "" } {
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $newNode] [list name $newname]
    } elseif { [$domNode @editable_name ""] eq "unique" } {
        set values ""
        foreach node [$domNode selectNodes "../blockdata"] {
            if { $node eq $newNode } { continue }
            lappend values [get_domnode_attribute $node name]
        }
        set basename [get_domnode_attribute $newNode name]
        set newname $basename
        if { ![regexp {(.*[^\d])(\d+)$} $newname {} basename idx] } {
            set idx ""
        }
        while { [lsearch -exact $values $newname] != -1 } {
            if { $idx eq "" } {
                set idx 2
            } else {
                incr idx
            }
            set newname $basename$idx
        }
        if { $newname ne $basename } {
            gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $newNode] [list name $newname]
        }
    }
    if { [set proc [$domNode @update_proc ""]] ne "" } {
        gid_groups_conds::eval_proc $proc $newNode
    }
}

proc gid_groups_conds::set_blockdataList { args } {
    variable doc

    set optional {
        { -ask_confirmation "" 0 }
        { -parent_win w "" }
    }
    set compulsory "xpBlockdata loadcases"
    parse_args $optional $compulsory $args

    set bdNodes [$doc selectNodes $xpBlockdata]

    if { [llength $bdNodes] < [llength $loadcases] } {
        if { $ask_confirmation } {
            set t [_ "Do you want to create/update loadcases (Create %d loadcases)?" \
                [expr {[llength $loadcases]-[llength $bdNodes]}]]
            set ret [snit_messageBox -type okcancel -default ok \
                -message $t -parent $parent_win]
            if { $ret == "cancel" } { error "" }
        }
        set n [lindex $bdNodes 0]
        foreach l [lrange $loadcases [llength $bdNodes] end] {
            gid_groups_conds::copy_block_data -copy_cond_groups -newname $l $n
        }
    } elseif { [llength $bdNodes] > [llength $loadcases] } {
        if { $ask_confirmation } {
            set t [_ "Do you want to update loadcases (Delete %d loadcases)?" \
                [expr {[llength $bdNodes]-[llength $loadcases]}]]
            set ret [snit_messageBox -type okcancel -default ok \
                -message $t -parent $parent_win]
            if { $ret == "cancel" } { error "" }
        }
        foreach n [lrange $bdNodes [llength $loadcases] end] {
            gid_groups_conds::delete [gid_groups_conds::nice_xpath $n]
        }
    }

    set bdNodes [$doc selectNodes $xpBlockdata]
    foreach n $bdNodes l $loadcases {
        if { $n eq "" } { break }
        gid_groups_conds::setAttributes [gid_groups_conds::nice_xpath $n] \
            [list name $l]
        foreach cndNode [$n selectNodes {.//condition}] {
            clear_condition [gid_groups_conds::nice_xpath $cndNode]
        }
    }
}

################################################################################
#    Mesh criteria
################################################################################

proc gid_groups_conds::change_default_meshing_criteria_all { what } {
    variable doc
    variable change_meshing_data_stack

    if { ![info exists change_meshing_data_stack] } { set change_meshing_data_stack "" }

    set err [catch { write_calc_data version } version]
    if { $err } { set version 1.0 }

    switch $what {
        init {
            if { "change_default_meshing_criteria_all" in $change_meshing_data_stack } {
                error "error in gid_groups_conds::change_default_meshing_criteria_all. Calling twice without clearing"
            }
            lappend change_meshing_data_stack change_default_meshing_criteria_all

            set groupsDict ""
            if { $version >= 1.4 } {
                set xp {//condition[not(ancestor::blockdata[@n="Internal data"]) and @ovm="none"]/group}
                append  xp {|//condition[not(ancestor::blockdata[@n="Internal data"]) and @ovm1="none"]/groupList/group[1]}
                append  xp {|//condition[not(ancestor::blockdata[@n="Internal data"]) and @ovm2="none"]/groupList/group[2]}

                foreach groupNode [$doc selectNodes $xp] {
                    set ov [$groupNode @ov ""]
                    if { $ov eq "" } {
                        set condNode [$groupNode selectNodes {ancestor::condition[1]}]
                        if { [$groupNode parentNode] == $condNode } {
                            set ov [$condNode @ov]
                        } elseif { [[$groupNode parentNode] nodeName] eq "groupList" } {
                            if { $groupNode == [lindex [[$groupNode parentNode] childNodes] 0] } {
                                set ov [$condNode @ov1]
                            } else {
                                set ov [$condNode @ov2]
                            }
                        }
                    }
                    dict set groupsDict [$groupNode @n] ${ov}_off
                }
            }
            set xp {//condition[not(ancestor::blockdata[@n="Internal data"]) and @ovm="element"]/group}
            append  xp {|//condition[not(ancestor::blockdata[@n="Internal data"]) and @ovm1="element"]/groupList/group[1]}
            append  xp {|//condition[not(ancestor::blockdata[@n="Internal data"]) and @ovm2="element"]/groupList/group[2]}

            foreach groupNode [$doc selectNodes $xp] {
                set ov [$groupNode @ov ""]
                if { $ov eq "" } {
                    set condNode [$groupNode selectNodes {ancestor::condition[1]}]
                    if { [$groupNode parentNode] == $condNode } {
                        set ov [$condNode @ov]
                    } elseif { [[$groupNode parentNode] nodeName] eq "groupList" } {
                        if { $groupNode == [lindex [[$groupNode parentNode] childNodes] 0] } {
                            set ov [$condNode @ov1]
                        } else {
                            set ov [$condNode @ov2]
                        }
                    }
                }
                dict set groupsDict [$groupNode @n] $ov
            }
            change_default_meshing_criteria init $groupsDict
        }
        clear {
            if { [llength $change_meshing_data_stack] == 0 } {
                error "error in gid_groups_conds::change_default_meshing_criteria_all. Calling clear without previous init"
            }
            if { [lindex $change_meshing_data_stack end] ne "change_default_meshing_criteria_all" } {
                error "error in gid_groups_conds::change_default_meshing_criteria_all. Calling clear and last meshing call was to [lindex $change_meshing_data_stack end]"
            }
            set change_meshing_data_stack [lrange $change_meshing_data_stack 0 end-1]
            change_default_meshing_criteria clear
        }
    }
}

proc gid_groups_conds::change_structured_divisions_lines { what args } {
    variable change_meshing_data_stack

    if { ![info exists change_meshing_data_stack] } { set change_meshing_data_stack "" }

    switch $what {
        init {
            if { "change_structured_divisions_lines" in $change_meshing_data_stack } {
                error "error in gid_groups_conds::change_structured_divisions_lines. Calling twice without clearing"
            }
            lappend change_meshing_data_stack change_structured_divisions_lines

            if { [llength $args] != 1 } {
                error "error in gid_groups_conds::change_structured_divisions_lines. Args: init groupsDict"
            }
            change_line_structured_divisions init [lindex $args 0]
        }
        clear {
            if { [llength $change_meshing_data_stack] == 0 } {
                error "error in gid_groups_conds::change_structured_divisions_lines. Calling clear without previous init"
            }
            if { [lindex $change_meshing_data_stack end] ne "change_structured_divisions_lines" } {
                error "error in gid_groups_conds::change_structured_divisions_lines. Calling clear and last meshing call was to [lindex $change_meshing_data_stack end]"
            }
            set change_meshing_data_stack [lrange $change_meshing_data_stack 0 end-1]
            if { [llength $args] != 0 } {
                error "error in gid_groups_conds::change_structured_divisions_lines. Args: clear"
            }
            change_line_structured_divisions clear
        }
    }
}

################################################################################
#    Generic select entities
################################################################################

proc gid_groups_conds::_keep_frame_on_top { t f } {
    if { ![winfo exists $f] } {
        bind $t <Configure> ""
        return
    }
    place $f -in $t -x 0 -y 0 -anchor nw -width \
        [winfo width $t] -height [winfo height $t]
    bind $t <Configure> [namespace code [list _keep_frame_on_top $t $f]]
}

proc gid_groups_conds::_select_apply_entities_label { w } {
    $w set_uservar_value label_apply $::WarnLineVar
}

#_local_axes_gid_apply
proc gid_groups_conds::select_apply_entities { args } {

    set optional {
        { -window_style small|same small }
        { -end_cmd script "" }
    }
    set compulsory "w cmd txt"
    parse_args $optional $compulsory $args

    set geometry [winfo geometry $w]
    set focus [focus -lastfor $w]
    set grabList [grab status $w]
    lappend grabList [grab current $w]
    grab release $w

    destroy $w.fapply
    set f [ttk::frame $w.fapply]
    label $f.l1 -wrap 200 -justify left -textvariable \
        [$w give_uservar label_apply] -height 3

    set b [ttk::button $f.b -text [_ "Finish"] -width 10 -command [namespace code \
        [list _select_apply_entities_end $w $f $f.b $end_cmd $geometry $focus \
        $grabList]]]
    grid $f.l1 -sticky nw -padx 3 -pady 3
    grid $b -padx 3 -pady "2 3" -sticky nw
    grid columnconfigure $f 1 -weight 1
    grid rowconfigure $f 2 -weight 1
    _keep_frame_on_top $w $f

    update idletasks

    if { $window_style eq "small" } {
        regexp {(\d+)x(\d+)\+([+-]?\d+)\+([+-]?\d+)} $geometry {} width - x y
        set height [winfo reqheight $f]
        wm geometry $w ${width}x$height+$x+$y
    }

    if { $txt ne "" } {
        $w set_uservar_value label_apply $txt
    } else {
        set cmd0 [namespace code [list _select_apply_entities_label $w]]
        trace add variable ::WarnLineVar write "$cmd0 ;#"
        bind $f <Destroy> [list trace remove variable ::WarnLineVar write "$cmd0 ;#"]
    }
    bind $b <Escape> "[list $b invoke]; break"
    bind $b <Return> [list $b invoke]
    raise $w
    focus .gid
    eval $cmd
    delayedop changefunc [list $b invoke]
}

proc gid_groups_conds::_select_apply_entities_end { w f b end_cmd geometry focus grabList } {

    set found [delayedop -cancel changefunc [list $b invoke]]
    if { $found } { GiD_Process escape escape }
    destroy $f
    wm geometry $w $geometry
    focus $focus
    switch [lindex $grabList 0] {
        local { grab [lindex $grabList 1] }
        global { grab -global [lindex $grabList 1] }
    }
    uplevel #0 $end_cmd
    $w set_uservar_value update 1 ;# to update the GUI
}

################################################################################
#    GiD Control-1 menu
################################################################################


proc gid_groups_conds::add_items_to_data_menu { menu whatuse type entity } {

    GidUtils::DisableGraphics

    switch $type {
        Points { set cnds point_groups }
        Lines { set cnds line_groups }
        Surfaces { set cnds surface_groups }
        Volumes { set cnds volume_groups }
        Nodes { set cnds point_groups }
        Elements { set cnds "line_groups surface_groups volume_groups" }
    }

    switch $type {
        Points - Lines - Surfaces - Volumes { set where geometry }
        Nodes - Elements { set where mesh }
    }

    set doc $gid_groups_conds::doc

    set submenu ""
    set needsseparator 0
    foreach cnd $cnds {
        set ret [GiD_Info Conditions $cnd $where $entity]
        set list ""
        foreach i $ret {
            #             if { $i in $list } { continue }
            foreach "num face - group" $i break
            if { $num == 0 } { continue }
            if { $num eq "E" } { set num $face }
            if { $submenu == "" } {
                set j 1
                while { [winfo exists $menu.m$j] } { incr j }
                set submenu [menu $menu.m$j -tearoff 0]
                set text [_ "Groups"]
                $menu add cascade -label $text -menu $submenu
                set needsseparator 1
            }
            set xp [format_xpath {//group[@n=%s and ancestor::condition]} $group]
            set domNodes [$doc selectNodes $xp]
            set txt [_ "Group: %s" $group]
            if { ![llength $domNodes] } {
                ContextualEntity::AddToMenu $submenu $txt \
                    [list gid_groups_conds::_add_items_to_data_menu_item $group]
            } else {
                set j 1
                while { [winfo exists $submenu.m$j] } { incr j }
                set subsubmenu [menu $submenu.m$j -tearoff 0]
                $submenu add cascade -label $txt -menu $subsubmenu
                ContextualEntity::AddToMenu $subsubmenu $txt \
                    [list gid_groups_conds::_add_items_to_data_menu_item $group]
                foreach domNode $domNodes {
                    set cndNode [$domNode selectNodes ancestor::condition]
                    set blockNode [$domNode selectNodes {ancestor::blockdata[@sequence='1']}]
                    if { $blockNode ne "" } {
                        set txt [concat [_ "Condition"]: [$blockNode @name] - [$cndNode @pn]]]
                    } else {
                        set txt [concat [_ "Condition"]: [$cndNode @pn]]
                    }
                    ContextualEntity::AddToMenu $subsubmenu $txt \
                        [list gid_groups_conds::_add_items_to_data_menu_itemD $group \
                        $domNode]
                }
            }
            lappend list $i
        }
    }
    if { $needsseparator } { $menu add separator }

    GidUtils::EnableGraphics
}

proc gid_groups_conds::_add_items_to_data_menu_item { group } {
    gid_groups_conds::open_groups .gid window_force window
    set wg $gid_groups_conds::gid_group_var
    set tree [$wg givetreectrl]

    $tree selection clear
    foreach item [$tree item children root] {
        if { [$tree item text $item 0] eq $group } {
            $tree selection add $item
            break
        }
    }
}

proc gid_groups_conds::_add_items_to_data_menu_itemD { group domNode } {

    set what [gid_groups_conds::open_conditions window_type]
    if { $what eq "none" } { set what window }

    set xpath [gid_groups_conds::nice_xpath $domNode]
    gid_groups_conds::open_conditions $what -select_xpath $xpath
}

################################################################################
#    create_update_group_with_external_surfaces
################################################################################

proc gid_groups_conds::create_update_group_with_external_surfaces { groupname } {

    set err [catch { add_new_group -type auto -exactname $groupname }]
    if { $err } {
        delete_entities_from_groups 1: [list [list "" $groupname]] surface
    }
    set entities [GiD_Geometry list -higherentity 1 surface]
    add_entities_to_groups $entities [list [list "" $groupname]] surface
}

################################################################################
#    Compressed subtree
################################################################################

proc gid_groups_conds::prepare_for_compress_subtree_xp { xpath } {
    variable doc

    foreach domNode [$doc selectNodes $xpath] {
        prepare_for_compress_subtree $domNode
    }
}

proc gid_groups_conds::prepare_for_compress_subtree { domNode } {
    if { [$domNode hasAttribute di_ic] } { return }
    if { [$domNode @sequence 0] != 1 } { return }

    set root [$domNode selectNodes /*]
    $domNode setAttribute di_ic 0
}

# proc gid_groups_conds::_compress_subtree_remove_tree_state { domNode } {
#     if { [$domNode hasAttribute tree_state] } {
#         $domNode removeAttribute tree_state
#     }
#     foreach n [$domNode childNodes] { _compress_subtree_remove_tree_state $n }
# }

proc gid_groups_conds::compress_subtree { domNode } {

    # di_ic= >=1: is compressed ; 0: is not compressed but could be

    # di_id identifier (>= 1)
    # di_refcount: number of compressed nodes that compare to this
    # di_refcount_total: number of historical compressed nodes that compare to this

    if { ![$domNode hasAttribute di_ic] } { return 0 }
    if { [$domNode @di_ic] != 0 } { return 1 }

    set nodeList [internal_data get_list compress_subtrees]

    lassign [list "" 0 0 0] cmpNode num id_max refcount_total_max
    foreach n [$nodeList childNodes] {
        set refcount_total [$n @di_refcount_total]
        if { $cmpNode eq "" || $refcount_total < [$cmpNode @di_refcount_total] } {
            set cmpNode $n
        }
        if { $refcount_total > $refcount_total_max } {
            set refcount_total_max $refcount_total
        }
        if { [$n @di_id] > $id_max } { set id_max [$n @di_id] }
        incr num
    }
    if { $cmpNode eq "" || ( $num == 1 && $refcount_total_max > 10 ) } {
        set cmpNode [$nodeList appendChildTag blockdata [list attributes() \
            di_id [incr id_max] di_refcount 0 di_refcount_total 0]]
        foreach n [$domNode childNodes] {
            $cmpNode appendChild [$n cloneNode -deep]
        }
    }
    lassign "" cmp_data data
    foreach n [$cmpNode childNodes] { append cmp_data [$n asXML] }
    foreach n [$domNode childNodes] { append data [$n asXML] }

    set delta [cu::rsync_create_delta -blocklen 180 -base64 1 \
        $cmp_data $data]

    foreach n [$domNode childNodes] { $n delete }

    $domNode appendChild [[$domNode ownerDocument] createTextNode $delta]

    $domNode setAttribute di_ic [$cmpNode @di_id]
    $cmpNode setAttribute di_refcount [expr {[$cmpNode @di_refcount]+1}]
    $cmpNode setAttribute di_refcount_total [expr {[$cmpNode @di_refcount_total]+1}]
    return 1
}

proc gid_groups_conds::uncompress_subtree { args } {

    set optional {
        { -mydoc mydoc ""}
    }
    set compulsory "domNode"
    parse_args $optional $compulsory $args
    if { [$domNode @di_ic 0] == 0 } {
        return 0
    }
    set nodeList [internal_data get_list -mydoc $mydoc compress_subtrees]
    set xp [format_xpath {blockdata[@di_id=%s]} [$domNode @di_ic]]
    set cmpNode [$nodeList selectNodes $xp]
    lassign "" cmp_data
    foreach n [$cmpNode childNodes] {
        append cmp_data [$n asXML]
    }
    set txtNode [$domNode selectNodes text()]
    set xml [cu::rsync_apply_patch $cmp_data [$txtNode nodeValue]]
    $txtNode delete
    $domNode appendXML <n>$xml</n>
    set n [$domNode selectNodes {*[1]}]
    foreach child [$n childNodes] { $domNode appendChild $child }
    $n delete
    $domNode setAttribute di_ic 0
    set refcount [expr {[$cmpNode @di_refcount]-1}]
    if { $refcount == 0 } {
        gid_groups_conds::delete $cmpNode
    } else {
        $cmpNode setAttribute di_refcount $refcount
    }
    return 1
}

################################################################################
#    import export materials
################################################################################

proc gid_groups_conds::import_export_materials { args } {
    variable doc
    variable ListboxCont

    set optional {
        { -boundary_conds_sel name "" }
    }
    set compulsory "wp xpath"
    parse_args $optional $compulsory $args

    set w $wp.iematerials
    destroy $w
    set w [dialogwin_snit $w -title [_ "Import/export materials"] -morebuttons \
        [list [_ "Apply"]] -style separator_right]
    set f [$w giveframe]

    set domNode_orig [$doc selectNodes $xpath]
    $w set_uservar_value domNode_orig $domNode_orig

    $w set_uservar_value haschanges_local 0
    $w set_uservar_value haschanges_remote 0

    set xp {string(/*/@version)}
    set v [$doc selectNodes $xp]
    set edoc [gid_groups_conds::parse_XML_data "<[[$doc documentElement] nodeName] version='$v'/>"]
    set root [$edoc documentElement]
    set sel_dom_nodes [$doc selectNodes $xpath]
    set domNode_new [$sel_dom_nodes cloneNode -deep]
    if { [llength $domNode_new] != 1 } {
        #what must do???
        return
    }
    $root appendChild $domNode_new
    set pNode [$domNode_orig selectNodes /*/procs]
    if { $pNode != "" } {
        $root appendChild [$pNode cloneNode -deep]
    }

    $domNode_new setAttribute state hidden_this
    boundary_conds $f.bc1 -domnodes $domNode_new \
        -edit_window_position up -height 300 -title [_ "Local materials"] \
        -title_active 0 -have_import_export 0

    if { $boundary_conds_sel ne "" } {
        foreach id [$boundary_conds_sel selection get] {
            set domNode [$boundary_conds_sel item element cget $id 0 e_text -data]
            if { [$domNode hasAttribute name] } {
                set xp [format_xpath {.//*[@name=%s]} [$domNode @name]]
            } else {
                set xp [format_xpath {.//*[@n=%s and @pn=%s]} [$domNode @n] [$domNode @pn]]
            }
            set n [$domNode_new selectNodes $xp]
            if { $n ne "" } {
                set id [$f.bc1 select_domNode $n]
                $f.bc1 item collapse $id
            }
        }
    }

    ttk::frame $f.f1

    set theme [gid_groups_conds::GetCurrentTheme]
    if { ![regexp "Tdyn" $theme {} curr_theme] } {
        ttk::button $f.f1.b1 -image $gid_groups_conds::images(forward-22) -command \
            [list gid_groups_conds::_import_export_materials_copy $w local $f.bc1 $f.bc2 $f]
        ttk::button $f.f1.b2 -image $gid_groups_conds::images(back-22) -command \
            [list gid_groups_conds::_import_export_materials_copy $w remote $f.bc1 $f.bc2 $f]
    } else {
        ttk::button $f.f1.b1 -image arrow_right  -command \
            [list gid_groups_conds::_import_export_materials_copy $w local $f.bc1 $f.bc2 $f]
        ttk::button $f.f1.b2 -image arrow_left -command \
            [list gid_groups_conds::_import_export_materials_copy $w remote $f.bc1 $f.bc2 $f]
    }
    grid $f.f1.b1 -padx 5 -pady 5
    grid $f.f1.b2 -padx 5 -pady 5

    ttk::frame $f.f2

    variable material_import_export_actions_dict
    ttk::label $f.f2.l1 -text [_ "Import/export to"]:

    cu::combobox $f.f2.cb -textvariable [$w give_uservar import_export_to] \
        -dict $material_import_export_actions_dict -values [dict keys $material_import_export_actions_dict] -state readonly

    ttk::button $f.f2.b1 -image $gid_groups_conds::images(filenew-16) -style Toolbutton \
        -command [list gid_groups_conds::_import_export_materials_copy_from_spd $w $f.bc2 $f] \
        -state disabled

    tooltip::tooltip $f.f2.b1 [_ "Import default materials"]

    $w set_uservar_value import_export_to ""
    $w set_uservar_value import_export_to_val ""

    grid $f.f2.l1 $f.f2.cb $f.f2.b1 -pady 5 -padx 2
    grid configure $f.f2.cb -sticky ew
    grid columnconfigure $f.f2 1 -weight 1

    boundary_conds $f.bc2 -domnodes "" \
        -edit_window_position up -height 300 -title [_ "Import/export"] \
        -title_active 0 -have_import_export 0 -can_delete_last_material 1

    $w add_trace_to_uservar import_export_to [list \
        gid_groups_conds::_import_export_materials_change_file $f $f.bc2 $f.f2.b1 $w]

    grid    x     x   $f.f2 -sticky ew
    grid $f.bc1 $f.f1 $f.bc2 -sticky nsew -padx 5 -pady 5

    ttk::frame $f.fwarn -borderwidth 0
    ttk::frame $f.fwarn.ff -borderwidth 8
    ttk::scrollbar $f.fwarn.yscroll -takefocus 0 \
        -command [list $f.fwarn.list yview]

    tk::listbox $f.fwarn.list -yscroll "$f.fwarn.yscroll set" -takefocus 0 \
        -relief flat -borderwidth 1
    set ListboxCont [list [_ "Import/export materials... \n"]]
    foreach i $ListboxCont {
        $f.fwarn.list insert end $i
    }
    $f.fwarn.list see end

    $f.fwarn.list configure -height 2

    grid $f.fwarn -sticky news -padx 4 -pady 0 -columnspan 3

    grid $f.fwarn.ff -sticky news -padx 2 -pady 2
    grid $f.fwarn.list $f.fwarn.yscroll -sticky news -padx 4 -pady 0

    grid columnconfigure $f.fwarn 0 -weight 1
    grid rowconfigure $f.fwarn 0 -weight 1

    grid columnconfigure $f "0 2" -weight 1
    grid rowconfigure $f 1 -weight 1

    tk::TabToWindow $f.bc1
    bind $w <Return> [list $w invokeok]
    set action [$w createwindow]
    while 1 {
        switch -- $action {
            -1 - 0 {
                if { [$w give_uservar_value haschanges_local] ||
                    [$w give_uservar_value haschanges_remote] } {
                    set ret [snit_messageBox -type okcancel -default ok \
                        -message [_ "Are you sure to discard all changes?"] \
                        -parent $w]
                    if { $ret eq "ok" } { break }
                } else {
                    break
                }
            }
            1 - 2 {
                set err [catch { _import_export_materials_save $w $f.bc1 $f.bc2 } errstring]
                if { $err } {
                    if { $errstring ne "" } {
                        snit_messageBox -message [_ "Could not save exported materials (%s)" \
                            "${errstring}\n$::errorInfo"] -parent $w
                    } else {
                        snit_messageBox -message [_ "Nothing to save"] -parent $w
                    }
                } elseif { $action == 1 } {
                    GiD_ModifiedFileFlag set 1
                    break
                }
            }
        }
        set action [$w waitforwindow]
    }
    set edoc [[lindex [$f.bc1 cget -domnodes] 0] ownerDocument]
    $edoc delete
    if { [$f.bc2 cget -domnodes] ne "" } {
        set edoc [[lindex [$f.bc2 cget -domnodes] 0] ownerDocument]
        $edoc delete
    }
    if { [$w exists_uservar default_spd_matNode] } {
        set edoc [[$w give_uservar_value default_spd_matNode] ownerDocument]
        $edoc delete
    }
    destroy $w
}

proc gid_groups_conds::_import_export_materials_change_file { f bc button w } {
    variable doc
    variable material_import_export_actions_dict

    set import_export_to [$w give_uservar_value import_export_to]
    $w set_uservar_value import_export_to_val $import_export_to

    if { $import_export_to ni "global_active global_inactive import_file export_file" } {
        return
    }

    if { [$bc cget -domnodes] ne "" } {
        if { [$w give_uservar_value haschanges_remote] } {
            set ret [snit_messageBox -type okcancel -default ok -message [_ "Are you sure to discard all changes?"] -parent $w]
            if { $ret eq "cancel" } {
                return
            }
        }
        set edoc [[lindex [$bc cget -domnodes] 0] ownerDocument]
        $edoc delete
        $w set_uservar_value haschanges_remote 0
    }
    if { $import_export_to eq "global_active" } {
        $button state !disabled
    } else {
        $button state disabled
    }

    switch $import_export_to {
        global_active - global_inactive {
            set f [gid_groups_conds::give_defaults_file]
            set file [file root $f]_materials[file extension $f]
        }
        import_file - export_file {
            set dir [gid_groups_conds::get_preference materials_dir ""]
            switch $import_export_to {
                import_file { set cmd tk_getOpenFile }
                export_file { set cmd tk_getSaveFile }
            }
            set file [$cmd -defaultextension .xml -filetypes [list \
                [list [_ "XML files"] [list ".xml" ".lam"]] \
                [list [_ "All files"] [list "*"]]] \
                -initialdir $dir -parent $w \
                -title [_ "Import/export file"]]
            if { $file eq "" } {
                $w set_uservar_value import_export_to ""
                $bc configure -domnodes ""
                return
            }
            gid_groups_conds::set_preference -local_global global materials_dir [file dirname $file]
            $w set_uservar_value import_export_to $file
        }
    }
    set edoc ""
    if { [file exists $file] } {
        set err [catch { open_XML_file $file } edoc]
        if { !$err } {
            if { [file extension $file] eq ".lam" } {
                $edoc delete
                set edoc ""
            } else {
                set xp {//container[@n="materials" or @n="Materials"]}
                if { [$edoc selectNodes $xp] eq "" } {
                    $edoc delete
                    set edoc ""
                }
            }
        }
    }
    if { $import_export_to eq "export_file" && $edoc ne "" } {
        set num [$edoc selectNodes {count(//blockdata[@n="material"])}]
        if { $num } {
            set t [_ "File '%s' already contains %d materials. Import them?" $file $num]
            set ret [snit_messageBox -type yesnocancel -default yes -message $t -parent $w]
            if { $ret == "cancel" } {
                $w set_uservar_value import_export_to ""
                $bc configure -domnodes ""
                return
            }
            if { $ret eq "no" } {
                $edoc delete
                set edoc ""
            }
        }
    }

    if { $edoc eq "" } {
        set xp {string(/*/@version)}
        set v [$doc selectNodes $xp]
        set edoc [gid_groups_conds::parse_XML_data "<[[$doc documentElement] nodeName] version='$v'/>"]
        set root [$edoc documentElement]
        set mNode [$root appendXML {<container n="materials" pn="Materials"/>}]
        if { $import_export_to in "global_active global_inactive" } {
            $root appendXML {
                <container n="materials" pn="Inactive materials" inactive="1"/>
            }
            set xp {container[@n="materials" and (not(@inactive) or @inactive!="1")]}
            set mNode [$root selectNodes $xp]
            set spd_edoc [gid_groups_conds::open_XML_file_gzip [gid_groups_conds::give_spd_file]]
            set xp {//container[@n="materials" or @n="Materials"]/*}
            foreach node [$spd_edoc selectNodes $xp] {
                $mNode appendChild $node
            }
            $spd_edoc delete
        }
        if { [file extension $file] eq ".lam" } {
            set matsXMLFile [file join [file dir [gid_groups_conds::give_spd_file]] xml materials.xml]
            set spd_edoc [gid_groups_conds::open_XML_file_gzip $matsXMLFile]
            set xp {//container[@n="materials_types" and @pn="Orthotropic"]}
            set baseNode [$spd_edoc selectNodes $xp]
            set newNode [$baseNode clone]
            $mNode appendChild $newNode
            set xp {//container[@n="laminate_type"]}
            set baseNode [$spd_edoc selectNodes $xp]
            set newNode [$baseNode clone]
            $mNode appendChild $newNode
            Composite::import_laminates_from $file lam $mNode
        }
    }

    set domNode_orig [$w give_uservar_value domNode_orig]
    set pNode [$edoc selectNodes /*/procs]
    if { $pNode ne "" } { $pNode delete }
    set pNode [$domNode_orig selectNodes /*/procs]
    if { $pNode != "" } {
        [$edoc documentElement] appendChild [$pNode cloneNode -deep]
    }

    switch $import_export_to {
        global_active {
            set xp {//container[(@n='materials' or @n='Materials') and (not(@inactive) or @inactive!="1")]}
        }
        global_inactive {
            set xp {//container[(@n='materials' or @n='Materials') and (@inactive="1")]}
        }
        import_file - export_file {
            set xp {//container[@n='materials' or @n='Materials']}
        }
    }

    set domNode [$edoc selectNodes $xp]
    $domNode setAttribute state hidden_this
    $bc configure -domnodes $domNode
}

proc gid_groups_conds::_import_export_materials_save { w bc_local bc_remote } {

    if { [$w give_uservar_value haschanges_local] } {
        set domNode_orig [$w give_uservar_value domNode_orig]
        set domNode_new [$bc_local cget -domnodes]

        foreach node [$domNode_orig childNodes] {
            $node delete
        }
        foreach node [$domNode_new childNodes] {
            $domNode_orig appendChild [$node cloneNode -deep]
        }
        actualize_conditions_window
        $w set_uservar_value haschanges_local 0
    }

    switch [$w give_uservar_value import_export_to] {
        global_active - global_inactive {
            set f [gid_groups_conds::give_defaults_file]
            set file [file root $f]_materials[file extension $f]
        }
        default {
            set file [$w give_uservar_value import_export_to]
            if { $file eq "" } { error "" }
        }
    }

    if { [$w give_uservar_value import_export_to_val] eq "export_file"} {
        set fout [open $file w]
        fconfigure $fout -encoding utf-8
        puts $fout "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
        set domNode [$bc_remote cget -domnodes]
        set edoc [$domNode ownerDocument]
        set procsNode [$domNode selectNodes /*/procs]
        if { $procsNode != "" } {
            set parentNode [$procsNode parentNode]
            $parentNode removeChild $procsNode
        }
        puts $fout [$edoc asXML -indent 2]
        if { $procsNode != "" } {
            $parentNode appendChild $procsNode
        }
        close $fout
    }
    $w set_uservar_value haschanges_remote 0
}

proc gid_groups_conds::_import_export_materials_copy { w from_local_remote bc_local bc_remote f } {

    switch $from_local_remote {
        local {
            lassign [list $bc_local $bc_remote] bc_from bc_to
        }
        remote {
            lassign [list $bc_local $bc_remote] bc_to bc_from
        }
    }
    set mats ""
    foreach id [$bc_from selection get] {
        set domNode [$bc_from item element cget $id 0 e_text -data]
        set xp {self::*[@n="material"]|ancestor::*[@n="material"]|*[@n="material"]}
        foreach i [$domNode selectNodes $xp] {
            if { $i ni $mats } { lappend mats $i }
            if { [$domNode selectNodes {value[@n="plies_list"]}] ne "" } {
                set v [$domNode selectNodes {string(value[@n="plies_list"]/@v)}]
                set xpm {//blockdata[@n="material"]}
                foreach vi [split $v ,] {
                    lassign $vi imat -
                    foreach matNode [$gid_groups_conds::doc selectNodes $xpm] {
                        set name [$matNode @name]
                        if { $imat eq $name } {
                            lappend mats $matNode
                            if { [$matNode selectNodes {container[@n="material_elastic_compos"]}] ne "" } {
                                set xpmat {container[@n="material_elastic_compos"]/value}
                                foreach plymatNode [$matNode selectNodes $xpmat] {
                                    set ismat [$plymatNode @ismaterial 0]
                                    if { !$ismat } { continue }
                                    set plymat [$plymatNode @v]
                                    set xpm {//blockdata[@n="material"]}
                                    foreach iNode [$gid_groups_conds::doc selectNodes $xpm] {
                                        set name [$iNode @name]
                                        if { $plymat eq $name } {
                                            lappend mats $iNode
                                            break
                                        }
                                    }
                                }
                            }
                            break
                        }
                    }
                }
            } elseif { [$domNode selectNodes {container[@n="material_elastic_compos"]}] ne "" } {
                set xpmat {container[@n="material_elastic_compos"]/value}
                foreach plymatNode [$matNode selectNodes $xpmat] {
                    set ismat [$plymatNode @ismaterial 0]
                    if { !$ismat } { continue }
                    set plymat [$plymatNode @v]
                    set xpm {//blockdata[@n="material"]}
                    foreach matNode [$gid_groups_conds::doc selectNodes $xpm] {
                        set name [$matNode @name]
                        if { $plymat eq $name } {
                            lappend mats $matNode
                            break
                        }
                    }
                }
            }
        }
    }
    _import_export_materials_copy_do $w $from_local_remote $mats $bc_to $f
}

proc gid_groups_conds::_import_export_materials_copy_from_spd { w bc_to f } {

    set spd_edoc [gid_groups_conds::open_XML_file_gzip [gid_groups_conds::give_spd_file]]

    set mats ""
    foreach i [$spd_edoc selectNodes {//*[@n="material"]}] {
        lappend mats $i
    }
    _import_export_materials_copy_do $w remote $mats $bc_to $f
    $spd_edoc delete
}

proc gid_groups_conds::_import_export_materials_copy_do { w from_local_remote mats bc_to f } {
    set domNode_to [$bc_to cget -domnodes]
    if { $domNode_to eq "" } {
        snit_messageBox -message [_ "Select a place where import/export materials"] -parent $w
        return
    }
    if { $mats eq "" } {
        snit_messageBox -message [_ "There are no selected materials to import/export"] -parent $w
        return
    }

    set islisted 0
    if { $from_local_remote eq "remote" } {
        if { ![$w exists_uservar default_spd_matNode] } {
            set file [gid_groups_conds::give_spd_file]
            set doc [gid_groups_conds::open_XML_file_gzip $file]
            foreach includeNode [$doc selectNodes {.//include}] {
                set active [$includeNode @active 1]
                if { !$active } { continue }
                set parent [$includeNode parent]
                set path [$includeNode @path ""]
                set file_ [file join [file dirname $file] $path]
                set xml [gid_groups_conds::open_XML_file_gzip $file_]
                foreach newnode [$xml childNodes] {
                    if { [$newnode nodeName] eq "" } {
                        continue
                    }
                    if { $includeNode eq "" } {
                        $parent appendChild $newnode
                    } else {
                        $parent insertBefore $newnode $includeNode
                    }
                }
                $parent removeChild $includeNode
            }
            set root_ [$doc documentElement]
            foreach elem [$root_ getElementsByTagName "appLink"] {
                set appid [$elem getAttribute "appid"]
                set xpk {.//hiddenfield[@n="activeapp"]}
                set active 0
                if { $appid eq [[$gid_groups_conds::doc selectNodes $xpk] @v ""] && $appid ne "" } {
                    set active 1
                }
                if {$active} {
                    set dir [GiD_Info problemtypepath]
                    set f_ [file join $dir apps $appid xml Main.spd]
                    set processedAppnode [gid_groups_conds::ProcessIncludesRecurse $f_ $dir]
                    $root_ insertBefore $processedAppnode $elem
                    $elem delete
                }
            }

            set xp {//container[@n="materials" or @n="Materials"]}
            set default_spd_matNode [$doc selectNodes $xp]
            if { [llength $default_spd_matNode] > 1 } {
                snit_messageBox -message [_ "There are %d container n=materials, expected 1. The first will be used" [llength $default_spd_matNode]] -parent $w
                set default_spd_matNode [lindex $default_spd_matNode 0]
            }

            $w set_uservar_value default_spd_matNode $default_spd_matNode
        } else {
            set default_spd_matNode [$w give_uservar_value default_spd_matNode]
        }
        set err 0
        foreach matNode_from $mats {
            set num 0
            if {[$matNode_from @pn ""] eq "laminate"} {
                set type_mat "laminate_type"
                set iNodes $matNode_from
                set pmatNode_from [$matNode_from selectNodes {ancestor::container[@n='materials_types']}]
            } else {
                set type_mat "materials_types"
                set iNodes [$matNode_from selectNodes container]
                if { $iNodes eq "" } {
                    set iNodes $matNode_from
                }
                set pmatNode_from [$matNode_from selectNodes {ancestor::container[@n='materials_types']}]
                if { $pmatNode_from eq "" } {
                    set pmatNode_from $matNode_from
                }
            }

            set namemat [$matNode_from @name ""]
            foreach cNode $iNodes {
                if {$pmatNode_from eq ""} {
                    set xp [format_xpath {(.//blockdata[@n="material"]/container[@n=%s])} [$cNode @n ""]]
                    foreach cNode_to [$default_spd_matNode selectNodes $xp] {
                        if {$cNode_to eq ""} {
                            set matNode_from_type [[$cNode parentNode] @pn ""]
                            if {[$cNode @pn ""] eq "laminate"} {
                                set xp [format_xpath {(./container[@pn=%s]/blockdata[@n="material"])} $matNode_from_type]
                            } else {
                                set xp [format_xpath {(./container[@pn=%s]/blockdata[@n="material"]/container[@n=%s])} $matNode_from_type [$cNode @n ""]]
                            }
                            foreach vNode [$cNode selectNodes {value|container/value}] {
                                set xpv [format_xpath {.//value[@n=%s]} [$vNode @n ""]]
                                set vNode_to [$cNode_to selectNodes $xpv]
                                if { $vNode_to ne "" } {
                                    incr num
                                }
                            }
                        } else {
                            foreach vNode [$cNode selectNodes {value|container/value}] {
                                set xp [format_xpath {.//value[@n=%s]} [$vNode @n ""]]
                                set vNode_to [$cNode_to selectNodes $xp]
                                if { $vNode_to ne "" } {
                                    incr num
                                }
                            }
                        }
                    }
                } else {
                    set matNode_from_type [$pmatNode_from @pn ""]
                    set xp [format_xpath {(./container[@pn=%s]/blockdata[@name=%s]/container[@n=%s])} $matNode_from_type $namemat [$cNode @n ""]]
                    set cNode_to [$default_spd_matNode selectNodes $xp]
                    if {$cNode_to eq ""} {
                        if {[$cNode @pn ""] eq "laminate"} {
                            set xp [format_xpath {(./container[@pn=%s]/blockdata[@n="material"])} $matNode_from_type]
                        } else {
                            set xp [format_xpath {(./container[@pn=%s]/blockdata[@n="material"]/container[@n=%s])} $matNode_from_type [$cNode @n ""]]
                            if { [$default_spd_matNode selectNodes $xp] == "" } {
                                set xp {./blockdata[@n="material"]}
                            }
                        }
                        foreach cNode_to [$default_spd_matNode selectNodes $xp] {
                            foreach vNode [$cNode selectNodes {value|container/value}] {
                                set xpv [format_xpath {.//value[@n=%s]} [$vNode @n ""]]
                                set vNode_to [$cNode_to selectNodes $xpv]
                                if { $vNode_to ne "" } {
                                    incr num
                                }
                            }
                        }
                    } else {
                        foreach vNode [$cNode selectNodes {value|container/value}] {
                            set xp [format_xpath {.//value[@n=%s]} [$vNode @n ""]]
                            set vNode_to [$cNode_to selectNodes $xp]
                            if { $vNode_to ne "" } {
                                incr num
                            }
                        }
                    }
                }
            }
            if { $num == 0 || [$matNode_from @name ""] eq "" } {
                set err 1
            }
            if { [$matNode_from @name ""] ne "" } {
                $f.fwarn.list insert end [_ "Material '%s' added to local materials list" [$matNode_from @name ""]]
                $f.fwarn.list see end
                set islisted 1
            }
        }
    }

    set yestoall 0
    set num 0
    foreach matNode_from $mats {
        set xp [format_xpath {.//*[@name=%s]} [$matNode_from @name ""]]
        set matNode_to [$domNode_to selectNodes $xp]
        if { $matNode_to ne "" && !$yestoall } {
            regsub -all {\s*tree_state="[^"]*"\s*} [$matNode_to asXML -indent none] {} m1
            regsub -all {\s*tree_state="[^"]*"\s*} [$matNode_from asXML -indent none] {} m2
            if { $m1 ne $m2 } {
                set w_in [dialogwin_snit $w._ask -okname [_ "Yes"] -morebuttons [list [_ "No"]] -repeat_answer_check 1 -entrytext [_ "Are you sure to update existing material '%s'?" [$matNode_from @name]]]
                set action [$w_in createwindow]
                set repeat_my_answer [$w_in give_repeat_my_answer]
                destroy $w_in
                switch -- $action {
                    -1 - 0 {
                        return
                    }
                    1 {
                        if { $repeat_my_answer } {
                            set yestoall 1
                        }
                    }
                    2 {
                        if { $repeat_my_answer } {
                            return
                        }
                        continue
                    }
                }
            }
        }
        if { $matNode_to ne "" } {
            $matNode_to delete
        }

        if { [$matNode_from selectNodes {parent::*/@n != "materials"}] } {
            set parentNode_from [$matNode_from parentNode]
            set xp [format_xpath {.//*[@n=%s and @pn=%s]} [$parentNode_from @n] [$parentNode_from @pn]]
            set parentNode_to [$domNode_to selectNodes $xp]
            if { $parentNode_to eq "" } {
                set parentNode_to [$domNode_to appendChild [$parentNode_from cloneNode]]
            }
        } else {
            set parentNode_to $domNode_to
        }
        $parentNode_to appendChild [$matNode_from cloneNode -deep]
        if { [$matNode_from @name ""] ne "" && $islisted == "0" } {
            $f.fwarn.list insert end [_ "Material '%s' added to import/export list" [$matNode_from @name ""]]
            $f.fwarn.list see end
        }

        $parentNode_to setAttribute tree_state open
        incr num
    }
    $bc_to actualize
    if { $num } {
        switch $from_local_remote {
            local {
                $w set_uservar_value haschanges_remote 1
            }
            remote {
                $w set_uservar_value haschanges_local 1
            }
        }
    }
}

proc gid_groups_conds::import_export_materials_get_global {} {
    variable doc

    set f [gid_groups_conds::give_defaults_file]
    set file [file root $f]_materials[file extension $f]
    set edoc ""
    if { [file exists $file] } {
        set err [catch { open_XML_file $file } edoc]
        if { !$err } {
            regexp {^\d+} [$doc selectNodes {string(/*/@version)}] v1
            regexp {^\d+} [$edoc selectNodes {string(/*/@version)}] v2
            set xp {//container[(@n="materials" or @n="Materials"]) and (not(@inactive) or @inactive!="1")]}
            if { $v1 != $v2 || [$edoc selectNodes $xp] eq "" } {
                $edoc delete
                set edoc ""
            }
        }
    }
    if { $edoc eq "" } { return }

    set xp {//container[@n="materials" or @n="Materials"]}
    set domNode [$doc selectNodes $xp]
    set xp {(//container[@n="materials" or @n="Materials"]) and (not(@inactive) or @inactive!="1")]}
    set domNode_global [$edoc selectNodes $xp]

    foreach node [$domNode childNodes] {
        $node delete
    }
    foreach node [$domNode_global childNodes] {
        $domNode appendChild [$node cloneNode -deep]
    }
    $edoc delete
}

################################################################################
#    Properties unicity
################################################################################

proc gid_groups_conds::check_properties_unicity { wp domNode_path } {
    variable doc

    return [check_properties_unicity_do $wp [$doc selectNodes $domNode_path]]

}
proc gid_groups_conds::check_properties_unicity_do { wp domNode } {

    switch [GiD_Info Project ViewMode] {
        "GEOMETRYUSE" {
            set geom_mesh geometry
            set ent_names [list [_ Point] [_ Line] [_ Surface] [_ Volume]]
        }
        "MESHUSE"  {
            set geom_mesh mesh
            set ent_names [list [_ "Node"] [_ "Line element"] [_ "Surface element"] [_ "Volume element"]]
        }
    }
    lassign [list 0 0 0] ok_to_all num_changes_groups num_changes_entities
    set grList ""
    foreach grNode [$domNode selectNodes .//group] {
        lappend grList [list [$grNode @n] $grNode]
    }
    set grList [lsort -index 0 $grList]
    set i_prev 0
    for { set i 1 } { $i < [llength $grList] } { incr i } {
        if { [lindex $grList $i 0] eq [lindex $grList $i_prev 0] } {
            set group_name [lindex $grList $i_prev 0]
            while { $i < [llength $grList]-1 && [lindex $grList $i+1 0] eq  $group_name } {
                incr i
            }
            if { !$ok_to_all } {
                destroy $wp._ask
                set txt [_ "Group '%s' is applied to several properties. It has to be erased from all except one. Which do you want to keep?" \
                    $group_name]
                set w [dialogwin_snit $wp._ask -title [_ "Duplicated groups"] -morebuttons [list [_ "OK to all"]]]
                set f [$w giveframe]

                ttk::label $f.l1 -text $txt -wraplength 400 -justify left
                grid $f.l1 -padx 2 -pady "2 20" -sticky w

                set idx 0
                foreach j [rangeF $i_prev $i] {
                    set cndNode [[lindex $grList $j 1] selectNodes ancestor::condition]
                    set pn [get_domnode_attribute $cndNode pn]
                    ttk::radiobutton $f.cb$idx -text [_ "Keep group '%s' in property '%s'" $group_name $pn] -variable \
                        [$w give_uservar condpos] -value $j
                    grid $f.cb$idx -sticky w -pady 2 -padx "5 60"
                    incr idx
                }
                $w set_uservar_value condpos $i_prev
                bind $w <Return> [list $w invokeok]
                set action [$w createwindow]
                set condpos [$w give_uservar_value condpos]
                destroy $w
                if { $action < 1 } {
                    return
                }
                if { $action == 2 } {
                    set ok_to_all 1
                }
            } else {
                set condpos 0
            }
            foreach j [rangeF $i_prev $i] {
                if { $j == $condpos } {
                    continue
                }
                set groupNode [lindex $grList $j 1]
                deleteN $groupNode
                lset grList $j [list "" ""]
            }
            incr num_changes_groups
        }
        set i_prev $i
    }
    set grList [lsort -index 0 $grList]

    foreach i [list point line surface volume] i_name $ent_names {
        set ents ""
        set group_ents ""
        foreach j [GiD_Info conditions ${i}_groups $geom_mesh] {
            lassign $j - num - group
            if { [set ipos [lsearch -index 0 -exact -sorted $grList $group]] != -1 } {
                dict lappend ents $num [lindex $grList $ipos]
            }
        }
        lassign "" history history_auto
        set do_the_same 0
        dict for "n v" $ents {
            if { [llength $v] < 2 } { continue }

            if { !$ok_to_all && ![dict exists $history_auto $v] } {
                destroy $wp._ask
                set txt [_ "%s '%s' is applied to several properties. It has to be unassigned from all except one. Which do you want to keep?" \
                    $i_name $n]
                set w [dialogwin_snit $wp._ask -title [_ "Duplicated properties for entities"] -morebuttons \
                    [list [_ "OK to all"]]]
                set f [$w giveframe]

                ttk::label $f.l1 -text $txt -wraplength 400 -justify left
                grid $f.l1 -padx 2 -pady "2 20" -sticky w

                if {  [dict exists $history $v] } {
                    $w set_uservar_value condpos [dict get $history $v]
                } else {
                    $w set_uservar_value condpos 0
                }
                set idx 0
                foreach j $v {
                    set group_name [lindex $j 0]
                    set cndNode [[lindex $j 1] selectNodes ancestor::condition]
                    set pn [get_domnode_attribute $cndNode pn]
                    ttk::radiobutton $f.cb$idx -text [_ "Keep %s '%s' in property '%s' group '%s'" \
                        $i_name $n $pn $group_name] -variable \
                        [$w give_uservar condpos] -value $idx
                    grid $f.cb$idx -sticky w -pady 2 -padx "5 60"
                    incr idx
                }
                ttk::checkbutton $f.cbk -text [_ "Do the same for similar entities"] -variable \
                    [$w give_uservar do_the_same $do_the_same]
                grid $f.cbk -sticky w -pady "12 2" -padx "5 60"
                bind $w <Return> [list $w invokeok]
                set action [$w createwindow]
                set condpos [$w give_uservar_value condpos]
                set do_the_same  [$w give_uservar_value do_the_same]
                destroy $w
                if { $action < 1 } { return }
                if { $action == 2 } {
                    set ok_to_all 1
                }
                if { $do_the_same } {
                    dict set history_auto $v $condpos
                } else {
                    dict set history $v $condpos
                }
            } else {
                if {  [dict exists $history_auto $v] } {
                    set condpos [dict get $history_auto $v]
                } elseif {  [dict exists $history $v] } {
                    set condpos [dict get $history $v]
                } else {
                    set condpos 0
                }
            }
            set idx 0
            foreach j $v {
                if { $idx != $condpos } {
                    dict lappend group_ents [lindex $j 0] $n
                }
                incr idx
            }
            incr num_changes_entities
        }
        dict for "group_name entities" $group_ents {
            delete_entities_from_groups $entities [list [list "" $group_name]] $i
        }
    }
    if { $num_changes_groups } {
        actualize_conditions_window
    }
    set num_changes [expr {$num_changes_groups+$num_changes_entities}]
    if { !$num_changes } {
        snit_messageBox -message [_ "Properties are OK. No duplicated entities"] -parent $wp
    } else {
        snit_messageBox -message [_ "Properties have been corrected. %d changes" $num_changes] -parent $wp
    }
}
proc gid_groups_conds::ChangedLanguage {} {
    gid_groups_conds::actualize_conditions_window
    draw_post::reopen_gui
}

#################################################################################
#    small utilities for drawing symbols and other operations
#################################################################################


proc gid_groups_conds::give_value_and_unit { valuesList prop } {

    set ipos [lsearch -index 0 $valuesList $prop]
    if { $ipos == -1 } { return "" }
    set elm [lindex $valuesList $ipos]
    return [lrange $elm 1 2]
}

proc gid_groups_conds::calc_yz_beam { xvec } {

    set epsilon 1e-11
    if { abs([lindex $xvec 0]) > $epsilon || abs([lindex $xvec 1]) } {
        if { abs([lindex $xvec 0]) > $epsilon } {
            set yvec [list [expr {-1*[lindex $xvec 1]/double([lindex $xvec 0])}] 1.0 0.0]
        } else {
            set yvec [list 1.0 0.0 0.0]
        }
        set yvec [m::unitLengthVector $yvec]
        set zvec [m::vectorprod3 $xvec $yvec]
        if { abs([lindex $zvec 2]) < $epsilon } {
            set yvec [m::scale -1.0 $yvec]
            set zvec [m::vectorprod3 $xvec $yvec]
        }
    } else {
        set yvec [list 1.0 0.0 0.0]
        set zvec [m::vectorprod3 $xvec $yvec]
    }
    return [list $yvec $zvec]
}

proc gid_groups_conds::calc_xy_shell { zvec } {

    set uy [list 0.0 1.0 0.0]
    set uz [list 0.0 0.0 1.0]

    set zvec [m::unitLengthVector $zvec]
    if { abs([lindex $zvec 0])<1.0/64.0 && abs([lindex $zvec 1])<1.0/64.0 } {
        set xvec [m::unitLengthVector [m::vectorprod3 $uy $zvec]]
    } else {
        set xvec [m::unitLengthVector [m::vectorprod3 $uz $zvec]]
    }
    set yvec [m::unitLengthVector [m::vectorprod3 $zvec $xvec]]
    return [list $xvec $yvec $zvec]
}

proc gid_groups_conds::give_OpenGL_mat { xvec yvec zvec } {
    # mat is 4x4 = 16 in column order (OpenGL matrix)
    set mat [concat $xvec 0.0]
    eval lappend mat $yvec 0.0
    eval lappend mat $zvec 0.0
    lappend mat 0.0 0.0 0.0 1.0
    return $mat
}

proc gid_groups_conds::give_LA_mat { mat } {

    set ret ""
    foreach "c1 c2 c3" $mat {
        lappend ret [list $c1 $c2 $c3]
    }
    return $ret
}

proc gid_groups_conds::draw_symbol_local_axes { valuesList geom_mesh ov \
    num pnts points ent_type center scale } {

    switch $geom_mesh {
        GEOMETRYUSE { set gm geometry }
        MESHUSE {
            error "not implemented yet"
            set gm mesh
        }
    }
    set pntsList ""
    switch $ov {
        point {
            set LAmat [gid_groups_conds::local_axes_matrix Point $gm $num]
            lappend pntsList [list $center ""]
        }
        line {
            set LAmat [gid_groups_conds::local_axes_matrix Line $gm $num]
            if { $ent_type eq "STLINE" } {
                set p [gid_groups_conds::parametric_line $num 0.5]
                set der [m::unitLengthVector [gid_groups_conds::deriv_parametric_line $num 0.5]]
                lappend pntsList [list $p $der]
            } else {
                lassign $points p1 p2
                set delta .1
                for { set t $delta } { $t <= 1.0+.5*$delta } { set t [expr {$t+$delta}] } {
                    set p [gid_groups_conds::parametric_line $num $t]
                    set der [m::unitLengthVector [gid_groups_conds::deriv_parametric_line $num $t]]
                    lappend pntsList [list $p $der]
                }
            }
        }
        surface {
            set LAmat [gid_groups_conds::local_axes_matrix Surface $gm $num]
            set normal  [m::unitLengthVector [gid_groups_conds::give_surface_normal $num]]
            lappend pntsList [list $center $normal]
        }
    }
    foreach pL $pntsList {
        lassign $pL p der
        if { $LAmat ne "" } {
            lassign "" xvec yvec zvec
            foreach "c1 c2 c3" $LAmat {
                if { $ov ne "line" } {
                    lappend xvec $c1
                }
                lappend yvec $c2
                lappend zvec $c3
            }
            if { $ov eq "line" } {
                set xvec $der
            }
        } else {
            switch $ov {
                line { set xvec $der; lassign [gid_groups_conds::calc_yz_beam $der] yvec zvec }
                surface { lassign [gid_groups_conds::calc_xy_shell $der] xvec yvec zvec }
            }
        }
        set mat [gid_groups_conds::give_OpenGL_mat $xvec $yvec $zvec]
        GiD_OpenGL draw -pushmatrix -translate $p -multmatrix $mat -scale [list $scale $scale $scale]
        ggc::draw::local_axes -draw_labels 1 -draw_style cylinders 1.0 "" ""
        GiD_OpenGL draw -popmatrix
    }
}

proc gid_groups_conds::parametric_line { num t } {
    variable is_draw_post

    if { !$is_draw_post } {
        return [GiD_Info parametric line $num coord $t]
    } else {
        return [lindex [ggc::manage_geometry_entities properties  line_coords_dercoords $num $t] 0]
    }
}

proc gid_groups_conds::deriv_parametric_line { num t } {
    variable is_draw_post

    if { !$is_draw_post } {
        return [GiD_Info parametric line $num deriv_t $t]
    } else {
        return [lindex [ggc::manage_geometry_entities properties  line_coords_dercoords $num $t] 1]
    }
}

proc gid_groups_conds::local_axes_matrix { ov geometry_mesh num } {
    variable doc
    variable is_draw_post

    if { !$is_draw_post } {
        set ret [GiD_Info conditions -localaxesmat ${ov}_Local_axes \
            $geometry_mesh $num]
        return [lindex $ret 0 3]
    } else {
        set ov [string tolower $ov]
        lassign [ggc::manage_geometry_entities list details $ov $num] - d_geom

        set xp {condition[@n="local_axes"]/groupList}
        set root [$doc documentElement]
        set grDict ""
        foreach nodeGroup [$root selectNodes $xp] {
            set n [$nodeGroup selectNodes {string(group[1]/@n)}]
            dict set grDict [ggc::manage_geometry_groups id $n] $nodeGroup
        }
        set nodeGroup ""
        foreach gr [dict get $d_geom GroupList] {
            if { [dict exists $grDict $gr] } {
                set nodeGroup [dict get $grDict $gr]
            }
        }
        if { $nodeGroup eq "" } {
            return ""
        }

        set local_axes_type [$nodeGroup selectNodes {string(value[@n="local_axes_type"]/@v)}]
        set swap_axes [$nodeGroup selectNodes {string(value[@n="swap_axes"]/@v)}]

        if { $local_axes_type eq "Automatic" } {
            switch $ov {
                line {
                    set vector [deriv_parametric_line $num 0.5]
                    set mat [ggc::calc::calculate_local_axes -swap_axes $swap_axes matrix xaxe $vector]
                }
                surface {
                    set vector [give_surface_normal $num]
                    set mat [ggc::calc::calculate_local_axes -swap_axes $swap_axes matrix zaxe $vector]
                }
                default {
                    error "error in gid_groups_conds::local_axes_matrix"
                }
            }
            return $mat
        } elseif { $local_axes_type eq "Tangent lines" } {
            # not implemented yes
            return ""
        } elseif { $local_axes_type eq "User defined" } {
            set mat [lrepeat 9 0.0]
            foreach axe [list x y z] col "0 1 2" {
                set vec [$nodeGroup selectNodes [format_xpath {string(value[@n=%s]/@v)} ${axe}paxe]]
                foreach v [split $vec ,] row "0 1 2" {
                    lset mat [expr {3*$row+$col}] $v
                }
            }
            return $mat
        }
        #puts $num---[nice_number_printList $mat]
    }
}

proc gid_groups_conds::give_point_coords { num } {
    variable is_draw_post

    if { !$is_draw_post } {
        return [lrange [GiD_Geometry get point $num] 1 3]
    } else {
        lassign [ggc::manage_geometry_entities list details point $num] - d_geom
        return [dict get $d_geom PointCoords]
    }
}

proc gid_groups_conds::give_line_points { num } {
    variable is_draw_post

    if { !$is_draw_post } {
        set list_ent [GiD_Info list_entities Lines $num]
        regexp {^\S+} $list_ent ent_type
        regexp {Points:\s+([0-9]+)\s+([0-9]+)} $list_ent {} P1 P2
    } else {
        lassign [ggc::manage_geometry_entities list geometry line $num] - d_geom
        lassign [dict get $d_geom ContourEntities] P1 P2
        if { [dict get $d_geom Degree] ==1 && [dict get $d_geom NumCtrlPoints] == 2 } {
            set ent_type STLINE
        } else {
            set ent_type NURBLINE
        }
    }
    return [list $P1 $P2 $ent_type]
}

proc gid_groups_conds::give_surface_normal { num } {
    variable is_draw_post

    if { !$is_draw_post } {
        set list_ent [GiD_Info list_entities Surfaces $num]
        regexp {Normal:\s+(\S+)\s+(\S+)\s+(\S+)} $list_ent {} x y z
        return [list $x $y $z]
    } else {
        lassign [ggc::manage_geometry_entities list geometry surface $num] - d_geom
        return [dict get $d_geom Normal]
    }
}

proc gid_groups_conds::local_axes_edit_end { dict ltype callback value } {

    set d ""
    switch $ltype {
        "select_vectorX" {
            dict set dict xpaxe [join $value ,]
            set axe_name xaxe
        }
        "select_vectorY" {
            dict set dict ypaxe [join $value ,]
            set axe_name yaxe
        }
        "select_vectorZ" {
            dict set dict zpaxe [join $value ,]
            set axe_name zaxe
        }
        "define_xz" {
            set mat $value
        }
    }
    if { ![info exists mat] } {
        foreach i [list xpaxe ypaxe zpaxe] {
            set $i [split [dict get $dict $i] ,]
            if { [catch { m::add [set $i] "0 0 0" }] } {
                set $i "0 0 0"
            }
        }
        set mat [ggc::calc::calculate_local_axes matrix $axe_name $xpaxe $ypaxe $zpaxe]
    } else {
        foreach i [list xpaxe ypaxe zpaxe] {
            set $i [list 0 0 0]
        }
    }
    foreach  i [list xpaxe ypaxe zpaxe] col "0 1 2" {
        foreach row "0 1 2" {
            lset $i $row [lindex $mat [expr {3*$row+$col}]]
        }
        dict set d $i [join [set $i] ,]
    }
    eval $callback [list $d ""]
}

proc gid_groups_conds::local_axes_edit { domNode dict ltype callback } {

    set d ""
    switch $ltype {
        "select_vectorX" - "select_vectorY" - "select_vectorZ" {
            draw_post::get_vector_gui -callback [list gid_groups_conds::local_axes_edit_end \
                $dict $ltype $callback]
        }
        define_xz {
            draw_post::get_local_axes_xz -callback [list gid_groups_conds::local_axes_edit_end \
                $dict $ltype $callback] start
        }
    }
}

proc gid_groups_conds::transfer_mesh_sizes_to_gid {} {
    variable doc

    GidUtils::DisableGraphics

    set has_reset 0

    set xp {/*/container[@n="mesh_sizes"]/condition[@n="mesh_sizes_entities"]/group}
    foreach groupNode [$doc selectNodes $xp] {
        if { !$has_reset } {
            GiD_Process MEscape Meshing reset Yes
            set has_reset 1
        }
        set ov [$groupNode @ov]
        set group_name [$groupNode @n]
        set vs ""
        foreach i [list mesh_size mesh_num_cells chordal_error size_transition mesh_criteria mesh_structuring] {
            dict set vs $i [$groupNode selectNodes [format_xpath {string(value[@n=%s]/@v)} $i]]
        }
        set entities ""
        if { $::CUSTOM_LIB_USE_NATIVE_GROUPS } {
            foreach num [GiD_EntitiesGroups get $group_name ${ov}s] {
                lappend entities $num
            }
        } else {
            foreach j [GiD_Info conditions ${ov}_groups geometry] {
                lassign $j - num - group
                if { $group ne $group_name } {
                    continue
                }
                lappend entities $num
            }
        }
        switch [dict get $vs mesh_structuring] {
            "Structured lines" {
                set m [dict get $vs mesh_num_cells]
                if { [string is integer -strict $m] } {
                    GiD_Process MEscape Meshing Structured $ov $m {*}$entities Escape Escape
                }
            }
            "Structured" {
                GiD_Process MEscape Meshing Structured $ov {*}$entities Escape Escape
            }
            "Semistructured volumes" {
                set m [dict get $vs mesh_num_cells]
                if { [string is integer -strict $m] } {
                    GiD_Process MEscape Meshing SemiStructured $ov $m {*}$entities Escape Escape
                }
            }
            "Semistructured master" {
                GiD_Process MEscape Meshing SemiStructured SetMaster {*}$entities Escape Escape
            }
            default {
                # Unstructured
                set m [dict get $vs mesh_size]
                if { [string is double -strict $m] } {
                    GiD_Process MEscape Meshing AssignSizes $ov $m {*}$entities Escape Escape
                }
            }
        }
    }
    GidUtils::EnableGraphics
}

proc gid_groups_conds::_update_menubutton_check_listbox { w n } {
    array set parents ""
    foreach v [$w cget -values] {
        foreach "name state parentName open_close" $v break
        if { $state eq "on" } {
            if { $parentName ne "" } {
                incr parents($parentName)
            } else {
                incr parents($name)
            }
        }
    }
    set text [_ "Selected items (%s)" [llength [array names parents]]]
    $w configure -text $text
    #     set entryvalues($n) $text
    #     $self _fill_data_window_dependencies -value_variables -is_pre_dependency $is_pre_dependency $valueNodes
}

##########################################################################################
#GiD BEGIN: new procedures required to allow customlib_native_groups
##########################################################################################

#implement GiD-Tcl events to sincronize data in both directions

proc gid_groups_conds::GiD_AfterChangeGroup { group property } {
    if { [gid_groups_conds::group_exists $group] } {
        if { $property == "color" } {
            set color [string range [GiD_Groups get color $group] 0 6] ;#remove last xx values of alpha
            set current_color_rgb [gid_groups_conds::get_color_rrggbb_from_symbolic_name [gid_groups_conds::get_group_att $group color]]
            if { $current_color_rgb != $color } {
                gid_groups_conds::change_group_att $group color $color
            }
        } elseif { $property == "visible" } {
            set value [GiD_Groups get visible $group]
            if { [gid_groups_conds::get_group_att $group onoff] != $value } {
                gid_groups_conds::change_group_att $group onoff $value
            }
        } elseif { $property == "state" } {

        } elseif { $property == "allowed_types" } {

        } elseif { $property == "allowed_element_types" } {

        } else {
        }
    }
}

proc gid_groups_conds::GiD_AfterChangeParentGroup { oldname newname } {
    if { [gid_groups_conds::group_exists $oldname] } {
        gid_groups_conds::rename_group -noactualize_gid_groups -undo 0 $oldname $newname
    }
}

proc gid_groups_conds::GiD_AfterCreateGroup { group } {
    if { ![gid_groups_conds::group_exists $group] } {
        set color [string range [GiD_Groups get color $group] 0 6] ;#remove last xx values of alpha
        gid_groups_conds::add_new_group -noundo -noactualize_gid_groups -color $color -exactname $group
    }
}

proc gid_groups_conds::GiD_AfterRenameGroup { oldname newname } {
    if { [gid_groups_conds::group_exists $oldname] } {
        gid_groups_conds::rename_group -noactualize_gid_groups -undo 0 $oldname $newname
    }

    if { 1 } {
        #I think that now it is not necessary this code, the first gid_groups_conds::rename_group act recursive with childs
        #group already was renamed by GiD, must ask by newname
        set prefix $newname//
        set len [string length $prefix]
        foreach descendant_group [Groups::ListDescendantGroups $newname normal] {
            set descendant_group_relative [string range $descendant_group $len end]
            set oldfullname [Groups::NameJoin $oldname $descendant_group_relative]
            if { [gid_groups_conds::group_exists $oldfullname] } {
                set newfullname [Groups::NameJoin $newname $descendant_group_relative]
                gid_groups_conds::rename_group -noactualize_gid_groups -undo 0 $oldfullname $newfullname
            }
        }
    }
}

proc gid_groups_conds::GiD_AfterDeleteGroup { group } {
    variable delete_gr_from_cond
    if { [gid_groups_conds::group_exists $group] } {
        gid_groups_conds::delete_group -noactualize_gid_groups -undo 0 \
            -delete_gr_from_cond $delete_gr_from_cond $group
    }
}

proc gid_groups_conds::set_doc_groups_from_gid_groups { } {
    foreach group [GiD_Groups list] {
        gid_groups_conds::GiD_AfterCreateGroup $group
    }
}

proc gid_groups_conds::set_gid_groups_from_doc_groups { } {
    variable doc
    set root [$doc documentElement]
    foreach domNode [$root selectNodes groups/group] {
        set group [$domNode @n]
        if { ![GiD_Groups exists $group] } {
            set color [gid_groups_conds::get_group_att $group color]
            #GiD_Process 'Groups create $group escape
            GiD_Groups create $group
            set color_rgb [gid_groups_conds::get_color_rrggbb_from_symbolic_name $color]
            #GiD_Process 'Groups Edit Color $group ${color_rgb}ff escape
            GiD_Groups edit color $group ${color_rgb}ff
        }
    }
}

proc gid_groups_conds::syncronize_gid_groups_and_doc_groups { } {
    gid_groups_conds::set_doc_groups_from_gid_groups
    gid_groups_conds::set_gid_groups_from_doc_groups
}

proc gid_groups_conds::get_color_rrggbb_from_symbolic_name { color } {
    if { [string index $color 0] != "#" } {
        if { [info command winfo] ne "" } {
            lassign [winfo rgb . $color] r16 g16 b16 ;#values from 0 to 65535
            set color_rgb [format "#%02x%02x%02x" [expr int($r16*255/65535)] [expr int($g16*255/65535)] [expr int($b16*255/65535)]]
        } else {
            # orange. when running without tk
            set color_rgb #ffa500
        }
    } else {
        set color_rgb $color
    }
    return $color_rgb
}

##########################################################################################
#GiD END: new procedures required to allow customlib_native_groups
##########################################################################################

#some useful procedures

proc gid_groups_conds::GetDomNodes {xpath } {
    return [[$gid_groups_conds::doc documentElement] selectNodes $xpath]
}

proc gid_groups_conds::AddConditionGroupOnNode {basenode groupid} {
    set prev [$basenode selectNodes "./group\[@n='$groupid'\]"]
    if { [llength $prev] } {
        return $prev
    }
    set newNode [gid_groups_conds::addF [$basenode toXPath] group [list n $groupid]]
    foreach val [$basenode childNodes] {
        if {[$val nodeName] eq "value"} {
            set newChild [$val cloneNode -deep]
            $newNode appendChild $newChild
        }
    }
    return $newNode
}

proc gid_groups_conds::AddConditionGroupOnXPath {xpath groupid} {
    set dom_nodes [gid_groups_conds::GetDomNodes $xpath]
    return [gid_groups_conds::AddConditionGroupOnNode $dom_nodes $groupid]
}


#get a list of dom nodes from a tree condition. The condition child nodes are created at runtime
#each [$xmlnode @n] is the group name, that is paired with a GiD group and its entities could be asked with GiD_EntitiesGroups
proc gid_groups_conds::GetGroupXmlNodesOfCondition { customlib_condition_name } {
    set group_xmlnodes ""
    set root_xmlnode [$::gid_groups_conds::doc documentElement] ;#xml document to get some tree data
    #maybe more efficient $root selectNodes $xpath, with the appropiated xpath
    foreach xmlnode [$root_xmlnode getElementsByTagName condition] {
        if { [$xmlnode @n] == $customlib_condition_name } {
            set group_xmlnodes [$xmlnode getElementsByTagName group]
            #assumed unique name
            break
        }
    }
    return $group_xmlnodes
}

#get the list of group names of a tree condition
proc gid_groups_conds::GetGroupNamesOfCondition { customlib_condition_name } {
    set groups [list]
    foreach xmlnode [PedFlow::GetGroupXmlNodesOfCondition $customlib_condition_name] {
        lappend groups [$xmlnode @n]
    }
    return $groups
}

#to get from a group dom node of a tree condition the value of a field named $field_name
proc gid_groups_conds::GetValueFromGroupXmlNode { group_xmlnode field_name } {
    set value ""
    set found 0
    foreach xmlnode [$group_xmlnode getElementsByTagName value] {
        if { [$xmlnode @n] == $field_name } {
            set value [$xmlnode @v]
            #assumed unique name
            set found 1
            break
        }
    }
    if { !$found } {
        error "field_name does not found"
    }
    return $value
}

proc gid_groups_conds::font_to_drawopengl_font { font } {
    if { $::tcl_platform(platform) ne "windows" } {
        set wght [font actual $font -weight]
        if { $wght eq "normal"} { set wght medium }
        set slant [string index [font actual $font -slant] 0]
        set fnt_family [ font actual $font -family]
        set fnt_size [ font metrics $font -ascent]

        set fnt_family "helvetica" ;# this should be present in all computers
        set com_fnt "-*-$fnt_family-$wght-$slant-*-*-$fnt_size-*-*-*-*-*-*-*"
    } else {
        set com_fnt ""
        lappend com_fnt [ font actual $font -family]
        lappend com_fnt [ font actual $font -size]
        lappend com_fnt [ font actual $font -weight]
        lappend com_fnt [ font actual $font -slant]
    }
    return $com_fnt
}

# function moved here because function_editor.tcl is only loaded if TK is loaded
proc gid_groups_conds::read_matrix_function { gNode } {
    set v [get_domnode_attribute $gNode v]
    set function [$gNode selectNodes {function}]
    set function_name ""
    if { $function ne "" } {
        foreach fNode [$function selectNodes {functionVariable}] {
            if { [$fNode @n] == "matrix_func" } {
                set v ""
                foreach Node [$fNode selectNodes {value}] {
                    set vv [$Node @v]
                    set function [$Node selectNodes {function}]
                    if { $function ne "" } {
                        set n [$Node selectNodes {./function/functionVariable/value}]
                        set vv [$n @v]
                    }
                    lappend v $vv
                }
            }
        }
    }
    return $v
}

# moved to end because regexp expressions are problematic to format the code from VSCode tcl plugin
proc gid_groups_conds::text_html_add_tag_do { cmd_handler text idx1 idx_end } {

    set rex {<(b|i|sub|sup|blockquote|ul|li|small|a|help)([^>]*?)>}

    set idx1 [$text search -regexp -count len1 $rex $idx1 $idx_end]
    if { $idx1 eq "" } {
        return ""
    }

    set rexF {^(.*?)<(b|i|sub|sup|blockquote|ul|li|small|a|help)([^>]*?)>}
    regexp $rexF [$text get $idx1 "$idx1 +$len1 c"] {} before bi rest
    set rex2 "</$bi>"
    set tagsList ""
    while 1 {
        set idx2 [$text search -regexp -count len2 $rex2 "$idx1 +$len1 c" $idx_end]
        if { $idx2 eq "" } {
            return [$text index "$idx1 +$len1 c"]
        }
        lappend tagsList {*}[lassign [text_html_add_tag_do $cmd_handler $text "$idx1 +$len1 c" $idx2] idx_in]
        if { $idx_in eq "" } {
            break
        }
    }

    lassign [list 0 0] deltaL deltaC
    regexp {^(\s*\n)?([ \t]*)(\S.*)} $before$bi$rest {} b sp txt
    incr deltaL [expr {-[regexp -all {\n} $b]}]
    incr deltaC [expr {-[string length $sp]-[string length $txt]}]

    set txt [$text get "$idx1 +$len1 c" $idx2]
    switch $bi {
        b { set tag bold }
        i { set tag italic }
        sub { set tag sub }
        sup { set tag sup }
        blockquote { set tag blockquote }
        ul {
            set tag ""
            regexp {^(\s*\n)?([ \t]*)(\S.*)} $txt {} b sp txt
            incr deltaL [expr {-[regexp -all {\n} $b]}]
            incr deltaC [expr {-[string length $sp]}]
        }
        li {
            set tag item
            set tag ""
            regexp {^(\s*\n)?([ \t]*)(\S.*)} $txt {} b sp txt
            incr deltaL [expr {-[regexp -all {\n} $b]}]
            incr deltaC [expr {-[string length $sp]}]

            set txt "\n\u2022 $txt\n"
            incr deltaC 2
            incr deltaL 1
        }
        small { set tag small }
        a {
            set href $txt
            set ret [regexp {href\s*=\s*["']([^'"]*)['"]} $rest {} href]
            if { !$ret } {
                regexp {href\s*=\s*(\S+)} $rest {} href
            }
            set href [xml_map_inv $href]

            if { [regexp {^http://} $href] } {
                set title "$href"
            } else {
                set title ""
            }
            set ret [regexp {title\s*=\s*["']([^'"]*)['"]} $rest {} title]
            if { !$ret } {
                regexp {title\s*=\s*(\S+)} $rest {} title
            }
            set title [xml_map_inv $title]

            set tag link$idx1
            $text tag configure $tag -foreground blue -underline 1
            if { $title ne "" } {
                gid_groups_conds::register_popup_help -text_tag $tag $text $title
            }
            set cmd $cmd_handler
            lappend cmd $href
            set cmd [string map [list % %%] $cmd]
            $text tag bind $tag <ButtonRelease-1> [list gid_groups_conds::text_html_add_tag_process_cmd $cmd]
            $text tag bind $tag <Enter> [list $text configure -cursor hand2]
            $text tag bind $tag <Leave> [list $text configure -cursor ""]
        }
        help {
            set href $txt
            set ret [regexp {href\s*=\s*["']([^'"]*)['"]} $rest {} href]
            if { !$ret } {
                regexp {href\s*=\s*(\S+)} $rest {} href
            }
            set href [xml_map_inv $href]
            set tag help$idx1
            $text tag configure $tag -underline 1
            gid_groups_conds::register_popup_help -text_tag $tag $text $href
            $text tag bind $tag <Enter> [list $text configure -cursor hand2]
            $text tag bind $tag <Leave> [list $text configure -cursor ""]
        }
    }
    set idx_end [$text index "$idx2 +$len2 c"]
    if { $bi in "ul li" } {
        while { [$text compare $idx1 > 1.0] && [$text search -regexp {\s+} "$idx1-1c" $idx1] ne "" } {
            set deltaL [expr {$deltaL-[regexp -all {\n} [$text get "$idx1-1c" "$idx1"]]}]
            set idx1 [$text index "$idx1-1c"]
            if { $bi eq "ul" && [$text search -regexp {\n} $idx1 "$idx1+1c"] ne "" } {
                break
            }
        }
    }
    $text replace $idx1 $idx_end $txt
    set idx_end_new [$text index "$idx1 +[string length $txt] c"]
    set ret [list $idx_end_new]
    lappend ret [list $idx1 $idx_end_new $tag]
    foreach i $tagsList {
        lassign $i i1 i2 t
        scan $i1 %d.%d l1 c1
        scan $i2 %d.%d l2 c2
        incr l1 $deltaL
        incr l2 $deltaL
        incr c1 $deltaC
        incr c2 $deltaC
        lappend ret [list $l1.$c1 $l2.$c2 $t]
    }
    return $ret
}

# gid_groups_conds::init_package
